"""Includes code for the ActiveSite and related models"""

# django standard imports
from django.db import models
from django.contrib.contenttypes.fields import GenericRelation
from simple_history.models import HistoricalRecords
from django.db.models import Prefetch

# atlas imports
from atlas import settings
import common.models


class ActiveSiteQuerySet(models.QuerySet):
    """Some predefined querysets for the Active Site model"""

    def annotate_residues(self, include_homologues=False, limit_to_references=False):
        """annotates with residue information"""
        queryset = common.models.Residue.objects.annotate_res_chains().annotate_res_sequences()
        if include_homologues:
            queryset = queryset.annotate_res_chains(is_reference=False, limit_to_references=limit_to_references)
            queryset = queryset.annotate_res_sequences(is_reference=False)

        return self.prefetch_related(Prefetch("residue_set", queryset=queryset, to_attr="residues"))


class ActiveSite(models.Model):
    """An active site is a group of residues and cofactors for each entry"""
    entry = models.ForeignKey("Entry", on_delete=models.CASCADE)
    history = HistoricalRecords()

    objects = ActiveSiteQuerySet.as_manager()

    def __str__(self):
        return "AS {0}".format(self.entry)

    def cofactor_count(self):
        active_site_cofactors = ActiveSiteCofactor.objects.filter(
            active_site=self)
        return ", ".join(active_site_cofactor.count_str() for active_site_cofactor in active_site_cofactors)

    def get_reference_residues(self):
        return self.entry.get_reference_residues()


class ActiveSiteCofactor(models.Model):
    """Links the ActiveSite objects to the correspondent Cofactor objects"""
    active_site = models.ForeignKey("ActiveSite", on_delete=models.CASCADE)
    cofactor = models.ForeignKey("Cofactor", on_delete=models.CASCADE)
    count = models.IntegerField(verbose_name="Count", default=1)
    roles = GenericRelation('Role', content_type_field="residue_ct", object_id_field="residue_id",
                            related_query_name="cofactor_roles")
    history = HistoricalRecords()

    objects = models.Manager()

    def __str__(self):
        return "{1}_{0}".format(self.active_site, self.cofactor)

    def count_str(self):
        return "{0} ({1})".format(self.cofactor.name.capitalize(), self.count)

    @property
    def molecule(self):
        return self.cofactor

    def get_mdl_str(self, label_extra=""):
        """Return the mdl mol contents for this cofactor"""
        return self.cofactor.get_mdl_str(label_extra=label_extra)
        # if self.cofactor.mol_file:
            # self.cofactor.mol_file.open()
            # mdl_str = self.cofactor.mol_file.read().decode("utf-8")
            # self.cofactor.mol_file.close()
            # no_atoms = int(mdl_str.splitlines()[3][:3])
            # str_to_add = "\n".join(["A    {0}\nchebi:{1}{2}".format(i+1, self.cofactor.chebi_id, label_extra)
                                    # for i in range(no_atoms)])
            # mdl_str = mdl_str.replace("M  END", str_to_add+"\nM  END")
        # else:
            # missing_filename = "{0}{1}".format(settings.RES_MOL_PATH, "missing.mol")
            # mdl_str = open(missing_filename, 'r').read()
            # mdl_str = mdl_str.replace('missing:',
                                      # 'missing:{}'.format(self.cofactor))
        # return mdl_str + "$$$$\n"
