"""Models related with the Arrow class"""

# python imports
import collections
import logging
from typing import List, Optional, DefaultDict, Tuple, Dict, FrozenSet, Union
import networkx as nx
import math

# django imports
from django.db import models
from rdkit import Chem
import simplejson as json
from django.db.models.query import QuerySet
from django.db.models.manager import BaseManager

# atlas imports
import common.others
from common.models import Entry
import common.models
import common.utils

logger = logging.getLogger("script")


class ArrowQuerySet(models.QuerySet):
    """Some predefined querysets for the arrow model"""

    def get_arrows_graph(self) -> "nx.DiGraph":
        """
        return a graph that represents curly arrows and their relationships
        """
        #: a graph that represents all the curly arrows and their relationship
        graph = nx.DiGraph()

        #: set that stores the mechanisms to be skipped: mechanisms containing metals and radical reactions
        mechs_to_skip: set[int] = set()

        # will process curly arrows one step at a time, so they are sorted and put in a dict
        step_to_arrows: DefaultDict["common.models.Step", list["Arrow"]] = collections.defaultdict(list)
        for arrow in self.order_by("step", "step_order"):
            step_to_arrows[arrow.step].append(arrow)

        for step, arrows in step_to_arrows.items():
            if step.mechanism_id in mechs_to_skip:
                continue
            #: a graph for each step to calculate some properties unique to the step
            step_graph = nx.DiGraph()

            charge_plus = {}
            charge_minus = {}
            for _, arrow in enumerate(arrows):
                # find atoms that gain and loose charge (electrons) with this curly arrow
                am_charge_diff = collections.defaultdict(int)
                # the json field does not work as expected so using eval instead
                arrow_info = eval(arrow.arrow)
                # print(arrow_info)
                # do not process mechanism if any arrow involves radical steps
                if arrow_info[2] == 1 or arrow.radical_in_step:
                    mechs_to_skip.add(arrow.step.mechanism_id)
                    continue

                # get env of entire arrow
                reactant_mol = Chem.MolFromSmarts(arrow.reactant_smarts)
                arrow_ams: set[int] = {atm for arrow_end in arrow_info[:2] for atm in arrow_end}
                arrow_env_join, _ = common.utils.get_mol_env(reactant_mol, arrow_ams)
                arrow_env_mols = Chem.GetMolFrags(arrow_env_join, asMols=True, sanitizeFrags=False)
                # for env in arrow_env_mols:
                    # print(Chem.MolToSmiles(env))

                # add env of arrow as a node
                mols_for_node = []
                am_to_atom = {}
                for mol in arrow_env_mols:
                    for atom in mol.GetAtoms():
                        am_to_atom[atom.GetAtomMapNum()] = atom
                        # these atom maps are here to make canonical smiles take into account the arrow info
                        new_am = 0
                        if atom.GetAtomMapNum() in arrow_info[0]:
                            new_am += 1
                        if atom.GetAtomMapNum() in arrow_info[1]:
                            new_am += 2
                        if new_am:
                            new_am += arrow_info[2]*10
                        atom.SetAtomMapNum(new_am)
                    # using smiles to be canonical
                    mols_for_node.append(Chem.MolToSmiles(mol))

                frozenset_mols_node = frozenset(mols_for_node)
                step_graph.add_node(frozenset_mols_node, arrow=arrow, arrow_env=arrow_env_mols)

                # find atoms that are electron defficient or charged to add connection points between curly arrows
                for am in arrow_info[0]:
                    am_charge_diff[am] += arrow_info[2]/len(arrow_info[0])
                for am in arrow_info[1]:
                    am_charge_diff[am] -= arrow_info[2]/len(arrow_info[1])

                for am, charge_diff in am_charge_diff.items():
                    if math.isclose(1, charge_diff):
                        charge_plus[am] = frozenset_mols_node
                        am_to_atom[am].SetProp("charge_plus", "1")
                    elif math.isclose(-1, charge_diff):
                        charge_minus[am] = frozenset_mols_node
                        am_to_atom[am].SetProp("charge_minus", "1")
                    elif math.isclose(0, charge_diff):
                        pass
                    else:
                        print("ERROR: something funny in curly arrow of step ", step)
                        print(am, charge_diff)

            # add graph edges to connect curly arrows based on charges
            for connection_point in charge_plus.keys() & charge_minus.keys():
                step_graph.add_edge(charge_minus[connection_point], charge_plus[connection_point], )

            # find the start and end nodes of each curly arrow chain. This must be done at the step level
            attributes: DefaultDict[FrozenSet, Dict[str, Union[int, str, bool]]] = collections.defaultdict(dict)
            for node in step_graph.nodes():
                attributes[node]["arrow"] = step_graph.nodes()[node]["arrow"]
                if step_graph.in_degree(node) == 0:
                    attributes[node]["start"] = True
                if step_graph.out_degree(node) == 0:
                    attributes[node]["end"] = True
            nx.set_node_attributes(step_graph, attributes)
            graph = nx.compose(graph, step_graph)

            for node, data in step_graph.nodes(data=True):
                # graph.add_node(node)
                graph.nodes()[node].setdefault("arrows", set())
                graph.nodes()[node]["arrows"].add(data["arrow"])
                for field in ("start", "end"):
                    if field in data:
                        graph.nodes()[node][field] = True

        graph.remove_edges_from(nx.selfloop_edges(graph))

        # code to visualise the graph. comment out when not debugging
        # colors = ["yellow" if "start" in d and "end" in d else "green" if "start" in d else "red"
                  # if "end" in d else "blue" for _, d in graph.nodes(data=True)]
        # labels = {node: str(list(node)) for node in graph.nodes()}
        # from matplotlib import pyplot as plt
        # nx.draw(graph, with_labels=True, labels=labels, node_color=colors)
        # plt.show()
        return graph


class ArrowRule(models.Model):
    rule = models.ForeignKey("Rule", on_delete=models.CASCADE)
    arrow = models.ForeignKey("Arrow", on_delete=models.CASCADE)
    arrow_no = models.IntegerField()
    reversed = models.BooleanField()


ArrowManager = BaseManager.from_queryset(ArrowQuerySet)


class Arrow(models.Model):
    """
    Each arrow object represents an arrow in the scheme xml
    file of a step in either the forward or backward direction
    """
    step = models.ForeignKey("Step", on_delete=models.CASCADE, related_name="arrows")
    rules = models.ManyToManyField("Rule", related_name="arrows", through=ArrowRule)
    is_reversed = models.BooleanField(default=False)
    step_order = models.IntegerField(default=0)
    arrow = models.JSONField()
    radical_in_step = models.BooleanField(default=False)
    reactant_smarts = models.TextField()
    product_smarts = models.TextField()

    objects = ArrowManager()

    class Meta:
        unique_together = ["step", "step_order", "is_reversed"]

    def __str__(self) -> str:
        return "{0.reaction_smarts}".format(self)

    @property
    def reaction_smarts(self) -> str:
        """returns a reaction smarts that can be read into rdkit"""
        return "{0.reactant_smarts}>>{0.product_smarts}".format(self)

    @classmethod
    def create_objects(cls, limit_to_entries: "Optional[QuerySet[Entry]]" = None) -> "List[Arrow]":
        """
        Create Arrow objects based on the mcsa entries. Does not transform molecules as it goes.
        Return the created objects.
        Deletes previously created arrows for the selected entries

        :param limit_to_entries: limit the creation of arrows to these entries
        :type limit_to_entries: "Optional[QuerySet[Entry]]"
        :rtype: "List[Arrow]"
        """
        # Selection of the entries, mechanisms, and steps
        entries = common.models.Entry.objects.mcsa()  # type: ignore
        entries = entries if limit_to_entries is None else entries.filter(id__in=limit_to_entries)
        logger.info("Creating Arrow objects for {} entries".format(entries.count()))
        mechanisms = common.models.Mechanism.objects.entries(entries).filter(csa=False, prediction=False)  # with_good_mechanism()  # type: ignore
        steps = (common.models.Step.objects.filter(mechanism__in=mechanisms).annotate_activesite()  # type: ignore
                 .order_by("mechanism__reaction__entry__macie_id", "mechanism__mechanism_id", "step_id"))

        # Deleting existing Arrow objects
        logger.info("Deleting {} existing Arrow objects for these entries".format(
            cls.objects.filter(step__in=steps).delete()))

        # Get info from step and save it as new arrows
        to_create = []
        for step in steps:
            logger.debug("Parsing arrows for step {}".format(step))
            print(step)
            arrows_info = step.scheme_xml_arrows_info
            radical_in_step = any([arrow[2] == 1 for arrow in arrows_info])
            # find mol product
            product = step.combined_mol
            for arrow_info in arrows_info:
                product = common.utils.get_mol_product_from_arrow(product, arrow_info)

            for step_order, arrow_info in enumerate(arrows_info):
                # todo: delete is_reversed and steo_order
                to_create.append(cls(
                    step_id=step.id,
                    radical_in_step=radical_in_step,
                    reactant_smarts=Chem.MolToSmarts(step.combined_mol),
                    product_smarts=Chem.MolToSmarts(product),
                    is_reversed=False,
                    step_order=step_order,
                    arrow=json.dumps(arrow_info, iterable_as_array=True)))
            step.scheme.close()

        # Commit changes to the database
        logger.info("Adding {} Arrow objects".format(len(to_create)))
        return cls.objects.bulk_create(to_create)
