# coding=utf-8
"""Includes models related with the CATH database"""

# python standard imports
import pathlib
import glob
import os
import textwrap
import gzip
from urllib.request import urlretrieve

# other python imports
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO

# django imports
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.db.models import Count, Prefetch
from django.utils.html import format_html

# atlas imports
from atlas import files
from common import link
import common.models


class CATHQuerySet(models.QuerySet):
    """Some predefined querysets for the CATH model"""

    def complete(self):
        """Shows only proper CATH superfamilies with 4 codes"""
        return self.exclude(cath4=0)

    def order(self):
        """Orders CATH domains. Do not do this by default because it is expensive"""
        return self.order_by("cath1", "cath2", "cath3", "cath4")

    def at_least(self, sequences, num=10):
        """Filters CATH domains that have at least 'num' of these sequences.
        These sequences are saved in attribute 'filtered_sequences'"""
        return self.filter(sequences__in=sequences).annotate(num=Count('sequences')).filter(num__gte=num)\
            .prefetch_related(Prefetch('sequences', queryset=sequences, to_attr="filtered_sequences"))

    def get_code2id(self):
        """Returns a dictionary of CATH code to database if pairs"""
        return {cath.code: cath.id for cath in self}

    def catalytic(self):
        """Returns CATH domains with catalytic residues in M-CSA"""
        return self.filter(domain__residue_chains__isnull=False).distinct("pk")

    def annotate_entries(self):
        """Annotate with entry information"""
        return self.prefetch_related(
                Prefetch(
                    "domain_set",
                    queryset=common.models.Domain.objects.filter(
                        residue_chains__is_reference=True)
                    .annotate_residue_chains(),
                    to_attr="domains"))


class CATH(models.Model):
    """Model for the CATH superfamilies"""
    name = models.CharField(max_length=1023, verbose_name="Name")
    cath1 = models.IntegerField(verbose_name="Class")
    cath2 = models.IntegerField(verbose_name="Architecture")
    cath3 = models.IntegerField(verbose_name="Topology")
    cath4 = models.IntegerField(verbose_name="Homologous Superfamily")
    search_cath = models.CharField(max_length=256, verbose_name="Search CATH", blank=True)

    trees = GenericRelation('Tree', content_type_field="reference_ct", object_id_field="reference_id",
                            related_query_name="cath")

    objects = CATHQuerySet.as_manager()

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        unique_together = ('cath1', 'cath2', 'cath3', 'cath4')

    def __repr__(self):
        return self.code

    def __str__(self):
        return self.code

    @property
    def code(self):
        """Returns formatted CATH superfamily code"""
        return "{0.cath1}.{0.cath2}.{0.cath3}.{0.cath4}".format(self)

    @property
    def code_topology(self):
        """Returns formatted CAT topology code"""
        return "{0.cath1}.{0.cath2}.{0.cath3}".format(self)

    @property
    def wrap_name(self):
        """"Returns the html wrapped version of the name"""
        return format_html("<br>".join(textwrap.wrap(self.name, width=40)))

    @property
    def cath_text_link(self):
        """Returns html text link to superfamily page of CATHdb"""
        return link.cath_code_2_cath(self.code)

    @property
    def cath_icon_link(self):
        """Returns html icon link to superfamily page of CATHdb"""
        return link.cath_code_2_cath(self.code, icon=True)

    @property
    def search_html_link(self):
        """"Returns html link to a superfamily search in the M-CSA search page"""
        return link.search_cath(self.code)

    @classmethod
    def create_fasta(cls, sequences, output_folder, min_sequences=10):
        """Creates fasta files for all CATH that have at least n of these sequences.
        Will delete any fasta existing files in the output_folder

        :type cls: CATH
        :param sequences: only include these sequences in the fasta files
        :type sequences: Sequence
        :param output_folder: fasta files will be created in this folder
        :type output_folder: str
        :param min_sequences: only write fasta files that have at least this number of sequences
        :type min_sequences: int
        """
        pathlib.Path(output_folder).mkdir(parents=True, exist_ok=True)
        for filename in glob.glob("{}/*.fasta".format(output_folder)):
            os.remove(filename)
        caths = CATH.objects.complete().at_least(sequences, min_sequences).order()
        for cath in caths:
            filename = "{}/{}.fasta".format(output_folder, cath.code.replace(".", "_"))
            records = []
            for seq in cath.filtered_sequences:
                if len(seq.sequence) >= 20:
                    records.append(SeqRecord(Seq(seq.sequence), seq.uniprot_id, description=seq.name))
            SeqIO.write(records, filename, "fasta")

    @classmethod
    def download_cath_names(cls):
        """Downloads the cath-names.txt file from the CATH website"""
        urlretrieve(
            "http://download.cathdb.info/cath/releases/daily-release/newest/cath-b-newest-names.gz",
            files.CATH_NAMES
        )

    @classmethod
    def create_objects(cls):
        """Creates CATH objects by reading the cath-names.txt file"""
        existing_caths = set(cls.objects.values_list('cath1', 'cath2', 'cath3', 'cath4', 'name'))
        in_file = set()
        with gzip.open(files.CATH_NAMES, 'rt', encoding='utf8') as cath_names_file:
            for line in cath_names_file:
                line = line.strip()
                code_name = line.split(' ', 1)
                if len(code_name) == 2:
                    code, name = code_name
                else:
                    code = code_name[0]
                    name = ""
                cath1, cath2, cath3, cath4 = (code+".0.0.0.0.").split(".")[:4]
                in_file.add((int(cath1), int(cath2), int(cath3), int(cath4), name))
        to_add = in_file - existing_caths
        print("Adding {} CATH".format(len(to_add)))
        print(to_add)
        for cath1, cath2, cath3, cath4, name in to_add:
            CATH.objects.update_or_create(cath1=cath1, cath2=cath2, cath3=cath3, cath4=cath4, defaults={'name': name})

    @classmethod
    def populate_search_cath(cls):  # TODO I am not sure if this is necessary
        """Populate Search CATH column in CATH table"""
        for cath in cls.objects.all():
            string = "{0.cath1}.-.-.-".format(cath)
            if cath.cath2:
                string += " {0.cath1}.{0.cath2}.-.-".format(cath)
            if cath.cath3:
                string += " {0.cath1}.{0.cath2}.{0.cath3}.-".format(cath)
            if cath.cath4:
                string += " {0.code}".format(cath)
            if cath.search_cath != string:
                cath.search_cath = string
                cath.save()


class DomainQuerySet(models.QuerySet):
    """Some predefined querysets for the Domain model"""

    def annotate_residue_chains(self):
        """Annotates with the reference residue_chains to reference_residue_chains attribute"""
        return self.prefetch_related(
                Prefetch(
                    "residue_chains",
                    queryset=common.models.ResidueChain.objects.reference().prefetch_related(
                        Prefetch(
                            "residue__active_site__entry",
                            queryset=common.models.Entry.objects.annotate_reactions().select_related("owner"),
                            to_attr="annotated_entry"
                        )),
                    to_attr="reference_residue_chains"))


class Domain(models.Model):
    """CATH domains"""
    chain = models.ForeignKey("Chain", on_delete=models.PROTECT)
    structure = models.ForeignKey("Structure", on_delete=models.PROTECT)
    cath = models.ForeignKey('CATH', on_delete=models.PROTECT)
    boundaries = models.CharField(max_length=255, blank=True)
    name = models.CharField(max_length=10, blank=True, verbose_name="Domain name")

    objects = DomainQuerySet.as_manager()

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        unique_together = ('structure', 'name')

    def __repr__(self):
        return "{}{} {}".format(self.structure.pdb_id, self.name, self.cath)

    @classmethod
    def download_cath_boundaries(cls):
        """Downloads the cath-domain-boundaries.txt file from the CATH website"""
        urlretrieve(
            "http://download.cathdb.info/cath/releases/latest-release/cath-classification-data/"
            "cath-domain-boundaries-seqreschopping.txt",
            files.CATH_BOUNDARIES
        )

    @classmethod
    def download_cath_domains(cls):
        """Downloads the cath-domain-list.txt file from the CATH website"""
        urlretrieve(
            "http://download.cathdb.info/cath/releases/latest-release/cath-classification-data/cath-domain-list.txt",
            files.CATH_DOMAINS
        )

    @classmethod
    def create_objects(cls):
        """Adds domain objects from CATH Domain list"""
        from common.models import Structure, Chain
        pdb_to_dbid = dict(Structure.objects.get_idx2id())
        cath_to_dbid = CATH.objects.get_code2id()
        existing = set(Domain.objects.values_list('name', 'structure_id', 'cath_id', 'boundaries'))
        in_file = set()
        with open(files.CATH_BOUNDARIES, 'r') as cath_boundaries_file:
            boundaries = {line[0:7]: line[8:]
                          for line in cath_boundaries_file.read().splitlines()}
        with open(files.CATH_DOMAINS, 'r') as cath_domain_file:
            for line in cath_domain_file:
                if not line.startswith("#"):
                    pdb = line[:4]
                    name = line[4:7]
                    cath = ".".join(line.split()[1:5])
                    if pdb in pdb_to_dbid and cath in cath_to_dbid:
                        in_file.add((name, pdb_to_dbid[pdb], cath_to_dbid[cath], boundaries[pdb+name]))
        to_add = in_file - existing
        print("Adding {} Domains".format(len(to_add)))
        for name, structure_id, cath_id, boundaries in to_add:
            Domain.objects.update_or_create(
                structure_id=structure_id,
                name=name,
                defaults={
                    'cath_id': cath_id,
                    'boundaries': boundaries,
                    'chain': Chain.objects.get(name=name[0], structure_id=structure_id)
                }
            )
