# coding=utf-8
"""Includes the models related with the EC classification"""

# python imports
import textwrap
from urllib.request import urlretrieve
from lxml import etree

# django imports
from django.db import models
from django.db.models import Q, TextField, Value as V, CharField
from django.db.models.functions import Concat, Cast
from django.utils.html import format_html

# atlas imports
from atlas import files
from common import link


class ECQuerySet(models.QuerySet):
    """Some predefined querysets for the EC model"""

    def complete(self):
        """Shows only EC codes with 4 codes"""
        return self.exclude(ec4=0)

    def current(self):
        """Filters only current ECs. Ignores preliminary, transferred and deleted"""
        return self.filter(is_transferred=False, is_deleted=False)

    def get_code2id(self):
        """Returns a dictionary of EC code to database id pairs"""
        return {ec.code: ec.id for ec in self}

    def codes(self, codes):
        """Returns ec objects with these codes"""
        return self.annotated_code().filter(Q(ec1234__in=codes) | Q(ec123__in=codes) | Q(ec12__in=codes))

    def annotated_code(self):
        """Return EC objects annotated with their parent levels"""
            # .annotate(char_ec1=Cast("ec1", TextField())).annotate(char_ec2=Cast("ec2", TextField()))\
            # .annotate(char_ec3=Cast("ec3", TextField())).annotate(char_ec4=Cast("ec4", TextField()))\
        return self\
            .annotate(char_ec1=Cast("ec1", CharField())).annotate(char_ec2=Cast("ec2", CharField()))\
            .annotate(char_ec3=Cast("ec3", CharField())).annotate(char_ec4=Cast("ec4", CharField()))\
            .annotate(ec12=Concat("char_ec1", V("."), "char_ec2"))\
            .annotate(ec123=Concat("char_ec1", V("."), "char_ec2", V("."), "char_ec3"))\
            .annotate(ec1234=Concat("char_ec1", V("."), "char_ec2", V("."), "char_ec3", V("."), "char_ec4"))


class EC(models.Model):
    """Enzyme Commission Number"""
    name = models.CharField(max_length=1023, verbose_name="Name")
    systematic_name = models.CharField(max_length=1023, verbose_name="Systematic Name", blank=True)
    description = models.TextField(verbose_name="Description", blank=True)
    ec1 = models.IntegerField(verbose_name="EC Class")
    ec2 = models.IntegerField(verbose_name="EC Subclass", null=True)
    ec3 = models.IntegerField(verbose_name="EC Subsubclass", null=True)
    ec4 = models.IntegerField(verbose_name="EC Serial", null=True)
    is_preliminary = models.BooleanField()
    is_deleted = models.BooleanField()
    is_transferred = models.BooleanField()
    search_ec = models.CharField(max_length=256, verbose_name="Search EC", blank=True)

    objects = ECQuerySet.as_manager()

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        ordering = ['is_preliminary', 'ec1', 'ec2', 'ec3', 'ec4']
        unique_together = ('ec1', 'ec2', 'ec3', 'ec4', 'is_preliminary')

    def __str__(self):
        return self.code

    @property
    def code(self):
        """The standard representation for EC numbers: 1.1.1.-"""
        ec_l = [self.ec1, self.ec2, self.ec3, self.ec4]
        ec_l = [c if c else '-' for c in ec_l]
        if self.is_preliminary:
            ec_l[3] = 'n'+str(ec_l[3])
        return "{}.{}.{}.{}".format(*ec_l)

    @property
    def code3(self):
        """The standard representation for EC numbers subsubclass: 1.1.1"""
        ec_l = [self.ec1, self.ec2, self.ec3]
        ec_l = [c if c else '-' for c in ec_l]
        if self.is_preliminary:
            ec_l[2] = 'n'+str(ec_l[2])
        return "{}.{}.{}".format(*ec_l)

    @property
    def search_link(self):
        """Returns an URL for the M-CSA search page search for this EC number"""
        return link.search_ec(self.code)

    @property
    def intenz_text_link(self):
        """Returns text link to intenz search page"""
        return link.ec_code_2_intenz(self.code)

    @property
    def intenz_icon_link(self):
        """Returns icon link to intenz search page"""
        return link.ec_code_2_intenz(self.code, icon=True)

    @property
    def explorenz_icon_link(self):
        """Returns icon link to explorenz search page"""
        return link.ec_code_2_explorenz(self.code, icon=True)

    @property
    def kegg_icon_link(self):
        """Return icon link to kegg search page"""
        return link.ec_code_2_kegg(self.code, icon=True)

    @property
    def rhea_icon_link(self):
        """Return icon link to rhea search page"""
        return link.ec_code_2_rhea(self.code, icon=True)

    @property
    def enzyme_icon_link(self):
        """Return icon link to ENZYME ExPASy search page"""
        return link.ec_code_2_enzyme(self.code, icon=True)

    @property
    def metacyc_icon_link(self):
        """Return icon link to METACYC EC page"""
        return link.ec_code_2_metacyc(self.code, icon=True)

    @property
    def brenda_icon_link(self):
        """Return link to the BRENDA search page"""
        return link.ec_code_2_brenda(self.code, icon=True)

    @property
    def wrap_name(self):
        """Wraps EC name for small boxes"""
        return format_html("<br>".join(textwrap.wrap(self.name, width=40)))

    @property
    def name3(self):
        """Build the name of the 3rd level EC using names from 1st, 2nd and 3rd levels"""
        return "{} {} {}".format(
            EC.objects.get(ec1=self.ec1, ec2=0).name,
            EC.objects.get(ec1=self.ec1, ec2=self.ec2, ec3=0).name.lower(),
            self.name.lower())

    @classmethod
    def populate_search(cls):
        """Populate search EC column in EC table"""
        for ec in cls.objects.all():
            string = "{0.ec1}.-.-.-".format(ec)
            if ec.ec2:
                string += " {0.ec1}.{0.ec2}.-.-".format(ec)
            if ec.ec3:
                string += " {0.ec1}.{0.ec2}.{0.ec3}.-".format(ec)
            if ec.ec4:
                string += " {0.code}".format(ec)
            if ec.search_ec != string:
                ec.search_ec = string
                ec.save()

    @classmethod
    def download_intenz_xml_file(cls):
        """Downloads intenz xml file with EC info"""
        urlretrieve("ftp://ftp.ebi.ac.uk/pub/databases/intenz/xml/ASCII/intenz.xml", files.INTENZ_XML)

    @classmethod
    def create_objects(cls):
        """Creates EC objects reading from the intenz xml file"""
        for ec1 in etree.parse(str(files.INTENZ_XML)).getroot():
            ec1_no = ec1.get("ec1")
            cls._create_an_object([ec1_no, 0, 0, 0], xml_obj=ec1)
            for ec2 in cls._find_children_with_tag(ec1, 'ec_subclass'):
                ec2_no = ec2.get("ec2")
                cls._create_an_object([ec1_no, ec2_no, 0, 0], xml_obj=ec2)
                for ec3 in cls._find_children_with_tag(ec2, 'ec_sub-subclass'):
                    ec3_no = ec3.get("ec3")
                    cls._create_an_object([ec1_no, ec2_no, ec3_no, 0], xml_obj=ec3)
                    for ec4 in cls._find_children_with_tag(ec3, 'enzyme'):
                        ec4_no = ec4.get("ec4")
                        ec4_obj = cls._create_an_object([ec1_no, ec2_no, ec3_no, ec4_no], xml_obj=ec4)
                        if ec4_obj is None:
                            continue
                        synonym_parent = cls._find_children_with_tag(ec4, 'synonyms')
                        if synonym_parent:
                            for syn in synonym_parent[0]:
                                Synonym.objects.get_or_create(ec=ec4_obj, name=syn.text)

    @classmethod
    def _create_an_object(cls, ecs, xml_obj=None):
        """Create or update ec_object from the ec codes and the ec xml object"""
        name_key = 'accepted_name' if ecs[3] else 'name'
        description = cls._find_child_text_with_tag(xml_obj, 'description')
        comments_parent = cls._find_children_with_tag(xml_obj, 'comments')
        is_preliminary = xml_obj.get("preliminary") == "true"
        if comments_parent:
            description += " ".join([c.text for c in comments_parent[0]])
        ec_obj, _ = EC.objects.update_or_create(
            ec1=ecs[0], ec2=ecs[1], ec3=ecs[2], ec4=ecs[3], is_preliminary=is_preliminary,
            defaults={
                'is_deleted': bool(cls._find_children_with_tag(xml_obj, 'deleted')),
                'is_transferred': bool(cls._find_children_with_tag(xml_obj, 'transferred')),
                'name': cls._find_child_text_with_tag(xml_obj, name_key),
                'description': description,
                'systematic_name': cls._find_child_text_with_tag(xml_obj, 'systematic_name'),
            })
        return ec_obj

    @staticmethod
    def _find_children_with_tag(elem, tag):
        """Returns children that contains this tag in xml object"""
        return [child for child in elem if etree.QName(child).localname == tag]

    @classmethod
    def _find_child_text_with_tag(cls, elem, tag):
        """Returns the text of the child with this tag of a xml object"""
        results = cls._find_children_with_tag(elem, tag)
        if results:
            return results[0].text
        else:
            return ''


class Synonym(models.Model):
    """Synonyms of Enzyme names"""
    ec = models.ForeignKey('EC', on_delete=models.CASCADE)
    name = models.CharField(max_length=1023, verbose_name="name")

    objects = models.Manager()

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        unique_together = ('ec', 'name')
