# coding=utf-8
"""Includes models related with the M-CSA entries"""

# python imports
import itertools
import pathlib
import pickle
import re

# django imports
from django.contrib.auth.models import User
from django.db import models
from simple_history.models import HistoricalRecords
from django.db.models import Prefetch
from django.db.models.query import QuerySet


# atlas imports
from common import link
import common.models

PUBLIC_USERNAME = 'macie'


class EntryContributor(models.Model):
    """Links entries with users that are contributors"""
    entry = models.ForeignKey("Entry", on_delete=models.CASCADE)
    user = models.ForeignKey(User, on_delete=models.CASCADE, verbose_name="Annotator")
    notes = models.TextField(verbose_name='Notes', blank=True)
    history = HistoricalRecords()


class EntryQuerySet(models.QuerySet):
    """Some predefined querysets for the Entry model"""

    def mcsa(self) -> 'QuerySet[Entry]':
        """Filters entries that are shown on the public website"""
        return self.filter(owner__username=PUBLIC_USERNAME)

    def mechanism(self, mechanism: bool = True) -> 'QuerySet[Entry]':
        """Filters Entries with mechanism that are shown on the public website"""
        return self.mcsa().filter(reaction__mechanism__csa=not mechanism).distinct()

    # TODO remove and use mechanism(mechanism=False) instead
    def catalytic_site(self):
        """Filters Entries without mechanism that are shown on the public website"""
        return self.mcsa().exclude(reaction__mechanism__csa="False").distinct()

    def residues(self, residues):
        """Filters entries that own these residues"""
        return self.filter(activesite__residue__in=residues)

    def residue_sequences(self, residue_sequences):
        """Filters entries that own these residues sequences"""
        return self.filter(activesite__residue__residuesequences__in=residue_sequences)

    def proteins(self, proteins):
        """Filters entries that have these proteins"""
        return self.filter(protein__in=proteins)

    def reactions(self, reactions):
        """Filters entries that have these reactions"""
        return self.filter(reaction__in=reactions)

    def mcsa_ids(self, mcsa_ids):
        """Filters entries that have these mcsa_ids"""
        return self.filter(macie_id__in=mcsa_ids)

    def with_pdb(self):
        """Include only entries associated with a PDB entry"""
        return self.filter(protein__pdb__in=common.models.Pdb.objects.filter(is_reference=True, protein__is_reference=True))

    def annotate_proteins(self, is_reference=True):
        """Annotates entries with the proteins, pdbs, sequence and structures"""
        return self.prefetch_related(
            Prefetch(
                "protein_set",
                queryset=common.models.Protein.objects.filter(is_reference=is_reference)
                               .prefetch_related("sequences").annotate_pdbs(),
                to_attr="proteins"))

    def annotate_reactions(self, is_reference=True, only_three_stars_mechanisms=True, **kwargs):
        """Annotates entries with the reactions"""
        queryset = common.models.Reaction.objects.filter(is_reference=is_reference)\
                                                 .select_related("ec_code")\
                                                 .annotate_mechanisms(only_three_stars=only_three_stars_mechanisms)
        for key, value in kwargs.items():
            if key.startswith("reaction_annotate") and value == True:
                method_to_call = getattr(queryset, re.sub("^reaction_", "", key))
                queryset = method_to_call(**kwargs)
        return self.prefetch_related(Prefetch("reaction_set", queryset=queryset, to_attr="reactions"))

    def annotate_activesites(self, include_homologues=False, limit_to_references=False):
        """Annotate entries with residue and cofactor information"""
        return self.prefetch_related(
            Prefetch(
                "activesite_set",
                queryset=common.models.ActiveSite.objects.prefetch_related("activesitecofactor_set__cofactor")
                               .annotate_residues(include_homologues=include_homologues, limit_to_references=limit_to_references),
                to_attr="activesites"))

    def evolution_study_sets(self):
        """Sorts entries in sets usefull to study enzyme evolution"""
        self = self.annotate_reactions(reaction_annotate_mechanisms=True).annotate_activesites()
        sets = {k: {"Same CATH": {}, "Same CAT": {}, "Different Topology": {}}
                                 for k in ("Different EC", "Same subsubclass", "Same EC")}
        for pair in itertools.combinations(self, 2):
            # find correct ec key
            ecs = [entry.reactions[0].ec_code for entry in pair]
            if ecs[0].code3 == ecs[1].code3:
                if ecs[0].code == ecs[1].code and "-" not in ecs[0].code:
                    ec_key = "Same EC"
                elif "-" not in ecs[0].code3:
                    ec_key = "Same subsubclass"
                else:
                    continue  # ignore these
            else:
                ec_key = "Different EC"

            # find correct cath key
            caths = []
            caths3 = []
            for entry in pair:
                caths.append(set([residue.res_chain[0].domain.cath.code for residue in entry.activesites[0].residues
                                  if residue.res_chain and residue.res_chain[0].domain]))
                caths3.append(set([residue.res_chain[0].domain.cath.code_topology for residue in entry.activesites[0].residues
                                  if residue.res_chain and residue.res_chain[0].domain]))

            if not caths[0] and not caths[1]:
                continue
            elif caths[0] == caths[1]:  # maybe not the best way to deal with proteins with more than one cath
                cath_key = "Same CATH"
            elif caths[0] & caths[1]:
                continue
            elif caths3[0] == caths3[1]:
                cath_key = "Same CAT"
            elif caths3[0] & caths3[1]:
                continue
            else:
                cath_key = "Different Topology"

            # mechanism_similarity similarity
            try:
                mechanism_similarity = common.others.fingerprint_similarity(
                        pair[0].reactions[0].mechanisms[0].bond_changes_fingerprint,
                        pair[1].reactions[0].mechanisms[0].bond_changes_fingerprint)
            except IndexError:  # no mechanism with good rating
                continue

            # reaction bond change similarity
            reaction_similarity = common.others.fingerprint_similarity(
                    *[entry.reactions[0].fingerprint.bond_changes_fingerprint for entry in pair]
            )
            sets[ec_key][cath_key][pair] = {}
            sets[ec_key][cath_key][pair]["reaction_similarity"] = reaction_similarity
            sets[ec_key][cath_key][pair]["mechanism_similarity"] = mechanism_similarity

        return sets


class Entry(models.Model):
    """The base model for M-CSA.

    A M-CSA entry is defined as a chemical reaction that happens in an
    active centre using the same mechanism.
    """
    macie_id = models.IntegerField(verbose_name="M-CSA ID")
    enzyme_name = models.CharField(max_length=255, verbose_name="Enzyme Name")
    description = models.TextField()
    notes = models.TextField(verbose_name='Notes', blank=True)
    creator_macie_id = models.IntegerField(verbose_name="User M-CSA ID", null=True)
    complete = models.BooleanField(default=False)
    interpro = models.CharField(max_length=255, verbose_name="InterPro", blank=True)
    slfd = models.CharField(max_length=255, blank=True, verbose_name="SFLD Family")
    in_metal_macie = models.BooleanField(default=False, verbose_name='In Metal MACiE')
    last_time_updated = models.DateField(null=True, blank=True)
    creator = models.ForeignKey(User, related_name='creator', null=True, on_delete=models.PROTECT)
    owner = models.ForeignKey(User, related_name='owner_entry', on_delete=models.PROTECT)
    contributors = models.ManyToManyField(User, through='EntryContributor', related_name='contributors')
    history = HistoricalRecords()
    # dummy entries are created to make predictions, they do not contain annotations
    dummy = models.BooleanField(default=False)

    objects = EntryQuerySet.as_manager()

    class Meta:
        ordering = ['macie_id']
        verbose_name_plural = "entries"
        unique_together = ('macie_id', 'owner')

    def __str__(self):
        return "{0}:{1.macie_id} - {1.enzyme_name}".format(
            self.owner.username.replace('macie', 'm-csa'), self)

    @property
    def mcsa_id(self):
        return self.macie_id

    @property
    def interpro_html_link(self):
        return link.interpro_id_2_interpro(self.interpro)

    @property
    def interpro_icon_link(self):
        return link.interpro_id_2_interpro(self.interpro, icon=True)

    @property
    def metal_macie_text_link(self):
        if self.in_metal_macie:
            return link.mcsa_id_2_metal_macie(self.macie_id)

    # usefull sets
    @classmethod
    def get_evolution_study_sets(cls, reread=False):
        pickle_file = pathlib.Path("/tmp/evolution_sets_pickle")
        if not pickle_file.exists() or reread:
            print("reading data from database and saving pickle")
            sets = cls.objects.mcsa().mechanism().evolution_study_sets()
            pickle.dump(sets, pickle_file.open(mode="wb"))
        else:
            print("reading data pickle")
            sets = pickle.load(pickle_file.open(mode="rb"))
        print("reading done")
        return sets

    def set_values_from_uniprot(self, protein, overwrite=False):
        """Use the Uniprot ID of a protein and set missing fields"""
        if overwrite or not self.enzyme_name:
            self.enzyme_name = protein.get_enzyme_name_from_uniprot()
        if overwrite or not self.description:
            self.description = protein.get_description_from_uniprot()

    def find_repeated_interpro(self):
        return [entry for entry in Entry.objects.mcsa() if self.interpro and self.interpro == entry.interpro and
                entry.id != self.id]

    def save(self, *args, **kwargs):
        if not self.macie_id:
            try:
                last = Entry.objects.filter(owner=self.owner).order_by('macie_id').last()
                last_id = last.macie_id + 1
            except AttributeError:
                last_id = 1
            try:
                macie_last = Entry.objects.filter(
                    owner=User.objects.get(username="macie"),
                    creator=self.owner).order_by('creator_macie_id').last()
                last_macie_id = macie_last.creator_macie_id + 1
            except (AttributeError, User.DoesNotExist) as e:
                last_macie_id = 1
            self.macie_id = max([last_id, last_macie_id])
            self.creator_macie_id = self.macie_id if not self.creator_macie_id \
                else self.creator_macie_id
            self.creator = self.owner if not self.creator else self.creator
        super(Entry, self).save(*args, **kwargs)

    # TODO this will not be necessary as there will not be non reference residues in residues table
    def get_reference_residues(self):
        return self.activesite_set.get().residue_set.filter(
            protein__is_reference=True,
            pdb__is_reference=True,
            protein__entry=self)

    def get_catalytic_cath(self):
        return common.models.CATH.objects.filter(domain__residue_chains__is_reference=True,
                                   domain__residue_chains__residue__in=self.get_reference_residues()).distinct()
