# coding=utf-8
"""Includes the Mechanism model"""

# django imports
from django.db import models
from django.urls import reverse
from django.utils.html import format_html
from simple_history.models import HistoricalRecords
from django.db.models import Prefetch

# atlas settings
from common.models import EvidenceType
import common.models


# TODO REMOVE use link
def make_link(url, placeholder, title="", target="_blank", text=None):
    if text is None:
        text = placeholder
    return format_html(
        '<a href="{}" title="{}" target="{}">{}</a>'.format(
            url.format(placeholder), title, target, text)
    )


class MechanismQuerySet(models.QuerySet):
    """Some predefined querysets for the Mechanism model"""

    def entries(self, entries):
        """Filters mechanisms with these entries"""
        return self.filter(reaction__entry__in=entries)

    def good_rating(self):
        """Filters mechanisms that have three stars"""
        return self.filter(rating=3, prediction=False)

    def with_good_mechanism(self):
        """Filters mechanism with detailed information and three stars"""
        return self.good_rating().filter(csa=False)

    def annotate_bib_references(self):
        """Annotates with the bibliographic references"""
        return self.filter(prediction=False).prefetch_related(
            Prefetch(
                "annotations",
                queryset=common.models.ReferenceType.objects.exclude(evidence_type__evidence_type="csa_annotation").exclude(evidence_type__evidence_type="primary reference")
                               .prefetch_related("evidence_type", "reference"),
                to_attr="bib_references"
                ))

    def annotate_steps(self):
        """Annotates with the steps and their fingerprints"""
        return self.filter(prediction=False).prefetch_related(
                Prefetch(
                    "step_set",
                    queryset=common.models.Step.objects.all().annotate_roles(),
                    to_attr="steps"
                    ))


class Mechanism(models.Model):
    """A mechanism proposal for this reaction in this enzyme"""
    reaction = models.ForeignKey("Reaction", on_delete=models.PROTECT)
    mechanism_id = models.IntegerField(verbose_name="Mechanism ID")
    mechanism_text = models.TextField(verbose_name="Mechanism Introduction Text", blank=True)
    curator_notes = models.TextField(verbose_name="Curator Notes", blank=True)
    reference = models.ManyToManyField('Reference', through='ReferenceType')  # todo change to references
    REVERSIBLE_CHOICES = (
        ("Unknown", "Unknown"),
        ("yes", "yes"),
        ("no", "no"),
    )
    reversible = models.CharField(max_length=10, choices=REVERSIBLE_CHOICES, default="unknown")
    RATING_CHOICES = (
        (1, "1"),
        (2, "2"),
        (3, "3"),
    )
    rating = models.IntegerField(choices=RATING_CHOICES, default=3)
    protein = models.ForeignKey("Protein", on_delete=models.PROTECT, null=True) # is this necessary?
    pdb = models.ForeignKey("Pdb", null=True, on_delete=models.PROTECT)
    complete = models.BooleanField(default=False, verbose_name="Complete")
    csa = models.BooleanField(default=False, verbose_name="Catalytic Site Only")
    prediction = models.BooleanField(default=False)
    history = HistoricalRecords()

    objects = MechanismQuerySet.as_manager()

    class Meta:
        ordering = ['mechanism_id']

    def __repr__(self):
        return (
            "Mechanism: {0.id} {0.reaction.entry.owner.username}."
            "{0.reaction.entry.macie_id}.{0.mechanism_id}".format(self))

    def __str__(self):
        return (
            "Mechanism: {0.reaction.entry.owner.username}."
            "{0.reaction.entry.macie_id}.{0.mechanism_id}".format(self))

    def get_stars(self):
        return ''.join(['&#9734;' for _ in range(self.rating)])

    def clean(self):
        if self.mechanism_id:
            return self.mechanism_id
        else:
            current_mechanisms = Mechanism.objects.filter(reaction=self.reaction).order_by('-mechanism_id')
            if current_mechanisms:
                self.mechanism_id = current_mechanisms[0].mechanism_id + 1
            else:
                self.mechanism_id = 1

    @property
    def primary_references(self):
        return self.annotations.filter(evidence_type=EvidenceType.objects.get(id=17))  # TODO use custom manager

    @property
    def edit_link(self):
        url = reverse('common:edit_mechanism', args=(
            self.reaction.entry.owner.username,
            self.reaction.entry.macie_id,
            self.mechanism_id)
                      )
        text = "M{0}".format(self.mechanism_id)
        return make_link(url, "", title="Edit mechanism", target="", text=text)

    @property
    def bond_changes_fingerprint(self):
        """returns bond changes fingerprint which is built from the steps fingerprints

        Much faster for several mechanisms is they are annotated with steps first. Use mechanisms.annotate_steps()"""
        try:
            steps = self.steps
        except AttributeError:
            steps = self.step_set.all()

        fingerprint = [0] * len(common.models.Step.get_bond_changes_fingerprint_key())
        for step in steps:
            for i, v in enumerate(step.fingerprint.bond_changes_fingerprint):
                fingerprint[i] += v
        return fingerprint
