# coding=utf-8
"""Includes models related with small molecules"""

# python imports
import csv
import gzip
import os
import subprocess
from collections import defaultdict

# other python imports
from PIL import Image

# django imports
from django.core.files.base import ContentFile
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.utils.html import format_html
from django.urls import reverse
from simple_history.models import HistoricalRecords

# atlas imports
from atlas import settings
from atlas import files
from common import link
from common.validators import validate_chebi_id_exists


class Cofactor(models.Model):
    """A organic or metal cofactor"""
    # TODO group them into the 27 groups
    active_sites = models.ManyToManyField("ActiveSite", through="ActiveSiteCofactor")
    pdb_code = models.CharField(max_length=3, unique=True, verbose_name='PDB Code', db_index=True)
    name = models.CharField(max_length=255, blank=True, verbose_name='PDB Name', db_index=True)  # TODO pdb -> chebi
    chebi = models.CharField(max_length=10, verbose_name='ChEBI ID', db_index=True)  # make unique and main
    kegg = models.CharField(max_length=10, blank=True, verbose_name='KEGG Compound', db_index=True)
    is_metal = models.BooleanField()
    mol_file = models.FileField(verbose_name='Mol File', blank=True, upload_to='cofactor_mols/')
    history = HistoricalRecords()

    objects = models.Manager()

    def __str__(self):
        return "{0}: {1}".format(self.pdb_code, self.name)

    @property
    def formatted_str(self):
        return "{}".format(self.name)

    # TODO fix links
    @property
    def chebi_html_link(self):
        if self.chebi.startswith('X'):
            return format_html(
                '<a href="{0}" title="Not a ChEBI compound. Click to edit"'
                ' target="_blank">{1}</a>'.format(
                    reverse('common:edit_compound', args=(self.chebi, )),
                    self.chebi
                ))
        else:
            return format_html(
                '<a href="https://www.ebi.ac.uk/chebi/searchId.do?'
                'chebiId=CHEBI:{0}" target="_blank" title="Open in ChEBI">'
                '{0}</a>'.format(self.chebi))

    @property
    def kegg_html_link(self):
        return format_html(
            '<a href="http://www.genome.jp/dbget-bin/www_bget?cpd:'
            '{0}" title="Open in KEGG" target="_blank">{0}</a>'.format(
                self.kegg))

    @property
    def pdb_code_html_link(self):
        return format_html(
            '<a href="http://www.ebi.ac.uk/pdbe-srv/pdbechem/chemicalCompound/'
            'complete/{0}" title="Go to PDBe" target="_blank">{0}</a>'.format(
                self.pdb_code))

    def _set_mol_file_from_chebi(self):
        """Set the mol file from chebi website"""
        db_mol_name = os.path.basename(self.mol_file.name)
        if not db_mol_name.startswith(self.chebi):
            mol_filename = "{0}.mol".format(self.chebi)
            mol_file_text = validate_chebi_id_exists(self.chebi)
            self.mol_file.save(mol_filename, ContentFile(mol_file_text))
        else:
            print("{0} file already exists in the database.".format(
                self.mol_file))

    @property  # TODO delete and make uniform
    def chebi_id(self):
        return self.chebi

    def get_mdl_str(self, label_extra=""):
        """Return the mdl mol contents for this cofactor"""
        if self.mol_file:
            self.mol_file.open()
            mdl_str = self.mol_file.read().decode("utf-8")
            self.mol_file.close()
            no_atoms = int(mdl_str.splitlines()[3][:3])
            str_to_add = "\n".join(["A    {0}\nchebi:{1}{2}".format(i+1, self.chebi_id, label_extra)
                                    for i in range(no_atoms)])
            mdl_str = mdl_str.replace("M  END", str_to_add+"\nM  END")
        else:
            missing_filename = "{0}/{1}".format(settings.RES_MOL_PATH, "missing.mol")
            mdl_str = open(missing_filename, 'r').read()
            mdl_str = mdl_str.replace('missing:',
                                      'missing:{}'.format(self))
        return mdl_str + "$$$$\n"


class GroupCofactor(models.Model):  # change later to Cofactor
    """A table with all the cofactors groups"""
    name = models.CharField(max_length=215, db_index=True)
    is_metal = models.BooleanField()

    @classmethod
    def create_objects(cls):
        """Creates models by reading cofactor like file. the cofactor file has no unique identifiers
        so do not run this files blindly, or the table will end up with duplicates. Check only when link_to_compound
        fails due to unknown cofactor_name"""
        unique_names = set()
        with open(files.COFACTOR_LIKE, 'r') as cofactor_like_file:
            for line in cofactor_like_file:
                _, cofactor_name, het_code = line.strip().split(",")
                unique_names.add(cofactor_name)
        for name in unique_names:
            cls.objects.get_or_create(name=name, is_metal=False)

    @classmethod
    def link_to_compound(cls):
        """Updates the cofactor like column"""
        with open(files.COFACTOR_LIKE, 'r') as cofactor_like_file:
            for line in cofactor_like_file:
                _, cofactor_name, het_code = line.strip().split(",")
                cofactor = cls.objects.get(name=cofactor_name)
                for compound in Compound.objects.filter(pdbe_chem=het_code).exclude(cofactor=cofactor):
                    compound.cofactor = cofactor
                    compound.save()


class Compound(models.Model):
    """A small molecule"""
    chebi_id = models.CharField(max_length=10, verbose_name="ChEBI ID", unique=True, db_index=True)
    name = models.CharField(max_length=1023, db_index=True)
    description = models.CharField(max_length=2047)
    kegg = models.CharField(max_length=10, blank=True, db_index=True)
    pdbe_chem = models.CharField(max_length=3, db_index=True)
    cofactor = models.ForeignKey("GroupCofactor", on_delete=models.SET_NULL, null=True)

    # Choose or the other TODO
    mol_file = models.FileField(verbose_name='Mol File', blank=True, upload_to='compound_mols/')
    mol = models.TextField()
    image_file = models.FileField(verbose_name='Image File', blank=True, upload_to='compound_images/')

    objects = models.Manager()

    def __str__(self):
        return "{0.chebi_id}_{0.name}".format(self)

    @classmethod
    def read_structures(cls):
        """Returns chebi to mol structures dict by reading the chebi structures file"""
        chebi2mol = {}
        secondary_to_main = cls.read_secondary_names()
        with gzip.open(files.CHEBI_STRUCTURES, 'rt') as structures_file:
            csv_reader = csv.reader(structures_file)
            for words in csv_reader:
                chebi_id = words[1]
                if chebi_id in secondary_to_main:
                    chebi_id = secondary_to_main[chebi_id]
                if words[3] == 'mol':
                    chebi2mol[chebi_id] = words[2]
        return chebi2mol

    @classmethod
    def read_relation(cls, relations):
        """Returns dictionary with information of the selected relations"""
        related_dict = defaultdict(list)
        with open(files.CHEBI_OBO, 'r') as relation_file:
            chebi1 = ""
            for line in relation_file:
                if line == "\n":
                    chebi1 = ""
                elif line.startswith("id: CHEBI:"):
                    chebi1 = line.strip()[10:]
                elif line.startswith("relationship:") and line.split()[1] in relations:
                    chebi2 = line.split()[2].split(":")[1]
                    related_dict[chebi1].append(chebi2)
        return related_dict

    @classmethod
    def read_chebi_accessions(cls):
        """Reads chebi to kegg links from database_accession file"""
        secondary_to_main = cls.read_secondary_names()
        chebi2kegg = {}
        chebi2pdb = {}
        conjugate_relations = cls.read_relation(('is_conjugate_acid_of', 'is_conjugate_base_of'))
        with open(files.CHEBI_ACCESSION, 'r') as accession_file:
            for line in accession_file:
                words = line.split("\t")
                idx = words[4].strip()
                chebi_id = words[1]
                if chebi_id in secondary_to_main:
                    chebi_id = secondary_to_main[chebi_id]
                if words[3].startswith("KEGG"):
                    if chebi_id in chebi2kegg and chebi2kegg[chebi_id].startswith("C"):
                        continue  # do not replace Compounds by other databases (drug, gluc)
                    else:
                        chebi2kegg[chebi_id] = idx
                elif words[3] == "PDBeChem accession" and len(idx) <= 3:
                    chebi2pdb[chebi_id] = idx
                    for conjugate in conjugate_relations[chebi_id]:
                        if idx == "FAD":
                            print(conjugate)
                        if conjugate not in chebi2pdb:
                            chebi2pdb[conjugate] = idx
        return chebi2kegg, chebi2pdb

    @classmethod
    def read_secondary_names(cls):
        """Reads CHEBI secondary names and returns a secondary->main dict"""
        secondary_to_main = {}
        with gzip.open(files.CHEBI_SDF, 'rt') as sdf_file:
            current_id = None
            reading_main = reading_sec = False
            for line in sdf_file:
                if line.startswith("> <ChEBI ID>"):
                    reading_main = True
                elif reading_main:
                    current_id = line.strip()[6:]
                    reading_main = False
                elif line.startswith("> <Secondary ChEBI ID>"):
                    reading_sec = True
                elif reading_sec:
                    if line.startswith("CHEBI:"):
                        secondary_to_main[line.strip()[6:]] = current_id
                    else:
                        reading_sec = False
        return secondary_to_main

    @classmethod
    def create_objects(cls):
        """Creates all objects from CHEBI compounds.tsv.gz files"""
        chebi2kegg, chebi2pdb = cls.read_chebi_accessions()
        chebi2mol = cls.read_structures()
        existing = set(cls.objects.values_list("chebi_id", "name", "description", "kegg", "pdbe_chem", "mol"))
        with gzip.open(files.CHEBI_COMPOUNDS, 'rt') as compounds_file:
            compounds_file.readline()
            for no, line in enumerate(compounds_file):
                words = line.split("\t")
                name = words[5]
                if name == "null":
                    continue
                chebi_id = words[0]
                description = words[6] if words[6] != "null" else ""
                kegg = chebi2kegg[chebi_id] if chebi_id in chebi2kegg else ''
                pdbe_chem = chebi2pdb[chebi_id] if chebi_id in chebi2pdb else ''
                mol = chebi2mol[chebi_id] if chebi_id in chebi2mol else ''
                if (chebi_id, name, description, kegg, pdbe_chem, mol) not in existing:
                    cls.objects.update_or_create(
                        chebi_id=chebi_id,
                        defaults={"name": name, "description": description, "kegg": kegg, "pdbe_chem": pdbe_chem,
                                  "mol": mol}
                    )

    @property
    def chebi_html_link(self):
        """Returns text URL to this molecule in the ChEBI website"""
        return self.chebi_id if self.chebi_id.startswith("X") else link.chebi_id_2_chebi(self.chebi_id)

    @property
    def chebi_icon_link(self):
        """Returns icon URL to this molecule in the ChEBI website"""
        return self.chebi_id if self.chebi_id.startswith("X") else link.chebi_id_2_chebi(self.chebi_id, True)

    @property
    def kegg_html_link(self):
        """Returns icon URL to this molecule in the KEGG website"""
        return link.kegg_compound_id_2_kegg(self.kegg) if self.kegg else ""

    @property
    def search_html_link(self):
        """Returns an URL for the M-CSA search page search for this chebi id"""
        return link.search_chebi_id(self.chebi_id)

    @property
    def edit_compound_link(self):
        """Returns a link to the M-CSA page that edits a compound"""
        if self.chebi_id.startswith("X"):
            return format_html(
                '<a href="{0}" title="Click to edit the structure of the ChEBI '
                ' compound. You will affect all entries that link to this '
                'compound." target="_blank">edit</a>'.format(reverse('common:edit_compound', args=(self.chebi_id, ))))

    def set_png_from_mol(self):
        """Create png file from the mol field"""
        if not self.mol:
            return
        # print("converting {}".format(self.mol_file.name))
        png_filename = "{0}.png".format(self.chebi_id)
        images_path = 'compound_images/'
        self.image_file = "{}{}".format(images_path, png_filename)
        molconvert_command = ("echo '{}' | molconvert 'png:scale200,#00ffffff' > {}/{}".format(
            self.mol, settings.MEDIA_ROOT, self.image_file))
        subprocess.getoutput(molconvert_command)
        print("{} saved".format(png_filename))
        img = Image.open("{0}/{1}".format(settings.MEDIA_ROOT, self.image_file))
        cropped = img.crop(img.getbbox())
        cropped.save("{0}/{1}".format(settings.MEDIA_ROOT, self.image_file))
        self.save()

    def get_mdl_str(self, label_extra="", replace_mol=None):
        """Return the mdl mol contents for this compound"""
        mol = self.mol if replace_mol is None else replace_mol
        if not mol:
            missing_filename = "{0}/{1}".format(settings.RES_MOL_PATH, "missing.mol")
            mdl_str = open(missing_filename, 'r').read()
            mdl_str = mdl_str.replace("missing:", "{}".format(
                self.chebi_id))
            no_atoms = 1
        else:
            # mol_file.open()
            # mdl_str = mol_file.read().decode("utf-8")
            # mol_file.close()
            no_atoms = int(mol.splitlines()[3][:3])
        str_to_add = "\n".join(["A    {0}\nchebi:{1}{2}".format(i+1, self.chebi_id, label_extra)
                                for i in range(no_atoms)])
        mol = mol.replace("M  END", str_to_add+"\nM  END")
        return mol + "$$$$\n"


class CompoundType(models.Model):
    """Links the compound objects to their type and reaction"""
    reaction = models.ForeignKey("Reaction", on_delete=models.CASCADE)
    compound = models.ForeignKey("Compound", on_delete=models.PROTECT)
    count = models.IntegerField(verbose_name="Count")
    belongs_to = models.CharField(max_length=255, verbose_name="Belongs to", blank=True)
    COMPOUND_TYPE_CHOICES = (("reactant", "Reactant"), ("product", "Product"), ("cofactor", "Cofactor"))
    type = models.CharField(max_length=255, verbose_name="Type", choices=COMPOUND_TYPE_CHOICES, default="reactant")
    roles = GenericRelation('Role', content_type_field="residue_ct", object_id_field="residue_id",
                            related_query_name="compound_roles")
    history = HistoricalRecords()

    objects = models.Manager()

    class Meta:
        pass
        #ordering = ['-type']

    @property
    def molecule(self):
        return self.compound

    def __str__(self):
        return "{0}_{1}".format(self.type, self.compound)

    def get_mdl_str(self, label_extra="", replace_mol=None):
        """Return the mdl mol contents for this compound"""
        return self.compound.get_mdl_str(label_extra=label_extra, replace_mol=replace_mol)
