# coding=utf-8
"""Includes the models related with the chemical reactions"""

# python standard imports
import collections
import os
import subprocess
import tempfile
from rdkit import Chem
from typing import FrozenSet, List, Set, Tuple, Dict

# django imports
from django.core.files.base import File
from django.db import models
from django.db.models import Prefetch
from django.urls import reverse
from django.utils.html import format_html
from django.utils.functional import cached_property
from simple_history.models import HistoricalRecords
from django.contrib.postgres.fields import ArrayField

# atlas import
from common.models import Entry
import common.models
import common.others
from atlas import files, settings


# TODO REMOVE and use common.link instead
def make_link(url, placeholder, title="", target="_blank", text=None):
    if text is None:
        text = placeholder
    return format_html(
        '<a href="{}" title="{}" target="{}">{}</a>'.format(
            url.format(placeholder), title, target, text)
    )


class ReactionQuerySet(models.QuerySet):
    """Some predefined querysets for the Reaction model"""

    def ecs(self, ecs):
        """Filters Reaction with these ecs"""
        return self.filter(ec_code__in=ecs)

    def annotate_mechanisms(self, only_three_stars=True, **kwargs):
        """annotates reactions with the mechanisms"""
        queryset = common.models.Mechanism.objects.filter(prediction=False).order_by("mechanism_id").annotate_steps()
        if only_three_stars:
            queryset = queryset.good_rating()
        return self.prefetch_related(Prefetch("mechanism_set", queryset=queryset, to_attr="mechanisms"))


class Reaction(models.Model):
    """The chemical reaction catalysed by this enzyme's entry"""
    entry = models.ForeignKey("Entry", on_delete=models.CASCADE)
    reaction_id = models.IntegerField(verbose_name="Reaction ID")
    scheme = models.FileField(verbose_name="Reaction Scheme", blank=True, upload_to="reaction_schemes/")
    rxn = models.FileField(verbose_name="Reaction rxn", blank=True, null=True, upload_to="reaction_rxns/")
    compounds = models.ManyToManyField('Compound', through='CompoundType')
    is_reference = models.BooleanField(verbose_name="Is Reference?")
    notes = models.TextField(verbose_name='Notes', blank=True,)
    is_polymeric = models.BooleanField(default=False, verbose_name="Is a polymeric reaction")
    kegg = models.CharField(max_length=10, blank=True, verbose_name='KEGG Reaction')
    rhea = models.CharField(max_length=10, blank=True, verbose_name="Rhea ID")
    ec_code = models.ForeignKey('EC', verbose_name="EC Number", null=True, blank=True, on_delete=models.PROTECT)
    history = HistoricalRecords()

    objects = ReactionQuerySet.as_manager()

    molconvert_namespace = "{http://www.chemaxon.com}"

    def __str__(self):
        return "Reaction {0}".format(self.entry)

    @property
    def reactants(self):  # TODO IS THIS NECESSARY? should be custom manager
        return self._get_compounds_str(type="reactant")

    @property
    def products(self):  # TODO IS THIS NECESSARY?
        return self._get_compounds_str(type="product")

    def _get_compounds_str(self, type=None):
        compounds = self.compoundtype_set.filter(type=type)
        return " + ".join(r.compound.name for r in compounds)

    @property
    def edit_link(self):
        url = reverse('common:edit_reaction', args=(
            self.entry.owner.username,
            self.entry.macie_id,
            self.reaction_id)
                      )
        text = "R{0}".format(self.reaction_id)
        return make_link(url, "", title="Edit reaction", target="", text=text)

    def find_repeated_ecs(self):
        macie_entries = Entry.objects \
            .filter(owner__username="macie", complete=True) \
            .exclude(id=self.entry.id)
        macie_reactions = Reaction.objects.filter(entry__in=macie_entries)
        return macie_reactions.filter(entry__in=macie_entries, ec_code=self.ec_code).exclude(ec_code=None)

    def get_mols_str_from_rxn(self) -> Tuple[List[str], List[str]]:
        """Returns two lists of strs with the info for reactants and products"""
        if not self.rxn or self.rxn.name == "running":
            return ([], [])
        self.rxn.seek(0)
        rxn_str = self.rxn.read().decode("utf-8")
        try:
            no_reactants = int(rxn_str.splitlines()[4].split()[0])
        except IndexError:
            return ([], [])
        mol_strs = rxn_str.split("$MOL\n")[1:]
        return (mol_strs[:no_reactants], mol_strs[no_reactants:])

    def get_mols_from_rxn(self) -> Tuple[List[Chem.Mol], List[Chem.Mol]]:
        """
        Return two lists of rdkit mols. One for reactants other for products
        """
        return (self.label_to_rxn_mol[0].values(), self.label_to_rxn_mol[1].values())

    @cached_property
    def label_to_rxn_mol(self) -> Tuple[Dict[str, Chem.Mol], Dict[str, Chem.Mol]]:
        """
        Return two dicts mapping a label with the chebi id to the rxn mol. First dict for reactants second for products
        """
        output_dicts = ({}, {})
        r_atoms = ["G*", "* ", "R#", "R "]
        chebi_id_counts = collections.defaultdict(int)
        for no, group in enumerate(self.get_mols_str_from_rxn()):
            for mol_str in group:
                chebi_id = mol_str.split()[0].split(":")[1]
                chebi_id_counts[chebi_id] += 1
                label = "chebi:{}_{}".format(chebi_id, chebi_id_counts[chebi_id])
                for r_atom in r_atoms:
                    mol_str = mol_str.replace(r_atom, "C ")
                mol = Chem.MolFromMolBlock(mol_str, removeHs=False)
                if mol is not None:
                    Chem.SanitizeMol(mol, sanitizeOps=Chem.SanitizeFlags.SANITIZE_ALL
                                     ^ Chem.SanitizeFlags.SANITIZE_SETAROMATICITY)
                    # todo wip  this is necessary when there are r_atoms
                    # for some reason they get assigned implicit Hs which breaks the code
                    add_hs_on = [atom.GetIdx() for atom in mol.GetAtoms() if atom.GetNumImplicitHs()]
                    if add_hs_on:
                        mol = Chem.rdmolops.AddHs(mol, onlyOnAtoms=[0])
                    output_dicts[no][label] = mol
        return output_dicts

    def get_chiral_reaction_centres(self) -> Set[int]:
        """
        Return the atom map numbers of the rxn atoms that only change chirality in the reaction
        """
        AtomInfo = Tuple[int, int]
        groups_atoms: Tuple[Set[AtomInfo], Set[AtomInfo]] = (set(), set())
        for no, group in enumerate(self.get_mols_from_rxn()):
            for mol in group:
                for chiral_info in Chem.FindMolChiralCenters(mol, includeUnassigned=True):
                    if chiral_info[1] == "?":
                        continue
                    chiral_atom_map = mol.GetAtomWithIdx(chiral_info[0]).GetAtomMapNum()
                    groups_atoms[no].add((chiral_atom_map, chiral_info[1]))
        return {rc[0] for rc in groups_atoms[0] ^ groups_atoms[1]}

    def get_reaction_centres_atoms(self) -> Set[int]:
        """Return the atom map numbers of the rdkit mols extracted from the rxn that participate in the reaction"""
        BondInfo = Tuple[FrozenSet[int], float]
        groups_bonds: Tuple[Set[BondInfo], Set[BondInfo]] = (set(), set())
        for no, group in enumerate(self.get_mols_from_rxn()):
            for mol in group:
                for bond in mol.GetBonds():
                    groups_bonds[no].add((
                        frozenset((bond.GetBeginAtom().GetAtomMapNum(), bond.GetEndAtom().GetAtomMapNum())),
                        bond.GetBondTypeAsDouble()
                    ))
        different_bonds = groups_bonds[0] ^ groups_bonds[1]
        reaction_centres = set()
        for bond in different_bonds:
            reaction_centres.update(bond[0])
        reaction_centres.update(self.get_chiral_reaction_centres())

        return reaction_centres

    def rxn_not_empty(self):
        if self.rxn and not self.rxn.name == "running":
            self.rxn.seek(0)
            rxn_str = self.rxn.read().decode("utf-8")
            return bool(rxn_str)

    def get_bonds_from_rxn_str(self, rxn_str):
        header_no_lines = 3
        rxn_lines = rxn_str.splitlines()
        try:
            start_line = int(rxn_lines[header_no_lines][:3]) + 1 + header_no_lines
        except IndexError:
            return []
        end_line = start_line + int(rxn_lines[header_no_lines][3:6])
        return rxn_lines[start_line:end_line]

    def get_mapped_bonds(self, bond_lines, atoms_dict):
        """ Maps the bonds to the mapped atoms in atoms_dict """
        mapped_bonds = []
        for bond_line in bond_lines:
            # print(bond_line)
            atom1 = int(bond_line[:3])
            atom2 = int(bond_line[3:6])
            order = int(bond_line[6:9])

            mapped_bonds.append((
                atoms_dict[atom1][1],
                atoms_dict[atom2][1],
                order,
                atoms_dict[atom1][0],
                atoms_dict[atom2][0],
            ))
        return mapped_bonds

    def get_atoms_from_rxn_str(self, rxn_str):
        """Return a dict atoms {[index]: (C, mapped index), ...} """
        header_no_lines = 3
        rxn_lines = rxn_str.splitlines()
        try:
            no_atoms = int(rxn_lines[header_no_lines][:3])
        except IndexError:
            # rxn probably empty
            return {}
        atoms_end_no = header_no_lines + 1 + no_atoms
        atoms = {}
        for no, line in enumerate(rxn_lines[header_no_lines+1:atoms_end_no]):
            atoms[no+1] = (line[31], int(line[60:63]))
        return atoms

    def get_bond_changes_from_rxn(self):
        reactants_products_bonds = [[], []]
        for no, reactant_product_strs in enumerate(self.get_mols_str_from_rxn()):
            for mol_str in reactant_product_strs:
                atoms_dict = self.get_atoms_from_rxn_str(mol_str)
                bond_lines = self.get_bonds_from_rxn_str(mol_str)
                reactants_products_bonds[no].extend(
                    self.get_mapped_bonds(bond_lines, atoms_dict))
        bonds_dicts = [{(frozenset(bond[0:2]), bond[2]): bond for bond in bond_list}
                       for bond_list in reactants_products_bonds]
        changes = bonds_dicts[0].keys() ^ bonds_dicts[1].keys()

        return [{frozenset(bond[0:2]): bond for key, bond in bond_dict.items()
                 if key in changes} for bond_dict in bonds_dicts]

    @property
    def is_running(self):
        return bool(self.rxn.name == "running")

    @property
    def is_not_balanced(self):
        if not self.rxn or self.rxn.name == "running":
            return []
        rxn_reactants_products = self.get_mols_str_from_rxn()
        mapped_reactants_atoms = set()
        mapped_products_atoms = set()
        for mol_str in rxn_reactants_products[0]:
            mapped_reactants_atoms.update(
                [atom[1] for atom in
                 self.get_atoms_from_rxn_str(mol_str).values()])
        for mol_str in rxn_reactants_products[1]:
            mapped_products_atoms.update(
                [atom[1] for atom in
                 self.get_atoms_from_rxn_str(mol_str).values()])
        return mapped_reactants_atoms.symmetric_difference(
            mapped_products_atoms)

    def set_bond_changes_from_rxn(self):
        """Create bond changes objects from the bond changes in the rxn"""
        changed_bonds = self.get_bond_changes_from_rxn()
        bond_changes = []
        if not self.is_not_balanced:
            for key, broken_bond in changed_bonds[0].items():
                final_order = changed_bonds[1].pop(key, [0, 0, 0])[2]
                bond_change, _ = BondChange.objects.get_or_create(
                    reaction=self,
                    idx1=broken_bond[0],
                    idx2=broken_bond[1],
                    initial_order=broken_bond[2],
                    element1=broken_bond[3],
                    element2=broken_bond[4],
                    final_order=final_order,
                )
                bond_changes.append(bond_change.pk)
            for key, formed_bond in changed_bonds[1].items():
                bond_change, _ = BondChange.objects.get_or_create(
                    reaction=self,
                    initial_order=0,
                    idx1=formed_bond[0],
                    idx2=formed_bond[1],
                    final_order=formed_bond[2],
                    element1=formed_bond[3],
                    element2=formed_bond[4],
                )
                bond_changes.append(bond_change.pk)
        BondChange.objects.filter(reaction=self).exclude(
            pk__in=bond_changes).delete()
        return bond_changes

    def set_rdt_rxn(self):
        self.rxn.name = "running"
        self.save()
        rxn_str = self.get_rxn_str(add_hydrogens=True)
        # I am not sure if this is needed. Adding hydrogens was causing problems
        # sometimes the file created could not be read by RDT
        # rxn_str = self.get_rxn_str()
        tmp_f = tempfile.NamedTemporaryFile('w', delete=False, suffix=".rxn")
        tmp_f.write(rxn_str)
        tmp_f.close()
        os.chdir('/tmp/')
        command = "{0}/java/jdk-11/bin/java -jar {0}/RDT2.5/RDT2.5.0.jar -Q RXN -q '{1}' -j AAM -f TEXT -g".format(
            files.PROGRAMS_ROOT, tmp_f.name)
        print(command)
        rdt_output = subprocess.getoutput(command)
        print(rdt_output)
        os.chdir(settings.BASE_DIR)
        output_filename = "/tmp/ECBLAST_{}_AAM.rxn".format(
            os.path.basename(tmp_f.name).replace(".rxn", ""))
        rxn_filename = "{0.owner.username}_{0.macie_id}_r_{1}.rxn".format(
            self.entry, self.reaction_id)
        self.rxn.save(rxn_filename, File(open(output_filename)))
        return rdt_output

    def calculate_similarity(self, reaction2):
        """Return similarity lines from ECBLAST output"""
        output_name = "ECBLAST_{0}_{1}_COMPARE.txt".format(
            os.path.basename(self.rxn.name).replace(".rxn", ""),
            os.path.basename(reaction2.rxn.name).replace(".rxn", ""),
        )
        if not os.path.isfile("/tmp/{}".format(output_name)):
            os.chdir('/tmp/')
            command = (
                "java -jar {0}/RDT1.4/bin/RDT1.4.2.jar "
                "-Q RXN -q '{1}/{2.rxn.name}' -T RXN -t '{1}/{3.rxn.name}' -j "
                "COMPARE -f TEXT -g".format(
                    files.PROGRAMS_ROOT, settings.MEDIA_ROOT, self, reaction2)
            )
            subprocess.getoutput(command)
        grep_command = "grep Similarity /tmp/{}".format(output_name)
        grep_output = subprocess.getoutput(grep_command)
        return grep_output

    def _create_reaction_similarity(self, reaction2):
        grep_lines = self.calculate_similarity(reaction2)
        try:
            print(grep_lines)
            values = [line.split(":")[1] for line in grep_lines.splitlines()]
            reaction_similarity = self.add_reaction_similarity(
                reaction2,
                bond_change_similarity=values[0],
                reaction_center_similarity=values[1],
                structure_center_similarity=values[2]
            )
            return reaction_similarity
        except IndexError:
            return

    def set_rdt_rxn_and_bond_changes(self):
        self.set_rdt_rxn()
        self.set_bond_changes_from_rxn()

    def draft_scheme_from_compounds(self):
        """Get mrv string of the reaction scheme from list of compounds"""
        rxn_str = self.get_rxn_str()
        mrv_str = subprocess.getoutput("echo '{0}' | molconvert mrv -2 ".format(
            rxn_str))
        return mrv_str + "\n"

    def mrv_str_from_rxn(self):
        if not self.rxn or self.rxn.name == "running":
            return ''
        self.rxn.seek(0)
        rxn_str = self.rxn.read().decode("utf-8")
        mrv_str = subprocess.getoutput("echo '{0}' | molconvert mrv -2 ".format(rxn_str))
        return mrv_str

    def get_compounds(self, ctype=""):
        """Returns a list with the compounds of this reaction"""
        return [compoundtype.compound for compoundtype in
                self.compoundtype_set.filter(type__contains=ctype) for _ in
                range(compoundtype.count)]

    def get_rxn_str(self, add_hydrogens=False):
        """Returns a str with the contents of the RXN file for this reaction"""
        reactants = self.get_compounds(ctype="reactant")
        products = self.get_compounds(ctype="product")
        reaction_file_str = "$RXN\n\n\n\n{0:3d}{1:3d}\n".format(len(reactants), len(products))
        r_atoms = ["G*", "* ", "R#", "R1", "R "]
        for compound in reactants+products:
            reaction_file_str += "$MOL\nChEBI:{0} {1}".format(compound.chebi_id, compound.name)
            if not compound.mol:
                missing_filename = "{0}{1}".format(settings.RES_MOL_PATH, "missing.mol")
                missing_file_str = open(missing_filename, 'r').read()
                missing_file_str = missing_file_str.replace(
                    "missing:", "{}".format(compound.chebi_id))
                reaction_file_str += missing_file_str
            else:
                mol_str = compound.mol
                for r_atom in r_atoms:
                    mol_str = mol_str.replace(r_atom, "C ")
                mol_str = "\n".join([line for line in mol_str.splitlines() if\
                                     ("RGP" not in line and "STY" not in line and "SAL" not in line and "SMT" not in line)])
                if add_hydrogens:
                    mol_str = Chem.MolToMolBlock(Chem.AddHs(Chem.MolFromMolBlock(mol_str)))
                reaction_file_str += mol_str
            if not reaction_file_str.endswith("\n"):
                reaction_file_str += "\n"
        reaction_file_str = reaction_file_str.replace("'", "")
        # if add_hydrogens:
            # tmp_f = tempfile.NamedTemporaryFile(
                # 'w', delete=False, suffix=".rxn")
            # tmp_f.write(reaction_file_str)
            # tmp_f.close()
            # reaction_file_str = subprocess.getoutput("molconvert rxn:+H < {} ".format(tmp_f.name))
        return reaction_file_str

    def save(self, *args, **kwargs):
        if not self.reaction_id:
            try:
                last = Reaction.objects.filter(entry=self.entry).order_by(
                    'reaction_id').last()
                self.reaction_id = last.reaction_id + 1
            except AttributeError:
                self.reaction_id = 1
        super(Reaction, self).save(*args, **kwargs)

    def add_reaction_similarity(self, reaction2, bond_change_similarity,
                                reaction_center_similarity,
                                structure_center_similarity, symm=True):
        if self == reaction2:
            return
        reaction_similarity, created = ReactionSimilarity.objects.update_or_create(
            reaction1=self,
            reaction2=reaction2,
            defaults={
                'bond_change_similarity': bond_change_similarity,
                'reaction_center_similarity': reaction_center_similarity,
                'structure_center_similarity': structure_center_similarity,
            })
        if symm:
            reaction2.add_reaction_similarity(
                self, bond_change_similarity, reaction_center_similarity,
                structure_center_similarity, False)
        return reaction_similarity

    def remove_reaction_similarity(self, reaction2, symm=True):
        ReactionSimilarity.objects.filter(reaction1=self, reaction2=reaction2) \
            .delete()
        if symm:
            reaction2.remove_reaction_similarity(self, False)


class ReactionSimilarity(models.Model):
    """To save results of reactions similarity"""
    reaction1 = models.ForeignKey('Reaction', on_delete=models.CASCADE, related_name='similarity1')
    reaction2 = models.ForeignKey('Reaction', on_delete=models.CASCADE, related_name='similarity2')
    bond_change_similarity = models.FloatField(null=True)
    reaction_center_similarity = models.FloatField(null=True)
    structure_center_similarity = models.FloatField(null=True)

    class Meta:
        unique_together = ('reaction1', 'reaction2')

    def __str__(self):
        return ("{0.reaction1} <-> {0.reaction2} [{0.bond_change_similarity}, "
                "{0.reaction_center_similarity}, "
                "{0.structure_center_similarity}]".format(self))


class BondChange(models.Model):
    """Overall reaction bond changes"""
    reaction = models.ForeignKey(Reaction, on_delete=models.CASCADE)
    element1 = models.CharField(max_length=2)
    element2 = models.CharField(max_length=2)
    idx1 = models.CharField(verbose_name="Index 1", max_length=10)
    idx2 = models.CharField(verbose_name="Index 2", max_length=10)
    initial_order = models.CharField(max_length=3)
    final_order = models.CharField(max_length=3)
    history = HistoricalRecords()

    objects = models.Manager()

    def __str__(self):
        return "{0.atom_1}-{0.atom_2} {0.initial_order} {0.final_order}".format(
            self)

    @property
    def atom_1(self):
        return "{0.element1}.{0.idx1}".format(self)

    @property
    def atom_2(self):
        return "{0.element2}.{0.idx2}".format(self)


class ReactionFingerprint(models.Model):
    """Reaction fingerprints"""
    reaction = models.OneToOneField("Reaction", on_delete=models.CASCADE, related_name="fingerprint", primary_key=True)
    bond_changes_fingerprint = ArrayField(models.IntegerField())

    @classmethod
    def create_or_update_objects(cls):
        """Create and updates fingerprints based on the bondchange table"""
        reactions = common.models.Reaction.objects.prefetch_related("bondchange_set")
        fingerprint_key = common.models.Step.get_bond_changes_fingerprint_key()
        for reaction in reactions:
            fingerprint = [0]*len(fingerprint_key)
            for bondchange in reaction.bondchange_set.all():
                bond_set = frozenset([element if element in common.models.Step.EXPLICIT_ATOMS else "other"
                                      for element in [bondchange.element1, bondchange.element2]])
                initial_order = int(bondchange.initial_order)
                final_order = int(bondchange.final_order)
                if initial_order == 0 and final_order != 0:
                    # bond formed
                    key = (bond_set, True)
                elif initial_order != 0 and final_order == 0:
                    # bond cleaved
                    key = (bond_set, False)
                else:
                    continue
                fingerprint[fingerprint_key.index(key)] += 1
            obj, created = cls.objects.update_or_create(
                    reaction=reaction,
                    defaults={
                        "bond_changes_fingerprint": fingerprint,
                        },
                    )
