# coding=utf-8
"""Includes the models related with literature references"""

# python standard imports
import re
import json
from typing import Dict

# other python imports
from Bio import Entrez
import habanero

# django imports
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.html import format_html
from simple_history.models import HistoricalRecords

# atlas imports
from atlas import files
from common import validators


# TODO REMOVE use link
def make_link(url, placeholder, title="", target="_blank", text=None):
    if text is None:
        text = placeholder
    return format_html(
        '<a href="{}" title="{}" target="{}">{}</a>'.format(
            url.format(placeholder), title, target, text)
    )


class ReferenceType(models.Model):  # TODO change name to annotation
    """Links the reference object ot its type, the mechanism, and residues"""
    mechanism = models.ForeignKey("Mechanism", on_delete=models.CASCADE, related_name="annotations")
    reference = models.ForeignKey("Reference", on_delete=models.PROTECT, verbose_name="Reference", null=True)
    residues = models.ManyToManyField("Residue", verbose_name="Residues", blank=True, related_name="references")
    active_site_cofactors = models.ManyToManyField("ActiveSiteCofactor", blank=True, verbose_name="Cofactors")
    evidence_type = models.ForeignKey("EvidenceType", on_delete=models.PROTECT, verbose_name="Type of Evidence")
    notes = models.TextField(blank=True, verbose_name="Notes")
    history = HistoricalRecords()

    def __str__(self):
        return "{} {}".format(self.mechanism, self.reference)


class EvidenceType(models.Model):
    """The EvidenceType of the references for the mechanism"""
    evidence_type = models.CharField(max_length=255)
    eco_code = models.CharField(max_length=255)
    description = models.TextField()
    history = HistoricalRecords()

    def __str__(self):
        return self.evidence_type

    class Meta:
        ordering = ['evidence_type']


class Reference(models.Model):
    """A literature reference"""
    doi = models.CharField(max_length=255, verbose_name="DOI", validators=[validators.validate_doi_or_pmid_exists],
                           null=True, unique=True)
    pubmed_id = models.CharField(max_length=255, null=True, unique=True)
    title = models.TextField(max_length=255)
    authors = ArrayField(models.CharField(max_length=255), default=list)
    journal = models.CharField(max_length=255)
    volume = models.CharField(max_length=255)
    year = models.IntegerField(null=True)
    start_page = models.CharField(max_length=255)
    end_page = models.CharField(max_length=255)
    history = HistoricalRecords()

    doi_url = "http://dx.doi.org/{}"
    pmc_url = "http://europepmc.org/abstract/MED/{0}"
    doi_data = ""
    pm_summary: Dict[str, str] = {}

    def strip_doi_pmid_spaces(self):
        if self.pubmed_id and self.pubmed_id != self.pubmed_id.strip():
            self.pubmed_id = self.pubmed_id.strip()
        if self.doi and self.doi != self.doi.strip():
            self.doi = self.doi.strip()

    def set_values_from_doi_or_pmid(self, overwrite=False):
        self.strip_doi_pmid_spaces()
        if not self.pubmed_id and self.doi:
            self.set_pmid_from_doi(overwrite=overwrite)
        if self.pubmed_id:
            return self.set_values_from_pmid(overwrite=overwrite)
        elif self.doi:
            return self.set_values_from_doi(overwrite=overwrite)

    def set_pmid_from_doi(self, overwrite=False):
        if not self.pubmed_id or overwrite:
            try:
                Entrez.email = "ribeiro@ebi.ac.uk"
                handle = Entrez.esearch(db='pubmed', term="{}[DOI]".format(self.doi))
                pubmed_id = Entrez.read(handle)['IdList'][0]
                if pubmed_id and pubmed_id != self.pubmed_id:
                    self.pubmed_id = pubmed_id.strip()
                    self.save()
            except IndexError:
                print("{} not found in pubmed".format(self.doi))
        return self.pubmed_id

    def get_pm_summary(self):
        if self.pubmed_id and not self.pm_summary:
            try:
                Entrez.email = "ribeiro@ebi.ac.uk"
                handle = Entrez.esummary(db='pubmed', id=self.pubmed_id, retmode='xml')
                self.pm_summary = Entrez.read(handle)[0]
            except RuntimeError:
                self.pm_summary = {}
        return self.pm_summary

    def set_doi_from_pubmed(self, overwrite=False):
        if not self.doi or overwrite:
            self.get_pm_summary()
            if self.pm_summary and 'DOI' in self.pm_summary and self.pm_summary['DOI'] != self.doi:
                self.doi = self.pm_summary['DOI'].strip()
                self.save()
        return self.doi

    def set_title_from_pubmed(self, overwrite=False):
        if not self.title or overwrite:
            self.get_pm_summary()
            if self.pm_summary:
                title = self.pm_summary['Title']
                if title != self.title:
                    self.title = title
                    self.save()
        return self.title

    def set_journal_from_pubmed(self, overwrite=False):
        if not self.journal or overwrite:
            self.get_pm_summary()
            if self.pm_summary and 'Source' in self.pm_summary and \
                    self.pm_summary['Source'].strip() != self.journal:
                self.journal = self.pm_summary['Source'].strip()
                self.save()
        return self.journal

    def set_authors_from_pubmed(self, overwrite=False):
        if not self.authors or overwrite:
            self.get_pm_summary()
            if self.pm_summary and 'AuthorList' in self.pm_summary and \
                    self.pm_summary['AuthorList'] != self.authors:
                authors = self.pm_summary['AuthorList']
                self.authors = authors
                print(self.authors)
                self.save()
        return self.authors

    def set_volume_from_pubmed(self, overwrite=False):
        if not self.volume or overwrite:
            self.get_pm_summary()
            if self.pm_summary and 'Volume' in self.pm_summary and \
                    self.volume != self.pm_summary['Volume']:
                self.volume = self.pm_summary['Volume']
                self.save()
        return self.volume

    def set_year_from_pubmed(self, overwrite=False):
        if not self.year or overwrite:
            self.get_pm_summary()
            if self.pm_summary and 'PubDate' in self.pm_summary:
                year = int(self.pm_summary['PubDate'].split()[0])
                if year and year != self.year:
                    print(self.year, "|", year)
                    self.year = year
                    self.save()
        return self.year

    def set_pages_from_pubmed(self, overwrite=False):
        if (not self.start_page) or overwrite:
            self.get_pm_summary()
            if self.pm_summary and 'Pages' in self.pm_summary:
                pages_spl = self.pm_summary['Pages'].split("-")
                start_page = pages_spl[0]
                if len(pages_spl) > 1:
                    end_page = start_page[:-(len(pages_spl[1]))] + pages_spl[1]
                else:
                    end_page = ""
                if start_page and start_page != self.start_page:
                    self.start_page = start_page
                    self.save()
                if end_page and end_page != self.end_page:
                    self.end_page = end_page
                    self.save()
        return self.start_page, self.end_page

    def set_values_from_pmid(self, overwrite=False):
        self.set_doi_from_pubmed(overwrite=overwrite)
        self.set_title_from_pubmed(overwrite=overwrite)
        self.set_journal_from_pubmed(overwrite=overwrite)
        self.set_authors_from_pubmed(overwrite=overwrite)
        self.set_volume_from_pubmed(overwrite=overwrite)
        self.set_pages_from_pubmed(overwrite=overwrite)
        self.set_year_from_pubmed(overwrite=overwrite)

    def get_doi_data(self):
        if self.doi and not self.doi_data:
            try:
                self.doi_data = json.loads(habanero.cn.content_negotiation(
                    ids=self.doi, format="citeproc-json"))
            except (json.decoder.JSONDecodeError, TypeError):
                print("JSON habanero Error: {}".format(self.doi))
        return self.doi_data

    def set_title_from_doi(self, overwrite=False):
        if not self.title or overwrite:
            self.get_doi_data()
            if self.doi_data:
                title = self.doi_data['title']
                if title != self.title:
                    self.title = title
                    self.save()
        return self.title

    def set_journal_from_doi(self, overwrite=False):
        abbreviations = read_journal_abbreviations()
        journal_abb = ""
        if not self.journal or overwrite:
            data = self.get_doi_data()
            try:
                journal_issns = data['ISSN']
            except KeyError:
                print("no issn in json: {}".format(self.doi))
                return self.journal
            for journal_issn in journal_issns:
                if journal_issn in abbreviations:
                    journal_abb = abbreviations[journal_issn]
                    break
            if journal_abb and journal_abb != self.journal:
                self.journal = journal_abb
                self.save()
            elif (not journal_abb) and journal_issns:
                print("issn not found in abbreviations: {}".format(
                    journal_issns))
        return self.journal

    def set_authors_from_doi(self, overwrite=False):
        if not self.authors or overwrite:
            self.get_doi_data()
            authors = []
            if self.doi_data.get('author'):
                for author in self.doi_data.get('author'):
                    initials = "".join(
                        [given[0] for given in author['given'].split()])
                    authors.append("{} {}".format(author['family'], initials))
                if authors != self.authors:
                    self.authors = authors
                    self.save()
        return self.authors

    def set_volume_from_doi(self, overwrite=False):
        if not self.volume or overwrite:
            self.get_doi_data()
            if self.doi_data and 'volume' in self.doi_data and \
                    self.doi_data['volume'] != self.volume:
                self.volume = self.doi_data['volume']
                self.save()
        return self.volume

    def set_year_from_doi(self, overwrite=False):
        if not self.year or overwrite:
            self.get_doi_data()
            if self.doi_data:
                try:
                    year = self.doi_data['published-print']['date-parts'][0][0]
                except KeyError:
                    try:
                        year = self.doi_data['published-online'][
                            'date-parts'][0][0]
                    except KeyError:
                        year = None
                if year and year != self.year:
                    self.year = year
                    self.save()
        return self.year

    def set_pages_from_doi(self, overwrite=False):
        if not self.start_page or overwrite:
            self.get_doi_data()
            if self.doi_data and 'page' in self.doi_data:
                pages = self.doi_data['page'].split('-')
                try:
                    start_page, end_page = pages
                except ValueError:
                    start_page = pages[0]
                    end_page = None
                if start_page and start_page != self.start_page:
                    self.start_page = start_page
                    self.save()
                if end_page and end_page != self.end_page:
                    self.end_page = end_page
                    self.save()
        return self.start_page, self.end_page

    def set_values_from_doi(self, overwrite=False):
        """Set all reference fields from DOI"""
        self.set_title_from_doi(overwrite=overwrite)
        self.set_journal_from_doi(overwrite=overwrite)
        self.set_authors_from_doi(overwrite=overwrite)
        self.set_volume_from_doi(overwrite=overwrite)
        self.set_year_from_doi(overwrite=overwrite)
        self.set_pages_from_doi(overwrite=overwrite)

    def __str__(self):
        if self.year and self.authors:
            string = "({}, {}) {}".format(
                self.authors[0], self.year, self.title)
        else:
            string = "{}, {}".format(self.doi, self.pubmed_id)
        return string

    @property
    def doi_or_pmid_link(self):
        return self.doi_html_link if self.doi_html_link else self.pmid_html_link

    @property
    def doi_html_link(self):
        if not self.doi:
            return ""
        return make_link(Reference.doi_url, self.doi)

    @property
    def pmid_html_link(self):
        if not self.pubmed_id:
            return ""
        return make_link(Reference.pmc_url, self.pubmed_id,
                         title="Open in Europe PMC")

    @property
    def formatted_str(self):
        output = ""
        if self.authors:
            output += self.authors[0]
            if len(self.authors) > 1:
                output += " <i>et al.</i>"
        if self.year:
            output += " ({0.year}),".format(self)

        if self.journal:
            output += " <em>{0.journal}</em>,".format(self)
        if self.volume:
            output += " <strong>{0.volume}</strong>,".format(self)
        if self.start_page:
            output += " {0.start_page}-{0.end_page}. ".format(self)
        output += re.sub("[{].*?[}]", "", self.title)
        if output[-1] not in ".!?":
            output += "."
        if self.doi:
            output += " DOI:{0}.".format(self.doi_html_link)
        if self.pubmed_id:
            output += " PMID:{0}.".format(self.pmid_html_link)
        return format_html(output)
        # output = (
            # " {1} ({0.year}), "
            # "<em>{0.journal}</em>, <strong>{0.volume}</strong>, "
            # "{0.start_page}-{0.end_page}. "
            # "{2}"
            # "".format(self, authors, re.sub("[{].*?[}]", "", self.title)))
        # if output[-1] not in ".!?":
            # output += "."
        # if self.doi:
            # output += " DOI:{0}.".format(self.doi_html_link)
        # if self.pubmed_id:
            # output += " PMID:{0}.".format(self.pmid_html_link)
        # return format_html(output)


# TODO should be inside Reference as a class method
def read_journal_abbreviations():
    """Reads a journal abbreviations file and returns a abbreviations dict"""
    abbreviations_dict = {}
    with open(files.JOURNAL_ABBREVIATIONS_FILE, 'r') as abbreviations_file:
        for line in abbreviations_file:
            if line.startswith("-----"):
                abbreviation = ""
            elif line.startswith("ISSN (Print):"):
                issn_print = line.split(":")[1].strip()
                abbreviations_dict[issn_print] = abbreviation
            elif line.startswith("ISSN (Online):"):
                issn_online = line.split(":")[1].strip()
                abbreviations_dict[issn_online] = abbreviation
            elif line.startswith("MedAbbr:"):
                abbreviation = line.rstrip("\n").split(":")[1].strip()
    return abbreviations_dict
