# coding=utf-8
"""Includes models related with catalytic residues"""

# python imports
import bisect
import collections
import copy
import urllib

# django imports
from django.db.models import F, Prefetch
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models, transaction
from simple_history.models import HistoricalRecords

# atlas imports
from atlas import settings
import common.models
from common.models import mix
import atlas.settings


class ResidueChainQuerySet(models.QuerySet):
    """Some predefined querysets for the ResidueChain model"""

    def reference(self, is_reference=True):
        """Filters reference residue chains"""
        return self.filter(is_reference=is_reference)


class ResidueChain(models.Model):
    """Matching of the catalytic residue in this chain"""
    residue = models.ForeignKey('Residue', on_delete=models.CASCADE, related_name="residue_chains")
    chain = models.ForeignKey('Chain', related_name="residue_chains", on_delete=models.PROTECT)
    resid = models.IntegerField(null=True)
    auth_resid = models.IntegerField(null=True)
    code = models.CharField(max_length=3, verbose_name="Three-letter code")
    is_reference = models.BooleanField(default=False, db_index=True, null=True)
    domain = models.ForeignKey('Domain', related_name="residue_chains", null=True, on_delete=models.PROTECT)

    objects = ResidueChainQuerySet.as_manager()

    class Meta:
        unique_together = ('residue', 'chain')

    def __str__(self):
        return ("auth_resid:{0.auth_resid} (resid:{0.resid}) {0.code} "
                "{0.chain.name:>2} {0.chain.structure.pdb_id}".format(self))


    # WIP
    @classmethod
    def create_non_reference_objects(cls):
        """Creates non_reference ResidueChains using the alignments"""
        #remove these filters
        # entry = common.models.Entry.objects.mcsa().get(macie_id=2)
        # residues = common.models.Residue.objects.filter(active_site__entry=entry)
        # print(residues)




        # create new chains with resid
        reference_residue_chains = cls.objects.filter(is_reference=True).select_related("chain__aa_sequence", "chain__structure")
        # .filter(residue__in=residues) # remove filter

        for entry in common.models.Entry.objects.mcsa().order_by("macie_id").filter(macie_id__gte=935):
            residues = common.models.Residue.objects.filter(active_site__entry=entry)
            print(entry)
            # this could be outside but here for saving ram
            existing_residue_chains = {}
            for res_chain in cls.objects.filter(residue__in=residues):
                existing_residue_chains[(res_chain.residue_id, res_chain.chain_id)] = res_chain

            sequence_to_chains = collections.defaultdict(set)
            for chain in common.models.Chain.objects.\
                    filter(aa_sequence__chains__residue_chains__residue__in=residues).\
                    select_related("structure"):
                sequence_to_chains[chain.aa_sequence_id].add((chain, chain.structure))

            msa_ref_seq_to_alignment = collections.defaultdict(dict)
            for msa in common.models.MultiAASeqAlignment.objects.\
                    filter(reference_sequence__chains__residue_chains__residue__in=residues).\
                    prefetch_related("reference_sequence__chains__structure", "sequence__chains__structure"):  # remove filter
                msa_ref_seq_to_alignment[msa.reference_sequence.id][msa.sequence.id] = msa

            mapping = {}
            for ref_residue_chain in reference_residue_chains.filter(residue__in=residues):
                residue_chains_to_create = []
                print("reference", ref_residue_chain)
                ref_sequence = ref_residue_chain.chain.aa_sequence
                try:
                    ref_alignment = msa_ref_seq_to_alignment[ref_sequence.id][ref_sequence.id]
                except KeyError:
                    print("ref alignment not found", ref_residue_chain.residue.active_site.entry)
                    continue

                ref_align_resid0 = ref_residue_chain.get_ref_resid0_in_multi_alignment(ref_alignment)
                for sequence_id, msa in msa_ref_seq_to_alignment[ref_sequence.id].items():
                    if sequence_id != ref_sequence.id:
                        resid, code = msa.get_non_reference_resid_code(ref_align_resid0)
                        for chain, structure in sequence_to_chains[sequence_id]:
                            if (ref_residue_chain.residue_id, chain.id) not in existing_residue_chains:
                                if structure.pdb_id not in mapping:
                                    mapping[structure.pdb_id] = structure.pdb_chain_resid_to_auth_dict
                                if mapping[structure.pdb_id] is None:
                                    print("structure obsolete or needs update", structure.pdb_id)
                                    continue

                                residue_chains_to_create.append(cls(
                                    residue_id=ref_residue_chain.residue_id,
                                    chain_id=chain.id,
                                    is_reference=False,
                                    resid=resid,
                                    code=code,
                                    auth_resid=mapping[structure.pdb_id].get((chain.name, resid), None)
                                ))
                            else:
                                old_res_chain = existing_residue_chains[(ref_residue_chain.residue_id, chain.id)]
                                if old_res_chain.is_reference:
                                    print("This should not happen")
                                    exit
                                if old_res_chain.resid != resid or old_res_chain.code != code:
                                    print(old_res_chain, resid, code)
                                    if structure.pdb_id not in mapping:
                                        mapping[structure.pdb_id] = structure.pdb_chain_resid_to_auth_dict
                                    if mapping[structure.pdb_id] is None:
                                        print("structure obsolete or needs update", structure.pdb_id)
                                        continue
                                    old_res_chain.resid = resid
                                    old_res_chain.auth_resid = mapping[structure.pdb_id].get((chain.name, resid), None)
                                    old_res_chain.code = code
                                    old_res_chain.save()
                                    if old_res_chain.auth_resid is None:
                                        print("could not map auth_resid", old_res_chain)

                print("Creating {} new residue_chain objects".format(len(residue_chains_to_create)))
                cls.objects.bulk_create(residue_chains_to_create)


    def get_ref_resid0_in_multi_alignment(self, ref_alignment):
        """return the position of the reference resid in the MultiAASeqAlignment"""
        counter = 1
        result = None
        for i, letter in enumerate(ref_alignment.alignment):
            if letter == '-':
                continue
            elif counter == self.resid:
                result = i + ref_alignment.pad_left
                break
            else:
                counter += 1
        # if result is None or\
            # ref_alignment.alignment[result-ref_alignment.pad_left] != common.others.AA_3TO1[self.residue.reference_residue_sequence.code]:
            # print(result, self)
        return result

    # @classmethod
    # def create_objects(cls):
        # """Creates all the residue sequences based on the MultiAlignment objects"""
        # for no, sequence in enumerate(common.models.Sequence.objects.reference()):
            # ref_residue_sequences = cls.objects.reference().filter(sequence=sequence)
            # non_ref_residue_sequences = cls.objects.reference(False).same_residue_as(ref_residue_sequences)
            # non_ref_dict = {(reseq.sequence_id, reseq.residue_id): (reseq.resid, reseq.code)
                            # for reseq in non_ref_residue_sequences}
            # msa = common.models.MultiAlignment.objects.with_reference(sequence.uniprot_id).exclude(alignment="")
            # if not msa:
                # continue
            # ref_msa = common.models.MultiAlignment.objects.filter(sequence=sequence, reference_sequence=sequence).get()
            # print(no, ref_msa, "number of alignments: ", msa.count())
            # ref_resids0 = []
            # for ref_residue_sequence in ref_residue_sequences:
                # ref_resids0.append((ref_residue_sequence.get_ref_resid0_in_multi_alignment(
                    # ref_malignment=ref_msa, ref_resid=ref_residue_sequence.resid)))
            # to_create = []
            # for alignment in msa.reference(False).select_related("sequence"):
                # resid_codes = alignment.get_non_reference_resid_codes(ref_resids0)
                # for no, ref_residue_sequence in enumerate(ref_residue_sequences):
                    # if (alignment.sequence_id, ref_residue_sequence.residue_id) not in non_ref_dict:
                        # to_create.append(cls(
                                # uniprot_id=alignment.sequence.uniprot_id,
                                # sequence_id=alignment.sequence_id,
                                # reference_sequence=sequence,#?? what is this?
                                # residue_id=ref_residue_sequence.residue_id,
                                # is_reference=False,
                                # resid=resid_codes[no][0],
                                # code=resid_codes[no][1],
                            # ))
                    # elif non_ref_dict[(alignment.sequence_id, ref_residue_sequence.residue_id)] != resid_codes[no]:
                        # print("updating", resid_codes[no][0], resid_codes[no][1], non_ref_dict[(alignment.sequence_id, ref_residue_sequence.residue_id)])

                        # cls.objects.filter(
                            # is_reference=False,  # just to make sure i am not changing references
                            # sequence_id=alignment.sequence_id,
                            # residue_id=ref_residue_sequence.residue_id).update(
                                # resid=resid_codes[no][0],
                                # code=resid_codes[no][1],
                        # )

            # print("Creating {} residue sequence objects".format(len(to_create)))
            # cls.objects.bulk_create(to_create, batch_size=5000)
        # return


            
    # old - using old alignment system
    # @classmethod
    # def create_non_reference_objects(cls):
        # """Creates objects using the alignments"""
        # reference_residue_chains = cls.objects.filter(is_reference=True)\
            # .select_related("chain__structure", "chain__aa_sequence")

        # existing = set(cls.objects.values_list("residue_id", "chain_id"))

        # sequence_to_chains = collections.defaultdict(list)
        # for aa_sequence_id, chain_id in common.models.Chain.objects.values_list("aa_sequence_id", "id"):
            # sequence_to_chains[aa_sequence_id].append(chain_id)

        # mapping = {}

        # for reference_residue_chain in reference_residue_chains:
            # other_chains_ids = [
                # chain_id for chain_id in sequence_to_chains[reference_residue_chain.chain.aa_sequence_id]
                # if chain_id != reference_residue_chain.chain_id]
            # for identical_chain_id in other_chains_ids:
                # if (reference_residue_chain.residue_id, identical_chain_id) not in existing:
                    # print()
                    # print(reference_residue_chain)
                    # # the second time is not copying the reference but the previous identical
                    # identical_residue_chain = copy.copy(reference_residue_chain)
                    # identical_residue_chain.pk = None
                    # identical_residue_chain.chain_id = identical_chain_id
                    # identical_residue_chain.is_reference = False
                    # identical_residue_chain.domain = None
                    # identical_residue_chain.auth_resid = None
                    # identical_residue_chain.save()
                    # identical_residue_chain.domain = identical_residue_chain.get_domain()
                    # if identical_residue_chain.chain.structure.pdb_id not in mapping:
                        # try:
                            # mapping[identical_residue_chain.chain.structure.pdb_id] =\
                                    # identical_residue_chain.chain.structure.pdb_chain_resid_to_auth_dict
                        # except urllib.error.URLError:
                            # # obsolete entry
                            # mapping[identical_residue_chain.chain.structure.pdb_id] = {}
                    # identical_residue_chain.auth_resid = mapping[identical_residue_chain.chain.structure.pdb_id].get(
                            # (identical_residue_chain.chain.name, identical_residue_chain.resid), None)
                    # identical_residue_chain.save()
                    # print(identical_residue_chain)
            # # find chains that have the exact same sequence


    # prefetch_related("chain__aa_sequence") to avoid extra query
    @property
    def get_code(self):
        """Returns one letter code based on aa_sequence and pdb resid"""
        if self.resid is None:
            return None
        return self.chain.aa_sequence.sequence[self.resid-1]

    @property
    def simple_str(self):
        """An alias based on the residue_chain data"""
        return "{}{}{}".format(
            self.code if not self.residue.ptm else self.residue.ptm,
            self.auth_resid if self.auth_resid else self.resid,
            self.residue.assembly_chain_name
        )

    def get_code3(self):
        """Returns three letter code based on aa_sequence and pdb resid"""
        return common.others.AA_1TO3.get(self.get_code, self.get_code)

    def get_reference_resid(self, auth=False):
        """Retrieve resid or auth_resid from reference_residue_sequence and sifts"""
        key = (self.residue.reference_sequence.uniprot_id,
               self.residue.reference_residue_sequence.resid)
        value = self.chain.structure.sequence_resid_chains_dict.get(key, None)
        return value[2 if auth else 1] if value is not None else None

    def get_auth_resid(self, use_assembly_chains=True):
        """Retrieve auth_resid from resid and sifts"""
        if use_assembly_chains:
            # should this ever be used?
            auth_resid = self.chain.structure.chain_resid_to_auth_dict[(self.chain.name, self.resid)]
        else:
            print(self.chain.structure.pdb_chain_resid_to_auth_dict)
            print(self.chain)
            auth_resid = self.chain.structure.pdb_chain_resid_to_auth_dict[(self.chain.name, self.resid)]
        return auth_resid

    def get_domain(self):
        """Finds the domain this residueChain belongs to"""
        chain_domains = common.models.Domain.objects.filter(
            name__startswith=self.chain.name, structure=self.chain.structure
        ).exclude(boundaries='')
        # print(chain_domains, [c.boundaries for c in chain_domains], self.resid)
        for chain_domain in chain_domains:
            for boundary in chain_domain.boundaries.split(","):
                start, stop = [int(x) for x in boundary.split("-")]
                if self.resid in range(start, stop+1):
                    return chain_domain
        return None

    def save(self, *args, overwrite=False, **kwargs):
        if self.is_reference is True:
            self.residue.residue_chains.filter(is_reference=True) \
                .exclude(id=self.id).update(is_reference=False)
            if self.resid is None or overwrite:
                self.resid = self.get_reference_resid()
            if self.auth_resid is None or overwrite:
                self.auth_resid = self.get_reference_resid(auth=True)
            if self.code == "" or overwrite:
                self.code = self.get_code3()
            if self.domain is None or overwrite:
                self.domain = self.get_domain()
        super(ResidueChain, self).save(*args, **kwargs)


class ResidueQuerySet(models.QuerySet):
    """Some predefined querysets for the Residue model"""

    def entries(self, entries):
        """Filters residues that belong to these entries"""
        return self.filter(active_site__entry__in=entries)

    def codes(self, codes):
        """Filters residues that have this code in their reference ResidueSequence"""
        residue_sequences = ResidueSequence.objects.reference().filter(code__in=codes)
        return self.filter(residue_sequences__in=residue_sequences)

    def functions(self, functions):
        """Filters residues that have this function in their roles"""
        return self.filter(roles__function__in=functions)

    def locations(self, locations):
        """Filters residues that act through this location"""
        return self.filter(function_location__in=locations)

    def roles(self, roles):
        """Filters residues that have these roles"""
        return self.filter(roles__in=roles)
    
    def get_conservation_homo_filter(self, min_no_res = 10,
            only_experimental=False, by_residue=False, by_role=False,
            homologues_filter=[], filtered_models={}):
        """get conservation in swissprot sequences"""
        print(homologues_filter)
        queryset = ResidueSequence.objects.filter(residue__in=self, is_reference=False)
        if only_experimental:
            queryset = queryset.filter(sequence__function_experimental_evidence=only_experimental)
        print("start sequences")
        sequences = set(queryset.values_list("sequence", flat=True))
        sequences.update(queryset.values_list("reference_sequence", flat=True))
        sequences = common.models.Sequence.objects.filter(id__in=sequences).prefetch_related(
                Prefetch("ecs", to_attr="all_ecs"))
        seq_ecs = collections.defaultdict(set)
        for seq in sequences:
            for ec in seq.all_ecs:
                seq_ecs[seq.id].add((ec.ec1, ec.ec2, ec.ec3, ec.ec4))
                seq_ecs[seq.id].add((ec.ec1, ec.ec2, ec.ec3))
        print("end_sequences")
        residues = collections.defaultdict(lambda: collections.defaultdict(list))
        print(queryset.count())
        for resseq in queryset:
            sequence_ecs = seq_ecs[resseq.sequence_id]
            reference_sequence_ecs = seq_ecs[resseq.reference_sequence_id]
            if not reference_sequence_ecs:
                continue
            if [ec for ec in sequence_ecs & reference_sequence_ecs if len(ec)==4]:
                if "same_enzyme" in homologues_filter:
                    residues[resseq.residue_id]["same_enzyme"].append(resseq.code)
                    residues[resseq.residue_id]["all"].append(resseq.code)
            elif sequence_ecs & reference_sequence_ecs:
                if "other_substrate" in homologues_filter:
                    residues[resseq.residue_id]["other_substrate"].append(resseq.code)
                    residues[resseq.residue_id]["all"].append(resseq.code)
            elif not sequence_ecs:
                if "nonenzymes" in homologues_filter:
                    residues[resseq.residue_id]["nonenzymes"].append(resseq.code)
                    residues[resseq.residue_id]["all"].append(resseq.code)
            elif "other_reaction" in homologues_filter:
                residues[resseq.residue_id]["other_reaction"].append(resseq.code)
                residues[resseq.residue_id]["all"].append(resseq.code)
        print("finished")
        conservations = collections.defaultdict(lambda: collections.defaultdict(list))
        annotated = self.annotate_res_sequences(is_reference=True).annotate_roles(
                queryset=filtered_models['roles'] if 'roles' in filtered_models else None).distinct()

        for residue in annotated:
            if len(residues[residue.id]["all"]) < 1:
                continue
            for group, resseqs in residues[residue.id].items():
                conserved_count = sum(code == residue.res_seq[0].code for code in resseqs)
                conserved_fraction = conserved_count/len(resseqs)
                if by_residue:
                    if not by_role:
                        conservations[group][residue.res_seq[0].code].append(conserved_fraction)
                functions_added = set()
                if by_role:
                    for role in residue.ann_roles:
                        if role.function.id not in functions_added:
                            functions_added.add(role.function.id)
                            if by_residue:
                                conservations[group][(residue.res_seq[0].code, "all")].append(conserved_fraction)
                                conservations[group][("all", role.function.function)].append(conserved_fraction)
                                conservations[group][("all", "all")].append(conserved_fraction)
                                conservations[group][(residue.res_seq[0].code, role.function.function)].append(conserved_fraction)
                            else:
                                conservations[group][role.function.function].append(conserved_fraction)
        output = {}
        for group, group_conservations in conservations.items(): 
            output[group] = {k: v for k, v in group_conservations.items() if len(conservations["all"][k]) > min_no_res}

        return output

    def get_avg_conservation(self, **kwargs):
        if not "homologues_filter" in kwargs:
            kwargs['homologues_filter'] = ["same_enzyme", "other_substrate", "other_reaction", "nonenzymes"]
        conservation_groups = self.get_conservation_homo_filter(**kwargs)
        output = {}

        for group, conservation in conservation_groups.items():
            print(group, conservation.keys())
            output[group] = {k: sum(v)/len(v) for k, v in conservation.items()}
        # order by the value in same_function
        sorting_key = 'same_enzyme' if 'same_enzyme' in output else 'all'
        for group, group_dict in output.items():
             output[group] = collections.OrderedDict((sorted(group_dict.items(), key=lambda x: -output[sorting_key][x[0]])))
        #print(output)
        return output

    def annotate_roles(self, queryset=None, distinct=["function", "residue"]):
        """Annotates with the residue roles."""
        #return self.prefetch_related("roles__function")
        queryset = common.models.Role.objects if queryset is None else queryset
        return self.prefetch_related(
                Prefetch(
                    "roles", 
                    queryset=queryset.prefetch_related("function__group"),
                    to_attr="ann_roles"))

    def annotate_references(self, csa_annotation=False):
        """Annotates with the bibliographic references"""
        queryset = common.models.ReferenceType.objects.prefetch_related("evidence_type")
        if not csa_annotation:
            queryset = queryset.exclude(evidence_type__evidence_type="csa_annotation")
        return self.prefetch_related(
            Prefetch(
                "references",
                queryset=queryset,
                to_attr="refs"
                ))

    def annotate_res_chains(self, is_reference=True, to_attr_name=None, limit_to_references=False):
        """Annotates with the residue chains (residues in chain)"""
        if to_attr_name is None:
            to_attr_name = "res_chain" if is_reference else "non_ref_res_chains"

        queryset = ResidueChain.objects.filter(is_reference=is_reference).select_related(
                    "chain", "domain", "domain__cath", "chain__structure", "residue__pdb"
                    )
        if not is_reference and limit_to_references:
            # want to retrieve homologues that are other mcsa entries
            queryset = queryset.filter(chain__structure__pdb__is_reference=True)

        return self.prefetch_related(Prefetch("residue_chains", queryset=queryset, to_attr=to_attr_name))

    def annotate_res_sequences(self, is_reference=True, only_experimental=False, homologues_filter=[], uniprot_id=None):
        """Annotates with the residue sequences (residues in swissprot)"""
        if uniprot_id is not None:
            queryset = ResidueSequence.objects.filter(sequence__uniprot_id=uniprot_id)
            return self.prefetch_related(Prefetch("residue_sequences", queryset=queryset, to_attr="res_seq"))
        queryset = ResidueSequence.objects.filter(is_reference=is_reference).select_related("sequence")
        if only_experimental:
            queryset = queryset.filter(sequence__function_experimental_evidence=only_experimental)
        if is_reference:
            return self.prefetch_related(Prefetch("residue_sequences", queryset=queryset, to_attr="res_seq"))
        else:
            return self.prefetch_related(Prefetch("residue_sequences", queryset=queryset, to_attr="non_ref_res_seq"))


class Residue(models.Model):
    """This model represent a catalytic residue"""
    active_site = models.ForeignKey("ActiveSite", on_delete=models.CASCADE)
    pdb = models.ForeignKey('Pdb', null=True, on_delete=models.SET_NULL) # to delete?
    protein = models.ForeignKey('Protein', on_delete=models.CASCADE)
    x_sequence = models.ForeignKey('Sequence', null=True, on_delete=models.PROTECT)  # TO DELETE
    domain = models.ForeignKey('Domain', null=True, on_delete=models.PROTECT)  # TO DELETE
    code = models.CharField(max_length=3, verbose_name="Code")  # to delete/rename
    ptm = models.CharField(max_length=3, blank=True)
    resid = models.IntegerField(verbose_name="Residue Number")  # delete/rename
    assembly_chain_name = models.CharField(max_length=3, verbose_name="Assembly Chain")
    FUNCTION_LOCATION_CHOICES = (
        ("side_chain", "Side Chain"),
        ("post_translation_modification", "Post-Translation Modification"),
        ("main_chain_amine", "Main Chain Amine"),
        ("main_chain_carbonyl", "Main Chain Carbonyl"),
        ("main_chain_c_terminus", "Main Chain C-terminus"),
        ("main_chain_n_terminus", "Main Chain N-terminus"),
        ("main_chain", "Main Chain")
    )
    FUNCTION_LOCATION_ABBREVIATIONS = {
        "post_translation_modification": "ptm",
        "main_chain_amine": "main-N",
        "main_chain_carbonyl": "main-C",
        "main_chain_c_terminus": "C-term",
        "main_chain_n_terminus": "N-term",
        "main_chain": "main"
    }
    function_location = models.CharField(max_length=255, verbose_name="Function Location",
                                         choices=FUNCTION_LOCATION_CHOICES)
    roles = GenericRelation('Role', content_type_field="residue_ct", object_id_field="residue_id",
                            related_query_name="residue_roles")
    chains = models.ManyToManyField('Chain', through='ResidueChain', related_name='residues')
    history = HistoricalRecords()

    objects = ResidueQuerySet.as_manager()

    class Meta:
        unique_together = ('resid', 'assembly_chain_name', 'pdb', 'active_site', 'function_location')
        #ordering = ["assembly_chain_name", "resid"]

    def __str__(self):
        res_str = r"{0} {1}".format(
            self.simple_str, " ({})".format(self.function_location_abb) if self.function_location_abb else "")
        return res_str

    def save(self, *args, **kwargs):
        # clear residue labels cache if residue info changes
        if "all_residue_labels" in self.__class__.__dict__:
            del self.__class__.all_residue_labels
        super(Residue, self).save(*args, **kwargs)

    @property
    def function_location_abb(self):
        try:
            return self.FUNCTION_LOCATION_ABBREVIATIONS[self.function_location]
        except KeyError:
            return ""

    @property
    def reference_residue_sequence(self):
        try:
            return self.res_seq[0]
        except AttributeError:
            # Annotate residues with annotate_res_sequences to make this query faster
            return self.residue_sequences.get(is_reference=True)

    @property
    def reference_sequence(self):
        return self.reference_residue_sequence.sequence

    @property
    def reference_residue_chain(self):
        try:
            return self.res_chain[0]
        except (AttributeError):
            try:
                # Annotate residues with annotate_res_chains to make this query faster
                return self.residue_chains.get(is_reference=True)
            except ResidueChain.DoesNotExist:
                return None
        except (IndexError): # is annotated but there is not reference chain 
            return None

    @property
    def reference_chain(self):
        return self.reference_residue_chain.chain if self.reference_residue_chain is not None else None

    @property
    def reference_domain(self):
        if self.reference_residue_chain is not None:
            return self.reference_residue_chain.domain
        else:
            return None

    @property
    def reference_structure(self):
        if self.reference_residue_chain:
            return self.reference_residue_chain.chain.structure

    @property
    def str_fast(self):
        """return a string label that uses only info from the residue table"""
        return "{0.code}{0.resid}{0.assembly_chain_name}{1}".format(
            self,
            " ({})".format(self.function_location_abb) if self.function_location_abb else ""
        )

    @property
    def simple_str(self):
        """To be used in the schemes alias, by default: Code(or ptm)-auth_resid-assemblychain"""
        res_seq, res_chain = self.reference_residue_sequence, self.reference_residue_chain
        return "{}{}{}{}".format(
            res_seq.code if not self.ptm else self.ptm,
            res_chain.auth_resid if res_chain and res_chain.auth_resid else res_seq.resid,
            self.assembly_chain_name,
            " ({})".format(self.function_location_abb) if self.function_location_abb else ""
        )

    @property
    def formatted_seq_resid(self):
        """Returns the residue name and Uniprot Sequence in a nicely formatted way"""
        return "{0.code}{0.resid}{1}".format(
            self.reference_residue_sequence,
            " ({})".format(self.function_location_abb) if self.function_location_abb else ""
            )
    
    @property
    def formatted_pdb_resid(self):
        """Returns the residue name and Uniprot Sequence in a nicely formatted way. Faster if queryset is annotated
        Will use first res_seq regardless of it being the reference or not"""
        if self.reference_residue_chain is None:
            return ""
        else:
            res_chain = self.reference_residue_chain
            return "{0}{1.auth_resid}{2}{1.chain.name}{3}{4}".format(
                # self.reference_residue_chain.code if not self.ptm else self.ptm,
                res_chain.code if not self.ptm else self.ptm,
                res_chain,#self.reference_residue_chain,
                "({})".format(res_chain.resid) if res_chain.resid != res_chain.auth_resid else "",
                "({})".format(self.assembly_chain_name) if self.assembly_chain_name != res_chain.chain.name else "",
                " ({})".format(self.function_location_abb) if self.function_location_abb else "",
                )

    @property
    def formatted_str(self):
        return self.formatted_pdb_resid if self.formatted_pdb_resid else self.formatted_seq_resid
        
    @property
    def mcs_label(self):
        """returns appropriate key for MOL_TEMPLATES dict"""
        if self.function_location in ["side_chain", "post_translation_modification"]:
            return self.reference_residue_sequence.code
        else:
            return self.function_location

    @classmethod
    def get_all_residue_labels(cls):
        """return a dictionary with all the possible labels pointing to a residue. caches the results"""
        try:
            return cls.all_residue_labels
        except AttributeError:
            cls.all_residue_labels = collections.defaultdict(lambda: collections.defaultdict(set))
            for residue in cls.objects.all().annotate_res_chains().annotate_res_sequences():
                cls.all_residue_labels[residue.active_site_id][residue.simple_str].add(residue)
            return cls.all_residue_labels

    def get_mdl_str(self, alias=None):
        """Return the mdl mol contents for this residue function location"""
        if self.function_location in ["side_chain", "post_translation_modification"]:
            filename = "side_chain_{0}.mol".format(self.code.lower())
            to_replace = self.code.title()
        else:
            filename = "{0}.mol".format(self.function_location)
            to_replace = "RRR"
        alias = self.simple_str if alias is None else alias
        try:
            with open(settings.RES_MOL_PATH/ filename, 'r') as res_mol_file:
                res_mol_str = res_mol_file.read()
                res_mol_str = res_mol_str.replace(to_replace, alias)
            no_atoms = int(res_mol_str.splitlines()[3][:3])
            str_to_add = ""
            for i in range(no_atoms):
                if not "A    {0}\n".format(i+1) in res_mol_str:
                    str_to_add += "\nA{0:5d}\nres:{1}".format(i+1, alias)
            res_mol_str = res_mol_str.replace("\nM  END", str_to_add+"\nM  END")
            return res_mol_str + "$$$$\n"
        except FileNotFoundError:
            return ""

    def get_litemol_selection_str(self):
        if self.reference_residue_chain is None:
            return None
        selection_str = "{{asymId: '{0}', seqNumber: {1} }}".format(self.assembly_chain_name, 
                                                                    self.reference_residue_chain.resid)
        return selection_str

    # def find_ref_resid0_in_aaalignment(self, aa_sequence):
        # alignment = common.models.AAAlignment.objects.get(
            # sequence1=self.reference_chain.chain.aa_sequence,
            # sequence2=aa_sequence
        # )
        # # find position of ref_residue in alignment
        # ref_align_resid0 = self.reference_residue_chain.resid - alignment.align1from - 1
        # if ref_align_resid0 < 0:
            # return None
        # ref_align_points_and_dots = []
        # for i, letter in enumerate(alignment.align_seq1):
            # if letter in ".-":
                # ref_align_points_and_dots.append(
                    # i - len(ref_align_points_and_dots))
        # ref_align_additions = bisect.bisect_right(
            # ref_align_points_and_dots, ref_align_resid0)
        # ref_align_resid0 += ref_align_additions
        # if ref_align_resid0 >= len(alignment.align_seq1):
            # return None
        # return ref_align_resid0

    # def conserved_in_chain(self, chain, overwrite=False):
        # ref_residue_chain = self.residuechain_set.get(is_reference=True)
        # res2_residue_chain = self.set_residue_chain(chain, overwrite)
        # if (not ref_residue_chain) or (not res2_residue_chain) \
                # or ref_residue_chain.code != res2_residue_chain.code:
            # return False
        # else:
            # return True

    # def set_residue_chain(self, chain, overwrite=False):
        # """Finds were the residue is in a chain and creates ResidueChain object"""
        # res2_residue_chain = self.residuechain_set.filter(chain=chain)
        # if res2_residue_chain and not overwrite:
            # return res2_residue_chain.get()
        # ref_residue_chain = self.residuechain_set.get(is_reference=True)
        # alignment = common.models.AAAlignment.objects.get(
            # sequence1=ref_residue_chain.chain.aa_sequence,
            # sequence2=chain.aa_sequence
        # )
        # ref_align_resid0 = self.find_ref_resid0_in_aaalignment(
            # chain.aa_sequence)

        # if ref_align_resid0 is not None:
            # points_and_dots2 = []
            # for i, letter in enumerate(alignment.align_seq2):
                # if letter in ".-":
                    # points_and_dots2.append(i-len(points_and_dots2))
            # additions_2 = bisect.bisect_right(points_and_dots2, ref_align_resid0)
            # seq2_resid0 = alignment.align2from + ref_align_resid0 - additions_2
            # resid = seq2_resid0+1
        # else:
            # resid = None
        # return ResidueChain.objects.update_or_create(
            # residue=self,
            # chain=chain,
            # defaults={'resid': resid, 'is_reference': False})[0]


class ResidueSequenceQuerySet(models.QuerySet, mix.AtlasQuerySetMixIn):
    """Some predefined querysets for the ResidueSequence model"""

    def reference(self, is_reference=True):
        """Filters reference residue sequences that are not enzymes"""
        return self.filter(is_reference=is_reference)

    def codes(self, codes):
        """Filters reference residue sequences that have this code"""
        return self.filter(code__in=codes)

    def residues(self, residues):
        """Filters residue sequences that belong to these residues"""
        return self.filter(residue__in=residues)

    def sequences(self, sequences):
        """Filters residue sequences that belong to these sequences"""
        return self.filter(sequence__in=sequences)

    def same_residue_as(self, residue_sequences):
        """Filters all residue sequences that have the same residue as these residue sequences"""
        return self.filter(residue__residue_sequences__in=residue_sequences)

    def common_20(self):
        """Filters residue sequences that have a code of the 20 common amino acids"""
        return self.filter(code__in=common.others.AA_20_CODE3)

    def distinct_code_entry(self):
        """Filters residue sequences to include unique code and entries"""
        return self.distinct("code", "residue__active_site__entry")

    def same_residue_value_list(self, *args, one_count_per_entry=False, only_functional_annotation=True):
        """Applies value list function to residue sequences that belong to these sequences and have same function"""
        these_residues = set(self.values_list("residue"))
        related = ResidueSequence.objects.filter(sequence__in_swissprot=True, residue__in=these_residues)\
            .filter(sequence__ecs=F("residue__active_site__entry__reaction__ec_code"))\
            .distinct("id")\
            #.filter(sequence__alignment__identity__gte=0.10) # not perfect
        if only_functional_annotation:
            related = related.filter(sequence__function_experimental_evidence=True)
        if one_count_per_entry:
            related = related.values_list("residue__active_site__entry", *args).distinct(
                "residue__active_site__entry", *args)

        return related.values_list(*args)

    def propensity(self):
        cat_res_frequency = self.frequency('code')
        aa_frequency = common.models.Sequence.objects.filter(residue_sequences__in=self).aa_frequency()
        return {k: cat_res_frequency[k]/aa_frequency[k] for k in cat_res_frequency}

    def sequences2resseq_dict(self):
        """returns a dictionary where the key is a tuple (ref_sequence, sequence) and the values a list of residue sequences"""
        # using uniprot_id field, could be using sequence__uniprot_id with some additional cost
        ref_resseq = dict(ResidueSequence.objects.filter(is_reference=True).values_list("residue_id", "uniprot_id"))
        seq2resseq = collections.defaultdict(list)
        for resseq in self:
            seq2resseq[(ref_resseq[resseq.residue_id], resseq.uniprot_id)].append(resseq.resid)
        return seq2resseq


class ResidueSequence(models.Model):
    """Links the catalytic residues with reference sequences and homologues"""
    resid = models.IntegerField(null=True)
    code = models.CharField(max_length=3, verbose_name="Three-letter code")
    is_reference = models.BooleanField(default=False, db_index=True)
    residue = models.ForeignKey("Residue", on_delete=models.CASCADE, related_name="residue_sequences")
    sequence = models.ForeignKey("Sequence", on_delete=models.PROTECT, related_name="residue_sequences")
    reference_sequence = models.ForeignKey("Sequence", null=True, on_delete=models.PROTECT, related_name="homologous_residue_sequences")
    uniprot_id = models.CharField(max_length=20) # out of desperation

    objects = ResidueSequenceQuerySet.as_manager()

    class Meta:
        unique_together = ('residue', 'sequence')
        index_together = ('is_reference', 'residue', 'sequence')

    def __str__(self):
        return "{} {} {}".format(self.resid.__str__(), self.code, self.sequence.__str__())

    @property
    def code_resid(self):
        return "{0.code}{0.resid}".format(self)

    @property
    def code_resid_location(self):
        output = self.code_resid
        if self.residue.function_location in Residue.FUNCTION_LOCATION_ABBREVIATIONS:
            output += "({})".format(Residue.FUNCTION_LOCATION_ABBREVIATIONS[self.residue.function_location])
        return output

    def mutation_color(self, ref_res_seq=None):
        """Returns a color based on the mutation state with respect to the reference residue sequence"""
        ref_res_seq = ref_res_seq if ref_res_seq is not None else self.residue.reference_residue_sequence
        if self.code == ref_res_seq.code:
            color = 'green'
        elif self.residue.function_location not in {'side_chain', 'post_translation_modification'}:
            # function is in the main chain
            color = 'orange'
        else:
            color = 'red'
        return atlas.settings.COLORS_HEX[color]

    @classmethod
    def create_objects(cls):
        """Creates all the residue sequences based on the MultiAlignment objects"""
        for no, sequence in enumerate(common.models.Sequence.objects.reference()):
            ref_residue_sequences = cls.objects.reference().filter(sequence=sequence)
            non_ref_residue_sequences = cls.objects.reference(False).same_residue_as(ref_residue_sequences)
            non_ref_dict = {(reseq.sequence_id, reseq.residue_id): (reseq.resid, reseq.code)
                            for reseq in non_ref_residue_sequences}
            msa = common.models.MultiAlignment.objects.with_reference(sequence.uniprot_id).exclude(alignment="")
            if not msa:
                continue
            ref_msa = common.models.MultiAlignment.objects.filter(sequence=sequence, reference_sequence=sequence).get()
            print(no, ref_msa, "number of alignments: ", msa.count())
            ref_resids0 = []
            for ref_residue_sequence in ref_residue_sequences:
                ref_resids0.append((ref_residue_sequence.get_ref_resid0_in_multi_alignment(
                    ref_malignment=ref_msa, ref_resid=ref_residue_sequence.resid)))
            to_create = []
            for alignment in msa.reference(False).select_related("sequence"):
                resid_codes = alignment.get_non_reference_resid_codes(ref_resids0)
                for no, ref_residue_sequence in enumerate(ref_residue_sequences):
                    if (alignment.sequence_id, ref_residue_sequence.residue_id) not in non_ref_dict:
                        to_create.append(cls(
                                uniprot_id=alignment.sequence.uniprot_id,
                                sequence_id=alignment.sequence_id,
                                reference_sequence=sequence,#?? what is this?
                                residue_id=ref_residue_sequence.residue_id,
                                is_reference=False,
                                resid=resid_codes[no][0],
                                code=resid_codes[no][1],
                            ))
                    elif non_ref_dict[(alignment.sequence_id, ref_residue_sequence.residue_id)] != resid_codes[no]:
                        print("updating", resid_codes[no][0], resid_codes[no][1], non_ref_dict[(alignment.sequence_id, ref_residue_sequence.residue_id)])

                        cls.objects.filter(
                            is_reference=False,  # just to make sure i am not changing references
                            sequence_id=alignment.sequence_id,
                            residue_id=ref_residue_sequence.residue_id).update(
                                resid=resid_codes[no][0],
                                code=resid_codes[no][1],
                        )

            print("Creating {} residue sequence objects".format(len(to_create)))
            cls.objects.bulk_create(to_create, batch_size=5000)
        return

    def get_ref_resid0_in_multi_alignment(self, ref_malignment=None, ref_resid=None):
        """Uses MultiAlignment"""
        if ref_malignment is None:
            reference_sequence = self.residue.reference_sequence
            non_ref_malignment = common.models.MultiAlignment.objects.get(
                reference_sequence=reference_sequence, sequence=self.sequence)
            ref_malignment = common.models.MultiAlignment.objects.get(
                reference_sequence=reference_sequence, sequence=reference_sequence)
        if ref_resid is None:
            ref_resid = self.residue.reference_residue_sequence.resid
        counter = 1
        result = None
        for i, letter in enumerate(ref_malignment.alignment):
            if letter == '-':
                continue
            elif counter == ref_resid:
                result = i + ref_malignment.pad_left
                break
            else:
                counter += 1
        if result is None or\
            ref_malignment.alignment[result-ref_malignment.pad_left] != common.others.AA_3TO1[self.residue.reference_residue_sequence.code]:
            print(result, self, non_ref_malignment)
        return result

    def save(self, *args, **kwargs):
        if self.is_reference is True:
            self.residue.residue_sequences.filter(is_reference=True) \
                .exclude(id=self.id).update(is_reference=False)
            if self.code == '':
                self.code = common.others.AA_1TO3[
                    self.sequence.sequence[self.resid-1]]
        super(ResidueSequence, self).save(*args, **kwargs)
