# coding=utf-8
"""Includes models related with Functions, Roles and Components"""

# core python imports
import os

# other python imports
from rdkit import Chem
from rdkit.Chem import Draw
from rdkit.Chem import rdDepictor


# django imports
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from simple_history.models import HistoricalRecords

# atlas imports
from common.models import mix
import common.models
import atlas


class FunctionQuerySet(models.QuerySet):
    """Some predefined querysets for the Role model"""

    def function_types(self, function_types):
        """Filters function that have these types"""
        return self.filter(function_type__in=function_types)

    def names(self, functions):
        """Filters function that have these names"""
        return self.filter(function__in=functions)

    def groups(self, groups):
        """Filter functions that belong to these groups"""
        return self.filter(group__in=groups)


class Function(models.Model):
    """Functions of residues and cofactors in the catalytic step"""
    FUNCTION_TYPE_CHOICES = (
        ('interaction', 'Interaction'),
        ('reactant', 'Reactant'),
        ('spectator', 'Spectator'),
    )
    function_type = models.CharField(max_length=255, verbose_name="Function type", choices=FUNCTION_TYPE_CHOICES)
    function = models.CharField(max_length=255)
    group = models.ForeignKey('self', on_delete=models.SET_NULL, null=True, related_name="elements")
    description = models.TextField()
    emo = models.CharField(max_length=255, blank=True)
    history = HistoricalRecords()

    objects = FunctionQuerySet.as_manager()

    class Meta:
        ordering = ['-function_type', 'function']

    def __str__(self):
        return "{0.function_type} : {0.function}".format(self)


class StepComponent(models.Model):
    """Links the step object with its chemical Component objects"""
    step = models.ForeignKey('Step', on_delete=models.CASCADE)
    component = models.ForeignKey('Component', on_delete=models.PROTECT)
    limit = models.Q(model='residue') | models.Q(model='cofactor') | models.Q(model='CompoundType')
    molecule1 = GenericForeignKey('molecule1_ct', 'molecule1_id') # not used, to delete
    molecule2 = GenericForeignKey('molecule2_ct', 'molecule2_id')
    molecule1_ct = models.ForeignKey(ContentType, on_delete=models.SET_NULL, limit_choices_to=limit, related_name='%(class)s_molecule1_ct',
                                     null=True)
    molecule1_id = models.IntegerField(null=True)
    molecule2_ct = models.ForeignKey(ContentType, on_delete=models.SET_NULL, limit_choices_to=limit, related_name='%(class)s_molecule2_ct',
                                     null=True)
    molecule2_id = models.IntegerField(null=True)
    automated = models.BooleanField()
    element1 = models.CharField(verbose_name="Element 1", max_length=2)
    element2 = models.CharField(verbose_name="Element 1", max_length=2)
    idx1 = models.CharField(verbose_name="Index 1", max_length=10)
    idx2 = models.CharField(verbose_name="Index 2", max_length=10)
    history = HistoricalRecords()

    objects = models.Manager()

    @property
    def serial1(self):
        return int(self.idx1)+1 if self.idx2 else 0

    @property
    def serial2(self):
        return int(self.idx2)+1 if self.idx2 else 0

    @property
    def atom1(self):
        return "{0.molecule1}:{0.element1}.{0.serial1}".format(self)

    @property
    def atom2(self):
        return "{0.molecule2}:{0.element2}.{0.serial2}".format(self)

    def __str__(self):
        return "{0.component} {0.atom1} -> {0.atom2}".format(self)


class Component(models.Model):
    """A chemical component of the reaction step"""
    component = models.CharField(unique=True, max_length=255)
    description = models.TextField()
    chemical = models.BooleanField()
    ingold = models.BooleanField()
    aromatic = models.BooleanField(null=True)
    image_example = models.FileField(verbose_name='Image Example', blank=True, upload_to='component_images/')
    reaction_type_choices = (
        ("addition", "Addition"),
        ("substitution", "Substitution"),
        ("elimination", "Elimination"),
    )
    reaction_type = models.CharField(max_length=255, choices=reaction_type_choices, blank=True)
    bond_fission_type_choices = (
        ("homolytic", "Homolytic"),
        ("nucleophilic", "Nucleophilic"),
        ("electrophilic", "Electrophilic"),
    )
    bond_fission_type = models.CharField(max_length=255, choices=bond_fission_type_choices, blank=True)
    molecularity_choices = (
        ("intramolecular", "Intramolecular"),
        ("bimolecular", "Bimolecular"),
        ("unimolecular", "Unimolecular"),
    )
    molecularity = models.CharField(max_length=255, choices=molecularity_choices, blank=True)
    history = HistoricalRecords()

    objects = models.Manager()

    class Meta:
        ordering = ['-ingold', 'reaction_type', 'bond_fission_type', 'molecularity', 'component']

    def __str__(self):
        string = ""
        if self.ingold:
            string = "ingold: "
        string += self.component
        return string


class RoleQuerySet(models.QuerySet, mix.AtlasQuerySetMixIn):
    """Some predefined querysets for the Role model"""

    def functions(self, functions):
        """Filters roles that have this function in their roles"""
        return self.filter(function__in=functions)

    def residues(self, residues):
        """Filters roles that have this residue"""
        return self.filter(residue_roles__in=residues)

    def distinct_residue(self):
        """Returns only roles with distinct residue and function"""
        return self.distinct("residue_roles", "function")

    def steps(self, steps):
        """Filters roles that belong to this step"""
        return self.filter(step__in=steps)

    def target_atoms(self, target_atoms):
        """Filters roles that have this target atom"""
        return self.filter(target_atom__in=target_atoms)

    def target_molecules(self, target_molecules):
        """Filters roles that have this target molecule"""
        return self.filter(target_molecule__in=target_molecules)

    def automated(self):
        """Filters automated roles"""
        return self.filter(automated=True)

    def ignored(self):
        """Filters ignored roles"""
        return self.filter(ignore=True)

    def filter_out_ignored(self):
        """Filters ignored roles"""
        return self.exclude(ignore=True)

    # def ignore_rare(self, field="target_env", no=1):
    #     """Filters out roles that are not common based on field"""
    #     counter = collections.Counter(self.values_list(field, flat=True))
    #     return self.filter(**{field+"__in": set([k for k, v in counter.items() if v > no])})


class Role(models.Model):
    step = models.ForeignKey("Step", on_delete=models.CASCADE, related_name="roles")
    residue = GenericForeignKey("residue_ct", "residue_id")
    limit = models.Q(model="residue") | models.Q(model="ActiveSiteCofactor") | models.Q(model="CompoundType")
    residue_ct = models.ForeignKey(ContentType, on_delete=models.PROTECT, limit_choices_to=limit)
    residue_id = models.IntegerField()
    function = models.ForeignKey(Function, on_delete=models.PROTECT)
    automated = models.BooleanField()
    ignore = models.BooleanField(default=False)  # for automated roles that are wrong
    history = HistoricalRecords()
    target_atom = models.CharField(max_length=255, default="")
    source_atom = models.CharField(max_length=255, default="")
    target_molecule = models.CharField(max_length=255, default="")
    source_smiles = models.CharField(max_length=1023, default="")
    target_smiles = models.CharField(max_length=1023, default="")
    source_env = models.CharField(max_length=1023, default="")
    target_env = models.CharField(max_length=1023, default="")

    objects = RoleQuerySet.as_manager()

    class Meta:
        unique_together = ('step', 'residue_ct', 'residue_id', 'function')

    def __str__(self):
        return "{0.residue}:{0.function} {1}".format(self, "(auto)" if self.automated else "")


    @property
    def target_smiles_png(self):
        filename = atlas.settings.ENV_PNG_PATH+self.target_env+".png"
        if not os.path.isfile(filename):
            print(filename, " does not exist. Creating it")
            self.create_smiles_png(filename, field="target_env")
        return filename

    def create_smiles_png(self, filename, field="target_target_env"):
        """create a smiles png for the field name after the smiles string"""
        smiles = getattr(self, field)
        mol = Chem.MolFromSmiles(smiles, sanitize=False)
        Chem.SanitizeMol(mol, sanitizeOps=Chem.SanitizeFlags.SANITIZE_NONE)
        #target_atom = [atom.GetIdx() for atom in mol.GetAtoms() if atom.GetIsotope() == 200]
        #[atom.SetIsotope(0) for atom in mol.GetAtoms()]
        #mol = Draw.rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=False)
        rdDepictor.Compute2DCoords(mol)
        drawer = Draw.rdMolDraw2D.MolDraw2DCairo(100, 100)
        drawer.DrawMolecule(mol, highlightAtoms=[0])
        drawer.FinishDrawing()
        png = drawer.GetDrawingText()
        f = open(filename, 'wb')
        f.write(png)
        f.close()

