"""Models related with the smarts Rules class"""

# python imports
import collections
import logging
import itertools
from rdkit.Geometry.rdGeometry import Point2D
from rdkit import Chem
from rdkit import RDLogger
import re
from typing import Optional, Dict, List, Tuple
from lxml import etree  # type: ignore
import networkx as nx

# django imports
from django.db import models
from django.db.models import QuerySet
from django.utils.functional import cached_property

# atlas imports
from common.models import Entry, Arrow
import common.others
import atlas.files

RDLogger.DisableLog('rdApp.*')

# type defintions
ArrowList = List[Arrow]

logger = logging.getLogger("script")

H_TWO_BONDS_PATTERNS = [r"-\[#1&[^\]]*\]-", r"\[#1&[^\]]*\]\("]
REGEX_H_TWO_BONDS = [re.compile(pattern) for pattern in H_TWO_BONDS_PATTERNS]


class EntryRule(models.Model):
    """relationship could be found at runtime, but is saved in db for easier filtering and searching"""
    rule = models.ForeignKey("Rule", on_delete=models.CASCADE, related_name="entry_rules")
    entry = models.ForeignKey("Entry", on_delete=models.CASCADE, related_name="entry_rules")
    exclusive_in_entry = models.BooleanField(
        help_text="True if at least one arrows of the rule is unique to entry")

    @classmethod
    def create_and_update_objects(cls):
        """create relationships between rule and entries including the exclusive_in_entry field"""
        rules = Rule.objects.all().prefetch_related("arrowrule_set__arrow__step__mechanism__reaction__entry")
        existing_objs = {(obj.rule_id, obj.entry_id): obj for obj in cls.objects.all()}
        to_create = []
        to_update = []
        for rule in rules:
            # find wich rules are exclusive to an entry
            arrow_no_to_entries = collections.defaultdict(set)
            entries_exclusive = {}
            for arrowrule in rule.arrowrule_set.all():
                entry = arrowrule.arrow.step.mechanism.reaction.entry
                arrow_no_to_entries[arrowrule.arrow_no].add(entry)
                entries_exclusive[entry] = False
            for arrow_no, entries in arrow_no_to_entries.items():
                if len(entries) == 1:
                    entries_exclusive[next(iter(entries))] = True
            for entry, is_exclusive in entries_exclusive.items():
                if not (existing_obj := existing_objs.get((rule.id, entry.id), None)):
                    to_create.append(cls(rule=rule, entry=entry, exclusive_in_entry=is_exclusive))
                elif existing_obj.exclusive_in_entry != is_exclusive:
                    existing_obj.exclusive_in_entry = is_exclusive
                    to_update.append(existing_obj)

        print(f"Updating/Creating EntryRule relations for {len(rules)} rules")
        print(f"Updating {len(to_update)} Entry-Rule relations")
        print(f"Creating {len(to_create)} Entry-Rule relations")

        cls.objects.bulk_update(to_update, ['exclusive_in_entry', ])
        cls.objects.bulk_create(to_create)


class StepRule(models.Model):
    """relationship could be found at runtime, but is saved in db for easier filtering and searching"""
    rule = models.ForeignKey("Rule", on_delete=models.CASCADE, related_name="step_rules")
    step = models.ForeignKey("Step", on_delete=models.CASCADE, related_name="step_rules")
    complete_in_step = models.BooleanField(
        help_text="True if all the arrows of the rule are seen in the step")
    is_reversed = models.BooleanField(
        null=True,
        default=None,
        help_text="If all the arrows are seen in the same step is this forward or reverse direction?")

    @classmethod
    def create_and_update_objects(cls):
        """create relationships between rule and steps including the complete_in_step_field"""
        rules = Rule.objects.all().prefetch_related("arrowrule_set__arrow__step")
        existing_rule_steps = {(obj.rule_id, obj.step_id): obj for obj in cls.objects.all()}
        print("There are {} existing step rule object relations".format(len(existing_rule_steps)))
        to_create = []
        to_update = []
        for rule in rules:
            # find in which steps the rule is found completely or partially
            # step_to_arrow_nos = collections.defaultdict(set)
            step_to_direction_to_arrows = collections.defaultdict(lambda: collections.defaultdict(set))
            number_of_arrows = len(rule.rule_arrows)
            for arrowrule in rule.arrowrule_set.all():
                step_to_direction_to_arrows[arrowrule.arrow.step_id][arrowrule.reversed].add(arrowrule.arrow_no)
                # step_to_arrow_nos[arrowrule.arrow.step_id].add((arrowrule.arrow_no, arrowrule.reversed))
            complete_step_ids = set()
            partial_step_ids = set()
            for step_id, direction_to_arrows in step_to_direction_to_arrows.items():
                is_complete = False
                is_reversed_in_step = None
                for is_reversed, arrow_numbers in direction_to_arrows.items():
                    if len(arrow_numbers) == number_of_arrows:
                        # print(arrow_numbers, is_reversed, step_id)
                        is_complete = True
                        is_reversed_in_step = is_reversed
                if is_complete:
                    complete_step_ids.add((step_id, is_reversed_in_step))
                else:
                    partial_step_ids.add((step_id, is_reversed_in_step))

            # flag what to create or update
            for (step_id, is_reversed) in complete_step_ids | partial_step_ids:
                print(is_reversed)
                complete_in_step = (step_id, is_reversed) in complete_step_ids
                if not (existing_obj := existing_rule_steps.get((rule.id, step_id), None)):
                    to_create.append(
                        cls(rule=rule, step_id=step_id, complete_in_step=complete_in_step, is_reversed=is_reversed))
                else:
                    update = False
                    if existing_obj.complete_in_step != complete_in_step:
                        existing_obj.complete_in_step = complete_in_step
                        update = True
                    if existing_obj.is_reversed != is_reversed:
                        existing_obj.is_reversed = is_reversed
                        update = True
                    if update:
                        to_update.append(existing_obj)

        print(f"Updating/Creating StepRule relations for {len(rules)} rules")
        print(f"Updating {len(to_update)} Step-Rule relations")
        print(f"Creating {len(to_create)} Step-Rule relations")

        cls.objects.bulk_update(to_update, ['complete_in_step', 'is_reversed'])
        cls.objects.bulk_create(to_create)


class RuleQuerySet(models.QuerySet):
    """Some predefined querysets for the Rule model"""

    def annotate_entry_ids(self) -> 'QuerySet[Rule]':
        """adds a entry_ids list to each rule with the entries where this rule is seen"""
        return self.prefetch_related("steps__mechanism__reaction__entry")

    def complete_steps(self):
        return self.steps.filter(step_rules__complete_in_step=True)


class Rule(models.Model):
    """
    Smarts rules derived from the Arrow objects and can be applied to explore new mechanisms
    """
    reaction_smarts = models.CharField(max_length=10000, db_index=True)
    radical_in_step = models.BooleanField(default=False)
    rule_arrows = models.JSONField()
    steps = models.ManyToManyField("Step", related_name="rules", through="StepRule")
    version = models.CharField(max_length=512, db_index=True, blank=False, null=False)

    objects = RuleQuerySet.as_manager()

    class Meta:
        unique_together = ("reaction_smarts", "version")

    def __str__(self) -> str:
        return "{0.reaction_smarts}".format(self)

    def save(self, *args, **kwargs) -> None:
        super(Rule, self).save(*args, **kwargs)

    @property
    def reactant_smarts(self) -> str:
        """Returns the reactant smarts pattern"""
        return self.reaction_smarts.split(">>")[0]

    @property
    def product_smarts(self) -> str:
        """Returns the product smarts pattern"""
        return self.reaction_smarts.split(">>")[1]

    @property
    def num_reactants(self) -> int:
        """returns number of reactants"""
        return self.reactant_smarts.count(".") + 1

    @property
    def simple_reaction_smarts(self) -> str:
        return common.utils.remove_atom_mapping_from_smarts(self.reaction_smarts)

    @property
    def simple_reaction_smiles(self) -> str:
        return ">>".join([Chem.MolToSmiles(Chem.MolFromSmarts(smarts), canonical=False)
                          for smarts in self.simple_reaction_smarts.split(">>")])

    @property
    def canonical_reaction_smiles(self) -> str:
        try:
            return ">>".join([Chem.MolToSmiles(Chem.MolFromSmarts(smarts))
                              for smarts in self.simple_reaction_smarts.split(">>")])
        except Exception as exception:
            print("error generating canonical_reaction_smiles", exception)
            return self.simple_reaction_smarts

    @property
    def has_double_bond_hydrogen(self):
        return any([bool(pattern.search(self.reaction_smarts)) for pattern in REGEX_H_TWO_BONDS])

    @classmethod
    def create_objects(cls, limit_to_entries: "Optional[QuerySet[Entry]]" = None, version=version) -> "List[Rule]":
        """
        Create Rule objects based on the Arrow objects in the database. Return the created Rules
        """
        from common.models import ArrowRule
        # selecting entries and arrows
        entries = common.models.Entry.objects.mcsa().mechanism()  # type: ignore
        entries = entries.filter(id__in=limit_to_entries) if limit_to_entries is not None else entries
        arrows = common.models.Arrow.objects.filter(step__mechanism__reaction__entry__in=entries)\
            .select_related("step__mechanism__reaction__entry__owner")
            # .filter(step__step_id__in=(1,))\
            # .filter(step__mechanism__mechanism_id=2)\
        print("Arrow count", arrows.count())
        logger.info("Creating rule objects based on the arrows of {} entries".format(entries.count()))

        # to make sure rules are unique
        r_canonical_smiles_to_arrows: "Dict[str, List[ArrowList]]" = collections.defaultdict(list)
        r_smarts_to_rule: "Dict[str, Rule]" = {
            rule.reaction_smarts: rule for rule in cls.objects.filter(version=version)}
        r_canonical_smiles_to_r_smarts: "Dict[str, str]" = {rule.canonical_reaction_smiles: rule.reaction_smarts
                                                            for rule in cls.objects.filter(version=version)}
        # get arrows graph
        graph = arrows.get_arrows_graph()
        print("no nodes", len(graph.nodes()))

        # find paths that correspond to each rule
        paths = []
        for subgraph_nodes in nx.weakly_connected_components(graph):
            start_nodes = [node for node in subgraph_nodes if "start" in graph.nodes[node]]
            end_nodes = [node for node in subgraph_nodes if "end" in graph.nodes[node]]
            for start, end in itertools.product(start_nodes, end_nodes):
                if start != end:
                    # this does not allow for repeated nodes such
                    paths.extend(list(nx.all_simple_paths(graph, source=start, target=end, cutoff=10)))

        paths = [(False, path) for path in paths]
        print("number of linear paths", len(paths))
        # add circular paths
        paths.extend([(True, path) for path in nx.simple_cycles(graph) if len(path) <= 10])
        print("number of paths including circular", len(paths))

        # each path of curly arrows is a potential rule
        for is_circular, path in paths:
            # create all single-step rules but limit combined rules to X arrows
            step_sets = []
            for node in path:
                step_sets.append(set())
                for arrow in graph.nodes()[node].get("arrows"):
                    step_sets[-1].add(arrow.step.id)
            # print(is_circular, bool(set.intersection(*step_sets)), len(path))
            if not(bool(set.intersection(*step_sets))) and len(path) > 6:
                # print("skip")
                continue

            # print(path)
            if is_circular:
                # change path until last arrow is creating a new bond, so last mol already exists
                n = 0
                while n < len(path):
                    if len(graph.nodes()[path[-1]].get("arrow_env")) == 2:
                        break
                    else:
                        path = [path[-1],] + path[0:-1]
                    n += 1
            # print("new path", len(path))
            #: list of lists of mols, each list of mols correspond to an arrow
            #: but  the same path of arrows can lead to more than one rule
            reactant_mols_lists = []
            arrow_objs = []
            for arrow_no, curly_arrow in enumerate(path):

                is_last_circular_arrow = arrow_no + 1 == len(path) and is_circular
                arrow_objs.append([])

                # only associate this rule with this arrow if there is a pair of arrows from the same step
                # since this is the criterion to connect the curly arrows in the first place
                arrows_current = graph.nodes()[curly_arrow].get("arrows")
                # print("arrows_current", arrows_current)
                try:
                    step_arrows_next = {a.step.id for a in graph.nodes()[path[arrow_no+1]].get("arrows")}
                except IndexError:
                    step_arrows_next = set()
                try:
                    step_arrows_previous = {a.step.id for a in graph.nodes()[path[arrow_no-1]].get("arrows")}
                except IndexError:
                    step_arrows_previous = set()
                for arrow in arrows_current:
                    if arrow.step.id in step_arrows_next | step_arrows_previous:
                        arrow_objs[-1].append(arrow)

                # save mols existing from processing the prevoius curly arrows
                # previous_mols = reactant_mols[:]

                # the are all the mols touching the current curly arrow
                current_arrow_mols = [Chem.RWMol(m) for m in graph.nodes()[curly_arrow].get("arrow_env")]

                # save information regarding the curly arrow that touch each arrow
                # coded as follows
                # tens digit - 2 if full arrow (carries two electrons), 1 if half arrow (carries one electron)
                # ones digit - 1 arrow out (touches base), 2 arrow in (touches tip), 3 both
                for current_mol in current_arrow_mols:
                    for atom in current_mol.GetAtoms():
                        if (am := atom.GetAtomMapNum()):
                            arrow_size = str(am//10)
                            in_out_code = am % 10
                            if in_out_code % 2:
                                atom.SetProp(f"arrow_{arrow_no+1}_out", arrow_size)
                            if in_out_code//2:
                                atom.SetProp(f"arrow_{arrow_no+1}_in", arrow_size)
                # print("current mol", Chem.MolToSmiles(current_mol))

                # first curly arrow
                # print("arrow no", arrow_no+1)
                if not reactant_mols_lists:
                    # reactant_mols.extend(current_arrow_mols)
                    reactant_mols_lists.append(current_arrow_mols)
                else:
                    # merge mols as necessary, change reactant_mols in place
                    reactant_mols_lists = merge_mols(current_arrow_mols, reactant_mols_lists, is_last_circular_arrow)
                # print("rmols", [Chem.MolToSmiles(m) for ms in reactant_mols_lists for m in ms])

                for reactant_mols in reactant_mols_lists:
                    for mol in reactant_mols:
                        for atom in mol.GetAtoms():
                            # keep information about charge_plus (base of curly arrow only of first arrow
                            if atom.HasProp("charge_plus") and not atom.HasProp("arrow_1_out"):
                                atom.ClearProp("charge_plus")
                            if atom.HasProp("charge_minus") and not atom.HasProp(f"arrow_{arrow_no+1}_in"):
                                atom.ClearProp("charge_minus")

            for reactant_mols in reactant_mols_lists:
                if is_circular:
                    for mol in reactant_mols:
                        for atom in mol.GetAtoms():
                            if atom.HasProp("charge_plus"):
                                atom.SetProp(f"arrow_{arrow_no+1}_in", arrow_size)
                            # print(atom.GetSymbol(), atom.GetPropsAsDict())

                combined_reactants = Chem.RWMol(common.utils.combine_mols_list(reactant_mols))

                try:
                    combined_reactants.UpdatePropertyCache()
                except:
                    print("error in molecule", Chem.MolToSmiles(combined_reactants))
                    continue

                # remove extra atoms for rule version 0.3
                for atom in combined_reactants.GetAtoms():
                    if atom.HasProp("rc") and (atom.HasProp("charge_plus") or atom.HasProp("charge_minus")):
                        for n_atom in atom.GetNeighbors():
                            n_atom.SetProp("env_term", "1")
                            for n2_atom in n_atom.GetNeighbors():
                                n2_atom.SetProp("env_term", "1")
                to_delete = []
                for atom in combined_reactants.GetAtoms():
                    if not atom.HasProp("rc") and not atom.HasProp("env_term"):
                        to_delete.append(atom.GetIdx())
                to_delete.sort(reverse=True)
                for i in to_delete:
                    combined_reactants.RemoveAtom(i)



                # Add atom maps
                for atom in combined_reactants.GetAtoms():
                    # print(atom, atom.GetPropsAsDict())
                    atom.SetIsotope(0)
                    # if atom.HasProp("rc"):
                    atom.SetAtomMapNum(atom.GetIdx()+1)
                    atom.SetQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(atom.GetAtomicNum()))
                    # if not atom.HasProp("charge_plus") and not atom.HasProp("charge_minus"):
                        # print("not", atom.GetPropsAsDict())
                    # else:
                        # print("yes", atom.GetPropsAsDict())
                    # atoms in the second shell that are either C or H
                    if atom.HasProp("2nd_after_c") and not atom.HasProp("rc_or_1st"):
                        if atom.GetAtomicNum() in (1, 6) and atom.GetDegree() == 1:
                            atom.ExpandQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(1 if atom.GetAtomicNum() == 6 else 6),
                                            how=Chem.rdchem.CompositeQueryType.COMPOSITE_OR)
                    atom.ExpandQuery(Chem.rdqueries.FormalChargeEqualsQueryAtom(atom.GetFormalCharge()),
                                     how=Chem.rdchem.CompositeQueryType.COMPOSITE_AND)
                print("combined reactants", Chem.MolToSmiles(combined_reactants))
                try:
                    products, rule_arrows = common.utils.get_mol_product_from_props(combined_reactants)
                except KeyError:
                    print(arrow_objs)
                    continue

                print("combined reactants", Chem.MolToSmiles(products))

                try:
                    products.UpdatePropertyCache()
                except:
                    print("error in molecule. products: ", Chem.MolToSmiles(products))
                    continue
                # for atom in combined_reactants.GetAtoms():
                    # print(atom.GetSymbol(), atom.GetPropsAsDict())
                # print(Chem.MolToSmarts(products), rule_arrows)
                for atom in products.GetAtoms():
                    atom.SetQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(atom.GetAtomicNum()))
                    if atom.HasProp("2nd_after_c") and not atom.HasProp("rc_or_1st"):
                        if atom.GetAtomicNum() in (1, 6)  and atom.GetDegree() == 1:
                            atom.ExpandQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(1 if atom.GetAtomicNum() == 6 else 6),
                                            how=Chem.rdchem.CompositeQueryType.COMPOSITE_OR)
                    atom.ExpandQuery(Chem.rdqueries.FormalChargeEqualsQueryAtom(atom.GetFormalCharge()),
                                     how=Chem.rdchem.CompositeQueryType.COMPOSITE_AND)
                    # print(atom.DescribeQuery())
                reaction_smarts = f"{Chem.MolToSmarts(combined_reactants)}>>{Chem.MolToSmarts(products)}".replace("&", "")
                print(reaction_smarts)
                reverse_r_smarts = f"{Chem.MolToSmarts(products)}>>{Chem.MolToSmarts(combined_reactants)}".replace("&", "")
                reverse_r_arrows = []
                for rule_arrow in rule_arrows[::-1]:
                    reverse_r_arrows.append([rule_arrow[1], rule_arrow[0], rule_arrow[2]])
                # print(reaction_smarts, rule_arrows)
                # print("same")

                for reversed, (r_smarts, r_arrows) in enumerate(
                        ((reaction_smarts, rule_arrows), (reverse_r_smarts, reverse_r_arrows))):
                    # print(r_smarts)

                    rule = r_smarts_to_rule.get(r_smarts,
                                                cls(reaction_smarts=r_smarts, rule_arrows=r_arrows, version=version))

                    canonical_smiles = rule.canonical_reaction_smiles
                    # print(canonical_smiles)
                    # if the canonical smiles is repeated, do not save this rule
                    # but still add the arrow_chain info so the other rules can be associated with this Arrow
                    if canonical_smiles not in r_canonical_smiles_to_r_smarts:
                        # print("repeated canonical smiles", canonical_smiles)
                        # else:
                        r_smarts_to_rule[r_smarts] = rule
                    r_canonical_smiles_to_arrows[canonical_smiles].append([arrow_objs, bool(reversed)])
                    r_canonical_smiles_to_r_smarts[canonical_smiles] = r_smarts

        to_create = [rule for rule in r_smarts_to_rule.values() if not rule.pk]
        print("rules to create", len(to_create))
        rules_created = cls.objects.bulk_create(to_create)
        logger.info("Created {} new rules".format(len(r_smarts_to_rule)))

        arrow_rules = []
        existing_rule_arrow = set(cls.arrows.through.objects.values_list("rule_id", "arrow_id", "arrow_no", "reversed"))
        # not reversing each curly arrow individually
        for r_smarts, rule in r_smarts_to_rule.items():
            for arrow_sets, reversed in r_canonical_smiles_to_arrows[rule.canonical_reaction_smiles]:
                for arrow_no, curly_arrow in enumerate(arrow_sets):
                    arrow_no += 1
                    for arrow in curly_arrow:
                        if (rule.id, arrow.id, arrow_no+1, reversed) not in existing_rule_arrow:
                            arrow_rules.append(ArrowRule(
                                rule_id=rule.id, arrow_id=arrow.id, arrow_no=arrow_no, reversed=reversed))

        print("Creating arrow rules", len(ArrowRule.objects.bulk_create(arrow_rules, batch_size=1000)))
        return rules_created

    @classmethod
    def create_objects_0_2(cls, limit_to_entries: "Optional[QuerySet[Entry]]" = None, version=version) -> "List[Rule]":
        """
        Create Rule objects based on the Arrow objects in the database. Return the created Rules
        """
        from common.models import ArrowRule
        # selecting entries and arrows
        entries = common.models.Entry.objects.mcsa().mechanism()  # type: ignore
        entries = entries.filter(id__in=limit_to_entries) if limit_to_entries is not None else entries
        arrows = common.models.Arrow.objects.filter(step__mechanism__reaction__entry__in=entries)\
            .select_related("step__mechanism__reaction__entry__owner")
            # .filter(step__step_id__in=(1,))\
            # .filter(step__mechanism__mechanism_id=2)\
        print("Arrow count", arrows.count())
        logger.info("Creating rule objects based on the arrows of {} entries".format(entries.count()))

        # to make sure rules are unique
        r_canonical_smiles_to_arrows: "Dict[str, List[ArrowList]]" = collections.defaultdict(list)
        r_smarts_to_rule: "Dict[str, Rule]" = {
            rule.reaction_smarts: rule for rule in cls.objects.filter(version=version)}
        r_canonical_smiles_to_r_smarts: "Dict[str, str]" = {rule.canonical_reaction_smiles: rule.reaction_smarts
                                                            for rule in cls.objects.filter(version=version)}
        # get arrows graph
        graph = arrows.get_arrows_graph()
        print("no nodes", len(graph.nodes()))

        # find paths that correspond to each rule
        paths = []
        for subgraph_nodes in nx.weakly_connected_components(graph):
            start_nodes = [node for node in subgraph_nodes if "start" in graph.nodes[node]]
            end_nodes = [node for node in subgraph_nodes if "end" in graph.nodes[node]]
            for start, end in itertools.product(start_nodes, end_nodes):
                if start != end:
                    # this does not allow for repeated nodes such
                    paths.extend(list(nx.all_simple_paths(graph, source=start, target=end, cutoff=10)))

        paths = [(False, path) for path in paths]
        print("number of linear paths", len(paths))
        # add circular paths
        paths.extend([(True, path) for path in nx.simple_cycles(graph) if len(path) <= 10])
        print("number of paths including circular", len(paths))

        # each path of curly arrows is a potential rule
        for is_circular, path in paths:
            # print(".", end="")
            # create all single-step rules but limit combined rules to X arrows
            step_sets = []
            for node in path:
                step_sets.append(set())
                for arrow in graph.nodes()[node].get("arrows"):
                    step_sets[-1].add(arrow.step.id)
            # print(is_circular, bool(set.intersection(*step_sets)), len(path))
            if not(bool(set.intersection(*step_sets))) and len(path) > 6:
                # print("skip")
                continue

            # print(path)
            if is_circular:
                # change path until last arrow is creating a new bond, so last mol already exists
                n = 0
                while n < len(path):
                    if len(graph.nodes()[path[-1]].get("arrow_env")) == 2:
                        break
                    else:
                        path = [path[-1],] + path[0:-1]
                    n += 1
            # print("new path", len(path))
            #: list of lists of mols, each list of mols correspond to an arrow
            #: but  the same path of arrows can lead to more than one rule
            reactant_mols_lists = []
            arrow_objs = []
            for arrow_no, curly_arrow in enumerate(path):

                is_last_circular_arrow = arrow_no + 1 == len(path) and is_circular
                arrow_objs.append([])

                # only associate this rule with this arrow if there is a pair of arrows from the same step
                # since this is the criterion to connect the curly arrows in the first place
                arrows_current = graph.nodes()[curly_arrow].get("arrows")
                # print("arrows_current", arrows_current)
                try:
                    step_arrows_next = {a.step.id for a in graph.nodes()[path[arrow_no+1]].get("arrows")}
                except IndexError:
                    step_arrows_next = set()
                try:
                    step_arrows_previous = {a.step.id for a in graph.nodes()[path[arrow_no-1]].get("arrows")}
                except IndexError:
                    step_arrows_previous = set()
                for arrow in arrows_current:
                    if arrow.step.id in step_arrows_next | step_arrows_previous:
                        arrow_objs[-1].append(arrow)

                # save mols existing from processing the prevoius curly arrows
                # previous_mols = reactant_mols[:]

                # the are all the mols touching the current curly arrow
                current_arrow_mols = [Chem.RWMol(m) for m in graph.nodes()[curly_arrow].get("arrow_env")]

                # save information regarding the curly arrow that touch each arrow
                # coded as follows
                # tens digit - 2 if full arrow (carries two electrons), 1 if half arrow (carries one electron)
                # ones digit - 1 arrow out (touches base), 2 arrow in (touches tip), 3 both
                for current_mol in current_arrow_mols:
                    for atom in current_mol.GetAtoms():
                        if (am := atom.GetAtomMapNum()):
                            arrow_size = str(am//10)
                            in_out_code = am % 10
                            if in_out_code % 2:
                                atom.SetProp(f"arrow_{arrow_no+1}_out", arrow_size)
                            if in_out_code//2:
                                atom.SetProp(f"arrow_{arrow_no+1}_in", arrow_size)
                # print("current mol", Chem.MolToSmiles(current_mol))

                # first curly arrow
                # print("arrow no", arrow_no+1)
                if not reactant_mols_lists:
                    # reactant_mols.extend(current_arrow_mols)
                    reactant_mols_lists.append(current_arrow_mols)
                else:
                    # merge mols as necessary, change reactant_mols in place
                    reactant_mols_lists = merge_mols(current_arrow_mols, reactant_mols_lists, is_last_circular_arrow)
                # print("rmols", [Chem.MolToSmiles(m) for ms in reactant_mols_lists for m in ms])

                for reactant_mols in reactant_mols_lists:
                    for mol in reactant_mols:
                        for atom in mol.GetAtoms():
                            # keep information about charge_plus (base of curly arrow only of first arrow
                            if atom.HasProp("charge_plus") and not atom.HasProp("arrow_1_out"):
                                atom.ClearProp("charge_plus")
                            if atom.HasProp("charge_minus") and not atom.HasProp(f"arrow_{arrow_no+1}_in"):
                                atom.ClearProp("charge_minus")

            for reactant_mols in reactant_mols_lists:
                print("path")
                if is_circular:
                    print("is_circular")
                    for mol in reactant_mols:
                        for atom in mol.GetAtoms():
                            if atom.HasProp("charge_plus"):
                                atom.SetProp(f"arrow_{arrow_no+1}_in", arrow_size)
                            # print(atom.GetSymbol(), atom.GetPropsAsDict())

                combined_reactants = common.utils.combine_mols_list(reactant_mols)

                try:
                    combined_reactants.UpdatePropertyCache()
                except:
                    print("error in molecule", Chem.MolToSmiles(combined_reactants))
                    continue

                # Add atom maps
                for atom in combined_reactants.GetAtoms():
                    # print(atom, atom.GetPropsAsDict())
                    atom.SetIsotope(0)
                    # if atom.HasProp("rc"):
                    atom.SetAtomMapNum(atom.GetIdx()+1)
                    atom.SetQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(atom.GetAtomicNum()))
                    # if not atom.HasProp("charge_plus") and not atom.HasProp("charge_minus"):
                        # print("not", atom.GetPropsAsDict())
                    # else:
                        # print("yes", atom.GetPropsAsDict())
                    # atoms in the second shell that are either C or H
                    if atom.HasProp("2nd_after_c") and not atom.HasProp("rc_or_1st"):
                        if atom.GetAtomicNum() in (1, 6) and atom.GetDegree() == 1:
                            atom.ExpandQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(1 if atom.GetAtomicNum() == 6 else 6),
                                            how=Chem.rdchem.CompositeQueryType.COMPOSITE_OR)
                    atom.ExpandQuery(Chem.rdqueries.FormalChargeEqualsQueryAtom(atom.GetFormalCharge()),
                                     how=Chem.rdchem.CompositeQueryType.COMPOSITE_AND)
                print("combined reactants", Chem.MolToSmiles(combined_reactants))
                try:
                    products, rule_arrows = common.utils.get_mol_product_from_props(combined_reactants)
                except KeyError:
                    print(arrow_objs)
                    continue

                print(rule_arrows)
                print("combined reactants", Chem.MolToSmiles(products))

                try:
                    products.UpdatePropertyCache()
                except:
                    print("error in molecule. products: ", Chem.MolToSmiles(products))
                    continue
                # for atom in combined_reactants.GetAtoms():
                    # print(atom.GetSymbol(), atom.GetPropsAsDict())
                # print(Chem.MolToSmarts(products), rule_arrows)
                for atom in products.GetAtoms():
                    atom.SetQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(atom.GetAtomicNum()))
                    if atom.HasProp("2nd_after_c") and not atom.HasProp("rc_or_1st"):
                        if atom.GetAtomicNum() in (1, 6)  and atom.GetDegree() == 1:
                            atom.ExpandQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(1 if atom.GetAtomicNum() == 6 else 6),
                                            how=Chem.rdchem.CompositeQueryType.COMPOSITE_OR)
                    atom.ExpandQuery(Chem.rdqueries.FormalChargeEqualsQueryAtom(atom.GetFormalCharge()),
                                     how=Chem.rdchem.CompositeQueryType.COMPOSITE_AND)
                    # print(atom.DescribeQuery())
                reaction_smarts = f"{Chem.MolToSmarts(combined_reactants)}>>{Chem.MolToSmarts(products)}".replace("&", "")
                print(reaction_smarts)
                reverse_r_smarts = f"{Chem.MolToSmarts(products)}>>{Chem.MolToSmarts(combined_reactants)}".replace("&", "")
                reverse_r_arrows = []
                for rule_arrow in rule_arrows[::-1]:
                    reverse_r_arrows.append([rule_arrow[1], rule_arrow[0], rule_arrow[2]])
                # print(reaction_smarts, rule_arrows)
                # print("same")

                for reversed, (r_smarts, r_arrows) in enumerate(
                        ((reaction_smarts, rule_arrows), (reverse_r_smarts, reverse_r_arrows))):
                    # print(r_smarts)

                    rule = r_smarts_to_rule.get(r_smarts,
                                                cls(reaction_smarts=r_smarts, rule_arrows=r_arrows, version=version))

                    canonical_smiles = rule.canonical_reaction_smiles
                    # print(canonical_smiles)
                    # if the canonical smiles is repeated, do not save this rule
                    # but still add the arrow_chain info so the other rules can be associated with this Arrow
                    if canonical_smiles not in r_canonical_smiles_to_r_smarts:
                        # print("repeated canonical smiles", canonical_smiles)
                        # else:
                        r_smarts_to_rule[r_smarts] = rule
                    r_canonical_smiles_to_arrows[canonical_smiles].append([arrow_objs, bool(reversed)])
                    r_canonical_smiles_to_r_smarts[canonical_smiles] = r_smarts

        to_create = [rule for rule in r_smarts_to_rule.values() if not rule.pk]
        print("rules to create", len(to_create))
        rules_created = cls.objects.bulk_create(to_create)
        logger.info("Created {} new rules".format(len(r_smarts_to_rule)))

        arrow_rules = []
        existing_rule_arrow = set(cls.arrows.through.objects.values_list("rule_id", "arrow_id", "arrow_no", "reversed"))
        # not reversing each curly arrow individually
        for r_smarts, rule in r_smarts_to_rule.items():
            for arrow_sets, reversed in r_canonical_smiles_to_arrows[rule.canonical_reaction_smiles]:
                for arrow_no, curly_arrow in enumerate(arrow_sets):
                    arrow_no += 1
                    for arrow in curly_arrow:
                        if (rule.id, arrow.id, arrow_no+1, reversed) not in existing_rule_arrow:
                            arrow_rules.append(ArrowRule(
                                rule_id=rule.id, arrow_id=arrow.id, arrow_no=arrow_no, reversed=reversed))

        print("Creating arrow rules", len(ArrowRule.objects.bulk_create(arrow_rules, batch_size=1000)))
        return rules_created

    def _get_mol(self, product=False):
        """return mol of reactant or product with final bonds and a good conformer, and rule_arrows mapped to idxs"""
        mol = Chem.RWMol(Chem.MolFromSmarts(self.reactant_smarts if not product else self.product_smarts))
        am_to_idx = {atom.GetAtomMapNum(): atom.GetIdx() for atom in mol.GetAtoms()}

        # find the indexes mentioned in the rule_arrows
        rule_arrows_idx = []
        for arrow in self.rule_arrows:
            rule_arrows_idx.append([])
            for base_or_tip in arrow[:2]:
                rule_arrows_idx[-1].append([])
                for am in base_or_tip:
                    rule_arrows_idx[-1][-1].append(am_to_idx[am])
            rule_arrows_idx[-1].append(arrow[2])

        if product:
            # give same position as reactant
            map_to_position = {atom.GetAtomMapNum(): self.reactant_mol[0].GetConformer().GetAtomPosition(atom.GetIdx())
                               for atom in self.reactant_mol[0].GetAtoms()}
            mol.AddConformer(Chem.Conformer(mol.GetNumAtoms()))
            for atom in mol.GetAtoms():
                mol.GetConformer().SetAtomPosition(atom.GetIdx(), map_to_position[atom.GetAtomMapNum()])
        # create new bonds if necessary
        for arrow in rule_arrows_idx:
            if len(arrow[1]) == 2 and not mol.GetBondBetweenAtoms(arrow[1][0], arrow[1][1]):
                mol.AddBond(arrow[1][0], arrow[1][1], Chem.rdchem.BondType.ZERO)

            if product:
                if len(arrow[0]) == 2 and not mol.GetBondBetweenAtoms(arrow[0][0], arrow[0][1]):
                    mol.AddBond(arrow[0][0], arrow[0][1], Chem.rdchem.BondType.ZERO)

        # compute conformer
        if not product:
            mol.UpdatePropertyCache(strict=False)
            Chem.rdDepictor.Compute2DCoords(mol)

        return mol, rule_arrows_idx

    @cached_property
    def reactant_mol(self):
        """return mol based on the reactant smarts with bonds to be formed added"""
        return self._get_mol(product=False)

    @cached_property
    def product_mol(self):
        """return mol based on the reactant smarts with bonds to be formed added"""
        return self._get_mol(product=True)

    def _get_svg(self, product=False) -> str:
        """return the svg text of reactants or products. If reactants it includes reaction arrows"""
        mol, rule_arrows_idx = self._get_mol(product=product)
        atoms_to_highlight = [idx for arrow in rule_arrows_idx for end in arrow[:2] for idx in end]
        # print("atoms to highlight", atoms_to_highlight, "product",product )

        # hide atom maps
        for atom in mol.GetAtoms():
            atom.SetAtomMapNum(0)
            if atom.GetAtomicNum() == 0:
                # atom.SetProp("atomLabel", "C/H")
                atom.SetProp("atomLabel", "")

        mc = Chem.Draw.rdMolDraw2D.PrepareMolForDrawing(mol, forceCoords=False)
        drawer = Chem.Draw.rdMolDraw2D.MolDraw2DSVG(300, 300)
        opts = drawer.drawOptions()
        opts.highlightRadius =0.4
        opts.setHighlightColour((0.8, 0.8, 0.8))
        opts.includeAtomTags = True
        opts.atomHighlightsAreCircles = True
        drawer.DrawMolecule(mc, highlightAtoms=atoms_to_highlight)
        drawer.FinishDrawing()
        svg_str = drawer.GetDrawingText()
        # logger.debug(svg_str)

        svg = etree.fromstring(svg_str.encode())

        defs = etree.SubElement(svg, "defs")
        marker = etree.SubElement(defs, "marker", {
            "id": "arrowhead",
            "viewBox": "0 0 10 10",
            "refX": "7",
            "refY": "5",
            "markerUnits": "strokeWidth",
            "markerWidth": "8",
            "markerHeight": "6",
            "orient": "auto",
            })
        etree.SubElement(marker, "path", {
            "d": "M 0 0 L 10 5 L 0 10 z",
            "stroke": "none",
            "fill": "black",
            })

        namespace = "http://www.w3.org/2000/svg"
        # space out zero bonds
        for bond in svg.findall("{{{}}}path".format(namespace)):
            style = bond.get("style")
            if style is not None:
                style = style.replace("stroke-dasharray:1,1.33333", "stroke-dasharray:2,5")
                style = style.replace("stroke-dasharray:2,2", "stroke-dasharray:2,5")
                bond.set("style", style)

        if not product:
            # this is necessary because rdkit has different behaviours for some reason
            # circle_or_ellipse = "ellipse" if svg.findall("{{{}}}ellipse".format(namespace)) else "circle" 
            atom_to_coords = {}
            for selector in ("ellipse", "circle"):
                atom_highlights = svg.findall(f"{{{namespace}}}{selector}")
                for hl in atom_highlights:
                    atom_to_coords[hl.get("class")] =\
                            Point2D(float(hl.get("cx")), float(hl.get("cy")))

            # print(atom_to_coords)
            for arrow_no, arrow in enumerate(rule_arrows_idx):
                coords = [[], []]
                arrow = arrow[:2]
                if product:
                    arrow.reverse()
                for no, end in enumerate(arrow):
                    for idx in end:
                        selector1 =  f"atom-{idx}"
                        selector2 =  f"atom-selector atom-{idx}"
                        selector = selector1 if selector1 in atom_to_coords else selector2
                        coords[no].append(atom_to_coords[selector])
                    # find mid point between atoms if arow tip/end points to bond
                    coords[no] = sum(coords[no], Point2D(0, 0))/len(coords[no])

                vector = coords[1] - coords[0]
                normal = Point2D(vector[1], -vector[0])

                # numbers over arrows
                # etree.SubElement(svg, "text", {
                    # "x": str((coords[0][0] + coords[1][0])/2 + normal[0]),
                    # "y": str((coords[0][1] + coords[1][1])/2 + normal[1]),
                    # "font-family": "Verdana",
                    # "font-size": "10",
                    # "fill": "black",
                # }).text = str(arrow_no+1)

                etree.SubElement(svg, "path", {
                    "d": "M {} {} C {} {}, {} {}, {} {}".format(coords[0][0] + normal[0]/8, coords[0][1] + normal[1]/8,
                                                                coords[0][0] + normal[0]/1,
                                                                coords[0][1] + normal[1]/1,
                                                                coords[1][0] + normal[0]/1,
                                                                coords[1][1] + normal[1]/1,
                                                                coords[1][0] + normal[0]/8, coords[1][1] + normal[1]/8,
                                                                ),
                    "marker-end": "url(#arrowhead)",
                    "class": "curly-arrow",
                    "stroke-width": "2",
                    "stroke": "black",
                    "fill": "transparent"})

        return etree.tostring(svg, pretty_print=True).decode()

    @cached_property
    def product_svg(self):
        """return the svg str of the rule products in the same configuration as the reactants"""
        return self._get_svg(product=True)

    @cached_property
    def reactant_svg(self):
        """return the svg str of the rule reactants with reaction arrows"""
        return self._get_svg(product=False)

    @classmethod
    def save_rules_svg(cls):
        for rule in cls.objects.all():
            with open(f"{atlas.files.RULE_SVG_FOLDER}{rule.id}_r.svg", 'w') as svg_file:
                svg_file.write(rule.reactant_svg)
            with open(f"{atlas.files.RULE_SVG_FOLDER}{rule.id}_p.svg", 'w') as svg_file:
                svg_file.write(rule.product_svg)



def add_last_arrow_info(new_mols, old_mols):
    """add last_arrow_info to circular paths"""
    # mark atoms to be superimposed with a different isotope
    # if molecule does not have any marked atom then it is a new molecule
    for old_mol in old_mols:
        for atom in old_mol.GetAtoms():
            if atom.HasProp("charge_minus"):
                atom.SetIsotope(atom.GetAtomicNum()+100)
                old_mol_merge = old_mol
                print(atom.GetSymbol(), "charge_minus")
            else:
                atom.SetIsotope(atom.GetAtomicNum())

    for new_mol in new_mols:
        for atom in new_mol.GetAtoms():
            if atom.HasProp("charge_plus"):
                atom.SetIsotope(atom.GetAtomicNum()+100)
                new_mol_merge = new_mol
                print(atom.GetSymbol(), "charge_plus")
            else:
                atom.SetIsotope(atom.GetAtomicNum())

    # merge molecules
    mcs = Chem.rdFMCS.FindMCS((new_mol_merge, old_mol_merge), atomCompare=Chem.rdFMCS.AtomCompare.CompareIsotopes)
    mcs_mol = Chem.MolFromSmarts(mcs.smartsString)

    # now add any atoms missing on current_arrow_mol
    old_map = old_mol_merge.GetSubstructMatch(mcs_mol)
    new_map = new_mol_merge.GetSubstructMatch(mcs_mol)
    old_to_new = {idx_p: idx_c for idx_p, idx_c in zip(old_map, new_map)}
    atoms_to_add = []
    for atom in old_mol_merge.GetAtoms():
        if atom.GetIdx() not in old_to_new:
            atoms_to_add.append(atom)

    # add new atoms that do not exist on the old mol
    for atom in atoms_to_add:
        idx = new_mol_merge.AddAtom(atom)
        old_to_new[atom.GetIdx()] = idx

    # add new bonds that do not exist on the old mol
    for bond in old_mol_merge.GetBonds():
        p_atom_beg = old_to_new[bond.GetBeginAtomIdx()]
        p_atom_end = old_to_new[bond.GetEndAtomIdx()]
        if not new_mol_merge.GetBondBetweenAtoms(p_atom_beg, p_atom_end):
            new_mol_merge.AddBond(p_atom_beg, p_atom_end, bond.GetBondType())

    # copy arrow data
    for atom in old_mol_merge.GetAtoms():
        for prop, value in atom.GetPropsAsDict().items():
            if "arrow" in prop:
                new_mol_merge.GetAtomWithIdx(old_to_new[atom.GetIdx()]).SetProp(prop, str(value))
    # print("adding", Chem.MolToSmiles(current_arrow_mol))


    # replace the old mols with the merged ones and add new mols to the old_mols list
    old_mols[old_mols.index(old_mol_merge)] = new_mol_merge
    for new_mol in new_mols:
        if new_mol is not new_mol_merge:
            old_mols.append(new_mol)


def merge_mols(new_mols, old_mols_lists, is_last_circular_arrow=False):
    """compare mols from two different arrows and merges as necessary

    the atom in the tip of the previous arrow is the same atoms as the base of
    the current one and so this is the same molecule and should be merged"""
    # print("\nMERGING")

    old_mols_lists_to_return = []
    for old_mols in old_mols_lists:
        for old_mol in old_mols:
            # print("old_mol", Chem.MolToSmiles(old_mol))
            for atom in old_mol.GetAtoms():
                new_isotope = atom.GetAtomicNum()
                if atom.HasProp("charge_minus"):
                    new_isotope += 100
                    old_mol_merge = old_mol
                    atom.SetAtomMapNum(111)
                else:
                    atom.SetAtomMapNum(0)
                if (charge := atom.GetFormalCharge()):
                    # 1 2 3 4 5 positive charge
                    # 9 8 7 6 negative charge
                    # digit saved in the thousands
                    new_isotope += ((10+charge) % 10)*1000
                atom.SetIsotope(new_isotope)
            # print("old mol", Chem.MolToSmiles(old_mol))

        for new_mol in new_mols:
            # print("new mol", Chem.MolToSmiles(new_mol))
            for atom in new_mol.GetAtoms():
                new_isotope = atom.GetAtomicNum()
                if atom.HasProp("charge_plus"):
                    new_isotope += 100
                    new_mol_merge = new_mol
                    # print(atom.GetSymbol(), "charge_plus")
                    atom.SetAtomMapNum(111)
                else:
                    atom.SetAtomMapNum(0)
                if (charge := atom.GetFormalCharge()):
                    # 1 2 3 4 5 positive charge
                    # 9 8 7 6 negative charge
                    # digit saved in the thousands
                    new_isotope += ((10+charge) % 10)*1000
                atom.SetIsotope(new_isotope)
            # print("new mol", Chem.MolToSmiles(new_mol))

        # merge molecules
        # to avoid greedy mcs, just superimpose env around common atom.
        common_rc_env_old, _ = common.utils.get_mol_env(old_mol_merge, {111, })
        common_rc_env_new, _ = common.utils.get_mol_env(new_mol_merge, {111, })

        # now add any atoms missing on current_arrow_mol
        old_maps = old_mol_merge.GetSubstructMatches(common_rc_env_old, uniquify=False)
        new_maps = new_mol_merge.GetSubstructMatches(common_rc_env_old, uniquify=False)
        
        # print("old", Chem.MolToSmiles(common_rc_env_old), Chem.MolToSmiles(old_mol_merge))
        # print("new", Chem.MolToSmiles(common_rc_env_new), Chem.MolToSmiles(new_mol_merge))
        # print(old_maps)
        # print(new_maps)

        new_mols_merged = []
        # random limit - is there a better way to do this?
        for old_map in old_maps[:10]:
            for new_map in new_maps[:10]:
                old_to_new = {}
                # ignore if superimposed atoms do not have the same isotope (not taken into account in 
                # getsubstructmatches

                to_break = False
                for idx_old, idx_new in zip(old_map, new_map):
                    old_to_new[idx_old] = idx_new
                    # old_atom = old_mol_merge.GetAtomWithIdx(idx_old)
                    # new_atom = new_mol_merge.GetAtomWithIdx(idx_new)
                    # print(old_atom.GetSymbol(), new_atom.GetSymbol())
                    # print(old_atom.GetIsotope(), new_atom.GetIsotope())
                    # print(old_atom.GetAtomMapNum(), new_atom.GetAtomMapNum())
                    if old_mol_merge.GetAtomWithIdx(idx_old).GetAtomMapNum() !=\
                            new_mol_merge.GetAtomWithIdx(idx_new).GetAtomMapNum():
                        to_break = True
                # print("to break", to_break)
                if to_break:
                    continue
                    # break

                mol_to_save = Chem.RWMol(new_mol_merge)
                old_mol_copy = Chem.RWMol(old_mol_merge)
                # old_to_new = {idx_p: idx_c for idx_p, idx_c in zip(old_map, new_map)}
                for atom in mol_to_save.GetAtoms():
                    atom.SetIsotope(atom.GetAtomicNum())
                    atom.SetAtomMapNum(0)
                for atom in old_mol_copy.GetAtoms():
                    atom.SetIsotope(atom.GetAtomicNum())
                    atom.SetAtomMapNum(0)
                for no, (idx_p, idx_c) in enumerate(old_to_new.items()):
                    # print(idx_p, idx_c)
                    old_mol_copy.GetAtomWithIdx(idx_p).SetIsotope(atom.GetAtomicNum()+(no+1)*100)
                    mol_to_save.GetAtomWithIdx(idx_c).SetIsotope(atom.GetAtomicNum()+(no+1)*100)

                mcs = Chem.rdFMCS.FindMCS((mol_to_save, old_mol_copy), atomCompare=Chem.rdFMCS.AtomCompare.CompareIsotopes, timeout=1)
                mcs_mol = Chem.MolFromSmarts(mcs.smartsString)
                # print(mcs.smartsString)
                # now add any atoms missing on current_arrow_mol
                old_maps_2 = old_mol_copy.GetSubstructMatches(mcs_mol, uniquify=False)
                new_maps_2 = mol_to_save.GetSubstructMatches(mcs_mol, uniquify=False)
                for new_map_2 in new_maps_2:
                    for old_map_2 in old_maps_2:
                        old_to_new_2 = {idx_p: idx_c for idx_p, idx_c in zip(old_map_2, new_map_2)}
                        mol_to_save_2 = Chem.RWMol(mol_to_save)
                        atoms_to_add = []
                        for atom in old_mol_copy.GetAtoms():
                            if atom.GetIdx() not in old_to_new_2:
                                atoms_to_add.append(atom)

                        # add new atoms that do not exist on the old mol
                        for atom in atoms_to_add:
                            idx = mol_to_save_2.AddAtom(atom)
                            old_to_new_2[atom.GetIdx()] = idx

                        # add new bonds that do not exist on the old mol
                        for bond in old_mol_copy.GetBonds():
                            p_atom_beg = old_to_new_2[bond.GetBeginAtomIdx()]
                            p_atom_end = old_to_new_2[bond.GetEndAtomIdx()]
                            if not mol_to_save_2.GetBondBetweenAtoms(p_atom_beg, p_atom_end):
                                mol_to_save_2.AddBond(p_atom_beg, p_atom_end, bond.GetBondType())

                        # copy atom properties including arrow data
                        for atom in old_mol_copy.GetAtoms():
                            for prop, value in atom.GetPropsAsDict().items():
                                mol_to_save_2.GetAtomWithIdx(old_to_new_2[atom.GetIdx()]).SetProp(prop, str(value))
                                # if "arrow" in prop or "charge" in prop:

                        for atom in mol_to_save_2.GetAtoms():
                            if atom.HasProp("charge_plus"):
                                atom.SetIsotope(111)
                            else:
                                atom.SetIsotope(0)  # to make canonical later
                        new_mols_merged.append(mol_to_save_2)
                # replace the old mols with the merged ones and add new mols to the old_mols list
        if not is_last_circular_arrow:
            for new_mol in new_mols:
                if new_mol is not new_mol_merge:
                    old_mols.append(new_mol)

        # make unique based on canonical smiles
        new_mols_merged = {Chem.MolToSmiles(mol): mol for mol in new_mols_merged}.values()
        for new_mol in new_mols_merged:
            old_mols_copy = []
            for old_mol in old_mols:
                # not creating new mols
                if old_mol is old_mol_merge:
                    old_mols_copy.append(new_mol)
                else:
                    old_mols_copy.append(Chem.Mol(old_mol))
            old_mols_lists_to_return.append(old_mols_copy)

    return old_mols_lists_to_return
