# coding=utf-8
"""Includes models related to protein Sequence (Uniprot Id) and PFAM domains"""

# python standard imports
import collections
import csv
import io
import glob
import gzip
import pathlib
import random
import shutil
import subprocess
import os
import urllib
import urllib.request
import pickle
from pathlib import Path

# other python imports
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqUtils import IUPACData

# django imports
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django import db
from django.db.models import Q, Count, F, OuterRef, Subquery
from django.utils.html import format_html

# atlas imports
from atlas import files
from common import others, link, validators
import common.models
import common.stats


class SequenceQuerySet(models.QuerySet):
    """Some predefined querysets for the Sequence model"""

    # ## the following definitions were used in pseudoenzyme paper
    # ## do not change without good reason
    def ec(self):
        """Filters sequences that have ecs"""
        return self.filter(ecs__in=common.models.EC.objects.current()).distinct("pk")

    def cat_kw(self, catalytic=True, id_set=False):
        """Filters sequences that have catalytic keywords"""
        catalytic_sequences = self.filter(keywords__in=common.models.Keyword.objects.catalytic())
        if catalytic:
            query = catalytic_sequences.distinct("pk")
        else:
            query = self.reviewed().exclude(id__in=catalytic_sequences).distinct("pk")
        if id_set:
            return set(query.values_list("pk", flat=True))
        else:
            return query

    def cat_go(self, enables=True, only_experimental=False, id_set=False):
        """"Filters sequence that link to catalytic go terms."""
        associations = common.models.SequenceGo.objects.catalytic()
        if only_experimental:
            associations = associations.experimental()
        if enables:
            associations = associations.enables()
            query = self.filter(id__in=associations.values_list("sequence_id"))
        else:
            associations = associations.enables(False)
            # exclude sequence that have a enabled catalytic annotation experimental or not
            query = self.filter(id__in=associations.values_list("sequence_id"))\
                .exclude(id__in=common.models.SequenceGo.objects.catalytic().enables().values_list("sequence_id"))

        if id_set:
            return set(query.values_list("pk", flat=True))
        else:
            return query

    def catalytic_annotation(self, id_set=False):
        reviewed = self.reviewed(id_set=True)
        cat_set = reviewed & (Sequence.cat_kw_set() | Sequence.cat_exp_go_set())
        if id_set:
            return cat_set
        else:
            return self.filter(id__in=cat_set)

    def enzymes(self, id_set=False):
        """Definition used in pseudoenzymes paper to categorize sequences in 'implicit' enzymatic annotation"""
        ids = self.catalytic_annotation(id_set=True) - set(self.filter(is_certain_nonenzyme_manual=True)
                                                           .values_list("pk", flat=True))
        if id_set:
            return ids
        else:
            return Sequence.objects.filter(id__in=ids)

    def problematic_pfams(self, id_set=False):
        """sequences that belong to a pfam were annotation is not consistent"""
        query = self.filter(pfams__idx__in=[
            12,         # "Heat shock protein 70 family",
            118,        # "Chaperonin Cpn60/TCP-1 family" P61112
            9,          # "Transcription factor, GTP-binding domain"
            550,        # "Phosphopantetheine binding ACP domain"
            1228,       # "Glycine radical domain"
            4,          # "ATPase, AAA-type, core"
            25,         # "Small GTPase superfamily, ARF/SAR type"
            148,        # 'Nitrogenase/oxidoreductase, component 1'
            503,        # "Guanine nucleotide binding protein (G-protein)"
            2337, 607,  # pol-gag
            10411,      # DsbC protein family
            4055,       # Radical SAM     Q8CBB9
            3088,       # Strictosidine synthase P68175
            85,         # Thioredoxin
            225,        # Kinesin motor domain
            6,          # ATP synthase alpha/beta subunits

        ])
        query |= self.filter(pfams__isnull=True)
        if id_set:
            return set(query.values_list("pk", flat=True))
        else:
            return query

    def uncharacterized(self, id_set=False):
        """annotation is insuficient, in particular to decide about pseudoenzymes."""
        query = self.reviewed().filter((Q(description='', caution='', misc='') |
                                        Q(name__istartswith="Uncharacterized")))
        if id_set:
            return set(query.values_list("pk", flat=True))
        else:
            return query

    def sure_annotation(self):
        "remove and use curated instead"
        return self.curated()

    def curated(self, id_set=False):
        "includes only enzymes that are characterised and pseudoenzymes that where manually checked"
        result_set = self.catalytic_annotation(id_set=True)
        result_set -= Sequence.uncharacterized_set()
        result_set |= set(self.filter(is_certain_nonenzyme_manual=True).values_list("pk", flat=True))
        if id_set:
            return result_set
        else:
            return self.filter(id__in=result_set)

    def experimental(self, id_set=False):
        """Filters sequences that have experimental functional evidence"""
        reviewed = self.reviewed()
        result_set = set(reviewed.filter(function_experimental_evidence=True).values_list("pk", flat=True))
        result_set.update(reviewed.cat_go(only_experimental=True).values_list("pk", flat=True))
        result_set.update(reviewed.cat_go(only_experimental=True, enables=False).values_list("pk", flat=True))
        if id_set:
            return result_set
        else:
            return self.filter(id__in=result_set)

    # ### end of definitions used in the paper ###

    def set_is_potential_nonenzyme(self):

        reviewed = Sequence.objects.reviewed()
        # these are considered nonenzymes without further checking
        sure_nonenzymes = Sequence.inactive_set()
        sure_nonenzymes |= set(reviewed.cat_go(only_experimental=True, enables=False)
                               .exclude(id__in=reviewed.cat_go(only_experimental=True)).values_list("pk", flat=True))
        print("sure nonenzymes: ", len(sure_nonenzymes))
        Sequence.objects.filter(id__in=sure_nonenzymes).update(
                is_potential_nonenzyme=True,
                is_certain_nonenzyme_manual=True)
        # flag homologues that are potential nonenzymes
        homologues_ids = Sequence.reviewed_homologue_set()
        homologues = Sequence.objects.reviewed().homologue()
        non_enzymes = homologues_ids - Sequence.objects.catalytic_annotation(id_set=True)
        non_enzymes -= Sequence.uncharacterized_set()
        non_enzymes -= sure_nonenzymes
        print("potential nonenzymes: ", len(non_enzymes))

        flag_words = ["lack", "not conserved", "regulatory", "lost", "no.*activity", "no.*catalytic", "inactive",
                      "weak.*activity", "any.*activity", "little.*activity", "defective"]
        non_enzymes &= set(homologues.filter(
                Q(description__iregex="|".join(flag_words)) |
                Q(misc__iregex="|".join(flag_words)) |
                Q(caution__iregex="|".join(flag_words))).values_list("pk", flat=True))

        print("after filtering for words: ", len(non_enzymes))
        Sequence.objects.filter(id__in=non_enzymes).update(is_potential_nonenzyme=True)
        return

    def inactive(self, id_set=False):
        """Filters sequences with a name that start with Inactive"""
        query = self.filter(
            Q(name__istartswith="inactive") |
            Q(name__istartswith="probable inactive") |
            Q(name__istartswith="putative inactive") |
            Q(name__istartswith="probably inactive")).distinct("pk")
        if id_set:
            return set(query.values_list("pk", flat=True))
        else:
            return query

    def reviewed(self, reviewed=True, id_set=False):
        """Filters sequences that are in SwissProt"""
        query = self.filter(in_swissprot=reviewed)
        if id_set:
            return set(query.values_list("pk", flat=True))
        else:
            return query

    def get_idx2id(self):
        """Returns values list uniprot id -> database id"""
        return self.values_list('uniprot_id', 'id')

    def get_idx2species_name(self):
        """Returns values list uniprot id -> species_name"""
        return self.values_list('uniprot_id', 'species__scientific_name')

    def get_idx2species_taxid(self):
        """Returns values list uniprot id -> species_taxid"""
        return self.values_list('uniprot_id', 'species__taxid')

    def get_idx2obj(self, only=None):
        """Returns dict uniprot id -> obj"""
        if only is not None:
            self = self.only(*only)
        return {seq.uniprot_id: seq for seq in self}

    def one_cath(self):
        """Filters sequences that are associated with only one CATH domain"""
        return self.annotate(num=Count('caths')).filter(num=1)

    def reference(self):
        """Filters all the sequences that are references for at least one residue in M-CSA published entries"""
        from common.models import ResidueSequence, Entry
        reference_residue_sequences = ResidueSequence.objects.filter(
            is_reference=True, residue__active_site__entry__in=Entry.objects.mcsa())
        return self.filter(residue_sequences__in=reference_residue_sequences).distinct()

    def homologue(self, id_set=False):
        """Filters all the sequences that are references or homologues to references"""
        query = self.filter(alignment__isnull=False)
        if id_set:
            return set(query.values_list("pk", flat=True))
        else:
            return query.distinct("pk")

    def uniprot_ids(self, uniprot_ids):
        """Filters sequences that have one of these uniprot_ids"""
        return self.filter(uniprot_id__in=uniprot_ids)

    def aa_frequency(self):
        """Returns a dict with the frequencies of all aminoacids in these sequences"""
        all_sequences_string = ''.join(self.values_list('sequence', flat=True))
        aa1_counter = collections.Counter(all_sequences_string)
        return {IUPACData.protein_letters_1to3[k]: int(v)/len(all_sequences_string)
                for k, v in aa1_counter.items() if k in IUPACData.protein_letters_1to3}

    def conservation_by_active_site(self, **kwargs):
        """this function does not need to be here except for the creation of a multibarplot"""
        return common.stats.conservation_by_active_site(**kwargs)

    def conservation_by_residue_role(self, **kwargs):
        """this function does not need to be here except for the creation of a multibarplot"""
        return common.stats.conservation_by_residue_role(**kwargs)

    def species_count(self, **kwargs):
        pickle_file = Path("/home/ze/protein_pickle.p")
        if not pickle_file.exists():
            print("reading data from database and saving pickle")
            protein_sets = common.models.Tree.get_protein_sets()
            pickle.dump(protein_sets, pickle_file.open(mode="wb"))
        else:
            print("reading data pickle")
        tag_verbose = {
            "kw": "Combined",
            "exp": "Experimental",
            "filtered_kw": "Filtered Combined",
            "filtered_exp": "Filtered Experimental",
            }
        protein_sets = pickle.load(pickle_file.open(mode="rb"))
        protein_sets = {k: v for k, v in protein_sets.items() if k in ["kw", "filtered_exp"]}
        print("reading done")
        output = collections.OrderedDict()
        keys = set()
        for tag, (all_set, enzyme_set) in protein_sets.items():
            enzymes = Sequence.objects.filter(id__in=enzyme_set)
            nonenzymes = Sequence.objects.filter(id__in=all_set-enzyme_set)
            for enz_tag, sequences in {"Enzymes in ": enzymes, "Pseudoenzymes in ": nonenzymes}.items():
                domain = []
                euk = set(('Metazoa', 'Viridiplantae', 'Fungi'))
                for lineage in sequences.values_list("species__lineage", flat=True):
                    if lineage[0] != "Eukaryota":
                        domain.append(lineage[0])
                        keys.add(lineage[0])
                    elif lineage[1] in euk:
                        domain.append(lineage[1])
                        keys.add(lineage[1])

                counter = collections.Counter(domain)
                for k in keys:
                    v = counter.get(k, 0)
                    if k not in output:
                        output[k] = collections.OrderedDict()
                    output[k][enz_tag+tag_verbose[tag]] = v/sum(counter.values())

        return output

    def get_conservation(self, limit_references=10, limit_homologues=1000, sort=True):  # todo make limits random
        """Returns the conservation score of all residues in these sequences."""
        self = self[:20]  # TODO do this better
        ref_subquery = MultiAlignment.objects.filter(
             reference_sequence__in=OuterRef('reference_sequence'),
             sequence__in=OuterRef('reference_sequence'))

        alignments = MultiAlignment.objects.filter(
              reference_sequence__in=self,  # ,random.choices(self.values_list("id", flat=True), k=limit_references),
              sequence__in_swissprot=True,
              sequence__ecs=F("reference_sequence__ecs"),
             )\
            .annotate(ref_alignment=Subquery(ref_subquery.values("alignment")[:1]))\
            .annotate(ref_pad_left=Subquery(ref_subquery.values("pad_left")[:1]))
        print(alignments.count())

        all_residues_conservation = []
        conservation = collections.defaultdict(lambda: collections.defaultdict(int))
        conservation_code = collections.defaultdict(lambda: collections.defaultdict(int))
        no_seqs_per_ref = collections.defaultdict(int)
        no_seqs_per_ref_code = collections.defaultdict(lambda: collections.defaultdict(int))

        for alignment in alignments.filter(id__in=random.choices(alignments.values_list("id", flat=True), k=100)):
            print('.', end="", flush=True)
            no_seqs_per_ref[alignment.reference_sequence_id] += 1
            for no, l in enumerate(alignment.ref_alignment):
                no_seqs_per_ref_code[alignment.reference_sequence_id][l] += 1
                no += alignment.ref_pad_left
                no_other = no - alignment.pad_left
                if len(alignment.alignment) > no_other >= 0 and l == alignment.alignment[no_other]:
                    conservation[alignment.reference_sequence_id][no] += 1
                    conservation_code[alignment.reference_sequence_id][l] += 1

        all_residues_conservation = []
        for ref in conservation:
            for count in conservation[ref].values():
                all_residues_conservation.append(count/no_seqs_per_ref[ref])
        if sort:
            all_residues_conservation.sort()

        all_conservations_code = {}
        for code1, code in others.AA_1TO3.items():
            all_conservations_code[code] = []
            for ref in conservation_code:
                if conservation_code[ref][code1]:
                    all_conservations_code[code].append(conservation_code[ref][code1]/no_seqs_per_ref_code[ref][code1])
            if sort:
                all_conservations_code[code].sort()
        print(all_conservations_code)

        return all_conservations_code

    def get_avg_conservation(self):
        return {k: sum(v)/len(v) for k, v in self.get_conservation().items() if v}


class Sequence(models.Model):
    """A sequence defined as a uniprot id"""
    uniprot_id = models.CharField(
        unique=True,
        max_length=10,
        validators=[validators.validate_uniprot_id_exists],
        verbose_name="Uniprot ID",
        db_index=True
    )
    name = models.CharField(max_length=511, blank=False, verbose_name="Name", db_index=True)
    species = models.ForeignKey("Species", on_delete=models.PROTECT, null=True)
    sequence = models.TextField()
    description = models.TextField()
    caution = models.TextField()
    misc = models.TextField()
    in_swissprot = models.BooleanField(db_index=True)
    in_reference_proteome = models.BooleanField(db_index=True, default=True)
    function_experimental_evidence = models.BooleanField(db_index=True)

    keywords = models.ManyToManyField("Keyword", related_name='sequences')
    ecs = models.ManyToManyField('EC')
    caths = models.ManyToManyField('CATH', related_name='sequences', through="SequenceCATH")
    gos = models.ManyToManyField('GoTerm', related_name='sequences', through="SequenceGo")
    pfams = models.ManyToManyField("PFAM", related_name="sequences", through="SequencePFAM")
    homologuesx = models.ManyToManyField('self', through='MultiAlignment', related_name='homologuesx+',
                                         symmetrical=False)
    trees = GenericRelation('Tree', content_type_field="reference_ct", object_id_field="reference_id",
                            related_query_name="sequence")
    is_potential_nonenzyme = models.BooleanField(db_index=True, null=True)
    is_certain_nonenzyme_manual = models.BooleanField(db_index=True, null=True)

    _uniprot_id = ''
    _uniprot_info = ''

    objects = SequenceQuerySet.as_manager()

    def __repr__(self):
        return self.uniprot_id

    def __str__(self):
        return self.uniprot_id

    # ## some useful sets
    reviewed_set_data = None

    @classmethod
    def reviewed_set(cls):
        if cls.reviewed_set_data is None:
            cls.reviewed_set_data = cls.objects.reviewed(id_set=True)
        return cls.reviewed_set_data

    manual_nonenzyme_set_data = None

    @classmethod
    def manual_nonenzyme_set(cls):
        if cls.manual_nonenzyme_set_data is None:
            cls.manual_nonenzyme_set_data = set(cls.objects.filter(is_certain_nonenzyme_manual=True)
                                                .values_list("pk", flat=True))
        return cls.manual_nonenzyme_set_data

    experimental_set_data = None

    @classmethod
    def experimental_set(cls):
        if cls.experimental_set_data is None:
            cls.experimental_set_data = cls.objects.experimental(id_set=True)
        return cls.experimental_set_data

    cat_kw_set_data = None

    @classmethod
    def cat_kw_set(cls):
        if cls.cat_kw_set_data is None:
            cls.cat_kw_set_data = cls.objects.cat_kw(id_set=True)
        return cls.cat_kw_set_data

    cat_exp_go_set_data = None

    @classmethod
    def cat_exp_go_set(cls):
        if cls.cat_exp_go_set_data is None:
            cls.cat_exp_go_set_data = cls.objects.cat_go(enables=True, only_experimental=True, id_set=True)
        return cls.cat_exp_go_set_data

    cat_go_set_data = None

    @classmethod
    def cat_go_set(cls):
        if cls.cat_go_set_data is None:
            cls.cat_go_set_data = cls.objects.cat_go(enables=True, id_set=True)
        return cls.cat_go_set_data

    uncharacterized_set_data = None

    @classmethod
    def uncharacterized_set(cls):
        if cls.uncharacterized_set_data is None:
            cls.uncharacterized_set_data = cls.objects.uncharacterized(id_set=True)
        return cls.uncharacterized_set_data

    inactive_set_data = None

    @classmethod
    def inactive_set(cls):
        if cls.inactive_set_data is None:
            cls.inactive_set_data = cls.objects.reviewed().inactive(id_set=True)
        return cls.inactive_set_data

    curated_set_data = None

    @classmethod
    def curated_set(cls):
        if cls.curated_set_data is None:
            cls.curated_set_data = cls.objects.curated(id_set=True)
        return cls.curated_set_data

    homologue_set_data = None

    @classmethod
    def reviewed_homologue_set(cls):
        if cls.homologue_set_data is None:
            cls.homologue_set_data = cls.objects.reviewed().homologue(id_set=True) - cls.problematic_pfams_set()
        return cls.homologue_set_data

    enzyme_set_data = None

    @classmethod
    def enzyme_set(cls):
        if cls.enzyme_set_data is None:
            cls.enzyme_set_data = cls.objects.reviewed().enzymes(id_set=True)
        return cls.enzyme_set_data

    problematic_pfams_set_data = None

    @classmethod
    def problematic_pfams_set(cls):
        if cls.problematic_pfams_set_data is None:
            cls.problematic_pfams_set_data = cls.objects.reviewed().problematic_pfams(id_set=True)
        return cls.problematic_pfams_set_data

    # ## end of useful sets definitions

    @classmethod
    def update_objects_swissprot(cls):
        """Adds new SwissProt sequences to the sequence table. Change in_swissprot field for existing entries"""
        with open(files.SWISSPROT_INDEX, 'r') as ids_file:
            uniprot_ids = set([line.split(":")[1].strip() for line in ids_file])
        # sanity check
        if len(uniprot_ids) < 550000:
            print("Could not read SwissProt file properly. Number of uniprot ids read: {}".format(len(uniprot_ids)))
            raise

        # removing obsolete
        existing = set(Sequence.objects.reviewed().values_list("uniprot_id", flat=True))
        for obsolete_id in existing - uniprot_ids:
            obsolete = Sequence.objects.get(uniprot_id=obsolete_id)
            print("obsolete ",  obsolete)
            if not common.models.Pdb.objects.filter(structure__sequences=obsolete) and not obsolete.residue_set.all() \
                    and not obsolete.protein_set.all() and not obsolete.residue_sequences.filter(is_reference=True):
                # obsolete.residue_sequences.filter(is_reference=False).delete()
                # print("Deleting obsolete accession id: {}. {}".format(obsolete_id, obsolete.delete()))
                pass
            else:
                print("Obsolete Uniprot ID used as reference: ", obsolete_id)

        # promoting trembl entries to swissprot
        promote = cls.objects.filter(uniprot_id__in=uniprot_ids, in_swissprot=False)
        print("Promoting {} entries to swissprot".format(promote.count()))
        promote.update(in_swissprot=True)
        cls.create_or_update_sequences_batch(uniprot_ids, in_swissprot=True)

    @classmethod
    def create_dat_gz_index(cls, swissprot: bool):
        """Creates a index for the dat gz file of SwissProt or Trembl"""
        dat_gz_file = files.SWISSPROT_DAT_GZ if swissprot else files.TREMBL_DAT_GZ
        print("Creating index for ", dat_gz_file)
        index_file = files.SWISSPROT_INDEX if swissprot else files.TREMBL_INDEX
        index_release = files.UNIPROT_INDEX_RELEASE.with_name("idx_release" +"_sp" if swissprot else "_up")
        current_release = files.UNIPROT_RELEASE
        assert current_release.exists(), "Cannot access {}".format(current_release)
        overwrite = True
        if index_release.exists():
            if index_release.read_text() == current_release.read_text():
                overwrite = False
        if index_file.exists() and not overwrite:
            print("updated index file already exists")
            return True
        else:
            assert dat_gz_file.exists(), "Cannot access {}".format(dat_gz_file)
            command = (
                """zgrep -b -A1 '^ID' {} | sed 's/:ID .*/:/' | sed '/^--/d' |"""
                """sed 's/^.*-AC\s*/ /' | awk '/:$/ {{ printf("%s", $0); next }}1' |tr -d " " |"""
                """tr  ";" ":" | sed 's/:$//' > {}""".format(dat_gz_file, index_file))
            print("Creating new index for {}".format(dat_gz_file))
            subprocess.call(command, shell=True)
            shutil.copyfile(current_release, index_release)
            print("Index created")
            return True

    @classmethod
    def create_or_update_sequences_batch(cls, uniprot_ids, in_swissprot: bool, create_only=False):
        """Adds the uniprot_ids in the list to the database or updates existing ones

        :param uniprot_ids: a container of uniprot_ids strings
        :param in_swissprot: if these sequences belong to swissprot or trembl

        """
        print("Told to add/update {} sequences".format(len(uniprot_ids)))
        if create_only:
            uniprot_ids = cls.check_for_new_sequences(uniprot_ids)
            print("Adding sequences not on M-CSA yet: {}".format(len(uniprot_ids)))
        index = files.SWISSPROT_INDEX if in_swissprot else files.TREMBL_INDEX
        data = files.SWISSPROT_DAT_GZ if in_swissprot else files.TREMBL_DAT_GZ
        records = cls.read_sequence_records(list(uniprot_ids), index, data)
        new_sequences = cls.create_and_update_sequences_from_records(records, in_swissprot=in_swissprot)
        print("{} sequences added to database".format(len(new_sequences)))
        return new_sequences

    @classmethod
    def check_for_new_sequences(cls, uniprot_ids) -> set:
        """
        Returns a set of the uniprot_ids that are not on the database yet

        :param uniprot_ids: container of the uniprot_ids to be tested
        :return: a set of the uniprot_ids that are not on the database yet
        """
        return set(uniprot_ids) - set(cls.objects.values_list("uniprot_id", flat=True))

    @classmethod
    def read_sequence_records(cls, uniprot_ids, index_file, dat):
        """
        Reads a list of Uniprot ids and returns Bio Seq records

        :param uniprot_ids: the input container with the uniprot ids
        :param index_file: the index file to use
        :param dat: the data file to use
        :return: Bio Seq records
        """
        print("Reading Index")
        index = cls.read_uniprot_index(index_file, uniprot_ids)
        print("Index Read")
        to_read_tuples = []
        for uniprot_id in uniprot_ids:
            if uniprot_id in index:
                to_read_tuples.append(index[uniprot_id])
        print("{} sequences found in index".format(len(to_read_tuples)))
        to_read_tuples = sorted(to_read_tuples, key=lambda tup: tup[0])
        with gzip.open(dat, 'rb') as dat_file:
            previous_end = 0
            data = []
            for no, pair in enumerate(to_read_tuples):
                jump = pair[0]-previous_end
                dat_file.seek(jump, 1)
                data.append(dat_file.read(pair[1]))
                previous_end = sum(pair)
        return SeqIO.parse(io.StringIO(b''.join(data).decode("utf-8")), "swiss")

    @classmethod
    def read_uniprot_index(cls, filename, uniprot_ids=None):
        """Reads a index file with the byte offsets of the uniprot ids

        :param filename: the index path and filename
        :param uniprot_ids: returns only the byte offset of these uniprot ids. Returns all by default
        """
        index_dict = {}
        if uniprot_ids is not None:
            uniprot_ids = set(uniprot_ids)
        with open(filename, 'r') as index_file:
            previous_entries = []
            for no, line in enumerate(index_file):
                words = line.strip().split(":")
                index = int(words[0])
                these_uniprot_ids = words[1:]
                for entry in previous_entries:
                    index_dict[entry][1] = index - index_dict[entry][0]
                previous_entries = []
                for uniprot_id in these_uniprot_ids:
                    if uniprot_ids is None or uniprot_id in uniprot_ids:
                        index_dict[uniprot_id] = [index, -1]
                        previous_entries.append(uniprot_id)
        return index_dict

    @classmethod
    def create_and_update_sequences_from_records(cls, records, in_swissprot: bool):
        """
        Creates and updates sequence objects from the record objects
        :param records: Bio Seq record with several sequences
        :param in_swissprot: if the sequences are in swissprot
        :return: the Sequence objects that were added to the database
        """
        species_ncbi_to_dbid = dict(common.models.Species.objects.get_idx2id())
        to_create = []
        exp_eco = set(common.models.EcoTerm.objects.experimental().values_list("idx", flat=True))
        fields_to_check = ["name", "species_id", "in_reference_proteome", "function_experimental_evidence",
                           "description", "caution", "misc"]
        if in_swissprot:
            existing = cls.objects.reviewed().get_idx2obj()
        else:
            existing = cls.objects.get_idx2obj()

        for no, record in enumerate(records):
            name = record.description
            name = name[name.find("Full=")+5:]
            name = name.split(";")[0]
            name = name.split(" {")[0]
            function_experimental_evidence = False
            description = caution = misc = ""
            if 'comment' in record.annotations:
                for line in record.annotations['comment'].splitlines():
                    if line.startswith('FUNCTION'):
                        description = line[10:line.find("{")]
                        start = 0
                        while True:
                            eco_position = line.find("ECO:", start)
                            if eco_position == -1:
                                break
                            elif line[eco_position+4:eco_position+11] in exp_eco:
                                function_experimental_evidence = True
                                break
                            else:
                                start = eco_position + 1
                    if line.startswith('CAUTION'):
                        caution = line[9:line.find("{")]
                        start = 0
                        while True:
                            eco_position = line.find("ECO:", start)
                            if eco_position == -1:
                                break
                            elif line[eco_position+4:eco_position+11] in exp_eco:
                                function_experimental_evidence = True
                                break
                            else:
                                start = eco_position + 1
                    if line.startswith('MISCELLANEOUS'):
                        start = 0
                        while True:
                            eco_position = line.find("ECO:", start)
                            if eco_position == -1:
                                break
                            elif line[eco_position+4:eco_position+11] in exp_eco:
                                function_experimental_evidence = True
                                break
                            else:
                                start = eco_position + 1
                        misc = line[15:line.find("{")]
            # in reference proteome
            in_reference_proteome = 'Reference proteome' in record.annotations.get('keywords', '')
            taxid = int(record.annotations['ncbi_taxid'][0])
            try:
                species_id = species_ncbi_to_dbid[taxid]
            except KeyError:
                species = common.models.Species.objects.create(taxid=taxid)
                species_id = species.id
                species_ncbi_to_dbid[taxid] = species_id

            if record.id in existing:
                current_record = existing[record.id]
                needs_update = False
                for field in fields_to_check:
                    new_value = eval(field)
                    if getattr(current_record, field) != new_value:
                        print("update field", field, new_value, getattr(current_record, field))
                        setattr(current_record, field, new_value)
                        needs_update = True
                if needs_update:
                    print("updating ", current_record)
                    current_record.save()
            else:
                to_create.append(Sequence(
                    uniprot_id=record.id,
                    name=name,
                    sequence=record.seq,
                    species_id=species_id,
                    in_swissprot=in_swissprot,
                    in_reference_proteome=in_reference_proteome,
                    function_experimental_evidence=function_experimental_evidence,
                    description=description,
                    caution=caution,
                    misc=misc,
                ))
        return Sequence.objects.bulk_create(to_create, batch_size=1000)

    # @classmethod
    # def create_sequences_from_records(cls, records, in_swissprot: bool):
    #     """
    #     Creates sequence objects from the record objects
    #     :param records: Bio Seq record with several sequences
    #     :param in_swissprot: if the sequences are in swissprot
    #     :return: the Sequence objects that were added to the database
    #     """
    #     species_ncbi_to_dbid = dict(Species.objects.get_idx2id())
    #     sequences_to_create = []
    #     for record in records:
    #         name = record.description
    #         name = name[name.find("Full=")+5: name.find(' {')]
    #         # function experimental evidence_type
    #         function_experimental_evidence = False
    #         if 'comment' in record.annotations:
    #             for line in record.annotations['comment'].splitlines():
    #                 if line.startswith('FUNCTION') and "ECO:0000269" in line:
    #                     function_experimental_evidence = True
    #         # in reference proteome
    #         in_reference_proteome = 'keywords' in record.annotations \
    #                                 and 'Reference proteome' in record.annotations['keywords']
    #         taxid = int(record.annotations['ncbi_taxid'][0])
    #         try:
    #             species_id = species_ncbi_to_dbid[taxid]
    #         except KeyError:
    #             species = Species.objects.create(taxid=taxid)
    #             species_id = species.id
    #             species_ncbi_to_dbid[taxid] = species_id

    #         sequences_to_create.append(Sequence(
    #             uniprot_id=record.id,
    #             name=name,
    #             sequence=record.seq,
    #             species_id=species_id,
    #             in_swissprot=in_swissprot,
    #             in_reference_proteome=in_reference_proteome,
    #             function_experimental_evidence=function_experimental_evidence,
    #         ))
    #     return Sequence.objects.bulk_create(sequences_to_create, batch_size=1000)

    @classmethod
    def _get_sequence_ec_pairs(cls, swissprot=True):
        """Find all uniprot <-> ec pairs in the UniProt dat files"""
        from common.models import EC
        seq_ids = dict(cls.objects.reviewed().get_idx2id())
        ecs_ids = EC.objects.get_code2id()
        all_pairs = set()
        original = set()
        with gzip.open(files.SWISSPROT_DAT_GZ, 'rt') as swissprot_dat_file:
            for line in swissprot_dat_file:
                ecs = []
                if line.startswith("ID"):
                    accession_ids = []
                if line.startswith('AC'):
                    line = line[2:].replace(" ", "").strip(";\n")
                    accession_ids += line.split(";")
                elif line.startswith('DE'):
                    line = line.replace(";", " ")
                    words = line.split()
                    for word in words:
                        if "EC=" in word:
                            ecs.append(word.replace("EC=", ""))
                all_pairs.update([(seq_ids[uniprot_id], ecs_ids[ec]) for ec in ecs for uniprot_id in accession_ids
                                 if uniprot_id in seq_ids and ec in ecs_ids])
                original.update([(uniprot_id, ec) for ec in ecs for uniprot_id in accession_ids
                                 if uniprot_id in seq_ids and ec in ecs_ids])
        return all_pairs  # , original

    @classmethod
    def link_to_ec(cls, swissprot=True):
        """Link link to EC table using the UniProt dat files"""
        sequence_ec_pairs = cls._get_sequence_ec_pairs()
        through = Sequence.ecs.through
        existing = set(through.objects.filter(sequence__in=Sequence.objects.reviewed())
                       .values_list('sequence_id', 'ec_id'))
        to_create = sequence_ec_pairs - existing
        print("Creating {} sequence <-> EC links".format(len(to_create)))
        through.objects.bulk_create(
            [through(sequence_id=seq_id, ec_id=ec_id) for seq_id, ec_id in to_create],
            batch_size=5000
        )

        to_delete = existing - sequence_ec_pairs
        print("Deleting {} sequence <-> EC links".format(len(to_delete)))
        for seq_id, ec_id in to_delete:
            through.objects.get(sequence_id=seq_id, ec_id=ec_id).delete()
    
    @classmethod
    def _get_sequence_keyword_pairs(cls, swissprot=True):
        """Find all uniprot <-> keyword pairs in the UniProt dat files"""
        seq_ids = dict(cls.objects.reviewed().get_idx2id())
        kw_ids = Keyword.objects.get_name2id()
        all_pairs = set()
        with gzip.open(files.SWISSPROT_DAT_GZ, 'rt') as swissprot_dat_file:
            for line in swissprot_dat_file:
                kws = []
                if line.startswith("ID"):
                    accession_ids = []
                if line.startswith('AC'):
                    line = line[2:].replace(" ", "").strip(";\n")
                    accession_ids += line.split(";")
                elif line.startswith('KW'):
                    words = line[5:-2].split(";")
                    for word in words:
                        kws.append(word.strip())
                all_pairs.update([(seq_ids[uniprot_id], kw_ids[kw]) for kw in kws for uniprot_id in accession_ids
                                 if uniprot_id in seq_ids])
        return all_pairs
    
    @classmethod
    def link_to_keyword(cls, swissprot=True):
        """Link to keyword table using the UniProt dat files"""
        sequence_kw_pairs = cls._get_sequence_keyword_pairs()
        through = Sequence.keywords.through
        existing = set(through.objects.filter(sequence__in=Sequence.objects.reviewed())
                       .values_list('sequence_id', 'keyword_id'))
        to_create = sequence_kw_pairs - existing
        print("Creating {} sequence <-> Keyword links".format(len(to_create)))
        through.objects.bulk_create([through(sequence_id=seq_id, keyword_id=kw_id) for seq_id, kw_id in to_create],
                                    batch_size=5000)

        to_delete = existing - sequence_kw_pairs
        print("Deleting {} sequence <-> Keyword links".format(len(to_delete)))
        for seq_id, kw_id in to_delete:
            through.objects.get(sequence_id=seq_id, keyword_id=kw_id).delete()

    @classmethod
    def _read_secondary_names(cls):
        name_dict = {}
        with open(files.UNIPROT_SECONDARY_NAMES, 'r') as secondary_names_file:
            for line in secondary_names_file:
                if line.startswith("__________"):
                    break
            for line in secondary_names_file:
                secondary, primary = line.split()
                name_dict[secondary] = primary
        return name_dict

    @classmethod
    def use_primary_accession_id(cls):
        """Make sure only primary accession ids are being used"""
        from common.models import Pdb
        secondary_names_dict = cls._read_secondary_names()
        sequence_ids_values = dict(cls.objects.get_idx2id())
        print(len(secondary_names_dict.keys() & sequence_ids_values.keys()))
        
        for uniprot_id in secondary_names_dict.keys() & sequence_ids_values.keys():
            primary_id = secondary_names_dict[uniprot_id]
            if primary_id not in sequence_ids_values:
                cls.objects.filter(uniprot_id=uniprot_id).update(uniprot_id=primary_id)
                sequence_ids_values[primary_id] = sequence_ids_values[uniprot_id]
            else:
                secondary = cls.objects.get(uniprot_id=uniprot_id)
                primary = cls.objects.get(uniprot_id=primary_id)
                if not Pdb.objects.filter(structure__sequences=secondary) and not secondary.residue_set.all() \
                        and not secondary.protein_set.all() \
                        and not secondary.residue_sequences.filter(is_reference=True):
                    # secondary.residue_sequences.filter(is_reference=False).delete()
                    # print("Deleting secondary accession id: {}".format(secondary), secondary.delete())
                    pass
                else:
                    # check these manually, switch directly the uniprot ids if primary does not link to anything
                    print("Secondary accession id is being used in a protein or has reference residue_sequence")
                    print("secondary:", secondary.protein_set.all())
                    print("primary:", primary.protein_set.all())
                    print()

    @classmethod
    def create_reference_fasta(cls):
        """Creates fasta files in the tmp folder for every reference sequence, deletes existing fasta files"""
        pathlib.Path(files.TMP_FASTA).mkdir(parents=True, exist_ok=True)
        for filename in glob.glob("{}/*.fasta".format(files.TMP_FASTA)):
            os.remove(filename)
        for seq in cls.objects.reference():
            fasta_name = "{}/{}.fasta".format(files.TMP_FASTA, seq.uniprot_id)
            SeqIO.write(seq.get_seq_record(), fasta_name, "fasta")

    @classmethod
    def create_flat_reference_fasta(cls):
        """Creates a fasta file with all the sequence references in the database"""
        records = []
        for seq in cls.objects.reference():
            if len(seq.sequence) >= 20:
                records.append(seq.get_seq_record())
        SeqIO.write(records, files.MCSA_FASTA, "fasta")

    def _set_uniprot_info(self):
        """Set _uniprot_info from the uniprot website"""
        if (not self._uniprot_info) or self._uniprot_id != self.uniprot_id:
            url = "http://www.uniprot.org/uniprot/{0}.txt".format(
                self.uniprot_id)
            self._uniprot_info = urllib.request.urlopen(url).read().decode(
                'utf-8')
        return self._uniprot_info

    def get_name_from_uniprot(self):
        """Get the enzyme name from uniprot website"""
        self._set_uniprot_info()
        for line in self._uniprot_info.splitlines():
            if (line.startswith("DE   RecName") or
                    line.startswith("DE   SubName")) and 'Full=' in line:
                init = line.index('Full=') + 5
                end = line.find('{')
                return line[init:end].strip(' ;')

    def get_description_from_uniprot(self):
        """Get intro text from uniprot website using the first uniprot id"""
        self._set_uniprot_info()
        description = ''
        reading = False
        for line in self._uniprot_info.splitlines():
            if reading:
                if '-!-' in line:
                    break
                description += line[8:]
            if line.startswith('CC   -!- FUNCTION:'):
                description += line[19:].strip('\n')
                reading = True
        description = description[:description.find('{')]
        return description

    def get_pdb_codes_from_sifts(self):
        """Get the PDB codes from a sifts uniprot_pdb file"""
        grep_command = ("grep -P '{0}\\t' {1}".format(
            self.uniprot_id, files.SIFTS_UNIPROT_PDB))
        grep_line = subprocess.run(grep_command, shell=True, stdout=subprocess.PIPE, universal_newlines=True).stdout
        if grep_line:
            pdbs = grep_line.split('\t')[1].strip().split(";")
            return pdbs
        else:
            return []

    def get_seq_record(self):
        """Return a Biopython Seq Record for this object"""
        return SeqRecord(Seq(self.sequence), self.uniprot_id, description=self.name)

    @property
    def uniprot_text_link(self):
        """Returns text URL to this sequence page in UniProt.org"""
        return link.uniprot_id_2_uniprot(self.uniprot_id)

    @property
    def uniprot_icon_link(self):
        """Returns URL with icon for this sequence page in UniProt.org"""
        return link.uniprot_id_2_uniprot(self.uniprot_id, icon=True)

    @property
    def pfam_icon_link(self):
        """Returns URL with icon for this sequence page in the PFAM website"""
        return link.uniprot_id_2_pfam_icon(self.uniprot_id)

    @property
    def funtree_icon_link(self):
        """Returns URL with icon for this sequence page in the funtree website"""
        return link.uniprot_id_2_funtree_icon(self.uniprot_id)

    @property
    def search_link(self):
        """Returns URL for the M-CSA search page searching for this sequence"""
        return link.search_sequence(self.uniprot_id)

    @property
    def intenz_text_links(self):
        """Returns URLs this sequence ecs codes for the ec code page in intenz"""
        return format_html(", ".join([ec.intenz_text_link for ec in self.ecs.all()]))


class PFAMQuerySet(models.QuerySet):
    """Some predefined querysets for the PFAM model"""

    def get_idx2id(self):
        """Returns values list uniprot id -> database id"""
        return self.values_list('idx', 'id')


class PFAM(models.Model):
    """A PFAM domain"""
    name = models.CharField(max_length=1023)
    idx = models.IntegerField(unique=True, db_index=True)

    objects = PFAMQuerySet.as_manager()

    def __repr__(self):
        return "PF{:05d}".format(self.idx)

    def __str__(self):
        return self.__repr__()


class SequenceDomainThrough(models.Model):
    """A base through class for domains that can be mapped into sequences"""
    sequence = models.ForeignKey('Sequence', on_delete=models.CASCADE)
    start = models.IntegerField()
    end = models.IntegerField()

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        abstract = True


class SequenceCATH(SequenceDomainThrough):
    """Through table that links Sequence to CATH"""
    cath = models.ForeignKey("CATH", on_delete=models.CASCADE)
    evalue = models.FloatField()

    objects = models.Manager()

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        unique_together = ("sequence", "cath", "start", "end")

    @classmethod
    def download_gene3d_file(cls):
        """Downloads GENE 3D file from cath"""
        urllib.request.urlretrieve(
            "ftp://orengoftp.biochem.ucl.ac.uk/gene3d/CURRENT_RELEASE/representative_uniprot_genome_assignments.csv.gz",
            files.GENE3D_REPRESENTATIVES
        )
        return

    @classmethod
    def create_objects(cls):
        """Update table using Gene3D file"""
        uniprot_to_id_dict = dict(Sequence.objects.reviewed().get_idx2id())
        cath_to_id_dict = common.models.CATH.objects.complete().get_code2id()
        info_gene3d = set()
        with gzip.open(files.GENE3D_REPRESENTATIVES, 'rt') as cath_file:
            csv_reader = csv.reader(cath_file)
            for no, words in enumerate(csv_reader):
                uniprot, _, _, _, _, cath, all_limits, evalue, _ = words
                for limits in all_limits.split(","):
                    start, end = limits.split("-")
                    try:
                        info_gene3d.add(
                            (uniprot_to_id_dict[uniprot], cath_to_id_dict[cath], int(start), int(end), float(evalue)))
                    except KeyError:
                        continue
        existing_pairs = set(cls.objects.values_list('sequence_id', 'cath_id', 'start', 'end', 'evalue'))
        to_delete = existing_pairs - info_gene3d
        print("Deleting {} cath<->Uniprot links".format(len(to_delete)))
        for t in to_delete:
            cls.objects.get(sequence_id=t[0], cath_id=t[1], start=t[2], end=t[3]).delete()
        to_create = info_gene3d - existing_pairs
        print("Adding {} new cath<->UniProt links".format(len(to_create)))
        cls.objects.bulk_create([cls(sequence_id=t[0], cath_id=t[1], start=t[2], end=t[3], evalue=t[4])
                                 for t in to_create], batch_size=5000)


class SequencePFAM(SequenceDomainThrough):
    """Through table that links Sequence to PFAM"""
    pfam = models.ForeignKey("PFAM", on_delete=models.CASCADE)

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        unique_together = ("sequence", "start", "end", "pfam")

    objects = models.Manager()

    @classmethod
    def create_ipr_index(cls):
        """Creates a index file of the interpro dat_gz file. Check md5sum of the ipr file.
        Only creates new index if there are changes.

        """
        index_file = pathlib.Path(files.INTERPRO_INDEX)
        md5_last_index = pathlib.Path(files.INTERPRO_DAT_GZ_INDEX_MD5)
        md5_current = pathlib.Path(files.INTERPRO_DAT_GZ_MD5)
        assert md5_current.exists(), "Cannot access {}".format(md5_current)
        overwrite = True
        if md5_last_index.exists():
            if md5_last_index.read_text() == md5_current.read_text():
                overwrite = False
        if index_file.exists() and not overwrite:
            print("ipr_dat_gz index already exists")
            return True
        else:
            dat_gz_file = pathlib.Path(files.INTERPRO_DAT_GZ)
            assert dat_gz_file.exists(), "Cannot access {}".format(dat_gz_file)

            command = (
                "zgrep -b '' {}|awk -F'[:\t]' '{{print $1\" \"$2}}'| uniq "
                "--skip-fields=1 > {}".format(dat_gz_file, index_file))
            print("Creating new index for ipr_dat_gz")
            subprocess.call(command, shell=True)
            print("Index for ipr_dat_gz created")
            shutil.copyfile(md5_current, md5_last_index)
            return True

    @classmethod
    def read_ipr_index(cls, uniprot_ids=None):
        """Returns a dict with the start and end positions of the uniprot id
        lines in the ipr_dat_gz file"""
        assert type(uniprot_ids) is set or type(uniprot_ids is dict)
        index = collections.OrderedDict()
        with open(files.INTERPRO_INDEX, 'r') as index_file:
            missing_last = False
            for line in index_file:
                position, uniprot_id = line.split()
                if missing_last:
                    key = next(reversed(index))
                    index[key][1] = int(position) - index[key][0]
                    missing_last = False
                if uniprot_id in uniprot_ids:
                    index[uniprot_id] = [int(position), -1]
                    missing_last = True
        return index

    @classmethod
    def read_dat_gz(cls, index):
        """Returns the relevant lines of the interpro dat file

        :param index is a dict where keys are uniprot_ids and values are a tuple with
         the start and end bytes of that uniprot in the interpro dat gz file.
         dict must be ordered
        :rtype: string
        :return: the lines of ipr_dat_gz file with the right uniprot ids
        """
        with gzip.open(files.INTERPRO_DAT_GZ, 'rb') as dat_file:
            previous_end = 0
            data = []
            no = 0
            for uniprot_id, start_end in index.items():
                jump = start_end[0] - previous_end
                no += 1
                dat_file.seek(jump, 1)
                data.append(dat_file.read(start_end[1]))
                previous_end = sum(start_end)
        return b"".join(data).decode("utf-8")

    @classmethod
    def create_objects(cls):
        """Creates SequencePFAM objects using InterPro files"""
        is_index_file_ok = cls.create_ipr_index()
        assert is_index_file_ok, "Cannot read or create interpro index file"
        sequence_idx2id = dict(Sequence.objects.reviewed().get_idx2id())
        index = cls.read_ipr_index(sequence_idx2id)
        print("Reading ipr_dat_gz file")
        data = cls.read_dat_gz(index)
        print("Finished reading ipr_dat_gz file")
        # first add missing pfams
        pfam_idx2id = dict(PFAM.objects.get_idx2id())
        lines_to_add = []
        pfams_to_add = {}
        for line in data.splitlines():
            words = line.split("\t")
            if words[3].startswith("PF"):
                lines_to_add.append(words)
                idx = int(words[3].replace('PF', ''))
                if idx not in pfam_idx2id and idx not in pfams_to_add:
                    pfams_to_add[idx] = PFAM(idx=idx, name=words[2])
        print("Creating {} pfam ids".format(len(pfams_to_add)))
        PFAM.objects.bulk_create(pfams_to_add.values())
        # then add links to sequence
        sequence_pfam_to_add = []
        pfam_idx2id = dict(PFAM.objects.get_idx2id())
        # noinspection PyUnresolvedReferences
        existing = set(cls.get_unique_together_values())
        for uniprot_id, _, _, pfam_idx, start, end in lines_to_add:
            start = int(start)
            end = int(end)
            sequence_id = sequence_idx2id[uniprot_id]
            pfam_id = pfam_idx2id[int(pfam_idx.replace('PF', ''))]
            if (sequence_id, start, end, pfam_id) not in existing:
                sequence_pfam_to_add.append(
                    cls(sequence_id=sequence_id, pfam_id=pfam_id,
                        start=start, end=end)
                )
        print("Creating {} sequence-pfam links".format(len(sequence_pfam_to_add)))
        cls.objects.bulk_create(sequence_pfam_to_add)


class MultiAlignmentQuerySet(models.QuerySet):
    """Some predefined querysets for the MultiAlignment model"""

    def with_reference(self, uniprot_id):
        """Returns all MultiAlignment objects values that have this uniprot as reference"""
        return self.filter(reference_sequence__uniprot_id=uniprot_id)

    def reference(self, is_reference=True):
        """Returns reference MultiAlignment objects (sequence is the same as reference sequences)"""
        if is_reference:
            return self.filter(reference_sequence=F("sequence"))
        else:
            return self.exclude(reference_sequence=F("sequence"))

    def annotate_with_ref(self):
        """Annotate with ref.alignment and ref.left_pad"""
        ref_subquery = MultiAlignment.objects.filter(
            reference_sequence__in=OuterRef('reference_sequence'), sequence__in=OuterRef('reference_sequence'))

        return self.annotate(
            ref_alignment=Subquery(ref_subquery.values("alignment")[:1]),
            ref_pad_left=Subquery(ref_subquery.values("pad_left")[:1])
            )

class MultiAlignment(models.Model):
    """A multi sequence alignment model that is created for every reference sequence in M-CSA"""
    reference_sequence = models.ForeignKey("Sequence", on_delete=models.CASCADE, related_name="reference_alignment")
    sequence = models.ForeignKey("Sequence", on_delete=models.CASCADE, related_name="alignment")
    pad_left = models.IntegerField()
    alignment = models.TextField()
    identity = models.FloatField()
    coverage = models.FloatField()


    objects = MultiAlignmentQuerySet.as_manager()

    # noinspection PyMissingOrEmptyDocstring
    class Meta:
        unique_together = ("sequence", "reference_sequence")

    def __repr__(self):
        return "{}-{}".format(self.reference_sequence, self.sequence)

    def __str__(self):
        return self.__repr__()

    @classmethod
    def create_sp_ref_fasta(cls):
        """Puts all the SwissProt and reference UniProt sequences in the same fasta file to be searched by phmmer"""
        # decompresses the swissprot fasta file 
        subprocess.call("gzip -d -c {} > {} ".format(files.SWISSPROT_FASTA, files.SP_REF_FASTA), shell=True)

        # adds all the proteins in the reference proteomes files to that file
        # TODO include only reference proteomes that have at least one structure in swissprot
        reference_with_swissprot = set(Sequence.objects.filter(in_reference_proteome=True, in_swissprot=True)\
                .values_list("species__taxid", flat=True))

        all_files = glob.glob("/nfs/ftp/pub/databases/uniprot/current_release/knowledgebase/reference_proteomes/*/*fasta.gz")
        for filepath in all_files:
            filename = pathlib.Path(filepath).name
            if not filename.endswith("_DNA.fasta.gz"):
                taxid = filename.split("_")[1].split(".")[0]
                if int(taxid) in reference_with_swissprot:
                    subprocess.call('gzip -d -c {} >> {} '.format(filepath, files.SP_REF_FASTA), shell=True)

    @property
    def reference_alignment(self):
        """returns reference alignment"""
        # annotate with reference alignment
        try:
            return self.ref_alignment
        except AttributeError:
            return MultiAlignment.objects.get(
                sequence=self.reference_sequence, reference_sequence=self.reference_sequence).alignment

    @property
    def reference_pad_left(self):
        """returns reference alignment"""
        # annotate with reference alignment
        try:
            return self.ref_pad_left
        except AttributeError:
            return MultiAlignment.objects.get(
                sequence=self.reference_sequence, reference_sequence=self.reference_sequence).pad_left

    def calculate_identity(self):
        """calculate identity between the two sequences alignment"""
        pad_diff = self.reference_pad_left - self.pad_left
        return sum(c1 == c2 != "-" for c1, c2 in zip(
            self.alignment[-pad_diff if pad_diff < 0 else 0:],
            self.reference_alignment[pad_diff if pad_diff > 0 else 0:])) / \
                (len(self.reference_alignment)-self.reference_alignment.count("-"))

    # @property
    # def ref_alignment_str(self):
    #     return "-"*self.reference_pad_left + self.reference_alignment

    # @property
    # def alignment_str(self):
    #     return "-"*self.pad_left + self.alignment

    # def calculate_coverage(self):
    #     """calculate identity between the two sequences alignment"""

    #     return (len(self.alignment_str[:len(self.ref_alignment_str)]) - 
    #             self.alignment_str[:len(self.ref_alignment_str)].count("-"))/(
    #         len(self.ref_alignment_str) - self.ref_alignment_str.count("-"))

    def get_non_reference_resid_codes(self, ref_resids0_in_ma):
        """Returns a list of tuples with resid and 3letter code"""
        output = []
        for ref_resid0_in_ma in ref_resids0_in_ma:
            if ref_resid0_in_ma is None:
                out = (None, 'out')
            else:
                resid0 = ref_resid0_in_ma-self.pad_left
                if resid0 < 0:
                    out = (None, '')
                elif resid0 >= len(str(self.alignment)):
                    out = (None, '')
                elif self.alignment[resid0] == '-':
                    out = (None, '')
                else:
                    resid = resid0 - self.alignment.count('-', 0, resid0) + 1
                    code1 = self.alignment[resid0]
                    try:
                        code3 = others.AA_1TO3[code1]
                    except KeyError:
                        code3 = ''
                    out = (resid, code3)
            output.append(out)
        return output

    @classmethod
    def create_objects(cls):
        """"Reads output mafft files and adds alignments to database """
        msa_output_files = glob.glob(files.MSA_FASTA_OUT + "/*.out.fasta")
        upid_to_dbid = dict(Sequence.objects.get_idx2id())
        sp_uniprot_ids = set(Sequence.objects.reviewed().values_list("uniprot_id", flat=True))
        for msa_output_file in msa_output_files:
            db.reset_queries()
            print(msa_output_file)
            if os.path.getsize(msa_output_file) == 0:
                continue
            reference_uniprot_id = os.path.basename(msa_output_file).replace(".out.fasta", "").split("_")[0]
            current_objects = cls.objects.with_reference(reference_uniprot_id)
            seqs = SeqIO.index(msa_output_file, 'fasta')
            sp_idxs = [idx for idx in seqs if idx in sp_uniprot_ids]
            if reference_uniprot_id not in sp_idxs:
                sp_idxs.insert(0, reference_uniprot_id)
            print("Read {} sequences from MSA file. {} in SwissProt".format(len(seqs), len(sp_idxs)))
            # Skip multiple sequence alignments that were already added to the database
            if set(sp_idxs) == set(current_objects.exclude(alignment="").values_list("sequence__uniprot_id", flat=True)):
                print("Already added, Skipping")
                continue
            print("Deleted: ", current_objects.filter(sequence__uniprot_id__in=sp_idxs).delete())
            alignments_to_create = []
            for no, seq_key in enumerate(sp_idxs):
                seq = seqs[seq_key]
                seq_lstrip = seq.seq.lstrip('-')
                pad_left = len(seq.seq) - len(seq_lstrip)
                seq_strip = seq_lstrip.rstrip('-')
                alignments_to_create.append(cls(
                    reference_sequence_id=upid_to_dbid[reference_uniprot_id],
                    sequence_id=upid_to_dbid[seq_key],
                    pad_left=pad_left,
                    alignment=seq_strip,
                    identity=0,
                    coverage=0,
                    
                ))
            print("creating {} alignments".format(len(alignments_to_create)))
            cls.objects.bulk_create(alignments_to_create, batch_size=1000)


class Interpro(models.Model):
    """A Interpro accession number"""
    interpro_id = models.IntegerField(unique=True, verbose_name="Interpro ID")
    uniprots = models.ManyToManyField("Sequence", related_name="interpros")
    name = models.CharField(max_length=1023)
    TYPE_CHOICES = (
        ("family", "Family"),
        ("homologous_superfamily", "Homologous Superfamily"),
        ("domain", "Domain"),
        ("ptm", "PTM"),
        ("repeat", "Repeat"),
        ("active_site", "Active Site"),
        ("binding_site", "Binding Site"),
        ("conserved_site", "Conserved Site"),
    )
    type = models.CharField(max_length=127, choices=TYPE_CHOICES)

    def __str__(self):
        return "IPR{:06d}".format(self.interpro_id)

    @classmethod
    def create_objects(cls):
        """Creates Interpro objects using information in interpro/entry.list"""
        existing = set(cls.objects.values_list("interpro_id", "name", "type"))
        with open(files.INTERPRO_ENTRY_LIST, "r") as entry_list_file:
            for line in entry_list_file:
                if line.startswith("IPR"):
                    ipr, entry_type, name = line.split("\t")
                    ipr = int(ipr[3:])
                    entry_type = entry_type.lower()
                    if (ipr, name, entry_type) not in existing:
                        cls.objects.update_or_create(interpro_id=ipr, defaults={"name": name, "type": entry_type})


class KeywordQuerySet(models.QuerySet):
    """Some predefined querysets for the CATH model"""

    def get_name2id(self):
        """Returns a dictionary of Keyword to database id"""
        return {kw.name: kw.id for kw in self}
    
    def catalytic(self):
        """Filter catalytic keywords using is_catalytic field"""
        return self.filter(is_catalytic=True)

    def filter_catalytic(self):
        """Filter all catalytic keywords"""
        top_catalytic = Keyword.objects.filter(name__in=[
            "Oxidoreductase", "Transferase", "Hydrolase", "Lyase", "Isomerase", "Ligase", "Translocase",
            "Motor protein",
            #"G-protein coupled receptor",
            "ATP synthesis",
            "Electron transport",
            #"Photoreceptor protein",
            #"Photoprotein", 
        ])
        return self.children_of(top_catalytic) | top_catalytic

    def children_of(self, parents):
        """Filter all keywords that are children of these parents
        This method is iterative"""
        keywords = parents.values_list("id", flat=True)
        ids = set(keywords)
        while keywords.count():
            keywords = Keyword.objects.filter(parents__in=keywords).values_list("id", flat=True)
            ids.update(keywords)
        return self.filter(id__in=ids)


class Keyword(models.Model):
    """A Swissprot knowledge base keyword"""
    idx = models.IntegerField(unique=True)
    name = models.CharField(max_length=1023, blank=False, unique=True)
    description = models.TextField()
    category = models.CharField(max_length=255, blank=False)
    parents = models.ManyToManyField("self", related_name="children", symmetrical=False)
    is_catalytic = models.BooleanField(db_index=True, default=False)

    objects = KeywordQuerySet.as_manager()

    def __repr__(self):
        return "KW-{:04d} {}".format(self.idx, self.name)

    def __str__(self):
        return self.__repr__()

    @classmethod
    def read_keyword_data(cls):
        """Reads keyword_data into a dictionairy"""
        data = collections.OrderedDict()
        with open(files.KEYWORDS, "r") as kw_file:
            for line in kw_file:
                if line.startswith("ID"):
                    name = line[5:].strip(".\n")
                    data[name] = collections.defaultdict(list)
                elif line.startswith("AC"):
                    data[name]["idx"] = int(line.split("-")[1].strip())
                elif line.startswith("DE"):
                    data[name]["description_parts"].append(line[5:].strip("\n"))
                elif line.startswith("CA"):
                    data[name]["category"] = line[5:].strip(".\n")
                elif line.startswith("HI"):
                    words = line[5:].strip(".\n").replace(":", ";").split(";")
                    data[name]["parents"].append(words[-2].strip())
        return data
           
    @classmethod
    def create_objects(cls):
        """Creates Keyword objects using information in keywlist.txt"""
        data = cls.read_keyword_data()
        for name, values in data.items():
            kw, _ = cls.objects.update_or_create(
                name=name,
                idx=data[name]["idx"],
                defaults={
                    "description": " ".join(data[name]["description_parts"]),
                    "category": data[name]["category"],
                }
            )
            for parent_name in data[name]["parents"]:
                parent = cls.objects.filter(name=parent_name)
                if parent:
                    kw.parents.add(parent.first())
        
        """sets the is_catalytic bool field"""
        cls.objects.filter_catalytic().update(is_catalytic=True)




#ignore_pfams = Sequence.objects.filter(pfams__idx__in=[
#    12, #"Heat shock protein 70 family",
#    118, #"Chaperonin Cpn60/TCP-1 family"
#    9, # "Transcription factor, GTP-binding domain"
#    550, # "Phosphopantetheine binding ACP domain"
#    1228, # "Glycine radical domain"
#    4, # "ATPase, AAA-type, core"
#    25, # "Small GTPase superfamily, ARF/SAR type"
#    148, # "Guanine nucleotide binding protein (G-protein)"
#    503, # 'Nitrogenase/oxidoreductase, component 1'
#    2337, 607, # pol-gag
#    ])

