# coding=utf-8
"""Contains the Species models and related classes"""

# django standard imports
import json
from urllib.request import urlopen
from urllib.error import HTTPError

# django imports
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.urls import reverse
from django.utils.html import format_html


class SpeciesQuerySet(models.QuerySet):
    """Some predefined querysets for the Species model"""

    def get_idx2id(self):
        """Returns values list ncbi taxid -> database id"""
        return self.values_list('taxid', 'id')


class Species(models.Model):
    """A biological species"""
    scientific_name = models.CharField(max_length=255, blank=False, verbose_name="Species")
    common_name = models.CharField(max_length=255, blank=True, verbose_name="Species Common")
    type_of_life = models.CharField(max_length=255, blank=True, verbose_name="Type of Life")
    taxid = models.IntegerField(verbose_name="NCBI TaxID", unique=True)
    lineage = ArrayField(models.CharField(max_length=255, blank=True), verbose_name="Lineage")

    objects = SpeciesQuerySet.as_manager()

    class Meta:
        verbose_name_plural = "Species"
        ordering = ['scientific_name']

    def __str__(self):
        output = self.scientific_name
        if self.common_name:
            output += " ({})".format(self.common_name)
        return output

    @classmethod
    def give_common_name(cls):
        species_missing_common = cls.objects.filter(common_name="")
        species_with_common = cls.objects.exclude(common_name="")
        print(species_missing_common.count())
        for species in species_missing_common:
            print(species, species.taxid, species.lineage)

            if species.lineage[0] in ["Bacteria", "Archaea"]:
                species.common_name = species.lineage[0]
                species.save()
            elif species.lineage[0] == "Viruses":
                species.common_name = "Virus"
                species.save()
            elif len(species.lineage) > 1 and species.lineage[1] == "Fungi":
                species.common_name = "Fungus"
                species.save()
            similar_species = species_with_common.filter(lineage=species.lineage)
            if similar_species:
                species.common_name = similar_species.first().common_name
                species.save()

    @classmethod
    def give_missing_type_of_life(cls):
        species_missing_type = cls.objects.filter(type_of_life="")
        print(species_missing_type.count())
        type_of_life_dict = {
            (0, "Bacteria"): "Bacteria",
            (1, "Fungi"): "Fungi",
            (0, "Archaea"): "Archaea",
            (0, "Viruses"): "Virus",
            (1, "Viridiplantae"): "Green Plants",
            (12, "Aves"): "Birds",
            (5, "Insecta"): "Insects",
            (6, "Mammalia"): "Mammals",
            (6, "Lepidosauria"): "Reptiles",
            (3, "Nematoda"): "Worms",
            (1, "Amoebozoa"): "Amoebea",
            (5, "Arachnida"): "Spiders",
            (6, "Amphibia"): "Amphibians",
            (5, "Cyclostomata"): "Fish",
            (5, "Chondrichthyes"): "Fish",
        }

        for species in species_missing_type:
            for key, value in type_of_life_dict.items():
                try:
                    if species.lineage[key[0]]==key[1]:
                        species.type_of_life = value
                        species.save()
                        print(value)
                        break
                except IndexError:
                    pass

    def get_from_ena(self, keys):
        """Get value from ena json dict using the taxid"""
        try:
            url = "http://www.ebi.ac.uk/ena/data/taxonomy/v1/taxon/tax-id/{}" \
                .format(self.taxid)
            response = urlopen(url)
            data = json.loads(response.read().decode('utf8'))
        except HTTPError:
            return ['']*len(keys)
        result = []
        for key in keys:
            try:
                value = data[key]
            except KeyError:
                value = ''
            result.append(value)
        return result

    @property
    def html_formatted_str(self):
        output = "<i>{}</i>".format(self.scientific_name)
        if self.common_name:
            output += " ({})".format(self.common_name)
        return format_html(output)

    @property
    def formatted_common_name(self):
        if self.common_name:
            return "({})".format(self.common_name)
        else:
            return ''

    # TODO change links to use new style
    @property
    def search_html_link(self):
        return format_html(
            '<a class="search" href="{0}?s={1}'
            '&species=on" '
            'title="Search for Entries in M-CSA with the same Species">'
            '<i>{2.scientific_name}</i></a> {2.formatted_common_name}'.format(
                reverse('common:search'),
                '+'.join(self.scientific_name.split()),
                self,
            ))

    @property
    def uniprot_icon_link(self):
        return format_html(
            '<a class="icon-link external-link" '
            'href="http://www.uniprot.org/taxonomy/{0.taxid}" '
            'target="_blank" title="Open in Uniprot/Taxonomy">'
            '<img class="inline-icon" src="/thornton-srv/m-csa/static/icons/'
            'uniprot.png" alt="Uniprot"></a>'.format(
                self))

    @property
    def uniprot_text_link(self):
        return format_html(
            '<a class="external-link" '
            'href="http://www.uniprot.org/taxonomy/{0.taxid}" '
            'target="_blank" title="Open in Uniprot/Taxonomy">'
            '{0.scientific_name}</a>'.format(
                self))

    def save(self, *args, **kwargs):
        if not self.pk:
            if not self.scientific_name or not self.common_name \
                    or not self.lineage:
                scientific_name, common_name, lineage = \
                    self.get_from_ena([
                        'scientificName', 'commonName', 'lineage'
                    ])
                lineage = lineage.split("; ")[:-1]
                if not self.scientific_name:
                    self.scientific_name = scientific_name
                if not self.common_name:
                    self.common_name = common_name
                if not self.lineage:
                    self.lineage = lineage
        super(Species, self).save(*args, **kwargs)
