# coding=utf-8
"""Contains the model for Step and related code"""

# python imports
import collections
import logging
from lxml import etree
import subprocess
from math import isclose
from rdkit.Chem.rdchem import RWMol, Atom, BondType, BondDir
from rdkit.Chem import AllChem
from rdkit.Geometry.rdGeometry import Point2D
from rdkit import Chem
import networkx as nx
from typing import Union, Dict, FrozenSet, List, DefaultDict, Optional, Tuple, Set

# django imports
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import ContentFile
from django.db import models
from django.db.models import Prefetch
from django.utils.functional import cached_property

from gm2m import GM2MField
from simple_history.models import HistoricalRecords

# atlas imports
import common.models
from common.others import (get_maximum_template_substructure,
                           METHOD_TO_RESIDUE_FUNCTIONS, is_heterolytic_bond_cleavage)
from common.models import ActiveSiteCofactor, CompoundType, Residue
import common.others
import common.utils

logger = logging.getLogger("script")
ArrowInfo = Tuple[FrozenSet[int], FrozenSet[int], int]
MoleculeDB = Union[Residue, CompoundType, ActiveSiteCofactor]
NonResidueMoleculeDB = Union[CompoundType, ActiveSiteCofactor]


def atom_array_to_atom_tags(root):
    tags = root.keys()
    tags.remove("atomID")
    for no, atomid in enumerate(root.attrib["atomID"].split()):
        atom = etree.Element("atom", id=atomid)
        for tag in tags:
            root_value = root.attrib[tag].split()[no]
            if root_value != "0":
                atom.attrib[tag] = root_value
        root.append(atom)
    for key in root.keys():
        root.attrib.pop(key)
    return root


class StepBondChange(models.Model):
    """Step level reaction bond changes"""
    # not very usefull, use Rules or fingerprint instead
    step = models.ForeignKey("Step", on_delete=models.CASCADE, verbose_name="Step")
    element1 = models.CharField(verbose_name="Element 1", max_length=2)
    element2 = models.CharField(verbose_name="Element 2", max_length=2)
    idx1 = models.CharField(verbose_name="Index 1", max_length=10)
    idx2 = models.CharField(verbose_name="Index 2", max_length=10)
    initial_order = models.CharField(verbose_name="Initial Bond Order", max_length=3)
    final_order = models.CharField(verbose_name="Final Bond Order", max_length=3)
    automated = models.BooleanField()
    history = HistoricalRecords()

    objects = models.Manager()

    @property
    def serial1(self):
        return int(str(self.idx1))+1

    @property
    def serial2(self):
        return int(str(self.idx2))+1

    @property
    def atom_1(self):
        return "{0.element1}.{0.serial1}".format(self)

    @property
    def atom_2(self):
        return "{0.element2}.{0.serial2}".format(self)

    def __str__(self):
        return "{0.atom_1} {0.atom_2} ({0.initial_order})->({0.final_order})".format(self)


class StepQuerySet(models.QuerySet):
    """Some predefined querysets for the Step model"""

    def functions(self, functions):
        """Filters Steps that have this role"""
        return self.filter(role__function__in=functions).distinct()

    def residues(self, residues):
        """Filters Steps that have this residue"""
        return self.filter(role__residue_roles__in=residues)

    def roles(self, roles):
        """Filters Steps that have this role"""
        return self.filter(roles__in=roles)

    def mechanisms(self, mechanisms):
        """Filters steps that belong to this mechanism"""
        return self.filter(mechanism__in=mechanisms)

    def annotate_roles(self):
        """Annotates with the residue roles."""
        return self.prefetch_related(
                Prefetch(
                    "roles",
                    queryset=common.models.Role.objects.prefetch_related("function__group"),
                    to_attr="ann_roles"))

    def annotate_activesite(self):
        """Add activesite to query"""
        return self.prefetch_related(
            Prefetch(
                "mechanism__reaction__entry__activesite_set",
                queryset=common.models.ActiveSite.objects.filter(),
                to_attr="pf_activesite"
            )
        )


class Step(models.Model):
    """A chemical step that is part of a mechanism proposal"""
    mechanism = models.ForeignKey("Mechanism", on_delete=models.CASCADE)
    step_id = models.IntegerField(verbose_name="Step ID")
    description = models.TextField(blank=True, verbose_name="Description")
    comment = models.TextField(blank=True)
    scheme = models.FileField(upload_to='schemes/')
    backup_scheme = models.FileField(upload_to='schemes/', default='', blank=True)
    scheme_product = models.FileField(upload_to='schemes/')
    scheme_svg = models.FileField(upload_to='schemes_svg/', default='', blank=True)
    structure_initial = models.FileField(blank=True)
    structure_final = models.FileField(blank=True)
    movie = models.FileField(blank=True)
    residues_roles = GM2MField(blank=True, through='Role')
    REVERSIBLE_CHOICES = (
        ("Unknown", "Unknown"),
        ("yes", "yes"),
        ("no", "no"),
    )
    reversible = models.CharField(max_length=10, choices=REVERSIBLE_CHOICES)
    is_product = models.BooleanField(verbose_name="Product", default=False)
    history = HistoricalRecords()

    objects = StepQuerySet.as_manager()

    class Meta:
        unique_together = ('mechanism', 'step_id')
        ordering = ["mechanism", "step_id"]

    def __str__(self):
        return (
            "Step: {0.mechanism.reaction.entry.owner.username}."
            "{0.mechanism.reaction.entry.macie_id}.{0.mechanism.mechanism_id}."
            "{0.step_id}".format(self)
        )

    def clean(self):
        if self.step_id:
            return self.step_id
        else:
            current_steps = Step.objects.filter(
                mechanism=self.mechanism).order_by('-step_id')
            if current_steps:
                self.step_id = current_steps[0].step_id + 1
            else:
                self.step_id = 1

    @cached_property
    def scheme_xml_namespace(self):
        """return the namespace of the scheme xml file"""
        return "{{{0}}}".format(self._get_scheme_xml_root().getroot().nsmap.get(None, ""))

    # scheme_namespace = "{http://www.chemaxon.com}"
    molconvert_namespace = "{http://www.chemaxon.com}"

    def _detect_array_mode(self):
        """Return True if atoms in scheme are in array mode, False otherwise"""
        if not self.scheme:
            return False
        self.scheme.seek(0)
        root = etree.parse(self.scheme).getroot()
        atom_array = root.findall(".//{0}atomArray".format(
            self.molconvert_namespace))[0]
        return bool("atomID" in atom_array.keys())

    def _atom_array_to_atom_tags(self):
        """Convert scheme from atom array mode to atom tags mode"""
        if self._detect_array_mode():
            self.scheme.seek(0)
            root = etree.parse(self.scheme).getroot()
            atom_array = root.findall(".//{0}atomArray".format(
                self.molconvert_namespace))[0]
            atom_array_to_atom_tags(atom_array)
            self.scheme.save(
                "{}.mrv".format(self),
                ContentFile(etree.tostring(root, pretty_print=True))
            )

    def _get_scheme_xml_root(self, product: bool = False) -> etree.ElementTree:
        """Return the etree object of the scheme or scheme_products file."""
        return self._scheme_product_xml_root if product else self._scheme_xml_root

    @cached_property
    def _scheme_xml_root(self) -> etree.ElementTree:
        """Return the etree object of the scheme file."""
        return etree.parse(self.scheme)

    @cached_property
    def _scheme_product_xml_root(self) -> etree.ElementTree:
        """Return the etree object of the scheme product file."""
        return etree.parse(self.scheme_product)

    def _get_scheme_xml_nodes(self, node_type: str, product: bool = False) -> List[etree.Element]:
        """
        Return a list of nodes in the xml with this node type

        :param node_type: the type of the nodes to be selected
        :param product: if True reads the xml of scheme_product
        """
        root = self._get_scheme_xml_root(product=product)
        return root.findall(".//{0}{1}".format(self.scheme_xml_namespace, node_type))

    def _get_scheme_xml_nodes_attribute(self, node_type: str, attrib: str) -> List[Optional[str]]:
        """
        return a list of this attribute for this nodes

        it will be in the same order as the nodes from _get_scheme_xml_nodes

        :param node_type: the type of nodes to be selected
        :param attrib: the attribute of each node to be selected
        """
        return [n.get(attrib) for n in self._get_scheme_xml_nodes(node_type)]

    def _get_scheme_xml_bond_info(self, coord: bool = True) -> Dict[FrozenSet[int], Dict[str, Union[str, int]]]:
        """
        return a dict with bond order information and stereo direction

        return looks like {frozenset(1,4): {"order": 1, "direction": "W"}}
        where the numbers in frozen set are the atom indexes

        :param coord: whether it should include coordination bonds (to metals)
        """
        bond_info: Dict[FrozenSet[int], Dict[str, Union[str, int]]] = {}
        bonds = self._get_scheme_xml_nodes("bond")
        for bond in bonds:
            if (not coord) and (bond.get("convention") == "cxn:coord" or bond.get("queryType") == "Any"):
                continue
            indexes_set = frozenset([int(no)-1 for no in bond.get("atomRefs2").replace("a", "").split()])
            bond_info[indexes_set] = {}
            stereo = bond.find("{}bondStereo".format(self.scheme_xml_namespace))
            if stereo is not None:
                bond_info[indexes_set]["direction"] = stereo.text
            bond_info[indexes_set]["order"] = int(bond.get("order", 1))  # default 1 for metal coordination bonds
        return bond_info

    @cached_property
    def scheme_xml_arrows_info(self) -> Set[ArrowInfo]:
        """
        return a set with information of each arrow in the scheme xml file.

        each arrow info looks like (frozenset({1, 2}), frozenset({2}), 2) where
        the first set are the atom map number for the atoms in the base of the arrow
        the second set are the atom map number for the atoms in the tip of the arrow
        the int in the last position of the tuple is the number of electrons
        returned as a set since the order of the arrows is arbitrary
        the scheme_xml_ordered_arrows_info function sorts this output if necessary
        """
        mrv_arrows = self._get_scheme_xml_nodes("MEFlow")
        arrows_info = set()
        for mrv_arrow in mrv_arrows:
            ends_info = []
            for end in mrv_arrow.getchildren():
                end_idx_set = set()
                for atom_info in end.values()[0].split():
                    atom_map = int(atom_info.split(".")[1].replace("a", ""))
                    end_idx_set.add(atom_map)
                ends_info.append(frozenset(end_idx_set))
            electrons = 2 if mrv_arrow.get('headFlags') is None else 1
            arrows_info.add((ends_info[0], ends_info[1], electrons))
        return arrows_info

    def _get_bond_index_order_changes(self):
        """Return list of tuples with the effect of the reaction on bonds

        each tuple corresponds to a change in bond order between two atoms:
        (atom1, atom2, change_in_atom1_charge, change_in_atom2_charge,
            change_in_bond_order)
        """
        reactions = self._get_reaction_indexes_from_scheme()
        changes = []
        for reaction in reactions:
            these_changes = None
            beg, end, no_electrons = reaction
            if reaction[2] == 1:  # single arrow
                atom_change = 0
                bond_change = 0.5
            else:
                bond_change = atom_change = 1

            if len(beg) == len(end) == 1:
                # charge transfer #TODO test
                if no_electrons == 2:
                    these_changes = [(beg.pop(), end.pop(), +2, -2, 0)]
                if no_electrons == 1:
                    these_changes = [(beg.pop(), end.pop(), +1, -1, 0)]
            elif len(beg) == 2 and len(end) == 1 and end[0] in beg:
                end_atom = end[0]
                beg.remove(end_atom)
                beg_atom = beg[0]
                these_changes = [(beg_atom, end_atom, +atom_change, -atom_change, -bond_change)]
            elif len(beg) == 1 and len(end) == 2 and beg[0] in end:
                beg_atom = beg[0]
                end.remove(beg_atom)
                end_atom = end[0]
                these_changes = [(beg_atom, end_atom, +atom_change, -atom_change, +bond_change)]
            elif len(beg) == len(end) == 2:
                common = set(beg).intersection(set(end)).pop()
                beg.remove(common)
                beg_atom = beg[0]
                end.remove(common)
                end_atom = end[0]
                these_changes = [
                    (beg_atom, common, +atom_change, 0, -bond_change),
                    (common, end_atom, 0, -atom_change, +bond_change)
                ]
            if these_changes:
                changes.extend(these_changes)
        return changes

    def draft_scheme_first_step(self):
        """Get mrv scheme from the reactants, residues, and cofactors"""
        entry = self.mechanism.reaction.entry
        active_site = entry.activesite_set.first()
        residues = common.models.Residue.objects.filter(active_site=active_site)\
            .annotate_res_sequences().annotate_res_chains()
        reactants = common.models.CompoundType.objects.filter(reaction=self.mechanism.reaction, type='reactant')\
            .select_related("compound")
        cofactors = ActiveSiteCofactor.objects.filter(active_site=active_site)
        mdl_list = []
        for molecule_queryset in [cofactors, reactants]:
            for molecule in molecule_queryset:
                this_mdl_str = molecule.get_mdl_str()
                for _ in range(molecule.count):
                    mdl_list.append(this_mdl_str)
        for residue in residues:
            mdl_list.append(residue.get_mdl_str())
        mdl_str = "".join(mdl_list).replace("'", "")
        # print(mdl_str)

        mrv_str = subprocess.getoutput("echo '{0}' | molconvert mrv -U -2".format(mdl_str))
        #  remove chebi:xxx from substrate and add it to mrvExtraLabel
        mrv_str = mrv_str.replace("\\u0020", " ")
        xml = bytes(bytearray(mrv_str, encoding="utf-8"))
        root = etree.fromstring(xml)
        atom_array = root.findall(".//{}atomArray".format(
            self.molconvert_namespace))[0]
        if "atomID" in atom_array.keys():
            atom_array = atom_array_to_atom_tags(atom_array)
        for atom in atom_array:
            try:
                alias = atom.attrib['mrvAlias']
                if alias.startswith("chebi") or alias.startswith("res"):
                    atom.attrib['mrvExtraLabel'] = alias
                    atom.attrib['mrvAlias'] = ""
            except KeyError:
                pass
        return etree.tostring(root)

    def _find_xml_bond(self, idx1, idx2):
        """Return the xml bond formed between two atoms"""
        root = self._get_scheme_xml_root()
        option1 = root.findall(".//{0}bond[@atomRefs2='a{1} a{2}']".format(
            self.scheme_xml_namespace, idx1+1, idx2+1))
        option2 = root.findall(".//{0}bond[@atomRefs2='a{1} a{2}']".format(
            self.scheme_xml_namespace, idx2+1, idx1+1))
        if option1:
            return option1[0]
        elif option2:
            return option2[0]
        else:
            return None

    radical_options = [None, "monovalent", "divalent", "trivalent quartet", "4 radical", "divalent1"]

    def xml_atom_add_charge(self, atom, charge):
        """Change the formalCharge of this atom by the 'charge' amount"""
        if not atom.get("formalCharge"):
            atom.attrib["formalCharge"] = '0'
        atom.attrib['formalCharge'] = str(int(atom.attrib['formalCharge']) + charge)
        return atom

    def xml_atom_add_radical(self, atom, add=True):
        """Adds or removes a radical from this atom"""
        atom.attrib['radical'] = 'monovalent' if atom.get('radical') in {None, ''} else ''
        self.xml_atom_add_charge(atom, -1 if add else 1)

    def xml_apply_arrows_on_atoms(self):
        """Calculates the effects of the arrows on the atoms and changes them accordingly"""
        xml_atoms = self._get_scheme_xml_nodes("atom")
        for no, arrow in enumerate(self.scheme_xml_arrows_info):
            group1, group2, no_electrons = arrow
            if no_electrons == 1:  # half arrow
                if len(group1) == 1:
                    self.xml_atom_add_radical(xml_atoms[next(iter(group1))-1], add=False)
                if len(group2) == 1:
                    self.xml_atom_add_radical(xml_atoms[next(iter(group2))-1], add=True)
                # this is necessary when there are equal single headed arrows (eg. m145 mechanism2 step4)
                # TODO change to double arrow
                # if arrows.count(arrow) == 2 and no == arrows.index(arrow):
                    # self.xml_atom_add_charge(xml_atoms[set(group2).difference(group1).pop()], -1)
            else:  # full arrow
                for atom_id in group1:
                    self.xml_atom_add_charge(xml_atoms[atom_id-1], 2//len(group1))
                for atom_id in group2:
                    self.xml_atom_add_charge(xml_atoms[atom_id-1], -2//len(group2))

    def xml_add_new_bond(self, idx1, idx2, order):
        """Adds a new bond to the xml scheme"""
        bond_array = self._get_scheme_xml_root().find(".//{0}bondArray".format(self.scheme_xml_namespace))
        last_bond_id = int(bond_array[-1].get('id').strip('b'))
        xml_bond = etree.Element('bond')
        xml_bond.attrib['id'] = 'b{0}'.format(last_bond_id+1)
        xml_bond.attrib['atomRefs2'] = "a{0} a{1}".format(idx1+1, idx2+1)
        xml_bond.attrib['order'] = str(order)
        bond_array.append(xml_bond)
        return xml_bond

    def xml_change_bond(self, idx1, idx2, order_change):
        """Change scheme xml bond by the order_change amount. Creates new bond if necessary"""
        xml_bond = self._find_xml_bond(idx1, idx2)
        if xml_bond is not None:
            if 'convention' in xml_bond.attrib:  # coordination bond should be considered a zero order bond
                xml_bond.attrib['order'] = "0"
            xml_bond.attrib['order'] = str(float(xml_bond.attrib['order']) + order_change)
        else:  # if order change >0 ?
            xml_bond = self.xml_add_new_bond(idx1, idx2, order_change)
        return xml_bond

    def xml_apply_arrows_on_bonds(self):
        """Calculates the effects of the arrows on the bonds and changes them accordingly."""
        xml_bonds = []
        for group1, group2, no_electrons in self.scheme_xml_arrows_info:
            if len(group1) == 2:
                idx1, idx2 = [atom_map - 1 for atom_map in group1]
                xml_bonds.append(self.xml_change_bond(idx1, idx2, -no_electrons/2))
            if len(group2) == 2:
                idx1, idx2 = [atom_map - 1 for atom_map in group2]
                xml_bonds.append(self.xml_change_bond(idx1, idx2, +no_electrons/2))
        # make bonds order integers and remove bonds with zero order
        for xml_bond in xml_bonds:
            if isclose(float(xml_bond.attrib['order']), 0):
                if xml_bond.getparent() is not None:
                    xml_bond.getparent().remove(xml_bond)
                continue
            xml_bond.attrib['order'] = str(int(float(xml_bond.attrib['order'])))

    def draft_scheme_from_previous_step(self, previous=1):
        """Get mrv str from a previous reaction scheme mrv file"""
        previous_step = Step.objects.get(mechanism=self.mechanism, step_id=self.step_id-previous)
        if not previous_step.scheme:
            return ""
        previous_step.xml_apply_arrows_on_atoms()
        previous_step.xml_apply_arrows_on_bonds()
        for xml_arrow in previous_step._get_scheme_xml_nodes("MEFlow"):
            xml_arrow.getparent().remove(xml_arrow)
        return etree.tostring(previous_step._get_scheme_xml_root(), pretty_print=True)

    def draft_scheme(self):
        """Return draft from scratch if first step or from previous step"""
        return self.draft_scheme_first_step() if self.step_id == 1 else self.draft_scheme_from_previous_step()

    def check_detected_roles(self, only_residues=True):
        """Returns roles that the code detects but are not in the database and vice versa"""
        detected_roles = self.detect_reactant_roles()
        functions = []
        for pair in METHOD_TO_RESIDUE_FUNCTIONS.values():
            functions.extend(pair)
        db_roles = common.models.Role.objects.filter(
            step=self,
            function__function__in=functions,
        )
        if only_residues:
            db_roles = db_roles.filter(residue_ct=ContentType.objects.get(model="residue"))
        overdetected_roles = []
        undetected_roles = db_roles
        for role in detected_roles:
            try:
                detected_in_db = db_roles.get(step=role.step, residue_id=role.residue_id, residue_ct=role.residue_ct,
                                              function=role.function)
                undetected_roles = undetected_roles.exclude(id=detected_in_db.id)
            except common.models.Role.DoesNotExist:
                if role.residue.function_location != "post_translation_modification":
                    overdetected_roles.append(role)
        no_ptm_undetected_roles = []
        for role in undetected_roles:  # exclude(residue_roles__function_location="ptm") does not work
            if role.residue.function_location != "post_translation_modification":
                no_ptm_undetected_roles.append(role)
        return no_ptm_undetected_roles, overdetected_roles

    BOND_CHANGES_FINGERPRINT_KEY = None
    EXPLICIT_ATOMS = ["C", "O", "N", "H", "S", "P"]

    @classmethod
    def get_bond_changes_fingerprint_key(cls):
        """Defines the key for the bond changes fingerprint and returns keys as a list"""
        if cls.BOND_CHANGES_FINGERPRINT_KEY is None:
            cls.BOND_CHANGES_FINGERPRINT_KEY = []
            bond_changes_fingerprint_atoms = cls.EXPLICIT_ATOMS + ["other", ]
            for i, atom1 in enumerate(bond_changes_fingerprint_atoms):
                for atom2 in bond_changes_fingerprint_atoms[i:]:
                    for formed in (True, False):
                        key = (frozenset((atom1, atom2)), formed)
                        cls.BOND_CHANGES_FINGERPRINT_KEY.append(key)
        return cls.BOND_CHANGES_FINGERPRINT_KEY

    def get_bond_changes_fingerprint(self):
        """returns bond changes fingerprint"""
        bond_changes_idx = self._get_bond_index_order_changes()
        bonds = self._get_scheme_xml_bond_info(coord=False)
        elements = self._get_scheme_xml_nodes_attribute("atom", "elementType")
        fingerprint_key = Step.get_bond_changes_fingerprint_key()
        fingerprint = [0]*len(fingerprint_key)
        for bond_change_idx in bond_changes_idx:
            if bond_change_idx[-1]:
                bond_set_idx = frozenset(bond_change_idx[:2])
                bond_set = frozenset([elements[idx] if elements[idx] in Step.EXPLICIT_ATOMS else "other"
                                      for idx in bond_change_idx[:2]])
                initial_order = bonds.get(bond_set_idx, {}).get("order", 0)
                change = bond_change_idx[-1]
                final_order = initial_order + change
                if initial_order == 0 and final_order != 0:
                    # bond formed
                    key = (bond_set, True)
                elif initial_order != 0 and final_order == 0:
                    # bond cleaved
                    key = (bond_set, False)
                else:
                    continue
                fingerprint[fingerprint_key.index(key)] += 1
        return fingerprint

    def detect_reactant_roles(self):
        """Returns a list of the  reactant roles based on the MRV arrows"""
        reactions = self.get_arrows()
        roles = []
        aliases = self._get_scheme_xml_nodes_attribute("atom", "mrvAlias")
        roles_pairs = []
        for reaction in reactions:
            self.idx_to_residue
            for method in METHOD_TO_RESIDUE_FUNCTIONS:
                method_return = method(reaction, self)
                if method_return:
                    roles_pair = []
                    idx1, idx2 = [int(label.split(".")[1]) for label in method_return]
                    for no, label in enumerate(method_return):
                        idx = int(label.split(".")[1])
                        mol = self.mrv_idx_to_mol[idx]
                        if label.startswith("H."):
                            for bond in mol.GetBonds():
                                if self.mrv_idx_to_mol_idx[idx] in [bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()]:
                                    mol_idx = bond.GetOtherAtomIdx(self.mrv_idx_to_mol_idx[idx])
                                    idx = self.mol_idx_to_mrv_idx[mol][mol_idx]
                                    break

                        if idx in self.idx_to_residue:
                            molecule = self.idx_to_residue[idx]
                            if method == is_heterolytic_bond_cleavage:
                                if idx1 in self.idx_to_residue and idx2 in self.idx_to_residue and\
                                        self.idx_to_residue[idx1] == self.idx_to_residue[idx2]:
                                    # at the moment cannot distinguish between electrofuge and nucleofuge if
                                    # both atoms can be attributed to the same molecule.
                                    pass
                                if molecule is not None:
                                    mol_tagged = Chem.rdchem.RWMol(self.mrv_idx_to_mol[idx])
                                    mol_tagged.GetAtomWithIdx(self.mrv_idx_to_mol_idx[idx]).SetIsotope(1000)
                                    mcs = get_mcs(molecule.mcs_label, mol_tagged)
                                    if same_as_template(molecule.mcs_label, mol_tagged):
                                        for substruct in mol_tagged.GetSubstructMatches(
                                                Chem.MolFromSmarts(mcs.smartsString)):
                                            if any([aliases[
                                                self.mol_idx_to_mrv_idx[mol][sub_idx]] == molecule.simple_str
                                                    for sub_idx in substruct]):
                                                molecule = None
                        else:
                            molecule = None
                        new_mol = Chem.rdchem.Mol(mol)
                        atom = new_mol.GetAtomWithIdx(self.mrv_idx_to_mol_idx[idx])
                        atom.SetIsotope(200)
                        smiles = Chem.MolToSmiles(new_mol, canonical=True, isomericSmiles=True)
                        roles_pair.append((molecule, METHOD_TO_RESIDUE_FUNCTIONS[method][no], smiles))

                        # SPECIAL CASES
                        # 1. when residue binds a proton and loses it in the same step. need to be labelled as relay
                        if METHOD_TO_RESIDUE_FUNCTIONS[method][no] == "proton acceptor":
                            if aliases[idx] and aliases[idx].endswith(".relay"):
                                roles_pair.append((molecule, "proton donor", smiles))
                    roles_pairs.append(roles_pair)
        for role_pair in roles_pairs:
            molecule, function, smiles = role_pair[0]
            molecule_target, function_target, smiles_target = role_pair[1]
            if molecule is not None:
                roles.append(common.models.Role(
                    step=self,
                    residue_ct=ContentType.objects.get_for_model(molecule),
                    residue_id=molecule.id,
                    function=common.models.Function.objects.get(function=function),
                    source_smiles=smiles,
                    target_smiles=smiles_target,
                    automated=True,
                ))
            if molecule_target is not None and function_target is not None:
                roles.append(common.models.Role(
                    step=self,
                    residue_ct=ContentType.objects.get_for_model(molecule_target),
                    residue_id=molecule_target.id,
                    function=common.models.Function.objects.get(function=function_target),
                    source_smiles=smiles_target,
                    target_smiles=smiles,
                    automated=True,
                ))

        return roles

    def update_reactant_roles_smiles(self):
        """adds source_smiles and target_smiles to existing roles"""
        for role in self.detect_reactant_roles():
            common.models.Role.objects.filter(
                step=role.step, residue_ct=role.residue_ct, residue_id=role.residue_id, function=role.function
            ).update(source_smiles=role.source_smiles, target_smiles=role.target_smiles)

    def update_rc_env_from_smiles(self):
        """updates atom and group fields using the smiles columns"""
        for role in common.models.Role.objects.filter(step=self).exclude(source_smiles="", target_smiles=""):
            source_mol = Chem.MolFromSmiles(role.source_smiles)
            target_mol = Chem.MolFromSmiles(role.target_smiles)
            [atom.SetIsotope(0) for atom in source_mol.GetAtoms() if atom.GetIsotope() != 200]
            [atom.SetIsotope(0) for atom in target_mol.GetAtoms() if atom.GetIsotope() != 200]

            source_mol = Chem.RemoveHs(source_mol)
            target_mol = Chem.RemoveHs(target_mol)

            source_atom = [atom for atom in source_mol.GetAtoms() if atom.GetIsotope() == 200][0]
            target_atom = [atom for atom in target_mol.GetAtoms() if atom.GetIsotope() == 200][0]

            role.source_atom = source_atom.GetSymbol()
            role.target_atom = target_atom.GetSymbol()
            source_frag_atoms = set()
            target_frag_atoms = set()
            if source_mol.GetNumHeavyAtoms() == 1:
                source_frag = Chem.RemoveHs(source_mol)
                source_frag_atoms.add(source_frag.GetAtoms()[0].GetIdx())
            else:
                frag = Chem.rdmolops.FindAtomEnvironmentOfRadiusN(source_mol, 1, source_atom.GetIdx())
                bonds = [source_mol.GetBondWithIdx(idx) for idx in frag]
                for bond in bonds:
                    source_frag_atoms.add(bond.GetBeginAtomIdx())
                    source_frag_atoms.add(bond.GetEndAtomIdx())

            if target_mol.GetNumHeavyAtoms() == 0:
                target_frag_atoms.update([atom.GetIdx() for atom in target_mol.GetAtoms()])
            elif target_mol.GetNumHeavyAtoms() == 1:
                frag = Chem.RemoveHs(target_mol)
                target_frag_atoms.add(frag.GetAtoms()[0].GetIdx())
            else:
                frag = Chem.rdmolops.FindAtomEnvironmentOfRadiusN(target_mol, 1, target_atom.GetIdx())
                bonds = [target_mol.GetBondWithIdx(idx) for idx in frag]
                for bond in bonds:
                    target_frag_atoms.add(bond.GetBeginAtomIdx())
                    target_frag_atoms.add(bond.GetEndAtomIdx())

            role.source_env = Chem.MolFragmentToSmiles(
                source_mol, source_frag_atoms, canonical=True, isomericSmiles=False, rootedAtAtom=source_atom.GetIdx())
            role.target_env = Chem.MolFragmentToSmiles(
                target_mol, target_frag_atoms, canonical=True, isomericSmiles=False, rootedAtAtom=target_atom.GetIdx())
            role.save()

    def get_non_reactant_roles(self):
        bonds = self._get_scheme_xml_nodes("bond")
        # roles for coordination to metals
        roles = []
        for bond in bonds:
            if bond.get("convention") == "cxn:coord":
                idx1, idx2 = bond.get("atomRefs2").replace('a', '').split()
                idx1, idx2 = (int(idx1)-1, int(idx2)-1)
                metal = self.get_molecule_from_idx(idx2)
                ligand = self.get_molecule_from_idx(idx1)
                if not ligand:
                    continue
                if isinstance(metal, ActiveSiteCofactor) and \
                        metal.cofactor.is_metal:
                    # noinspection PyShadowingBuiltins
                    function = common.models.Function.objects.get(function="metal ligand")
                    role = common.models.Role(
                        step=self,
                        residue_ct=ContentType.objects.get_for_model(ligand),
                        residue_id=ligand.id,
                        automated=True,
                        function=function
                    )
                    role.save()
                    roles.append(role)
        return roles

    def get_residues_from_alias(self, alias):
        """Identify all the residues with this alias"""
        active_site = self.mechanism.reaction.entry.pf_activesite[0]  # need to use annotate_activesite on step query
        these_alias = common.models.Residue.get_all_residue_labels()[active_site.id]
        return these_alias.get(alias, [])

    def get_molecule_from_label(self, label):
        """Identify a molecule from its extra label"""
        if label not in self.molecule_label_data:
            label_type, chebi_id = label.split(":")
            if "chebi" in label_type:
                try:
                    self.molecule_label_data[label] = common.models.CompoundType.objects.get(
                        reaction=self.mechanism.reaction,
                        compound__in=common.models.Compound.objects.filter(chebi_id=chebi_id))
                except Exception as e:
                    print(e)
                    try:
                        self.molecule_label_data[label] = ActiveSiteCofactor.objects.get(
                            active_site=common.models.ActiveSite.objects.get(
                                entry=self.mechanism.reaction.entry),
                            cofactor__in=common.models.Cofactor.objects.filter(chebi=chebi_id)
                        )
                    except (common.models.Cofactor.DoesNotExist, ActiveSiteCofactor.DoesNotExist):
                        self.molecule_label_data[label] = None
        return self.molecule_label_data[label]

    @cached_property
    def step_mol_to_db_molecule(self) -> Dict[Chem.Mol, MoleculeDB]:
        """map Chem.Mol objects from the scheme file to their database object"""
        step_mol_to_db_molecule_data = {}
        for mrv_idx, mol in self.mrv_idx_to_mol.items():
            if mol is not None:
                molecule = self.idx_to_molecule.get(mrv_idx, None)
                if molecule is not None:
                    step_mol_to_db_molecule_data[mol] = molecule
        return step_mol_to_db_molecule_data

    @cached_property
    def db_molecule_to_step_mol(self) -> Dict[MoleculeDB, Chem.Mol]:
        """map a compound database object to the Chem.Mol object of the scheme file"""
        return {v: k for k, v in self.step_mol_to_db_molecule.items()}

    @cached_property
    def db_non_residue_molecule_to_step_mol(self) -> Dict[NonResidueMoleculeDB, Chem.Mol]:
        """map a compound database object (excluding residues) to the Chem.Mol object of the scheme file"""
        return {k: v for k, v in self.db_molecule_to_step_mol.items()
                if isinstance(k, (ActiveSiteCofactor, CompoundType))}

    @cached_property
    def atom_map_to_coord(self) -> Dict[int, Point2D]:
        """map the atom_map_number of each atom in the rdkit Mol objects to a 2d coordinate"""
        position_x = self._get_scheme_xml_nodes_attribute("atom", "x2")
        position_y = self._get_scheme_xml_nodes_attribute("atom", "y2")
        out_dict = {}
        for idx, (x, y) in enumerate(zip(position_x, position_y)):
            assert isinstance(x, str)
            assert isinstance(y, str)
            out_dict[idx+1] = Point2D(float(x), float(y))
        return out_dict

    @cached_property
    def scheme_xml_atoms_element(self) -> List[str]:  # TODO test
        """return the elements of the atoms in the scheme xml"""
        return ["X" if e is None else e.replace("R", "C").replace("*", "C")
                for e in self._get_scheme_xml_nodes_attribute("atom", "elementType")]

    @cached_property
    def scheme_xml_atoms_alias(self) -> List[str]:  # TODO test
        """return the alias of the atoms in the scheme xml"""
        return ["" if a is None or any([pattern in a for pattern in self.IGNORE_ALIAS]) else a
                for a in self._get_scheme_xml_nodes_attribute("atom", "mrvAlias")]

    @cached_property
    def scheme_xml_atoms_label(self) -> List[str]:  # TODO test
        """return the alias of the atoms in the scheme xml"""
        return ["" if a is None else a for a in self._get_scheme_xml_nodes_attribute("atom", "mrvExtraLabel")]

    @cached_property
    def scheme_xml_atoms_r_group(self) -> List[str]:  # TODO test
        """return the alias of the atoms in the scheme xml"""
        return [a == "R" for a in self._get_scheme_xml_nodes_attribute("atom", "mrvAlias")]

    @cached_property
    def scheme_xml_atoms_charge(self) -> List[int]:  # TODO test
        """return the charge of the atoms in the scheme xml"""
        return [0 if c is None else int(c) for c in self._get_scheme_xml_nodes_attribute("atom", "formalCharge")]

    @cached_property
    def scheme_xml_atoms_radical(self) -> List[int]:  # TODO test
        """return the charge of the atoms in the scheme xml"""
        return [Step.radical_options.index(r) for r in self._get_scheme_xml_nodes_attribute("atom", "radical")]

    @cached_property
    def mrv_idx_to_mol_idx(self):
        """map atom idx in step scheme file to the atom idx of the respective mol in self.mols"""
        if not hasattr(self, "_mrv_idx_to_mol_idx"):
            self.mols
        return self._mrv_idx_to_mol_idx

    @cached_property
    def mrv_idx_to_mol(self):
        """map atom idx in step scheme file to the respective mol in self.mols"""
        if not hasattr(self, "_mrv_idx_to_mol"):
            self.mols
        return self._mrv_idx_to_mol

    @cached_property
    def mol_idx_to_mrv_idx(self):
        """map the atom idx of a mol in self.mols to the atom idx of the atom in step scheme file"""
        if not hasattr(self, "_mol_idx_to_mrv_idx"):
            self.mols
        return self._mol_idx_to_mrv_idx

    @cached_property
    def am_to_mol(self) -> Dict[int, Chem.Mol]:
        """return dict that maps a scheme atom map to a scheme mol"""
        am_to_mol = {}
        for mol in self.mols:
            for atom in mol.GetAtoms():
                am_to_mol[atom.GetAtomMapNum()] = mol
        return am_to_mol

    @cached_property
    def mols(self) -> List[Chem.Mol]:
        """
        Return a list of RDKit mols in based on the step mrv file
        """
        self._mrv_idx_to_mol: Dict[int, Chem.Mol] = {}
        self._mol_idx_to_mrv_idx: DefaultDict[Chem.Mol, Dict[int, int]] = collections.defaultdict(dict)
        self._mrv_idx_to_mol_idx: Dict[int, int] = {}
        # find how many molecules there are and creates a Mol object for each
        bond_info = self._get_scheme_xml_bond_info(coord=False)
        atom_idxs_set_groups = common.utils.consolidate_sets(
            [set(bonded_atoms) for bonded_atoms in bond_info] +  # molecules
            [{idx} for idx in range(len(self.scheme_xml_atoms_element))]  # lone atoms
            )
        mols = [RWMol() for atom_group in atom_idxs_set_groups]

        # create atoms and add them to the correct molecules
        for mol, atom_idx_set in zip(mols, atom_idxs_set_groups):
            for mrv_idx in atom_idx_set:
                self._mrv_idx_to_mol_idx[mrv_idx] = self._mols_add_atom(mol, mrv_idx)
                self._mrv_idx_to_mol[mrv_idx] = mol
                self._mol_idx_to_mrv_idx[mol][self._mrv_idx_to_mol_idx[mrv_idx]] = mrv_idx

        # create bonds
        for (idx1, idx2), info in bond_info.items():
            mol = self._mrv_idx_to_mol[idx1]
            bond_idx = mol.AddBond(
                    self._mrv_idx_to_mol_idx[idx1], self._mrv_idx_to_mol_idx[idx2], BondType.values[info["order"]])
            mol.GetBondWithIdx(bond_idx-1).SetBondDir(self.BOND_DIRECTION_MRV_TO_MOL.get(info.get("direction")))

        mol_to_mol_h = {}
        # clean up molecules
        for mol_idx, mol in enumerate(mols):
            this_coord_map = {atom.GetIdx(): self.atom_map_to_coord[atom.GetAtomMapNum()] for atom in mol.GetAtoms()}
            AllChem.Compute2DCoords(mol, coordMap=this_coord_map)
            try:
                # keep non aromatic atoms and bonds
                Chem.SanitizeMol(mol, sanitizeOps=Chem.SanitizeFlags.SANITIZE_ALL
                                 ^ Chem.SanitizeFlags.SANITIZE_SETAROMATICITY)
                mol_to_mol_h[mol] = Chem.rdmolops.AddHs(mol, addCoords=True)  # ADDHS
                AllChem.Compute2DCoords(mol_to_mol_h[mol], coordMap=this_coord_map)
            except Chem.rdchem.AtomValenceException:
                logger.warning("Could not sanitize: {}".format(str(self)))
                mol_to_mol_h[mol] = mol
            Chem.rdmolops.AssignChiralTypesFromBondDirs(mol, confId=0)

        # replace mol with mol h
        for mol, mol_h in mol_to_mol_h.items():
            for k, v in self._mrv_idx_to_mol.items():
                if v is mol:
                    self._mrv_idx_to_mol[k] = mol_h
            self._mol_idx_to_mrv_idx[mol_h] = self._mol_idx_to_mrv_idx.pop(mol)
        mols = [mol_to_mol_h[mol] for mol in mols]

        # add atom maps to new hydrogens
        previous_am = max(self.atom_map_to_coord)
        for mol in mols:
            for atom in mol.GetAtoms():
                if not atom.GetAtomMapNum():
                    previous_am += 1
                    atom.SetAtomMapNum(previous_am)
        return mols

    @cached_property
    def combined_mol(self):
        """return rdkit mol object with all the molecules in the scheme"""
        return common.utils.combine_mols_list(self.mols)

    def _mols_add_atom(self, mol: Chem.Mol, mrv_idx: int) -> int:
        """Add the atom to mol and return the idx of the atom in the mol object"""
        atom = Atom(self.scheme_xml_atoms_element[mrv_idx])
        atom.SetAtomMapNum(mrv_idx + 1)
        atom.SetFormalCharge(self.scheme_xml_atoms_charge[mrv_idx])
        atom.SetNumRadicalElectrons(self.scheme_xml_atoms_radical[mrv_idx])
        atom.SetProp("label", self.scheme_xml_atoms_label[mrv_idx])

        # if self.scheme_xml_atoms_alias[mrv_idx] != "":
            # atom.SetIsotope(100+atom.GetAtomicNum())  # this is used for substruct matching

        return mol.AddAtom(atom)

    IGNORE_ALIAS = ["R", "DNA", "Man", "Protein", "Base", "chebi", "TS", "CoA", "Cyt", "Ubi", "Fgl", "relay"]
    BOND_DIRECTION_MRV_TO_MOL = {"W": BondDir.BEGINWEDGE, "H": BondDir.BEGINDASH, None: BondDir.NONE}

    def find_extra_relay_roles(self):
        """A relay should not exist if residue is not both a donor and an acceptor"""
        for role in self.roles.filter(
                residue_ct=11, function__function__contains="relay", function__function_type="reactant"):
            acceptor = self.roles.filter(
                    function__group_id=role.function.group_id, residue_id=role.residue_id,
                    residue_ct=role.residue_ct, function__function__contains="acceptor")
            donor = self.roles.filter(
                    function__group_id=role.function.group_id, residue_id=role.residue_id,
                    residue_ct=role.residue_ct, function__function__contains="acceptor")
            if not (acceptor and donor):
                print(self, role)

    def add_missing_relay_roles(self):
        """If the same residue is both a donor and acceptor in the same step, it should also be a relay"""
        for role in self.roles.filter(
                residue_ct=11, function__function__contains="donor", function__function_type="reactant"):
            if "donor" in role.function.function:
                acceptor_roles = self.roles.filter(
                    function__group_id=role.function.group_id, residue_id=role.residue_id,
                    residue_ct=role.residue_ct, function__function__contains="acceptor")
                if acceptor_roles:
                    if not self.roles.filter(function__group_id=role.function.group_id, residue_id=role.residue_id,
                                             residue_ct=role.residue_ct, function__function__contains="relay"):
                        relay_function = common.models.Function.objects.get(
                                group_id=role.function.group_id, function__contains="relay")
                        new_role = common.models.Role.objects.create(
                                step=self, function=relay_function, residue_id=role.residue_id,
                                residue_ct=role.residue_ct, automated=True)
                        print(self, role, acceptor_roles.get(), new_role)

    @cached_property
    def db_residues(self) -> List[Residue]:
        """list of db residues detected in the scheme mrv file"""
        return list(set(self.idx_to_residue.values()))

    @cached_property
    def idx_to_residue(self) -> Dict[int, Residue]:
        """Return a dictionary that maps the atom indexes to the predicted db residue it belongs to"""
        idx_to_residue_data = {}

        for mrv_idx, mol_idx in self.mrv_idx_to_mol_idx.items():
            if not (alias := self.scheme_xml_atoms_alias[mrv_idx]):
                continue
            else:
                residues_with_label = self.get_residues_from_alias(alias)
                mol = self.mrv_idx_to_mol[mrv_idx]
                # print(alias, Chem.MolToSmarts(mol))
                for residue in residues_with_label:
                    if (substruct := get_maximum_template_substructure(residue.mcs_label, mol)):
                        for sub_idx in substruct:
                            # set to False if it is atom not in mrv such as added hydrogens 
                            if (other_mrv_idx := self.mol_idx_to_mrv_idx[mol].get(sub_idx, False)):
                                idx_to_residue_data[other_mrv_idx] = residue
                        break
        # print(idx_to_residue_data)
        return idx_to_residue_data

    @cached_property
    def idx_to_molecule(self) -> Dict[int, MoleculeDB]:
        """Returns a dictionary where each atom index points to the predicted molecule it belongs to
        should be merged with idx_to_residue.

        Coded to work only in the first step of each reaction (substrate changes afterwards)"""
        logger.info("Mapping mrv index of non residues in scheme to a db molecule")
        idx_to_molecule_data: Dict[int, MoleculeDB] = {}
        idx_to_molecule_data.update(self.idx_to_residue)
        unidentified_mols = set()
        identified_mols = set()  # any mol with an identified atom will be here
        for mrv_idx, mol in self.mrv_idx_to_mol.items():
            if mol not in identified_mols and mrv_idx not in idx_to_molecule_data:
                atom = mol.GetAtomWithIdx(self.mrv_idx_to_mol_idx[mrv_idx])
                if atom.GetAtomicNum() != 1:
                    unidentified_mols.add(mol)
            else:
                # one atom checked before was not identified but the current atoms is
                unidentified_mols.discard(mol)
                identified_mols.add(mol)

        reaction_compounds = {
            Chem.MolFromMolFile(reactant.compound.mol_file.file.name): [reactant.count, reactant, reactant.compound]
            for reactant in
            CompoundType.objects.filter(reaction=self.mechanism.reaction, type="reactant")}

        # do same for cofactors
        reaction_compounds.update({
            Chem.MolFromMolFile(cofactor.cofactor.mol_file.file.name): [cofactor.count, cofactor, cofactor.cofactor]
            for cofactor in
            ActiveSiteCofactor.objects.filter(active_site__entry=self.mechanism.reaction.entry)})

        for mol in unidentified_mols:
            logger.debug("mol not identified directly from labels in mrv curly arrow scheme: {}".format(
                Chem.MolToSmiles(mol)))
        if unidentified_mols:
            logger.info("Using common substructure matching")

        u_mol_to_molecule: Dict[Chem.Mol, MoleculeDB] = {}
        from common.predict_mechanism.predict_mechanism_util import find_matching_mols
        for rxn_mol, (u_mol, idx_mapping, atom_map_mapping) in\
                find_matching_mols(reaction_compounds, unidentified_mols).items():
            count, molecule, compound = reaction_compounds[rxn_mol]
            u_mol_to_molecule[u_mol] = molecule

        for u_mol, molecule in u_mol_to_molecule.items():
            for mrv_idx in self.mol_idx_to_mrv_idx[u_mol].values():
                idx_to_molecule_data[mrv_idx] = molecule

        return idx_to_molecule_data

    # @cached_property
    # def scheme_xml_ordered_arrows_info_old(self) -> List[ArrowInfo]:
        # """
        # return a ordered list with information of each arrow in the scheme xml file.

        # elements are of ArrowInfo type
        # arrows whose the tip does not touch the base of other arrows come first
        # """
        # # build graph with all the arrows in the step
        # graph = nx.DiGraph()
        # for arrow in self.scheme_xml_arrows_info:
            # for atom1 in arrow[0]:
                # for atom2 in arrow[1]:
                    # # do not add self loops
                    # if atom1 != atom2:
                        # # ignore middle atom when both base and tip have two atoms
                        # if not ({atom1, atom2} & arrow[0] & arrow[1] and len(arrow[0]) == len(arrow[1]) == 2):
                            # graph.add_edge(atom1, atom2)

        # # TODO not properly tested
        # unique_ordered_nodes = []
        # while graph.number_of_nodes():  # necessary when a node has no tips and two bases
            # for sub_graph_nodes in nx.weakly_connected_components(graph):
                # out_nodes = [x for x in graph.nodes()
                             # if graph.out_degree(x) == 0 and graph.in_degree(x) > 0 and x in sub_graph_nodes]
                # # if sub_graph_nodes and not out_nodes:
                # if not out_nodes:
                    # # circular graph, pick a heavy atom, to be an unstable intermediate with wrong valence
                    # out_nodes = [node for node in sub_graph_nodes
                                 # if self.scheme_xml_atoms_element[node-1] != "H" and graph.in_degree(node) > 0]
                # # print(out_nodes)
                # ordered_nodes = [node for edge in nx.edge_dfs(graph, source=out_nodes[0], orientation="reverse")
                                 # for node in reversed(list(edge))]
                # for node in ordered_nodes:
                    # if node not in unique_ordered_nodes and node != "reverse":
                        # unique_ordered_nodes.append(node)
            # # print(unique_ordered_nodes)
            # nx.draw(graph, with_labels=True, font_weight='bold')
            # from matplotlib import pyplot as plt
            # plt.show()
            # graph.remove_nodes_from(unique_ordered_nodes)
            # if graph.number_of_nodes() == 1:
                # node = list(graph.nodes())[0]
                # unique_ordered_nodes.append(node)
                # graph.remove_node(node)

        # ordered_arrows = sorted(self.scheme_xml_arrows_info, key=lambda x:
                                # unique_ordered_nodes.index([atom for atom in x[1] if atom in unique_ordered_nodes][0]))
        # return ordered_arrows

    @cached_property
    def scheme_xml_arrows_info_graph(self) -> List[ArrowInfo]:
        """
        return a ordered list with information of each arrow in the scheme xml file.

        elements are of ArrowInfo type
        arrows whose the tip does not touch the base of other arrows come first
        """
        # build graph with all the arrows in the step
        graph = nx.DiGraph()

        # add pairs of nodes corresponding to each tip of the curly arrows
        for arrow in self.scheme_xml_arrows_info:
            graph.add_nodes_from(arrow[:2])

        # connect nodes that share at least one atom
        for node in graph.nodes():
            for anode in graph.nodes():
                if node != anode and node & anode:
                    if not (graph.has_edge(anode, node) or graph.has_edge(node, anode)):
                        graph.add_edge(node, anode)
                        graph.add_edge(anode, node)

        # nx.draw(graph, with_labels=True, font_weight='bold')
        # from matplotlib import pyplot as plt
        # plt.show()
        return graph

    @cached_property
    def scheme_xml_reactant_product_for_arrows(self) -> List[Tuple[Chem.Mol, Chem.Mol, int, ArrowInfo]]:
        """
        get a list of reactants, products (as rdkit mol objects) and other info for each arrow in the step

        the numbering in the ArrowInfo uses the atom map numbers
        """
        # self.mols
        arrows_info = self.scheme_xml_ordered_arrows_info
        reactants_mol = self.combined_mol
        output = []
        for no, arrow_info in enumerate(arrows_info):
            products_mol = common.utils.get_mol_product_from_arrow(reactants_mol, arrow_info)
            output.append((reactants_mol, products_mol, no+1, arrow_info))
            reactants_mol = products_mol
        return output
