# coding=utf-8
"""Includes models related with PDB structures"""

# python standard imports
import collections
import gzip
import glob
import os
import json
import pathlib
import re
import ssl
from urllib.request import urlopen, urlretrieve
from urllib.parse import urlencode
from urllib.error import HTTPError, URLError
from typing import List, Dict
from rdkit import Chem
import time

# python other imports
import Bio
from Bio.PDB.Chain import Chain as BioChain
from Bio.PDB.Residue import Residue as BioResidue
from Bio.PDB.MMCIFParser import MMCIFParser
from lxml import etree
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB import PDBIO, Select
from Bio.PDB.Residue import Residue as PdbResidue
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO, AlignIO

# django imports
from django.db import models
from django.urls import reverse
from django.utils.html import format_html
from django.utils.functional import cached_property

# atlas imports
from atlas import files, settings
from common import link
from common.others import get_md5, AA_1TO3
import common.models


class MultiAASeqAlignment(models.Model):
    """A multi sequence alignment model that is created for every chain in reference structures in M-CSA"""
    reference_sequence = models.ForeignKey("AASequence", on_delete=models.CASCADE, related_name="reference_alignment")
    sequence = models.ForeignKey("AASequence", on_delete=models.CASCADE, related_name="alignment")
    pad_left = models.IntegerField()
    alignment = models.TextField()

    @classmethod
    def create_objects(cls, alignments_folder):
        """"Reads output mafft files and adds alignments to database"""
        msa_output_files = alignments_folder.glob("*.out.fasta")
        md5_to_dbid = dict(AASequence.objects.values_list("sequence_md5", "id"))
        for msa_output_file in msa_output_files:
            # db.reset_queries()
            print(msa_output_file)
            if os.path.getsize(msa_output_file) == 0:
                continue
            reference_sequence_md5 = msa_output_file.name.split(".")[0]
            current_objects = cls.objects.filter(reference_sequence__sequence_md5=reference_sequence_md5)
            alignment = AlignIO.read(str(msa_output_file), 'fasta')
            if {seq.name for seq in alignment} == set(current_objects.values_list("id", flat=True)):
                print("Already added, Skipping")
                continue
            if current_objects.count():
                print("Deleted: ", current_objects.delete())
            alignments_to_create = []
            for seq in alignment:
                seq_lstrip = seq.seq.lstrip('-')
                pad_left = len(seq.seq) - len(seq_lstrip)
                seq_strip = seq_lstrip.rstrip('-')
                alignments_to_create.append(cls(
                    reference_sequence_id=md5_to_dbid[reference_sequence_md5],
                    sequence_id=md5_to_dbid[seq.name],
                    pad_left=pad_left,
                    alignment=seq_strip,
                ))
            print("creating {} alignments".format(len(alignments_to_create)))
            cls.objects.bulk_create(alignments_to_create, batch_size=1000)

    def get_non_reference_resid_code(self, ref_align_resid0):
        """Return resid and 3letter code of residue in the alignment that corresponds to a certain reference residue"""
        # if ref_resid0_in_ma is None:
            # out = (None, 'out')
        # else:
        resid0 = ref_align_resid0-self.pad_left
        if 0 > resid0 or resid0 >= len(str(self.alignment)) or self.alignment[resid0] == '-':
            return None, ''
        else:
            resid = resid0 - self.alignment.count('-', 0, resid0) + 1
            code1 = self.alignment[resid0]
            try:
                code3 = AA_1TO3[code1]
            except KeyError:
                code3 = ''
                print("could not find this code in AA_1TO3: ", code1)
            return resid, code3

    # def get_non_reference_resid_codes(self, ref_resids0_in_ma):
        # """Returns a list of tuples with resid and 3letter code"""
        # output = []
        # for ref_resid0_in_ma in ref_resids0_in_ma:
            # if ref_resid0_in_ma is None:
                # out = (None, 'out')
            # else:
                # resid0 = ref_resid0_in_ma-self.pad_left
                # if resid0 < 0:
                    # out = (None, '')
                # elif resid0 >= len(str(self.alignment)):
                    # out = (None, '')
                # elif self.alignment[resid0] == '-':
                    # out = (None, '')
                # else:
                    # resid = resid0 - self.alignment.count('-', 0, resid0) + 1
                    # code1 = self.alignment[resid0]
                    # try:
                        # code3 = AA_1TO3[code1]
                    # except KeyError:
                        # code3 = ''
                    # out = (resid, code3)
            # output.append(out)
        # return output

class AASequenceQuerySet(models.QuerySet):
    """Some predefined querysets for the AASequence model"""

    def get_hash2id(self):
        """Returns values list sequence_md5 -> database id"""
        return self.values_list('sequence_md5', 'id')


class AASequence(models.Model):
    sequence = models.TextField()
    sequence_md5 = models.CharField(max_length=32, editable=False, unique=True)

    objects = AASequenceQuerySet.as_manager()

    def __str__(self):
        return "AA sequence: {}".format(self.pk)

    @classmethod
    def create_objects(cls, sequences):
        existing_sequence_md5 = set(cls.objects.values_list("sequence_md5", flat=True))
        sequences_to_create = []
        for sequence in sequences:
            md5 = get_md5(sequence)
            if md5 not in existing_sequence_md5:
                sequences_to_create.append(cls(sequence=sequence, sequence_md5=md5))
                existing_sequence_md5.add(md5)
        print("creating {} AASequence objects".format(len(sequences_to_create)))
        cls.objects.bulk_create(sequences_to_create)

    @classmethod
    def create_aa_sequences_fasta(cls, output_file):
        """Puts all the AASequence objects in the same fasta file to be searched by phmmer"""
        seqs = [SeqRecord(Seq(seq.sequence), seq.sequence_md5) for seq in cls.objects.all() if len(seq.sequence) >= 20]
        SeqIO.write(seqs, output_file, "fasta")

    @classmethod
    def create_reference_fasta(cls, output_folder, delete_old=False):
        """Creates fasta files for every reference structure sequence"""
        if delete_old:
            for filename in output_folder.glob("*.fasta"):
                print("deleting", filename)
                # os.remove(filename)
        for seq in cls.objects.filter(chain__residue_chains__is_reference=True).distinct("pk"):
            fasta_name = "{}/{}.fasta".format(output_folder, seq.sequence_md5)
            SeqIO.write(SeqRecord(Seq(seq.sequence), seq.sequence_md5), fasta_name, "fasta")

    def save(self, *args, **kwargs):
        self.sequence_md5 = get_md5(self.sequence)
        super(AASequence, self).save(*args, **kwargs)

    # def find_homologues(self, cutoff=0.000001, db="pdb", skip_existing=True):
        # """Find all sequence homologues, save them and create AAAlignment"""
        # print(self.sequence)
        # if len(str(self.sequence)) < 10 or len(collections.Counter(str(self.sequence))) <= 5:
            # return
        # existing_homologues = set(AASequence.objects.filter(
            # alignment1__sequence2=self).values_list(
            # 'chain__name', 'chain__structure__pdb_id').distinct())
        # # make request to phmmer
        # params = urlencode({
            # 'seqdb': db,
            # 'seq': self.sequence,
            # 'incE': cutoff,
            # 'E': cutoff
        # }).encode("utf-8")
        # request = urlopen(
            # 'https://www.ebi.ac.uk/Tools/hmmer/search/phmmer',
            # params
        # )
        # # get results from phmmer
        # results_url = request.geturl()
        # res_params = urlencode({
            # 'output': 'json',
            # 'range': '{},{}'.format(0, 100000)
        # })
        # modified_url = "{}?{}".format(results_url, res_params)
        # results = request.urlopen(modified_url).read()
        # hits = json.loads(results.decode('utf-8'))['results']['hits']
        # print("hits: {}".format(len(hits)))
        # # iterate over results and add them to the database
        # for hit in hits:
            # pdb2_id, chain_name = hit['acc'].split("_")
            # if (chain_name, pdb2_id) in existing_homologues and skip_existing:
                # continue
            # try:
                # aa_sequence2 = AASequence.objects.get(
                    # chain__name=chain_name,
                    # chain__structure__pdb_id=pdb2_id
                # )
                # self.add_aaalignment(
                    # sequence2=aa_sequence2,
                    # evalue=hit['evalue'],
                    # align1from=int(hit['domains'][0]['alihmmfrom'])-1,
                    # align2from=int(hit['domains'][0]['alisqfrom'])-1,
                    # align_seq1=hit['domains'][0]['alimodel'].upper(),
                    # align_seq2=hit['domains'][0]['aliaseq'],
                    # alignment=hit['domains'][0]['alimline'],
                # )
            # except AASequence.DoesNotExist:
                # continue

    # def add_aaalignment(self, sequence2, evalue, align1from, align2from, alignment, align_seq1, align_seq2, symm=True):
        # if self == sequence2:
            # return
        # aaalignment, created = AAAlignment.objects.update_or_create(
            # sequence1=self,
            # sequence2=sequence2,
            # defaults={
                # 'evalue': evalue,
                # 'align1from': align1from,
                # 'align2from': align2from,
                # 'alignment': alignment,
                # 'align_seq1': align_seq1.lower(),
                # 'align_seq2': align_seq2.lower(),
            # })
        # aaalignment.set_similarity_measures().save(update_fields=[
            # 'coverage', 'identity', 'similarity'])
        # if symm:
            # sequence2.add_aaalignment(
                # self, evalue, align2from, align1from, alignment,
                # align_seq2, align_seq1, False)
        # return aaalignment

    # def remove_aaalignment(self, sequence2, symm=True):
        # AAAlignment.objects.filter(
            # sequence1=self,
            # sequence2=sequence2,
        # ).delete()
        # if symm:
            # sequence2.remove_aaalignment(self, False)


class Chain(models.Model):
    structure = models.ForeignKey("Structure", on_delete=models.PROTECT)
    aa_sequence = models.ForeignKey("AASequence", on_delete=models.PROTECT, related_name="chains")
    sequences = models.ManyToManyField("Sequence", related_name="chains")
    name = models.CharField(max_length=10)

    objects = models.Manager()

    def __str__(self):
        return "{0.structure}_{0.name}".format(self)

    class Meta:
        unique_together = ("structure", "name")

    @classmethod
    def create_objects(cls):
        """Reads pdb_seqres and adds info to Chain and AASequence Table"""
        pdb_seq_dict = {}
        existing_chains = set(cls.objects.values_list('structure__pdb_id', 'name'))
        pdb_ids = set()
        with open(files.PDB_FASTA, 'r') as pdb_fasta_file:
            pdb_id = chain = ""
            for _, line in enumerate(pdb_fasta_file):
                if line.startswith(">") and "mol:protein" in line:
                    pdb_id, chain = line[1:].split()[0].split("_")
                elif "mol:na" in line:
                    break
                elif pdb_id and chain:
                    pdb_seq_dict[(pdb_id, chain)] = line.strip("\n")
                    pdb_ids.add(pdb_id)
                    pdb_id = chain = ""
                else:
                    print("Something is wrong with the file")
        pdb_seq_dict = {k: v for k, v in pdb_seq_dict.items()
                        if len(v) > 20 and k not in existing_chains}
        print("Adding pdb AA sequences")
        common.models.AASequence.create_objects(pdb_seq_dict.values())

        print("Adding {} new chains".format(len(pdb_seq_dict)))
        pdb_id2idx = dict(common.models.Structure.objects.filter(pdb_id__in=pdb_ids).get_idx2id())
        sequence_md5_to_id = dict(common.models.AASequence.objects.get_hash2id())
        chains_to_create = []
        for (pdb_id, chain_name), sequence in pdb_seq_dict.items():
            structure_id = pdb_id2idx.get(pdb_id, None)
            if structure_id is None:
                print("PDB not found: " + pdb_id)
            chains_to_create.append(cls(
                name=chain_name, structure_id=structure_id, aa_sequence_id=sequence_md5_to_id[get_md5(sequence)]))
        cls.objects.bulk_create(chains_to_create)

    @classmethod
    def link_to_sequences(cls):
        """Links chains with sequences"""
        chains_ids = cls.objects.values_list(
            'structure__pdb_id', 'name', 'id')
        chains_ids = {(pdb, chain): cid for pdb, chain, cid in chains_ids}
        sequences_ids = common.models.Sequence.objects.values_list('uniprot_id', 'id')
        sequences_ids = {uniprot_id: sid for uniprot_id, sid in sequences_ids}
        chain_uniprot_ids_tuples = set()
        with gzip.open(files.SIFTS_PDB_CHAIN_UNIPROT, 'rt') as pdb_chain_uniprot_file:
            pdb_chain_uniprot_file.readline()
            pdb_chain_uniprot_file.readline()
            for line in pdb_chain_uniprot_file:
                pdb_id, chain_letter, uniprot, _ = line.split('\t', 3)
                try:
                    chain_uniprot_ids_tuples.add(
                        (chains_ids[(pdb_id, chain_letter)], sequences_ids[uniprot])
                    )
                except KeyError:
                    pass
        chain_sequences_through = cls.sequences.through
        existing = set(chain_sequences_through.objects.values_list('chain_id', 'sequence_id'))
        to_add = chain_uniprot_ids_tuples - existing
        print(len(chain_uniprot_ids_tuples), len(to_add))
        print("Creating chain sequence links")
        # TODO delete old ones
        to_create = []
        for chain_id, sequence_id in to_add:
            to_create.append(
                chain_sequences_through(chain_id=chain_id, sequence_id=sequence_id))
        chain_sequences_through.objects.bulk_create(to_create, batch_size=10000)


class StructureQuerySet(models.QuerySet):
    """Some predefined querysets for the Structure model"""

    def get_idx2id(self):
        """Returns values list pdb id -> database id"""
        return self.values_list('pdb_id', 'id')


class Structure(models.Model):
    """A unique pdb structure table. To substitute the Pdb class long term"""
    pdb_id = models.CharField(max_length=4, verbose_name='PDB ID', unique=True)
    title = models.TextField()
    resolution = models.CharField(max_length=256)
    aa_sequences = models.ManyToManyField('AASequence', through='Chain', related_name="structures")
    sequences = models.ManyToManyField("Sequence", related_name="structures")
    biological_assembly = models.IntegerField(null=True)

    _cif_file = None
    _assembly_xml_tree = None
    _assembly = None
    _currenct_cif_assembly = None

    objects = StructureQuerySet.as_manager()

    def __str__(self):
        return self.pdb_id

    @property
    def pdbe_text_link(self):
        return link.pdb_id_2_pdbe(self.pdb_id)

    @property
    def search_html_link(self):
        return format_html(
            '<a class="search" href="{0}?s={1}&search=Search+M-CSA'
            '&protein__pdb__strcture__pdb_id=on" target="_blank" '
            'title="Search For M-CSA entries with the same Pdb">'
            '{1}</a>'.format(
                reverse('common:search'),
                self.pdb_id,
            )
        )

    @property
    def pdbe_text_link_res(self):
        return link.pdb_id_2_pdbe(self.pdb_id, res=self.resolution)

    @property
    def cath_text_link(self):
        return link.pdb_id_2_cath(self.pdb_id)

    @property
    def pdbsum_icon_link(self):
        return link.pdb_id_2_pdbsum(self.pdb_id, icon=True)

    @property
    def pdbe_icon_link(self):
        return link.pdb_id_2_pdbe(self.pdb_id, icon=True)

    @property
    def pdb_icon_link(self):
        return link.pdb_id_2_pdb(self.pdb_id, icon=True)

    @property
    def pfam_icon_link(self):
        return link.pdb_id_2_pfam_icon(self.pdb_id)

    @property
    def funtree_icon_link(self):
        return link.pdb_id_2_funtree_icon(self.pdb_id)

    _chain_resid_to_auth_dict = None

    @property
    def chain_resid_to_auth_dict(self):
        if self._chain_resid_to_auth_dict is None:
            self._chain_resid_to_auth_dict = self.get_sequence_resid_chains_dict(output_key="chain_resid")
        return self._chain_resid_to_auth_dict

    # _pdb_chain_resid_to_auth_dict = None  # uses pdb chain instead of assembly chains

    @property
    def pdb_chain_resid_to_auth_dict(self):
        # if self._pdb_chain_resid_to_auth_dict is None:
            # self._pdb_chain_resid_to_auth_dict = self.get_sequence_resid_chains_dict(
                    # output_key="chain_resid", use_assembly_chains=False)
        # return self._pdb_chain_resid_to_auth_dict
        return self.get_sequence_resid_chains_dict(output_key="chain_resid", use_assembly_chains=False)

    @classmethod
    def create_objects(cls):
        """Reads pdb index file and adds new pdbs to Structure table"""
        pdb_dict = {}
        db_pdb_ids = set(cls.objects.values_list('pdb_id', flat=True))
        with open(files.PDB_IDX, 'r') as pdb_idx_file:
            next(pdb_idx_file)
            next(pdb_idx_file)
            for line in pdb_idx_file:
                words = line.split("\t")
                pdb_id = words[0].lower()
                title = words[3]
                res = words[6]
                technique = words[7]
                if res == "NOT":
                    res = technique.lower()
                pdb_dict[pdb_id] = (title, res)
        new_pdb_ids = set(pdb_dict.keys()) - db_pdb_ids
        print("Adding {} new structures".format(len(new_pdb_ids)))
        new_structures = []
        for new_pdb_id in new_pdb_ids:
            new_structures.append(cls(
                pdb_id=new_pdb_id,
                title=pdb_dict[new_pdb_id][0],
                resolution=pdb_dict[new_pdb_id][1],
            ))
        Structure.objects.bulk_create(new_structures)

    @classmethod
    def add_related_uniprot(cls):
        """Adds TREMBL uniprot ids associated with PDBs to the Sequence table"""
        with gzip.open(files.SIFTS_UNIPROT_PDB, 'rt') as uniprot_pdb_file:
            uniprot_ids = [line.split("\t")[0] for line in uniprot_pdb_file][2:]
        common.models.Sequence.create_or_update_sequences_batch(uniprot_ids, False)

    @classmethod
    def link_to_sequences(cls):
        """Populate m2m structure_sequences table"""
        existing = set(cls.objects.values_list('pdb_id', 'sequences__uniprot_id'))
        nones = set(Structure.objects.values_list('pdb_id', 'sequences__uniprot_id').filter(sequences=None))
        print('Structures that are not associated with any Uniprot Sequence: {}'.format(len(nones)))
        in_file = set()
        with gzip.open(files.SIFTS_UNIPROT_PDB, 'rt') as uniprot_pdb_file:
            next(uniprot_pdb_file)
            next(uniprot_pdb_file)
            for line in uniprot_pdb_file:
                uniprot_id, pdb_ids = line.split("\t")
                pdb_ids = pdb_ids.strip('\n').split(";")
                for pdb_id in pdb_ids:
                    in_file.add((pdb_id, uniprot_id))
        to_add = in_file - existing
        # todo delete old ones
        print("Adding {} Structure-Sequence links".format(len(to_add)))
        # this could be made faster by adding to the intermediate table directly
        # see Chain.link_to_sequences to see how
        for (pdb_id, uniprot_id) in to_add:
            try:
                sequence = common.models.Sequence.objects.get(uniprot_id=uniprot_id)
            except common.models.Sequence.DoesNotExist:
                print("Sequence not found in database: {0} (maybe deleted from uniprot)".format(uniprot_id))
                continue
            structure = cls.objects.get(pdb_id=pdb_id)
            sequence.structures.add(structure)

    def get_sifts_file(self):
        """Returns the pdb file name. Downloads it if necessary"""
        sub_directory = files.SIFTS_FILES_PATH / "{}/".format(self.pdb_id[1:3])
        sub_directory.mkdir(exist_ok=True)
        sifts_filename = sub_directory / "{}.xml.gz".format(self.pdb_id)
        if not sifts_filename.exists():
            url = "ftp://ftp.ebi.ac.uk/pub/databases/msd/sifts/split_xml/{}/{}.xml.gz".format(self.pdb_id[1:3], self.pdb_id)
            print(url)
            try:
                urlretrieve(url, sifts_filename)
                print("Downloaded pdb_id sifts file {} to {}.".format(url, sifts_filename))
            except URLError:
                print("Could not download sifts file.".format(url))
                print("SIFTS XML file not found: {}. Obsolete entry or need to update the devserver files".format(sifts_filename))
                return None
        return sifts_filename

    def get_sequence_resid_chains_dict(self, output_key="uniprot_resid", use_assembly_chains=True):
        """Returns a dict where keys are Uniprot res and values are pdb res chains"""
        sifts_filename = self.get_sifts_file()
        if sifts_filename is None:
            return None
        with gzip.open(sifts_filename) as sifts_file:
            xml_str = sifts_file.read()
        xml_str = re.sub(b'\sxmlns="[^"]+"', b'', xml_str, count=1)
        root = etree.fromstring(xml_str)
        sequence_chain_dict = collections.defaultdict(list)
        chain_resid_to_auth_dict = {}
        for entity in root.findall(".//entity"):
            assembly_chains = set()
            for chain_group in self.assembly_chains_groups.values():
                if entity.attrib['entityId'] in chain_group:
                    assembly_chains.update(chain_group)
            for residue in entity.findall(".//residue"):
                uniprot_cross_rf = residue.find("./crossRefDb[@dbSource='UniProt']")
                if uniprot_cross_rf is None:
                    continue
                seq_resid = int(residue.attrib['dbResNum'])
                pdb_cross_ref = residue.find("./crossRefDb[@dbSource='PDB']")
                auth_resid = pdb_cross_ref.get('dbResNum')
                # in some rare cases auth_resid is not a number
                auth_resid = int(re.sub('[^0-9]', '', auth_resid)) if auth_resid != 'null' else None
                pdb_resname = pdb_cross_ref.get('dbResName').capitalize()
                u_resid = int(uniprot_cross_rf.attrib['dbResNum'])
                uniprot_id = uniprot_cross_rf.attrib['dbAccessionId']
                if not use_assembly_chains:
                    chains = {residue.find("./crossRefDb[@dbSource='PDB']").get('dbChainId'), }
                else:
                    chains = assembly_chains
                if (uniprot_id, u_resid) in sequence_chain_dict:
                    sequence_chain_dict[(uniprot_id, u_resid)][0].update(chains)
                else:
                    sequence_chain_dict[(uniprot_id, u_resid)] = [set(chains), seq_resid, auth_resid, pdb_resname]
                for chain in chains:
                    chain_resid_to_auth_dict[(chain, seq_resid)] = auth_resid

        return sequence_chain_dict if output_key == "uniprot_resid" else chain_resid_to_auth_dict

    @cached_property
    def sequence_resid_chains_dict(self):
        return self.get_sequence_resid_chains_dict(output_key="uniprot_resid", use_assembly_chains=True)

    @cached_property
    def assembly_chain_to_pdb_chain(self) -> Dict[str, str]:
        """return map of the auth_asym_id in mmmcif files to the label_asym_id"""
        data = self.mmcif_dict
        # return {label: data["_atom_site.auth_asym_id"][i] for i, label in enumerate(data["_atom_site.label_asym_id"])}
        return {label: data["_pdbe_chain_remapping.orig_auth_asym_id"][i]
                for i, label in enumerate(data["_pdbe_chain_remapping.new_label_asym_id"])}

    @property
    def pdb_chain_to_assembly_chains(self):
        pdb_to_assembly = {}
        for assembly_chain, pdb_chain in self.assembly_chain_to_pdb_chain.items():
            if pdb_chain in pdb_to_assembly:
                pdb_to_assembly[pdb_chain].append(assembly_chain)
            else:
                pdb_to_assembly[pdb_chain] = [assembly_chain, ]
        return pdb_to_assembly

    @cached_property
    def assembly_chains_groups(self):
        entity_id_to_assembly_chains = collections.defaultdict(list)
        for entity_id, asym_id in zip(
            self.mmcif_dict["_pdbe_chain_remapping.orig_label_asym_id"],
            self.mmcif_dict["_pdbe_chain_remapping.new_auth_asym_id"],
        ):
            entity_id_to_assembly_chains[entity_id].append(asym_id)
        return dict(entity_id_to_assembly_chains)

    def get_assembly_xml(self):
        if self._assembly_xml_tree is None:
            url = ("http://www.ebi.ac.uk/pdbe/static/entry/download/{}-assembly.xml".format(self.pdb_id))
            xml_str = urlopen(url).read()
            xml_str = re.sub(b'\sxmlns="[^"]+"', b'', xml_str, count=1)
            self._assembly_xml_tree = etree.fromstring(xml_str)
        return self._assembly_xml_tree

    def get_default_biological_assembly(self):
        return int(self.get_assembly_xml().find('./assembly[@prefered="True"]').attrib['id'])

    # def get_pdb_file(self):
        # """Returns the pdb file name. Downloads it if necessary"""
        # pdb_filename = pathlib.Path("{0}/pdb{1.pdb_id}.ent".format(files.PDB_FILES_PATH, self))
        # if not pdb_filename.exists():
            # Bio.PDB.PDBList().retrieve_pdb_file(self.pdb_id, pdir=files.PDB_FILES_PATH, file_format="pdb")
        # return pdb_filename

    # def get_pdb_parser(self):
        # """Get Biopython parser for pdb file"""
        # parser = Bio.PDB.PDBParser()
        # return parser.get_structure('pdb', self.get_pdb_file())

    @property
    def mmcif_dict(self):
        """return mmcif data as a dict. Just a wrapper to the MMCIF2Dict biopython function"""
        cif_file = gzip.open(self.get_assembly_cif_file(), mode="rt", encoding='utf-8')
        mmcif_dict = MMCIF2Dict(cif_file)
        cif_file.close()
        return mmcif_dict

    # this does not return the same objects as io
    # wip
    @cached_property
    def assembly_parser(self):
        """return the Biopython parser for the mmcif file of the biological assembly"""
        parser = MMCIFParser( QUIET=True)
        cif_file = gzip.open(self.get_assembly_cif_file(), mode="rt", encoding='utf-8')
        print(self.get_assembly_cif_file())
        assembly_parser = parser.get_structure('mmcif-{}'.format(self.pdb_id), cif_file)
        cif_file.close()
        return assembly_parser
 
    def residue_min_distance(self, res_a, res_b):
        """returns the min distance between the two residues. each residue is defined as (resid, chain) tuple"""
        minimum = 999
        for atom_a in res_a:
            for atom_b in res_b:
                minimum = min(minimum, atom_a-atom_b)
        return minimum

    @cached_property
    def assembly_het_residues(self):
        """return list of Biopython residue objects relative to HET residues in the structure"""
        het_residues = []
        for model in self.assembly_parser:
            for chain in model:
                for residue in chain:
                    if residue.get_id()[0].startswith('H') or residue.get_id()[0] == "W":
                        het_residues.append(residue)
        return het_residues

    @cached_property
    def assembly_het_labels_to_residues(self):
        return {self.get_label_from_residue(res): res for res in self.assembly_het_residues}

    @cached_property
    def assembly_het_labels_to_orig_chain_het_labels(self):
        return {self.get_label_from_residue(res, orig_chain=False): self.get_label_from_residue(res, orig_chain=True)
                for res in self.assembly_het_residues}

    @cached_property
    def assembly_orig_chain_het_labels_to_het_labels(self):
        return {v: k for k, v in self.assembly_het_labels_to_orig_chain_het_labels.items()}

    @cached_property
    def assembly_het_labels_to_mols(self):
        labels_to_mols = {}
        for label, residue in self.assembly_het_labels_to_residues.items():
            mol = self.get_mol_from_pdb_residue(residue)
            # Chem.rdDepictor.Compute2DCoords(mol)
            labels_to_mols[label] = mol
        return labels_to_mols

    def get_label_from_residue(self, residue, orig_chain=False):
        """a label usefull to identify residues"""
        parent_id = self.assembly_chain_to_pdb_chain[residue.parent.id] if orig_chain else residue.parent.id
        return "{},{},{},{}".format(parent_id, *residue.id)

    def get_residue_from_label(self, pdb_residue_label):
        """get the pdb residue for this label"""
        words = re.split(",", pdb_residue_label)
        chain = words[0]
        resid = (words[1], int(words[2]), words[3])
        return self.assembly_parser[0][chain][resid]

    def get_residue_center_of_geometry(self, residue):
        """return the coordinates corresponding to the center of geometry"""
        return sum([atom.coord for atom in residue])/len(residue)

    def get_same_assembly_chains(self, chain_name):
        """returns a list of the assembly chains that are based on the same original pdb chain"""
        for assembly_chain_group in self.assembly_chains_groups.values():
            if chain_name in assembly_chain_group:
                return assembly_chain_group
        return None

    def get_assembly_cif_file(self, assembly=None, overwrite=False):
        """Returns the mmcif filename and downloads it if necessary"""
        if assembly is None:
            if self.biological_assembly is None:
                self.biological_assembly = self.get_default_biological_assembly()
                self.save()
            assembly = self.biological_assembly

        directory = files.MMCIF_FILES_PATH / "{}/{}/".format(self.pdb_id[1:3], self.pdb_id)
        directory.mkdir(exist_ok=True, parents=True)
        filename = directory / "{}-assembly-{}.cif.gz".format(self.pdb_id, assembly)
        if not filename.exists() or overwrite: # this should only happen in the dev server (run setup_dev_files)
            url = "https://www.ebi.ac.uk/pdbe/static/entry/download/{}-assembly-{}.cif.gz".format(self.pdb_id, assembly)
            try:
                urlretrieve(url, filename)
                print("Downloaded pdb_id: {}, assembly: {} to {}.".format(self.pdb_id, assembly, filename))
            except HTTPError:
                print("Could not download pdb_id: {}, assembly: {} from {}.".format(self.pdb_id, assembly, url))
        if filename.exists():
            return filename
    
    def move_assembly_file_to_media(self):
        mmcif_media_dir = pathlib.Path(settings.MEDIA_ROOT) / "mmcif/"
        mmcif_filename = mmcif_media_dir / "{0.pdb_id}-assembly-{0.biological_assembly}.cif".format(self)
        if not mmcif_filename.exists():
            try:
                os.system("cd {0}; cp {1} .; gzip -d {1.stem}".format(mmcif_media_dir, self.get_assembly_cif_file()))
            except AttributeError as e:
                print(e)

    @property
    def cif_file_url(self):
        self.move_assembly_file_to_media()
        return "{0}mmcif/{1}-assembly-{2}.cif".format(settings.MEDIA_URL, self.pdb_id, self.biological_assembly)

    @cached_property
    def biopython_pdb_io(self):
        pdb_io = PDBIO()
        pdb_io.set_structure(self.assembly_parser)
        return pdb_io


    # def save_pdb_with_residue_selection(self, pdb_residues, new_filename):
        # tmp_folder = pathlib.Path("/tmp/")
        # tmp_filename = tmp_folder / 'test.pdb' #  .format(pdb_residue.id, pdb_residue.parent.id)

        # residues_id_chain = set()
        # old_chains = {}
        # # to avoid problems with chains with dashes
        # if (z_chains := [chain for chain in self.assembly_parser[0] if chain.get_id() == "z"]):
            # new_chain = z_chains[0]
        # else:
            # new_chain = BioChain("z")
            # self.assembly_parser[0].add(new_chain)

        # for pdb_residue in pdb_residues:
            # print("a", pdb_residue, id(pdb_residue))
            # old_chains[id(pdb_residue)] = pdb_residue.parent
            # old_chains[id(pdb_residue)].detach_child(pdb_residue.id)
            # new_chain.add(pdb_residue)
            # residues_id_chain.add((pdb_residue.id, pdb_residue.parent.id))

        # for pdb_residue, chain in old_chains.items():
            # print(pdb_residue, id(pdb_residue), chain)

        # self.biopython_pdb_io.save(str(tmp_filename), ResidueSelect(residues_id_chain))

        # for pdb_residue in pdb_residues:
            # print("b", pdb_residue, id(pdb_residue))
            # new_chain.detach_child(pdb_residue.id)
            # old_chains[id(pdb_residue)].add(pdb_residue)

        # self.assembly_parser[0].detach_child(new_chain.id)


    _mol_from_pdb_residue_data = {}
    def get_mol_from_pdb_residue(self, pdb_residue: PdbResidue, atom_selection="") -> Chem.Mol:
        """return a rdkit mol file for this pdb_residue"""
        if (pdb_residue.get_full_id(), atom_selection) not in self._mol_from_pdb_residue_data:
            tmp_folder = pathlib.Path("/tmp/")
            tmp_filename = tmp_folder / '{}_{}.pdb'.format(pdb_residue.id, pdb_residue.parent.id)
            # to avoid problems with chains with dashes
            if (z_chains := [chain for chain in pdb_residue.parent.parent if chain.get_id() == "z"]):
                new_chain = z_chains[0]
            else:
                new_chain = BioChain("z")
                pdb_residue.parent.parent.add(new_chain)
            old_chain = pdb_residue.parent
            old_chain.detach_child(pdb_residue.id)
            new_chain.add(pdb_residue)

            # print(pdb_residue.id, pdb_residue.parent.id, old_chain.id, atom_selection)
            self.biopython_pdb_io.save(str(tmp_filename), ResidueSelect(
                {(pdb_residue.id, pdb_residue.parent.id)}, atom_selection=atom_selection))

            new_chain.detach_child(pdb_residue.id)
            old_chain.add(pdb_residue)
            # ###

            pdb_residue_mol = Chem.MolFromPDBFile(str(tmp_filename), sanitize=False)
            self._mol_from_pdb_residue_data[(pdb_residue.get_full_id(), atom_selection)] = pdb_residue_mol

        return self._mol_from_pdb_residue_data[(pdb_residue.get_full_id(), atom_selection)]


class ResidueSelect(Select):
    """to select pdb output"""

    main_chain_atoms = {"N", "C", "O", "CA"}
    side_chain_exclude = {"N", "C", "O"}

    def __init__(self, residues_id_chain=set(), atom_selection=""):
        super(ResidueSelect, self).__init__()
        self.residues_id_chain = residues_id_chain
        self.atom_selection = atom_selection
        self.chain_ids = {chain_id for res, chain_id in residues_id_chain}

    def accept_chain(self, chain):
        if chain.id in self.chain_ids:
            return 1
        else:
            return 0

    def accept_residue(self, residue):
        if (residue.id, residue.parent.id) in self.residues_id_chain:
            return 1
        else:
            return 0

    def accept_atom(self, atom):
        if self.atom_selection == "main_chain":
            if atom.get_name() not in self.main_chain_atoms:
                return 0
        elif self.atom_selection == "side_chain":
            if atom.get_name() in self.side_chain_exclude:
                return 0
        return 1


class AAAlignment(models.Model):
    """Alignment of two aa sequences"""
    sequence1 = models.ForeignKey('AASequence', on_delete=models.CASCADE, related_name='alignment1')
    sequence2 = models.ForeignKey('AASequence', on_delete=models.CASCADE, related_name='alignment2')
    evalue = models.FloatField()
    identity = models.FloatField(default=0)
    similarity = models.FloatField(default=0)
    coverage = models.FloatField(default=0)
    align1from = models.IntegerField()
    align2from = models.IntegerField()
    align_seq1 = models.TextField()
    align_seq2 = models.TextField()
    alignment = models.TextField()

    objects = models.Manager()

    class Meta:
        unique_together = ("sequence1", "sequence2")

    def __str__(self):
        return "AA alignment: {0.evalue}".format(self)

    def set_similarity_measures(self):
        seq1_len = len(self.sequence1.sequence)
        self.coverage = len(self.align_seq1.replace("-", "").replace(".", ""))/seq1_len
        alignment_m_spaces = self.alignment.replace(" ", "")
        self.similarity = len(alignment_m_spaces)/seq1_len
        alignment_m_spaces_plus = alignment_m_spaces.replace("+", "")
        self.identity = len(alignment_m_spaces_plus)/seq1_len
        return self
