# coding=utf-8
"""Includes models related with the Tree class"""

# python core imports
import collections
import glob
from math import isclose
import pathlib
import tempfile
import os
from statistics import mean
from io import StringIO
import re

# others python imports
import Bio
import Bio.Phylo
from lxml import etree
from ete3 import PhyloTree as etetree
import numpy as np
from ete3 import NCBITaxa 
ncbi = NCBITaxa()

# django imports
from django.db import models
from django.db.models import Prefetch, Q
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import ContentFile, File
from django.utils.functional import cached_property
from django.db.models import F, Sum

# atlas imports
from common.models import CATH, Sequence, PFAM
from common import link
import common.models


def add_if_missing(parent_clade, branch_name, **kwargs):
    """Adds a new element to parent clade with branch name if it does not exist already"""
    branch = parent_clade.find(branch_name)
    if branch is None:
        branch = etree.SubElement(parent_clade, branch_name, **kwargs)
    return branch


def delete_old_add_again(parent_clade, branch_name, **kwargs):
    """Deletes element called branch name and adds a new one with the same name to parent clade"""
    branch = parent_clade.find(branch_name)
    if branch is not None:
        parent_clade.remove(branch)
    return etree.SubElement(parent_clade, branch_name, **kwargs)


class TreeCATHManager(models.Manager):
    """A manager that only includes Trees with a CATH reference"""

    def get_queryset(self):
        return super(TreeCATHManager, self).get_queryset().filter(reference_ct=ContentType.objects.get(model="cath"))


class TreeSequenceManager(models.Manager):
    """A manager that only includes Trees with a Sequence reference"""

    def get_queryset(self):
        return TreeQuerySet(self.model, using=self._db).filter(
            reference_ct=ContentType.objects.get(model="sequence"))


class TreeQuerySet(models.QuerySet):
    """Query set for the Tree model"""

    def distribution_pseudoenzyme(self, **kwargs):
        if not "protein_sets" in kwargs:
            kwargs["protein_sets"] = ["kw", "exp", "filtered_kw", "filtered_exp"]  # tree tags 
        output = collections.OrderedDict()
        for protein_set in kwargs["protein_sets"]:
            good_trees = self.filter(tag=protein_set)#.annotate(total=Sum(F("number_enzymes")+F("number_non_enzymes"))).filter(total__gte=10)
            print(protein_set, good_trees.count())
            enzymes = np.array(good_trees.values_list("number_enzymes", flat=True))
            non_enzymes = np.array(good_trees.values_list("number_non_enzymes", flat=True))
            ratio = non_enzymes/(enzymes + non_enzymes)
            bins = [0, 1E-10, 0.1, 0.2, 0.3, 1]
            bin_names = ["[{0:.0f}% - {1:.0f}%)".format(bins[no]*100, bins[no+1]*100) for no in range(len(bins)-1)]
            bin_names[0] = "only enzymes"
            bin_names[1] = "(0% - 10%)"
            values, b = np.histogram(ratio, bins)
            print(values)
            output[protein_set] = collections.OrderedDict(((bin_names[no], v/good_trees.count()) for no, v in enumerate(values)))
        #print(output)
        return output


class Tree(models.Model):
    """Phylogenetic trees that correspond to a multi sequence alignment"""
    xml_file = models.FileField(upload_to='xml_trees/')
    tag = models.CharField(max_length=255)
    number_enzymes = models.IntegerField(default=0, verbose_name="Number of enzymes")
    number_non_enzymes = models.IntegerField(default=0, verbose_name="Number of non-enzymes")
    gain_of_function = models.IntegerField(default=0, verbose_name="Number of Non-Enzyme -> Enzyme transitions")
    loss_of_function = models.IntegerField(default=0, verbose_name="Number of Enzyme <-> Non-Enzyme transitions")
    no_change_enzyme = models.IntegerField(default=0, verbose_name="Number of Enzyme <-> Enzyme transitions")
    no_change_pseudoenzyme = models.IntegerField(default=0, verbose_name="Number of Pseudoenzyme <-> Pseudoenzyme")
    change_first = models.IntegerField(default=0, verbose_name="Split in Last Common Ancestor")
    is_tree_ok = models.BooleanField(null=True)
    notes = models.TextField()

    reference = GenericForeignKey('reference_ct', 'reference_id')
    limit = models.Q(model='cath') | models.Q(model='sequence')
    reference_ct = models.ForeignKey(ContentType, on_delete=models.CASCADE, limit_choices_to=limit)
    reference_id = models.IntegerField()

    cath_reference = TreeCATHManager()
    objects = TreeSequenceManager()

    _tree = None
    _xml_has_changes = False

    def __repr__(self):
        return "Tree: {} ({})".format(self.reference.__repr__(), self.tag)

    def __str__(self):
        return self.__repr__()

    class Meta:
        unique_together = ['reference_ct', 'reference_id', 'tag']

    @classmethod
    def _cath_from_filename(cls, filename):
        """Returns CATH built from any filename that start with the CATH codes separated by underscores (1_10_100_10)"""
        basename = pathlib.Path(filename).stem.replace(".out", "")
        words = basename.split("_", maxsplit=5)
        cath1, cath2, cath3, cath4 = words[:4]
        cath = CATH.objects.get(cath1=cath1, cath2=cath2, cath3=cath3, cath4=cath4)
        return cath

    @classmethod
    def _sequence_from_filename(cls, filename):
        """Returns Sequence built from any filename that starts with the uniprot id"""
        from common.models import Sequence
        return Sequence.objects.get(uniprot_id=pathlib.Path(filename).stem.split(".")[0].split("_")[0])

    @property
    def filename_from_reference(self):
        """Returns a filename built from the reference model name and adding .xml (1_10_100_10.xml)"""
        if self.reference_ct.model == "cath":
            return "{0.cath1}_{0.cath2}_{0.cath3}_{0.cath4}.xml".format(self.reference)
        else:
            return "{}_{}.xml".format(self.reference.uniprot_id, self.tag)

    @classmethod
    def get_protein_sets(self):
        """breaks code if outside function"""
        reviewed = Sequence.objects.reviewed()
        return collections.OrderedDict((
            ("kw", [Sequence.reviewed_homologue_set(),
                Sequence.enzyme_set()&Sequence.reviewed_homologue_set()]),
            ("exp", [Sequence.experimental_set() & Sequence.reviewed_homologue_set(),
                Sequence.experimental_set() & Sequence.enzyme_set() & Sequence.reviewed_homologue_set()]),
            ("filtered_kw", [Sequence.curated_set() & Sequence.reviewed_homologue_set(),
                (Sequence.curated_set() & Sequence.reviewed_homologue_set()) - Sequence.manual_nonenzyme_set()]),
            ("filtered_exp", [Sequence.curated_set() & Sequence.experimental_set() & Sequence.reviewed_homologue_set(),
                (Sequence.experimental_set() & Sequence.curated_set() & Sequence.reviewed_homologue_set()) - Sequence.manual_nonenzyme_set()]),
        ))


    @classmethod
    def create_objects(cls, tag: str, tree_folder: str, reference_model="sequence", limit="", overwrite_tree_files=False) -> None:
        """
        Reads notung newick output files and import trees to database. Removes current tree files on the database
        with the same tag (but not tree objects).

        :param tag: identifies a different subset of alignments
        :param tree_folder: will add every tree in this folder to database
        :param reference_model: type of model to be used as reference. Can be Cath or Sequence
        :param limit: Run only for this sequence. Mostly for testing purposes.
        :return:
        """
        from common.models import Sequence

        try:
            sequences_set = cls.get_protein_sets()[tag][0]
        except KeyError:
            raise Exception("I do not recognize this tag: ", tag)

        reference = Sequence.objects.reference()
        print("reading allowed sequences")
        allowed_sequences = dict(Sequence.objects.filter(id__in=sequences_set).get_idx2id())
        allowed_sequences.update(dict(reference.get_idx2id()))
        print(len(allowed_sequences))

        tree_files = glob.glob(tree_folder+"/{}*.nw.rearrange.0.rooting.0".format(limit))
        for tree_file in tree_files:
            if os.path.getsize(tree_file) == 0:
                continue
            print(tree_file)
            reference_uniprot_id = pathlib.Path(tree_file).stem.split(".")[0]
            print(reference_uniprot_id)
            reference = cls._sequence_from_filename(tree_file)
            try:
                xml_file = cls.objects.get(
                            reference_ct=ContentType.objects.get(model="sequence"),
                            reference_id=reference.id,
                            tag=tag).xml_file
                if xml_file.storage.exists(xml_file.name) and not overwrite_tree_files:
                    continue
            except common.models.Tree.DoesNotExist:
                pass
            #nw_string = nw_string.replace("LOST", "LOST:0.0")  # adding dummy distances so ete3 does not complain
            #nw_string = re.sub(r'([rn]\d+)([,);])', r'\1:0.0\2', nw_string)
            #print(nw_string)
            #tree = Bio.Phylo.convert(tree_file, "newick", tree_file+"_good", "newick")
            nw_string = open(tree_file, "r").readlines()[0]
            nw_tree = etetree(nw_string, format=1)
            for node in nw_tree.traverse():
                if "_" in node.name:
                    node.name = node.name.split("_")[1]
            leaves_to_keep = set(nw_tree.get_leaf_names()) & set(allowed_sequences.keys())
            print("how many leaves_to_keep", len(leaves_to_keep))

            # remove sequences if not all catalytic residues map to something in these residues
            # very restrictive but gets rid of sequences with wrong domains
            resseq_dict = common.models.ResidueSequence.objects.filter(uniprot_id__in=leaves_to_keep).sequences2resseq_dict()
            for uniprot_id in leaves_to_keep.copy():
                if len(resseq_dict[reference_uniprot_id, uniprot_id]) != len(resseq_dict[reference_uniprot_id, reference_uniprot_id]):
                    leaves_to_keep.remove(uniprot_id)
                    continue
                if None in [resseq.resid for resseq in resseq_dict[reference_uniprot_id, uniprot_id]]:
                    leaves_to_keep.remove(uniprot_id)
            print("how many leaves to keep after removing Nones", len(leaves_to_keep))
            if len(leaves_to_keep) < 2:
                print("skipping")
                continue
            nw_tree.prune(leaves_to_keep, preserve_branch_length=True)
            tree = Bio.Phylo.read(StringIO(nw_tree.write(format=2)), "newick")

            if reference_model == "cath":
                reference = cls._cath_from_filename(tree_file)
            elif reference_model == "sequence":
                reference = cls._sequence_from_filename(tree_file)
            else:
                raise Exception("reference model must be either cath or sequence")
            tree = tree.as_phyloxml()
            with tempfile.NamedTemporaryFile('w') as tmpf:
                Bio.Phylo.write(tree, tmpf, 'phyloxml')
                tmpf.flush()
                tree_obj, _ = cls.objects.get_or_create(
                    reference_ct=ContentType.objects.get(model=reference_model),
                    reference_id=reference.id,
                    tag=tag)
                tree_obj.xml_file.delete()
                tree_obj.xml_file.save(tree_obj.filename_from_reference, File(open(tmpf.name)))

    @classmethod
    def annotate_trees(cls, tag, limit=""):
        """Annotate xml files with Uniprot, EC, GO, and PFAM info"""
        trees = cls.objects.filter(tag=tag)  # .filter(
        sequences = common.models.Sequence.objects
        if limit:
            trees = trees.filter(sequence__uniprot_id=limit)
        # reference_id=common.models.Sequence.objects.get(uniprot_id="Q62651").id)
        # go_terms_id_subset = set(GoTerm.objects.molecular_function().catalytic().values_list("id", flat=True))
        print("reading enzymes")
        enzymes = dict(Sequence.objects.filter(id__in=cls.get_protein_sets()[tag][1]).get_idx2id())
        print("reading refs")
        enzymes.update(dict(Sequence.objects.reference().get_idx2id()))  # some reference seqs are unreviewed
        for no, tree in enumerate(trees):
            if not tree.xml_file.storage.exists(tree.xml_file.name):
                print("tree not found: ", tree.xml_file.name)
                continue
            tree.annotate_tree(enzymes)  # go_terms_id_subset)


    def annotate_tree(self, enzymes, go_terms_subset=""):
        """Annotates a tree xml file with UniProt, EC, GO, and PFAM info"""
        enzymes = {seq.uniprot_id: seq.id for seq in self.sequences if seq.uniprot_id in enzymes}
        print("Annotating tree: {}".format(self.reference))
        self._add_is_enzyme_property(enzymes)
        self._add_residues_info_and_label()
        self._add_sequence_info()
        # self._add_info("go_info", self._uniprot_id_to_go_info(go_terms_subset))
        self._add_info("ec_info", self._uniprot_id_to_ec_info)
        codes = self._add_taxonomy_info()
        self._add_taxonomies(codes)
        self._add_label("ec_info", "EC")
        # self._add_label("go_info", "GO")
        self._add_domains_graph()
        self._add_graph(title="Is Enzyme", graph_type="pie")
        self._xml_has_changes = True
        self.save()

    def _add_residues_graph(self):
        pass

    def _add_residues_info_and_label(self):
        """Adds a property element to every sequence and catalytic residue
           returns list of sequences for which every residue is not aligned"""
        from common.models import ResidueSequence
        reference_residue_sequences = ResidueSequence.objects.reference().filter(sequence=self.reference)\
                                                                         .order_by("resid")
        not_aligned_count = collections.defaultdict(int)
        for ref_res_seq in reference_residue_sequences:
            residue_sequences = ResidueSequence.objects.same_residue_as([ref_res_seq, ])\
                .prefetch_related(Prefetch("sequence", queryset=Sequence.objects.all().only("uniprot_id")))\
                .prefetch_related("residue")
            code_name = "{0.code}{0.resid}".format(ref_res_seq)
            up_to_code_name = {res_seq.sequence.uniprot_id: "{0.code}{0.resid}  ".format(res_seq)
                               for res_seq in residue_sequences}
            #for res_seq in residue_sequences:
            #    if res_seq.resid is None:
            #        not_aligned_count[res_seq.sequence.uniprot_id] += 1
            up_to_code_name_color = {
                res_seq.sequence.uniprot_id: res_seq.mutation_color(ref_res_seq) for res_seq in residue_sequences}

            self._add_info(code_name, up_to_code_name)
            self._add_label(code_name, "")
            self._add_info(code_name+"_color", up_to_code_name_color)
            self._add_label(code_name+"_color", ref_res_seq.code_resid_location, type="color")

        #return [seq for seq, count in not_aligned_count.items() if count == reference_residue_sequences.count()]

    @property
    def number_sequences(self):
        """The number of sequences in this tree multi-sequence alignment"""
        return self.number_enzymes + self.number_non_enzymes

    @cached_property
    def ratio_enzymes(self):
        """The proportion of sequences in this tree that are enzymes"""
        return self.number_enzymes / self.number_sequences

    @property
    def tree(self):
        """Returns xml of the tree file"""
        if self._tree is None:
            self._tree = etree.parse(self.xml_file.file)
            query = "descendant-or-self::*[namespace-uri()!='']"
            for element in self._tree.xpath(query):
                element.tag = etree.QName(element).localname
        return self._tree

    @property
    def root(self):
        """Returns the root of the xml"""
        return self.tree.getroot()

    @cached_property
    def _sequence_nodes(self):
        """The terminal nodes (sequences) of this tree"""
        return [clade for clade in self.tree.findall("//clade") if clade.find("clade") is None]
    
    @cached_property
    def _branch_nodes(self):
        """The terminal nodes (sequences) of this tree"""
        return [clade for clade in self.tree.findall("//clade") if clade.find("clade") is not None]

    @property
    def _uniprot_ids(self):
        """Set of all the uniprot ids in this tree"""
        return set([clade.find('name').text for clade in self._sequence_nodes])

    @cached_property
    def sequences(self):
        """Return a Sequence query with all the sequences in the tree"""
        return Sequence.objects.filter(uniprot_id__in=self._uniprot_ids)\
            .prefetch_related("ecs")\
            .prefetch_related("sequencepfam_set__pfam")\
            .prefetch_related("gos") \
            .prefetch_related("species")

    def _uniprot_id_to_go_info(self, go_term_set=""):
        """Returns a dict that maps uniprot id to GO information"""
        return {seq.uniprot_id: ", ".join([go.code for go in seq.gos.all() if str(go.id) in go_term_set])
                for seq in self.sequences}

    @property
    def _uniprot_id_to_ec_info(self):
        """Returns a dict that maps uniprot id to EC information"""
        return {sequence.uniprot_id: ", ".join([ec.code for ec in sequence.ecs.all()]) for sequence in self.sequences}

    def _add_sequence_info(self):
        """Add sequence info to every clade (deletes existing info)"""
        seq_dict = {seq.uniprot_id: seq for seq in self.sequences}
        for clade in self._sequence_nodes:
            uniprot_id = clade.find("name").text
            db_sequence = seq_dict[uniprot_id]
            sequence = delete_old_add_again(clade, "sequence")
            etree.SubElement(sequence, "name").text = db_sequence.name
            link_uniprot = link.UNIPROT_URL.format(uniprot_id)
            etree.SubElement(sequence, "uri").text = link_uniprot
            domain_architecture = etree.SubElement(
                sequence, "domain_architecture", length=str(len(db_sequence.sequence)))
            for sequence_pfam in db_sequence.sequencepfam_set.all():
                domain = etree.SubElement(domain_architecture, "domain")
                domain.attrib['from'] = str(sequence_pfam.start)
                domain.attrib['to'] = str(sequence_pfam.end)
                domain.attrib['Description'] = sequence_pfam.pfam.name
                domain.attrib['confidence'] = str(1.0E-11)
                domain.text = str(sequence_pfam.pfam)
        return self.tree

    def _add_label(self, tag, label_name, type="text"):
        """Add label to tree using information from tag"""
        labels = add_if_missing(self.root, "labels", type=type)
        label = labels.find("label[@tag='{}']".format(tag))
        if label is not None:
            labels.remove(label)
        label = etree.SubElement(labels, "label", type=type, tag=tag)
        etree.SubElement(label, "name", show="1").text = label_name
        etree.SubElement(label, "data", tag="{}".format(tag))
        return

    COLORS = ['#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99', '#e31a1c',
              '#fdbf6f', '#ff7f00', '#cab2d6', '#6a3d9a', '#ffff99', '#b15928']

    def _add_domains_graph(self):
        """Adds domains graph to tree, needs PFAM information in sequence"""
        pfams = PFAM.objects.filter(sequences__in=self.sequences).distinct()
        domains = delete_old_add_again(self.root, "domains")
        for no, pfam in enumerate(pfams):
            domain = etree.SubElement(domains, "domain", name=str(pfam))
            etree.SubElement(domain, "color").text = self.COLORS[no % len(self.COLORS)]
            etree.SubElement(domain, "description").text = pfam.name

    #    <graph type="multibar">
    #    <graph type="pie">
    #    <graph type="heatmap">
    #    <graph type="binary">
    #    <graph type="boxplot">
    #    <graph type="pie">
    def _add_graph(self, title="", graph_type="pie"):
        """Adds graph to tree"""
        graphs = add_if_missing(self.root, "graphs")
        graph = graphs.find("graph")
        if graph is not None and "tag" in graph.attrib and graph.attrib['tag'] == "is_enzyme":
            graphs.remove(graph)
        graph = etree.SubElement(graphs, "graph", type=graph_type, tag="is_enzyme")
        name = etree.SubElement(graph, "name")
        name.text = title
        legend = etree.SubElement(graph, "legend", show="1")
        field = etree.SubElement(legend, "field")
        etree.SubElement(field, "name").text = title
        etree.SubElement(field, "color").text = "#66cc74"
        etree.SubElement(field, "background").text = "#cc6661"
        data = etree.SubElement(graph, "data", tag="property", ref="is enzyme info")
        etree.SubElement(data, "min").text = str(0)
        etree.SubElement(data, "max").text = str(1)

    def _add_info(self, tag, up_to_info):
        """Add go number information to this tree"""
        for clade in self._sequence_nodes:
            uniprot_id = clade.find("name").text
            tag_obj = delete_old_add_again(clade, tag)
            try:
                tag_obj.text = up_to_info[uniprot_id]
            except KeyError:
                tag_obj.text = 'X'

    def _add_taxonomy_info(self):
        """Add go number information to this tree"""
        up_to_species = {sequence.uniprot_id: sequence.species for sequence in self.sequences}
        codes = set()
        for clade in self._sequence_nodes:
            uniprot_id = clade.find("name").text
            taxonomy = delete_old_add_again(clade, "taxonomy")
            code = delete_old_add_again(taxonomy, "code")
            code.text = up_to_species[uniprot_id].common_name
            codes.add(code.text)
            name = delete_old_add_again(taxonomy, "name")
            name.text = up_to_species[uniprot_id].scientific_name
        return codes

    def _add_taxonomies(self, codes):
        """Add taxonomies element to tree with colours for taxonomy codes"""
        taxonomies = delete_old_add_again(self.root, "taxonomies")
        for no, code in enumerate(codes):
            taxonomy = etree.SubElement(taxonomies, "taxonomy", code=code)
            color = etree.SubElement(taxonomy, "color")
            color.text = self.COLORS[no % len(self.COLORS)]
        return

    def _add_is_enzyme_property(self, enzymes):
        """Add is enzyme property to every clade in the tree"""
        self._populate_stats_fields(enzymes)
        self._add_is_enzyme_property_sequence_nodes(enzymes)
        self._add_is_enzyme_to_clade(self.root.find("phylogeny"), enzymes)
        self._fix_ambiguous_enzyme_flag(self.root.find("phylogeny").find("clade"))
        self._set_branch_event_label(self.root.find("phylogeny").find("clade"))
        self._populate_branch_stats_fields()

    def _populate_branch_stats_fields(self):
        branch_values = [clade.find("property").get("branch_event_label") for clade in self._branch_nodes]
        self.gain_of_function = branch_values.count("gain")
        self.loss_of_function = branch_values.count("loss")
        self.no_change_enzyme = branch_values.count("enzyme")
        self.change_first = branch_values.count("change_first")
        self.no_change_pseudoenzyme = branch_values.count("pseudoenzyme")
        self.save()


    def _add_is_enzyme_property_sequence_nodes(self, enzymes):
        """Adds a property element to every sequence to tag them as enzyme or non enzyme"""
        for clade in self._sequence_nodes:
            idx = add_if_missing(clade, "id")
            idx.text = clade.find("name").text
            id_property = add_if_missing(clade, "property", applies_to="clade",
                                         datatype="xksd:boolean", ref="is enzyme info")
            id_property.text = str(int(idx.text in enzymes))

    def _populate_stats_fields(self, enzymes):
        """Populates the columns in the database related with the stats of this three (no of enzymes, etc)"""
        self.number_enzymes = len(enzymes)
        self.number_non_enzymes = self.sequences.count() - self.number_enzymes
        self.save()

    def _add_is_enzyme_to_clade(self, root, enzymes):
        """Adds enzyme flag to the tree starting from the root"""
        # enzyme enzyme flag to all terminal nodes
        for clade in root.findall("clade"):
            if clade not in self._sequence_nodes:
                self._add_is_enzyme_to_clade(clade, enzymes)

        # goes up the tree and calculates all the branches status as enzyme of nonenzyme
        # ratios close to 0 or  are rounded off

        ratio = mean([float(c.find("property").text) for c in root.findall("clade")])
        ratio_str = "0.5" if isclose(ratio, 0.5) else str(int(round(ratio)))

        new_property = add_if_missing(root, "property", applies_to="clade", datatype="xksd:boolean")
        new_property.text = ratio_str
        new_property.attrib["ref"] = "is enzyme info" if ratio_str == "0.5" else "is enzyme info hide"
    
    def _set_branch_event_label(self, root, parent_is_enzyme=None, transition_seen=False):
        """sets event flat to loss/gain/no change of function"""
        if root.find("property").text == "0.5":
            if parent_is_enzyme is not None:
                if parent_is_enzyme == "1":
                    label = "loss"
                else:
                    label = "gain"
                if transition_seen:
                    label += " sub"
                transition_seen = True
            else:
                label = "change_first"
        else:
            if root.find("property").text == "1":
                label = "enzyme"
            else:
                label = "pseudoenzyme"
        root.find("property").attrib["branch_event_label"] = label

        for clade in root.findall("clade"):
            if clade not in self._sequence_nodes:
                self._set_branch_event_label(clade, root.find("property").text, transition_seen=transition_seen)

    def _fix_ambiguous_enzyme_flag(self, root, parent_is_enzyme=None):
        """fixes cases where kids and parents both have 0.5"""
        if all([child.find("property").text == "0.5" for child in root.findall("clade")]):
            print("fixing ambiguous")
            if parent_is_enzyme is not None:
                root.find("property").text = parent_is_enzyme

        for clade in root.findall("clade"):
            if clade not in self._sequence_nodes:
                self._fix_ambiguous_enzyme_flag(clade, root.find("property").text)

    def save(self, *args, **kwargs):
        """Saves new xml file if necessary"""
        if self._xml_has_changes:
            filepath = pathlib.Path(self.xml_file.path)
            if filepath.is_file():
                filepath.unlink()
            self.xml_file.save(self.filename_from_reference, ContentFile(etree.tostring(self._tree)), False)
        super(Tree, self).save(*args, **kwargs)

def number_to_string(number):
    key = "abcdefghij"
    return "".join([key[int(n)] for n in str(number)])

