"""Models related with prediction of mechanisms"""

# python imports
import collections
import csv
import itertools
import json
import logging
import numpy as np
import pathlib
from collections import defaultdict
import subprocess
from typing import Dict, Tuple, List, Sequence, Set, Optional
from rdkit import Chem
Chem.rdDepictor.SetPreferCoordGen(False)
from lxml import etree
import networkx as nx
import threading


# django imports
from django.db import models
from django.utils.functional import cached_property

# atlas imports
import common.models
from common.models import Entry, Residue
import atlas
import atlas.settings
import common.utils
from atlas.files import MOLCONVERT_EXE
from common.utils import make_mol_equivalent
from common.models.step import atom_array_to_atom_tags
from prediction.predict_mechanism_util import find_matching_mols, mol_is_water_or_proton
from common.models import Rule, StepRule, EntryRule
from prediction.reaction_result import ReactionResult

# Type alias
Edge = Tuple[str, str]

logger = logging.getLogger("script")


class Prediction(models.Model):
    """The chemical reaction catalysed by this enzyme's entry"""
    entry = models.ForeignKey(Entry, on_delete=models.PROTECT)
    prediction_id = models.IntegerField()
    from_pdb = models.BooleanField(default=False)
    structure = models.ForeignKey(common.models.Structure, null=True, on_delete=models.SET_NULL)
    non_mutated_pdb_ids = models.JSONField(default=tuple)
    pdb_ligands_parity = models.JSONField(default=dict)
    residue_mappings = models.JSONField(default=dict)
    ligand_mappings = models.JSONField(default=dict)
    mechanism = models.ForeignKey(common.models.Mechanism, on_delete=models.PROTECT, related_name="prediction_set")
    thread_ident = models.CharField(default="", max_length=15)
    comments = models.TextField(blank=True)
    max_nodes = models.IntegerField(default=100)

    class State(models.TextChoices):
        TO_RUN = 'T'
        RUNNING = 'R'
        INTERRUPTED = 'I'
        ERROR = 'E'
        DONE = 'D'
        KILL = 'K'

    state = models.CharField(max_length=1, choices=State.choices, default=State.TO_RUN)
    progress_status = models.CharField(default="", max_length=100)

    class Meta:
        unique_together = ("entry", "prediction_id")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pdb_residue_distance_to_center = {}

    def save(self, *args, **kwargs):
        if not self.pk:
            self.mechanism = common.models.Mechanism.objects.create(reaction=self.reaction, mechanism_id=self.prediction_id+1000, prediction=True)
            common.models.Step.objects.create(mechanism=self.mechanism, step_id=1)
            common.models.Step.objects.create(mechanism=self.mechanism, step_id=2, is_product=True)
            # does not run for predictions_from_pdb
            if self.reaction.ec_code:
                self.non_mutated_pdb_ids = self._get_non_mutated_pdb_ids()
                self.pdb_ligands_parity = self._get_pdb_ligands_parity()
        super(Prediction, self).save(*args, **kwargs)

    @property
    def rules_to_try(self):
        rules = common.models.Rule.objects.filter(radical_in_step=False)
        rules = rules.filter(version="0.2", steps__mechanism__rating=3).distinct()
        rules = [rule for rule in rules if not rule.has_double_bond_hydrogen]
        return rules

    def is_running(self):

        if self.state == "R":
            if any([str(thread.ident) == self.thread_ident for thread in threading.enumerate()]):
                return True
            else:
                self.state = self.State.INTERRUPTED
                self.progress_status = ""
                self.save()
        return False

    @property
    def reaction(self):
        return common.models.Reaction.objects.get(entry=self.entry, is_reference=True)

    @property
    def step(self):
        return self.mechanism.step_set.get(step_id=1)

    @property
    def step_products(self):
        return self.mechanism.step_set.get(step_id=2)

    @property
    def db_residues(self):
        return common.models.Residue.objects.filter(active_site__entry=self.entry)

    @property
    def possible_structures(self):
        return sorted(common.models.Structure.objects.filter(pdb_id__in=self.pdb_ligands_parity), key=lambda s: (
            len(self.pdb_ligands_parity[s.pdb_id][1]),
            self.pdb_ligands_parity[s.pdb_id][0]), reverse=True)

    @cached_property
    def complete_rules(self):
        """ids of rules that are completely extracted from one step"""
        return set(Rule.objects.filter(step_rules__complete_in_step=True).values_list("id", flat=True))

    @cached_property
    def own_rules(self):
        """ids of the own rules"""
        step_rules = StepRule.objects.filter(complete_in_step=True, step__mechanism__reaction__entry=self.entry)
        return {step_rule.rule_id for step_rule in step_rules}

    def own_rules_json(self):
        return json.dumps(list(self.own_rules))

    @cached_property
    def only_own_rules(self):
        """ids of the rules that have at least one arrow exclusive of this entry"""
        entry_rules = EntryRule.objects.filter(exclusive_in_entry=True, entry=self.entry)
        return {entry_rule.rule_id for entry_rule in entry_rules}

    def only_own_rules_json(self):
        return json.dumps(list(self.only_own_rules))

    @cached_property
    def tried_own_rules(self):
        """set of the own rules ids that were tried (but may not appear in prediction"""
        return set([rule.id for rule in self.rules_to_try if rule.id in self.own_rules])

    @property
    def number_of_own_rules(self):
        """the total number of unique rules in the graph."""
        return len(self.own_rules)

    json_dict_to_output = None

    @cached_property
    def output(self):
        return PredictionOutput(self, self.json_dict_to_output)

    @cached_property
    def run(self):
        return PredictionRun(self)

    @cached_property
    def data_folder(self) -> pathlib.Path:
        media_path = pathlib.Path(atlas.settings.MEDIA_ROOT)
        data_folder = media_path / "prediction_results/{0.entry.owner.username}/{0.entry.macie_id}/{0.prediction_id}/".format(self)
        data_folder.mkdir(exist_ok=True, parents=True)
        return data_folder

    @cached_property
    def mol_folder(self) -> pathlib.Path:
        mol_folder = self.data_folder / "mol/"
        mol_folder.mkdir(exist_ok=True)
        return mol_folder

    @cached_property
    def svg_folder(self) -> pathlib.Path:
        svg_folder = self.data_folder / "svg"
        svg_folder.mkdir(exist_ok=True)
        return svg_folder

    @cached_property
    def scheme_mols_reactants(self):
        """return a list of mols based on the mols in the scheme

        changes the mols in step.mols. should do copy instead?"""
        scheme_mols = []
        atom_counter = collections.Counter()
        for mol in self.step.mols:
            for atom in mol.GetAtoms():
                atom_counter[atom.GetSymbol()] += 1
                if atom.GetSymbol() == "H":
                    atom.SetAtomMapNum(0)
                atom.SetIsotope(atom.GetAtomMapNum())
                # remove non relevant chiral information of rxn_mol to facilitate product detection
                if atom.GetAtomMapNum() not in self.rxn_chiral_reaction_centres_am:
                    atom.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)
            scheme_mols.append(mol)
        print("atoms in reactants", atom_counter)
        return scheme_mols

    @cached_property
    def scheme_mols_products(self):
        """return a list of mols based on the mols in the scheme_products"""
        combined_mol = Chem.MolFromMolFile(self.step_products.scheme.path, removeHs=False, sanitize=False)
        mols = []
        atom_counter = collections.Counter()
        for mol in Chem.rdmolops.GetMolFrags(combined_mol, asMols=True, sanitizeFrags=False):
            mol.UpdatePropertyCache()
            mol = Chem.rdmolops.AddHs(mol, addCoords=True)  # ADDHS
            for atom in mol.GetAtoms():
                atom_counter[atom.GetSymbol()] += 1
                if atom.GetSymbol() == "H":
                    atom.SetIsotope(0)
                atom.SetAtomMapNum(atom.GetIsotope())
                # atom.SetIsotope(atom.GetAtomMapNum())
                if atom.GetAtomMapNum() not in self.rxn_chiral_reaction_centres_am:
                    atom.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)
            mols.append(mol)
        print("atoms in products", atom_counter)
        return mols

    @cached_property
    def scheme_mols_products_from_pdb(self):
        """return a list of mols based on the mols in the scheme_products"""
        scheme_mols = []
        atom_counter = collections.Counter()
        for mol in self.step_products.mols:
            for atom in mol.GetAtoms():
                atom_counter[atom.GetSymbol()] += 1
                if atom.GetSymbol() == "H":
                    atom.SetAtomMapNum(0)
                atom.SetIsotope(atom.GetAtomMapNum())
                # remove non relevant chiral information of rxn_mol to facilitate product detection
                if atom.GetAtomMapNum() not in self.rxn_chiral_reaction_centres_am:
                    atom.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)
            scheme_mols.append(mol)
        print("atoms in reactants", atom_counter)
        return scheme_mols

    @cached_property
    def scheme_mols_products_draft(self):
        """
        return a list of mols to be used as draft of the products scheme based on scheme_mols_reactants
        """
        mols_products_draft = []

        am_to_2d_coord = {}
        for molecule, mol in self.db_molecule_to_scheme_mol.items():
            conformer_2d = mol.GetConformer(0)
            for atom in mol.GetAtoms():
                if atom.GetSymbol() != "H":
                    am_to_2d_coord[atom.GetAtomMapNum()] = conformer_2d.GetAtomPosition(atom.GetIdx())
            if isinstance(molecule, tuple) and isinstance(molecule[0], common.models.Compound):

                pass
            else:
                mols_products_draft.append(Chem.rdmolops.RemoveAllHs(mol))

        # add products of reaction
        for label, rxn_mol in self.reaction.label_to_rxn_mol[1].items():
            mol = Chem.RWMol(rxn_mol)
            # mol.AddConformer(mol.GetConformer(), assignId=True)
            conformer_2d = mol.GetConformer(0)
            mol_idx_to_am = {}
            # setup up scheme am mapping for the atoms of the product
            for atom in rxn_mol.GetAtoms():
                if atom.GetSymbol() != "H":
                    mol_idx_to_am[atom.GetIdx()] = self.rxn_am_to_scheme_am[atom.GetAtomMapNum()]
            for atom in mol.GetAtoms():
                if atom.GetAtomMapNum() not in self.rxn_chiral_reaction_centres_am:
                    atom.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)
                if atom.GetSymbol() == "H":
                    atom.SetAtomMapNum(0)
                else:
                    atom.SetAtomMapNum(mol_idx_to_am[atom.GetIdx()])
                    conformer_2d.SetAtomPosition(atom.GetIdx(), am_to_2d_coord[atom.GetAtomMapNum()])

            mols_products_draft.append(Chem.rdmolops.RemoveAllHs(mol))
        return mols_products_draft

    @property
    def draft_mrv_str_products(self):
        """list of all mols (catalytic residues, cofactors, substrates) as mol objects for products draft"""
        for no, mol in enumerate(self.scheme_mols_products_draft):
            for atom in mol.GetAtoms():
                atom.SetIsotope(atom.GetAtomMapNum())
                atom.SetAtomMapNum(0)
                # atom.SetIsotope(0)
        combined_mol = Chem.RWMol(common.utils.combine_mols_list(self.scheme_mols_products_draft))
        return Chem.MolToMolBlock(combined_mol)

    def scheme_mols_to_scheme_labels(self, mols):
        """map of the mols in the step scheme to their labels"""
        mols_to_labels = {}
        for mol in mols:
            labels = {label for atom in mol.GetAtoms() if atom.HasProp("label") and (label := atom.GetProp("label"))}
            if len(labels) > 1:
                print("WARNING: Mol in step scheme has more than one label: {}".format(labels))
            mols_to_labels[mol] = labels.pop()
        return mols_to_labels

    @cached_property
    def scheme_mols_to_scheme_labels_reactants(self):
        return self.scheme_mols_to_scheme_labels(self.scheme_mols_reactants)

    @cached_property
    def scheme_labels_to_scheme_mols_reactants(self):
        """map of labels in the steps scheme to the correponding mol objects"""
        return {v: k for k, v in self.scheme_mols_to_scheme_labels_reactants.items()}

    @property
    def db_rxn_substrates(self):
        return self.reaction.compoundtype_set.filter(type="reactant").select_related("compound")

    @cached_property
    def db_residue_to_possible_pdb_residues(self):
        """
        return possible pdb residues for each db residue
        """
        res_chain_ids_to_residues = collections.defaultdict(set)
        # a single PDB residue can map to two M-CSA residues (side chain and main chain)
        for res_chain in common.models.ResidueChain.objects.filter(
                residue__active_site__entry=self.entry, chain__structure=self.structure).select_related("chain", "residue"):
            chain_ids = self.structure.get_same_assembly_chains(res_chain.chain.name)
            if chain_ids is None:
                continue
            # if (assembly_chains := self.structure.get_same_assembly_chains(residue_chain.chain.name)) is None:
            # assembly_chains = self.structure.assembly_chains_groups["1"]
            for chain_id in chain_ids:
                res_chain_ids_to_residues[(res_chain.auth_resid, chain_id)].add(res_chain.residue)

        db_residue_to_possible_pdb_residues = defaultdict(list)
        for model in self.structure.assembly_parser:
            for chain in model:
                for pdb_residue in chain:
                    db_residues = res_chain_ids_to_residues.get((pdb_residue.id[1], chain.id), ())
                    for db_residue in db_residues:
                        db_residue_to_possible_pdb_residues[db_residue].append(pdb_residue)
        return db_residue_to_possible_pdb_residues

    # properties that got info from the chosen mappings
    @property
    def ligands_as_pdb_residues(self):
        """list of all the ligands (substrates and cofactors) as biopython PDB residue objects"""
        return [self.structure.get_residue_from_label(mapping_label) for mapping_label in self.ligand_mappings.values()
                if mapping_label != "None"]

    @property
    def cat_residue_as_pdb_residues(self):
        """list of all the catalytic residues as biopython PDB residue objects"""
        return [self.structure.get_residue_from_label(mapping_label) for mapping_label in self.residue_mappings.values()]

    @property
    def pdb_residues(self):
        """list of all the residues (catalytic residues, substrates and cofactors) as biopython PDB residue objects"""
        return self.ligands_as_pdb_residues + self.cat_residue_as_pdb_residues

    @property
    def draft_mrv_str_reactants(self):
        """list of all residues (catalytic residues, cofactors, substrates) as mol objects with no coordinates
        to be opened by marvin js box"""
        mdl_list = ""
        for ligand_label, db_ligand in self.ligand_label_to_db_ligand.items():
            if self.ligand_mappings[ligand_label] != "None":
                molecule, count = db_ligand
                if molecule.name == "hydron":
                    oxonium_mol = common.models.Compound.objects.get(name="oxonium").mol
                    mdl_list += molecule.get_mdl_str(label_extra="_{}".format(count),
                                                     replace_mol=oxonium_mol)
                else:
                    mdl_list += molecule.get_mdl_str(label_extra="_{}".format(count))
        # for catalytic residues
        for db_residue in self.db_residues:
            try:
                mdl_list += db_residue.get_mdl_str(self.db_residue_to_scheme_label[db_residue])
            except KeyError:
                print("no label found for this residue (not being added to scheme): ", db_residue)
        mdl_str = "".join(mdl_list).replace("'", "")
        mrv_str = subprocess.getoutput(f"echo '{mdl_str}' | {MOLCONVERT_EXE} mrv -U -2")
        #  remove chebi:xxx from substrate and add it to mrvExtraLabel
        mrv_str = mrv_str.replace("\\u0020", " ")
        xml = bytes(bytearray(mrv_str, encoding="utf-8"))
        root = etree.fromstring(xml)
        atom_array = root.findall(".//{}atomArray".format("{http://www.chemaxon.com}"))[0]
        if "atomID" in atom_array.keys():
            atom_array = atom_array_to_atom_tags(atom_array)
        for atom in atom_array:
            try:
                alias = atom.attrib['mrvAlias']
                if alias.startswith("chebi") or alias.startswith("res"):
                    atom.attrib['mrvExtraLabel'] = alias
                    atom.attrib['mrvAlias'] = ""
            except KeyError:
                pass
        return etree.tostring(root).decode("utf-8")

    @property
    def draft_mrv_from_pdb(self):
        """list of all residues (catalytic residues, cofactors, substrates) as mol objects with no coordinates
        to be opened by marvin js box in the input_from_pdb prediction page"""
        mdl_list = ""
        # for ligands
        for ligand_label, db_ligand in self.ligand_label_to_db_ligand.items():
            # if self.ligand_mappings[ligand_label] != "None":
            molecule, count = db_ligand
            mdl_list += molecule.get_mdl_str(label_extra="_{}".format(count))
        # for catalytic residues
        for db_residue in self.db_residues:
            mdl_list += db_residue.get_mdl_str(self.db_residue_to_scheme_label[db_residue])
            # label = self.get_reference_label_from_db_residue(db_residue)
            # mdl_list += db_residue.get_mdl_str(label)

        # convert mol to mrv
        mdl_str = "".join(mdl_list).replace("'", "")
        mrv_str = subprocess.getoutput(f"echo '{mdl_str}' | {MOLCONVERT_EXE} mrv -U -2")
        #  remove chebi:xxx from substrate and add it to mrvExtraLabel
        mrv_str = mrv_str.replace("\\u0020", " ")
        print("mrv strXX", mrv_str)
        xml = bytes(bytearray(mrv_str, encoding="utf-8"))
        root = etree.fromstring(xml)
        atom_array = root.findall(".//{}atomArray".format("{http://www.chemaxon.com}"))[0]
        if "atomID" in atom_array.keys():
            atom_array = atom_array_to_atom_tags(atom_array)
        for atom in atom_array:
            try:
                alias = atom.attrib['mrvAlias']
                if alias.startswith("chebi") or alias.startswith("res"):
                    atom.attrib['mrvExtraLabel'] = alias
                    atom.attrib['mrvAlias'] = ""
            except KeyError:
                pass
        return etree.tostring(root).decode("utf-8")

    def get_reference_label_from_db_residue(self, residue):
        return "{0}{1.code}{1.resid}{2}".format(
            "Main " if residue.function_location.startswith("m") else "", residue, residue.assembly_chain_name.upper())

    @cached_property
    def reference_label_to_db_residue(self):
        return {self.get_reference_label_from_db_residue(residue): residue for residue in self.db_residues}

    @cached_property
    def db_residue_to_mapping_label(self):
        return {res: self.residue_mappings[label] for label, res in self.reference_label_to_db_residue.items()}

    @cached_property
    def mapping_label_to_db_residue(self):
        return {label: res for res, label in self.db_residue_to_mapping_label.items()}

    @cached_property
    def db_residue_to_scheme_label(self):
        residue_to_scheme_label = {}
        for db_residue, mapping_label in self.db_residue_to_mapping_label.items():
            chain, _, resid, _ = mapping_label.split(",")
            residue_to_scheme_label[db_residue] = "{}{}{}{}".format(
                "Main" if db_residue.function_location.startswith("m") else "", db_residue.code, resid, chain)
        return residue_to_scheme_label

    @cached_property
    def scheme_label_to_db_residue(self):
        return {v: k for k, v in self.db_residue_to_scheme_label.items()}

    @cached_property
    def db_molecule_to_scheme_mol(self):
        db_molecule_to_mol = {}
        for label, mol in self.scheme_labels_to_scheme_mols_reactants.items():
            if label.startswith("res:"):
                # print(self.scheme_label_to_db_residue)
                db_molecule_to_mol[self.scheme_label_to_db_residue[label[4:]]] = mol
            elif label.startswith("chebi:"):
                db_molecule_to_mol[self.ligand_label_to_db_ligand[label]] = mol
        return db_molecule_to_mol

    def get_db_ligands(self, products=False):
        """list of database ligands (substrate and cofactors) that participate in reaction

        Each ligand is represented by a tuple (molecule, i) where i is the index of that ligand when there are more
        than one substrate/cofactor of the same type

        By default the list is for substrate and cofactors. If products is set to True then the list will be for
        products and cofactors
        """
        db_objects = []  # Substrates/Products and cofactors
        compound_type = "products" if products else "reactant"
        db_objects.extend(self.reaction.compoundtype_set.filter(type=compound_type).select_related("compound"))
        db_objects.extend(common.models.ActiveSiteCofactor.objects.filter(active_site__entry=self.entry)
                          .select_related("cofactor"))
        db_ligands = []
        chebi_id_to_count = collections.defaultdict(int)
        for db_object in db_objects:
            for i in range(db_object.count):
                chebi_id = db_object.molecule.chebi_id
                chebi_id_to_count[chebi_id] += 1
                db_ligands.append((db_object.molecule, chebi_id_to_count[chebi_id]))
        return db_ligands

    @cached_property
    def db_ligands(self):
        """list of database ligands (substate and cofactors) that participate in reaction

        Each ligand is represented by a tuple (molecule, i) where i is the index of that ligand when there are more
        than one substrate/cofactor of the same type
        """
        return self.get_db_ligands()

    @cached_property
    def db_ligands_products(self):
        """list of database ligands (products and cofactors) that are the product of the reaction

        Each ligand is represented by a tuple (molecule, i) where i is the index of that ligand when there are more
        than one product/cofactor of the same type
        """
        return self.get_db_ligands(products=True)

    @cached_property
    def ligand_label_to_db_ligand(self):
        return {"chebi:{0}_{1}".format(db_ligand[0].chebi_id, db_ligand[1]): db_ligand for db_ligand in self.db_ligands}

    @cached_property
    def db_ligand_to_db_mol(self):
        """maps a db_ligand tuple to a mol file. includes the reactants and cofactors"""
        return {db_ligand: Chem.MolFromMolFile(db_ligand[0].mol_file.path) for db_ligand in self.db_ligands}

    @cached_property
    def ligand_label_to_db_mol(self):
        """maps a db_ligand label to a mol file"""
        return {ligand_label: self.db_ligand_to_db_mol[db_ligand]
                for ligand_label, db_ligand in self.ligand_label_to_db_ligand.items()}

    @cached_property
    def db_molecule_to_pdb_residue(self):
        """Maps each residue, and ligand (cofactor or substrate, count) involved in reaction to the corresponding pdb_residue

        TODO should be possible to more than one pdb_residues
        """
        db_molecules_to_pdb_residues = {}
        print("structure", self.structure)

        for db_label, pdb_label in self.ligand_mappings.items():
            db_ligand = self.ligand_label_to_db_ligand[db_label]
            pdb_residue = self.structure.get_residue_from_label(pdb_label) if pdb_label != "None" else None
            db_molecules_to_pdb_residues[db_ligand] = pdb_residue

        for db_label, pdb_label in self.residue_mappings.items():
            db_residue = self.reference_label_to_db_residue[db_label]
            pdb_residue = self.structure.get_residue_from_label(pdb_label)
            db_molecules_to_pdb_residues[db_residue] = pdb_residue
        return db_molecules_to_pdb_residues

    @cached_property
    def pdb_residue_to_db_molecules(self):
        """map a pdb_residue to the db residue or ligands. Each pdb_residue can map to more than one ligand"""
        pdb_residue_to_db_molecules = collections.defaultdict(list)
        for db_molecule, pdb_residue in self.db_molecule_to_pdb_residue.items():
            if pdb_residue is not None:
                pdb_residue_to_db_molecules[pdb_residue].append(db_molecule)
        return pdb_residue_to_db_molecules

    @cached_property
    def db_molecule_to_pdb_residue_mol(self):
        """map each db_molecule (residue or ligand: cofactor and substrate) to a pdb_residue_mol"""
        db_molecule_to_pdb_residue_mol = {}
        for db_molecule, pdb_residue in self.db_molecule_to_pdb_residue.items():

            if pdb_residue is None:
                continue

            if isinstance(db_molecule, common.models.Residue):
                selection = "main_chain" if db_molecule.function_location.startswith("main") else "side_chain"
            else:   # is a ligand: substrate or cofactor
                selection = ""

            pdb_residue_mol = self.structure.get_mol_from_pdb_residue(pdb_residue, atom_selection=selection)
            # needs to be unique
            db_molecule_to_pdb_residue_mol[db_molecule] = pdb_residue_mol
        return db_molecule_to_pdb_residue_mol

    @cached_property
    def pdb_residue_mol_to_scheme_mols(self):
        """map each pdb_residue_mol to one or more scheme mols (see M32 for example)"""
        pdb_residue_mol_to_scheme_mols = collections.defaultdict(list)
        for db_molecule, scheme_mol in self.db_molecule_to_scheme_mol.items():
            pdb_residue_mol = self.db_molecule_to_pdb_residue_mol[db_molecule]
            pdb_residue_mol_to_scheme_mols[pdb_residue_mol].append(scheme_mol)
        return pdb_residue_mol_to_scheme_mols

    @cached_property
    def center_of_geometry(self):
        """the center of geometry of the catalytic residues.
        some of this functionality should be in structure"""
        if not self.residue_mappings:
            return None

        pdb_residue_centers_of_geometry = []
        for pdb_residue_label in self.residue_mappings.values():
            pdb_residue = self.structure.get_residue_from_label(pdb_residue_label)
            pdb_residue_centers_of_geometry.append(self.structure.get_residue_center_of_geometry(pdb_residue))

        return sum(pdb_residue_centers_of_geometry)/len(pdb_residue_centers_of_geometry)

    def distance_to_center_of_geometry(self, pdb_residue):
        """calculates the min distance of this pdb_residue to the center of geometry"""
        key = (pdb_residue.id, pdb_residue.parent.id)
        if key not in self.pdb_residue_distance_to_center:
            if self.center_of_geometry is None:
                self.pdb_residue_distance_to_center[key] = 0
            else:
                self.pdb_residue_distance_to_center[key] =\
                    min([np.linalg.norm(atom.coord - self.center_of_geometry) for atom in pdb_residue])

        return self.pdb_residue_distance_to_center[key]



    CognateData = Dict[Tuple[str, str], List[Tuple[float, str, List[Tuple[str, str, float]]]]]
    _cognate_data: CognateData = {}

    @classmethod
    def get_cognate_data(cls, data_file: pathlib.Path = atlas.files.COGNATE_DATA_FILE_TO_USE) -> CognateData:
        """
        Read data from the cognate file created by Jon.

        For each PDB id and EC number the output dictionary stores a list of
        similarities between a kegg reaction with that EC and all the ligands in
        that pdb. The individual similarity between each kegg cognate ligand PDB
        ligand is given in a list inside that list. Both lists are sorted so the
        most similar ligands appear first.  The actual file contains more
        information that is currently discarded.

        Stores the data in the _cognate_data variable for future calls.

        :output example:
            {('1uvn', '2.7.7.48'): [(1.0, 'R00441', [('GTP', 'C00044', 1.0), ('noMatch', 'C00046', 0.0)]),
                                   (1.0, 'R00444', [('GTP', 'C00201', 1.0), ('noMatch', 'C00046', 0.0)])]}
        """
        if cls._cognate_data == {}:

            logger.info("Start reading cognate similarity data from file: {}".format(data_file))

            with open(data_file, newline='') as csv_file:
                csv_reader = csv.DictReader(csv_file, delimiter=',')
                cls._cognate_data = defaultdict(list)

                # each row corresponds to a kegg reaction that has this EC number
                for row in csv_reader:
                    key = (row["pdb"], row["ec_kegg"])
                    sim_score = float(row["sim_score"])
                    kegg_reaction = row["keggR"]
                    matches = []
                    # each match corresponds to a (pdb ligand, native kegg ligand) pair
                    for ligand_pair_data in row["matches(boundPdb;cognateKegg;sim)"].split("_"):
                        pdb_ligand, kegg_cognate, sim = ligand_pair_data.split(";")
                        matches.append((pdb_ligand, kegg_cognate, float(sim)))
                    cls._cognate_data[key].append((sim_score, kegg_reaction, matches))

            logger.info("Read {} (pdb, ec) pairs from cognate similarity file".format(len(cls._cognate_data)))

        return cls._cognate_data

    @classmethod
    def pdb_to_ligands(cls, pdb_ids: Sequence[str], ec_code: str) -> Dict[str, Tuple[float, List[str]]]:
        """
        Find the pdb files that have ligands similar to the cognate ligands in this EC number.

        Return a dict of PDB ids that map to a list of ligands that are similar to the cognate ligands.
        Uses the ligands of the most similar kegg reaction to the given EC code. Uses Jon cognate data.

        :param pdb_ids: a collection of pdb_ids
        :param ec_code: an EC number like 1.1.1.1

        :output example:pred
            {'1est': (0.8, ['PIP', 'PEP'])}
        """
        cognate_data = cls.get_cognate_data()
        pdb_to_ligands_output: Dict[str, Tuple[float, List[str]]] = {}  # collections.defaultdict(list)
        for pdb_id in pdb_ids:
            kegg_reactions = cognate_data.get((pdb_id, ec_code), [])
            if kegg_reactions:
                # only considers first kegg reaction
                ligands_data = kegg_reactions[0][2]
                found_ligands = []
                for ligand_data in ligands_data:
                    ligand = ligand_data[0]
                    if ligand != "noMatch":
                        found_ligands.append(ligand)
                if found_ligands:
                    pdb_to_ligands_output[pdb_id] = (kegg_reactions[0][0], found_ligands)
        return pdb_to_ligands_output

    def _get_non_mutated_pdb_ids(self):
        """return set of homologous PDB Ids that have the same catalytic residues"""
        mutated_pdbs = set()
        all_pdbs = set()
        for residue in Residue.objects.filter(active_site__entry=self.entry).annotate_res_sequences():
            res_chains = common.models.ResidueChain.objects.filter(residue=residue)\
                .distinct("chain__structure").select_related("chain__structure", "residue")
            for res_chain in res_chains:
                pdb_id = res_chain.chain.structure.pdb_id
                all_pdbs.add(pdb_id)
                if res_chain.code != residue.res_seq[0].code and not residue.function_location.startswith("main"):
                    mutated_pdbs.add(pdb_id)
                    continue
                elif res_chain.auth_resid is None:
                    mutated_pdbs.add(pdb_id)
                    continue
        return tuple(all_pdbs-mutated_pdbs)

    def _get_pdb_ligands_parity(self):
        """return a dict that maps pdb ids to ligands in the PDB file and their similarities according to jon's data"""
        return Prediction.pdb_to_ligands(self.non_mutated_pdb_ids, ec_code=self.reaction.ec_code.code)

    @cached_property
    def step_rc_ams(self) -> Set[int]:
        """
        return a set with the scheme atoms maps of the reaction centres

        """
        return {step_am for rxn_am, step_am in self.rxn_am_to_scheme_am.items()
                if step_am != 0 and rxn_am in self.rxn_reaction_centres_am and not mol_is_water_or_proton(self.step.am_to_mol[step_am])}

    @cached_property
    def rxn_am_to_scheme_am(self) -> Dict[int, int]:
        """return the mapping of the atoms in the rxn to the atoms in the scheme"""
        am_map = {}
        for label, rxn_mol in self.reaction.label_to_rxn_mol[0].items():
            step_mol = self.scheme_labels_to_scheme_mols_reactants[label]
            _, _, map_map = find_matching_mols([rxn_mol, ], [step_mol, ])[rxn_mol]
            am_map.update(map_map)
        return am_map

    @cached_property
    def scheme_am_to_rxn_am_to(self) -> Dict[int, int]:
        """return the mapping of the atoms in the initial scheme to the reactants in the rxn"""
        return {v: k for k, v in self.rxn_am_to_scheme_am.items()}

    def get_rxn_products_clean_mols(self) -> List[Chem.Mol]:
        """
        return a list of rdkit mols that are the products of the overall rxn reaction
        """
        rxn_products_mols = self.reaction.get_mols_from_rxn()[1]
        clean_products = []
        for product in rxn_products_mols:
            for atom in product.GetAtoms():
                atom.SetAtomMapNum(0)
                atom.SetIsotope(0)
                # remove non relevant chiral information of rxn_mol to facilitate product detection
                if atom.GetAtomMapNum() not in self.rxn_chiral_reaction_centres_am:
                    atom.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)
            Chem.SanitizeMol(product)
            product = Chem.AddHs(product)
            clean_products.append(product)
        return clean_products

    @cached_property
    def rxn_reaction_centres_am(self):
        return self.reaction.get_reaction_centres_atoms()

    @cached_property
    def rxn_chiral_reaction_centres_am(self):
        return self.reaction.get_chiral_reaction_centres()


class PredictionRun:
    """
    Holds all the information necessary to do a mechanism prediction
    """
    def __init__(self, prediction):
        self.prediction = prediction

    def check_kill(self):
        # need to reread prediction from database
        self.prediction = Prediction.objects.get(id=self.prediction.id)
        if self.prediction.state == "K":
            self.prediction.state = "I"
            self.prediction.progress_status = ""
            self.prediction.save()
            return True
        else:
            return False

    def run_prediction_thread(self):
        if self.prediction.is_running():
            print("already running")
            return
        self.prediction.thread_ident = threading.get_ident()
        logger.info(f"### Thread id: {self.prediction.thread_ident} ###\n")
        self.prediction.state = Prediction.State.RUNNING
        self.prediction.progress_status = "Setting up search"
        self.prediction.save()
        logger.info("### Starting Setup ###\n")
        self.setup()
        logger.info("### Finished Setup ###\n")
        # remove all svg and mol files in folders
        self.prediction.data_folder.joinpath("info.json").unlink(missing_ok=True)
        self.prediction.data_folder.joinpath("network.json").unlink(missing_ok=True)
        for folder in (self.prediction.svg_folder, self.prediction.mol_folder):
            for conf_file in folder.iterdir():
                conf_file.unlink()
        if self.check_kill():
            return
        return_ok = self.apply_rules(self.prediction.rules_to_try, minimize=True)
        if return_ok:
            self.prediction.state = Prediction.State.DONE
            self.prediction.progress_status = ""
            self.prediction.save()

    def add_and_get_equivalent_mols(self, *mols):
        """set and get equivalent mol. if given a list sets all mols but only returns the last
        replace the mols in the list with the equivalent mols"""
        equivalent_mols = []
        for mol in mols:
            # smiles = Chem.MolToSmiles(mol)
            smiles = Chem.MolToSmiles(make_mol_equivalent(Chem.Mol(mol)))
            if smiles not in self.smiles_to_mol:
                self.smiles_to_mol[smiles] = mol
                for atom in mol.GetAtoms():
                    atom.SetProp("smiles", smiles)
            equivalent_mols.append(self.smiles_to_mol[smiles])
        return equivalent_mols

    def set_and_get_configuration_id(self, mols, **kwargs):
        """check if configuration is unique, if it is also add mols to graph. return configuration id
        extra kwargs will be added to the node
        """
        configuration_id = self.mols_2_conf_id.setdefault(frozenset(mols), len(self.mols_2_conf_id) + 1)
        if configuration_id not in self.G.nodes:
            self.G.add_node(configuration_id, **kwargs)
            self.G.nodes[configuration_id]["mols"] = set(mols)
        return configuration_id

    @cached_property
    def am_to_3d_coord(self):
        """return dict of atom maps in reactants configuration to the 3D coordinate"""
        am_to_3d_coord = {}
        for mol in self.initial_mols:
            conformer_3d = mol.GetConformer(1)
            for atom in mol.GetAtoms():
                if atom.GetSymbol() != "H":
                    am_to_3d_coord[atom.GetAtomMapNum()] = conformer_3d.GetAtomPosition(atom.GetIdx())
        return am_to_3d_coord

    @cached_property
    def ams_to_distance(self) -> Dict[Tuple[int, int], float]:
        """return distances betweeen all pairs of atoms in the configuration"""

        ams_to_distance = {}
        for mol_a in self.initial_mols:
            for atom_a in mol_a.GetAtoms():
                if (am_a := atom_a.GetAtomMapNum()) in self.am_to_3d_coord:
                    correction_a = 0
                    if len(atom_a.GetBonds()) == 1:
                        other_atom_a = atom_a.GetBonds()[0].GetOtherAtom(atom_a)
                        other_atom_a_idx = other_atom_a.GetIdx()
                        bond_length = Chem.rdMolTransforms.GetBondLength(mol_a.GetConformer(1), atom_a.GetIdx(), other_atom_a_idx)
                        correction_a = bond_length * 0.5
                        pos_a = self.am_to_3d_coord[other_atom_a.GetAtomMapNum()]
                    else:
                        pos_a = self.am_to_3d_coord[am_a]
                    for mol_b in self.initial_mols:
                        for atom_b in mol_b.GetAtoms():
                            if (am_b := atom_b.GetAtomMapNum()) in self.am_to_3d_coord:
                                correction_b = 0
                                if len(atom_b.GetBonds()) == 1:
                                    other_atom_b = atom_b.GetBonds()[0].GetOtherAtom(atom_b)
                                    other_atom_b_idx = other_atom_b.GetIdx()
                                    bond_length = Chem.rdMolTransforms.GetBondLength(mol_b.GetConformer(1), atom_b.GetIdx(), other_atom_b_idx)
                                    correction_b = bond_length * 0.5
                                    pos_b = self.am_to_3d_coord[other_atom_b.GetAtomMapNum()]
                                else:
                                    pos_b = self.am_to_3d_coord[am_b]
                                ams_to_distance[tuple(sorted((am_a, am_b)))] = max(
                                    0,
                                    (pos_a - pos_b).Length() - correction_a - correction_b
                                )
        return ams_to_distance

    def setup(self):

        self.smiles_to_mol = {}
        self.mols_2_conf_id = {}
        self.initial_mols = self.prediction.scheme_mols_reactants

        # find 3d coordinates for the reactants state
        self.add_3d_conformer_to_inital_mols_from_pdb()

        # initialise graph
        self.G = nx.MultiGraph()
        # print("mols_2_conf_id", self.mols_2_conf_id)

        # add to tracked and graph
        self.add_and_get_equivalent_mols(*self.initial_mols)
        self.set_and_get_configuration_id(self.initial_mols, is_initial_reactant=True)

        # get initial_product_mols
        if self.prediction.from_pdb:
            self.initial_product_mols = self.prediction.scheme_mols_products_from_pdb
        else:
            self.initial_product_mols = self.prediction.scheme_mols_products
        print([Chem.MolToSmiles(mol) for mol in self.initial_product_mols])

        # add to tracked and graph
        self.initial_product_mols = self.add_and_get_equivalent_mols(*self.initial_product_mols)
        self.set_and_get_configuration_id(self.initial_product_mols, is_initial_product=True)
        # print("p_configuration_id", p_configuration_id)


        # now copy 3d coordinates
        self.am_to_3d_coord
        self.ams_to_distance

        for mol in self.initial_product_mols:
            if len(mol.GetConformers()) == 1:
                mol.AddConformer(mol.GetConformer(), assignId=True)
                conformer_3d = mol.GetConformer(1)
                conformer_3d.Set3D(True)

                for atom in mol.GetAtoms():
                    if atom.GetSymbol() != "H":
                        conformer_3d.SetAtomPosition(atom.GetIdx(), self.am_to_3d_coord[atom.GetAtomMapNum()])


        # print("initial mols", self.initial_mols)
        # print("initial product mols", [Chem.MolToSmiles(mol) for mol in self.initial_product_mols])

        # print("reactants configuration: ", r_configuration_id)
        # print("products configuration: ", p_configuration_id)
        self.explore_from_r = True

    # @profile
    def apply_rules(self, rules: Sequence[Rule], minimize=False) -> bool:
        """
        exaustively apply rules to the initial configuration and all generated configurations

        :param rules: A sequence with all the rules to apply to the active site configuration
        """
        # self.rxn_templates = {rule.id: Chem.rdChemReactions.ReactionFromSmarts(rule.reaction_smarts) for rule in rules}

        logger.info("### Start applying {} rules to active site configuration ###\n".format(len(rules)))

        self.rxn_templates = {}
        self.rule_parts = {}
        self.product_rule_parts = {}
        self.reactant_smarts_rule_parts = {}
        self.product_smarts_rule_parts = {}
        self.rule_reactants_to_products = {}
        self.rule_am_to_part_id = {}
        self.rule_am_to_part_size = {}
        self.rule_to_bond_changes_am = {}
        self.rule_to_strict_ams = {}
        self.rule_to_bond_changes_am_heavy = {}
        self.rule_am_to_atom = {}
        self.rules_to_ignore = set()
        self.mol_to_part_matches = {}
        self.prediction.progress_status = "Preparing Rules"
        self.prediction.save()
        for rule in rules:
            self.rxn_templates[rule.id] = Chem.rdChemReactions.ReactionFromSmarts(rule.reaction_smarts)
            self.rxn_templates[rule.id].smarts = rule.reaction_smarts

            self.rule_reactants_to_products[rule.id] = {}
            self.rule_parts[rule.id] = []
            self.product_rule_parts[rule.id] = []
            self.reactant_smarts_rule_parts[rule.id] = []
            self.product_smarts_rule_parts[rule.id] = []
            self.rule_am_to_part_id[rule.id] = {}
            self.rule_am_to_part_size[rule.id] = {}
            self.rule_to_bond_changes_am_heavy[rule.id] = (set(), set())
            self.rule_am_to_atom[rule.id] = (dict(), dict())

            reactant_bonds_ams = set()
            for part_idx, part in enumerate(rule.reaction_smarts.split(">>")[0].split(".")):
                self.reactant_smarts_rule_parts[rule.id].append(part)
                part_mol = Chem.MolFromSmarts(part)
                self.rule_parts[rule.id].append(part_mol)
                for atom in part_mol.GetAtoms():
                    self.rule_am_to_part_id[rule.id][atom.GetAtomMapNum()] = part_idx
                    self.rule_am_to_atom[rule.id][0][atom.GetAtomMapNum()] = atom
                    self.rule_am_to_part_size[rule.id][atom.GetAtomMapNum()] = part_mol.GetNumHeavyAtoms()
                # self.mol_to_part_matches[part_mol] = dict()

                for bond in part_mol.GetBonds():
                    reactant_bonds_ams.add(
                        frozenset((bond.GetBeginAtom().GetAtomMapNum(), bond.GetEndAtom().GetAtomMapNum()))
                    )
                        # bond.GetBondTypeAsDouble()

            product_bonds_ams = set()
            for part_idx, part in enumerate(rule.reaction_smarts.split(">>")[1].split(".")):
                self.product_smarts_rule_parts[rule.id].append(part)
                part_mol = Chem.MolFromSmarts(part)
                self.product_rule_parts[rule.id].append(part_mol)
                for atom in part_mol.GetAtoms():
                    self.rule_am_to_atom[rule.id][1][atom.GetAtomMapNum()] = atom
                for bond in part_mol.GetBonds():
                    product_bonds_ams.add(
                        frozenset((bond.GetBeginAtom().GetAtomMapNum(), bond.GetEndAtom().GetAtomMapNum()))
                    )
                        # bond.GetBondTypeAsDouble()
            # new bonds and cleaved bonds
            bond_changes_am = (product_bonds_ams - reactant_bonds_ams, reactant_bonds_ams - product_bonds_ams)
            self.rule_to_bond_changes_am[rule.id] = bond_changes_am
            self.rule_to_strict_ams[rule.id] = {am for bonds in bond_changes_am for bond in bonds for am in bond}
            # use heavy atoms instead of H
            for bond_ams in self.rule_to_bond_changes_am[rule.id][0]:
                atoms_am = []
                for atom_am in bond_ams:
                    atom = self.rule_am_to_atom[rule.id][0][atom_am]
                    if atom.GetSymbol() == "H":
                        try:
                            reactant_bond = [bond for bond in self.rule_to_bond_changes_am[rule.id][1] for am in bond if am == atom_am][0]
                            atom_am = [am for am in reactant_bond if am != atom_am][0]
                        except IndexError:
                            print("problem with rule (ignoring): ", rule.id)
                            self.rules_to_ignore.add(rule)
                            # exit()
                        # print("new", atom_am)
                    atoms_am.append(atom_am)
                self.rule_to_bond_changes_am_heavy[rule.id][0].add(frozenset(atoms_am))
            # add templates f r possible intramolecular lyases
            # print(self.rule_to_bond_changes_am_heavy[rule.id])

        if self.check_kill():
            return False
        for rule in self.rules_to_ignore:
            rules.remove(rule)


        iteration = 1
        self.to_explore = {1, 2}
        self.explored = set()
        while start_node_id := self._get_next_node_to_explore():
            logger.info("Applying rules to configuration id: {}.".format(start_node_id))

            for mol in self.G.nodes[start_node_id]["mols"]:
                if mol not in self.mol_to_part_matches:
                    self.mol_to_part_matches[mol] = {}
                    for rule_parts in self.rule_parts.values():
                        for rule_part in rule_parts:
                            self.mol_to_part_matches[mol][rule_part] = []
                            matches = mol.GetSubstructMatches(rule_part)
                            for match in matches:
                                self.mol_to_part_matches[mol][rule_part].append(
                                    {rule_part.GetAtomWithIdx(rule_idx).GetAtomMapNum(): r_idx
                                     for rule_idx, r_idx in enumerate(match)})

            for no, rule in enumerate(rules):
                self._apply_rule_to_configuration(start_node_id, rule, minimize, iteration)
            iteration += 1
            if self.check_kill():
                return False

        self.prediction.progress_status = "Saving Output"
        self.prediction.save()
        self._write_network_json_file()
        return True

    def get_templates_for_duplicated_reactants(self, reactants, rule):
        """
        return a list of templates and reactants for duplicated reactants
        when the same reactant matches more than one rule parts the rxn_template and reactants need to be merged
        accordingly

        # 3 possibilities
        # 1.rule parts match the same strict RCs -> ignore
        # 2.rule parts match difference RCs altogehther -> add parentheses
        # 3.rule parts match different strict RCs but share some RCs -> create new rule with 
        """
        # print(f"duplicated reactants. rule: {rule.id}")
        #: each elements of the top level list corresponds to a unique reactant, and each sublist contains
        # all the reactant rule parts that match that reactant
        reactant_to_rule_parts_idx = collections.defaultdict(list)
        for rule_part_idx, reactant in enumerate(reactants):
            reactant_to_rule_parts_idx[reactant].append(rule_part_idx)

        leftside_new_rule_parts = []

        smarts_to_repeated_am_groups = {}
        for reactant, rule_parts_idx in reactant_to_rule_parts_idx.items():
            rule_parts = [self.rule_parts[rule.id][r_idx] for r_idx in rule_parts_idx]

            # if one reactant matches only one rule part
            if len(rule_parts) == 1:
                leftside_new_rule_parts.append([self.reactant_smarts_rule_parts[rule.id][rule_parts_idx[0]]])
            # if one reactant matches several rule_parts
            else:
                leftside_new_rule_parts.append([])
                rule_part_matches = [self.mol_to_part_matches[reactant][rule_part] for rule_part in rule_parts]
                for matches_combination in itertools.product(*rule_part_matches):
                    reactant_idx_to_rule_ams = collections.defaultdict(set)
                    for match in matches_combination:
                        for rule_am, reactant_atom_idx in match.items():
                            reactant_idx_to_rule_ams[reactant_atom_idx].add(rule_am)
                    repeated_am_groups = [list(ams) for ams in reactant_idx_to_rule_ams.values() if len(ams) > 1]
                    repeated_ams = {am for am_group in repeated_am_groups for am in am_group}
                                    
                    if self.rule_to_strict_ams[rule.id] & repeated_ams:
                        # print("rule parts match same strict RCS. ignoring")
                        # 1.rule parts match the same strict RCs -> ignore
                        continue

                    # 2. no merging necessary will add parentheses if it detects two rules parts for a reactant 
                    # 3.rule parts match different strict RCs but share some RCs -> create new rule by merging
                    # non strict RCs
                    # create left side of the reaction
                    new_rule_part = merge_rule_parts(rule_parts, repeated_am_groups)

                    smarts = Chem.MolToSmarts(new_rule_part).replace("&", "")
                    smarts_to_repeated_am_groups[smarts] = repeated_am_groups
                    leftside_new_rule_parts[-1].append(smarts)

        rxn_templates = []
        for rule_parts in itertools.product(*leftside_new_rule_parts):
            leftside_smarts = ".".join([f"({rule_part})" if "." in rule_part else rule_part
                                        for rule_part in rule_parts])
            repeated_ams = []
            for smarts in rule_parts:
                if (repeated := smarts_to_repeated_am_groups.get(smarts)):
                    repeated_ams.extend(repeated)
            p_rule_parts = merge_rule_parts(self.product_rule_parts[rule.id], repeated_ams)
            rigthside_smarts = Chem.MolToSmarts(p_rule_parts).replace("&", "")
            reaction_smarts = f"{leftside_smarts}>>{rigthside_smarts}"
            template = Chem.rdChemReactions.ReactionFromSmarts(reaction_smarts)
            template.smarts = reaction_smarts
            rxn_templates.append((template, list(reactant_to_rule_parts_idx.keys())))

        return rxn_templates

    def get_products_for_reactants_and_rule(self, reactants, rule):
        """Get the products for the application of a specific list of reactants and a rule"""
        # reactants x rule combinations are cached in self.rule_reactants_to_products
        if reactants not in self.rule_reactants_to_products[rule.id]:
            # check if two or more rule parts match the same reactant
            if len(reactants) != len(set(reactants)):
                rxn_templates = self.get_templates_for_duplicated_reactants(reactants, rule)
            else:
                rxn_templates = [(self.rxn_templates[rule.id], reactants)]
            # print(rule.id, len(rxn_templates))

            checked_products_possibilities = dict()
            for rxn_template, template_reactants in rxn_templates:
                max_products = 1
                for i, reactant in enumerate(reactants):
                    if len(self.mol_to_part_matches[reactant][self.rule_parts[rule.id][i]]) > 1:
                        # print("max_products")
                        max_products = 1000

                possible_prods = rxn_template.RunReactants(template_reactants, maxProducts=max_products)

                if max_products != 1:
                    # print("start", len(possible_prods))
                    possible_prods = {tuple({Chem.MolToSmiles(p) for p in ps}): ps for ps in possible_prods}.values()
                    # print("end", len(possible_prods))

                for products in possible_prods:
                    checked_products_reaction_data = self.get_reaction_data_for_products(
                        template_reactants, products, rule)
                    if checked_products_reaction_data is None:
                        pass
                    else:
                        checked_products, reaction_data = checked_products_reaction_data
                        checked_products_possibilities[frozenset(checked_products)] = reaction_data

            self.rule_reactants_to_products[rule.id][reactants] = checked_products_possibilities
        else:
            pass
        return self.rule_reactants_to_products[rule.id][reactants]

    def get_reaction_data_for_products(self, template_reactants, products, rule):
        """return checked products and reaction data for a given template with matching reactants"""

        # reaction_result = ReactionResult(template_reactants, products, rule, prediction_run=self)
        try:
            reaction_result = ReactionResult(template_reactants, products, rule, prediction_run=self)
        except KeyError as e:
            print("Error in reaction result, needs debugging", e)
            return
        checked_products = set()

        for product in reaction_result.products:
            try:
                 checked_products.add(self.add_and_get_equivalent_mols(product)[0])
                 product.UpdatePropertyCache()
            except Chem.rdchem.AtomValenceException as e:
                print("problem with rule: ", rule.id, e)
                print(Chem.MolToSmiles(product))
                print(".".join([Chem.MolToSmiles(mol) for mol in template_reactants]))
                # print("new rule", self.rxn_templates[rule.id].smarts)
                return

        max_bond_length = reaction_result.max_bond_length
        # if max_bond_length > 8: # WIP USE VARIABLE
            # return
        score = (max(4, max_bond_length) - 3)**2

        bonds_away_from_rc = reaction_result.get_bonds_away_from_rxn_rc()
        if bonds_away_from_rc == 1:
            score *= 0.50
            # score *= 0.25


        reaction_data = {
            "max_bond_length": round(max_bond_length, 2),
            "score": round(score, 2),
            "rcs_ams": list(reaction_result.reaction_centres_am),
            # "rcs_idx_r":  reaction_result.rcs_idx_r,
            # "rcs_idx_p":  rcs_idx_p,
        }
        return frozenset(checked_products), reaction_data

    # @profile
    def _apply_rule_to_configuration(self, node_id: int, rule: Rule, minimize: bool, iteration: int):
        """
        Apply a single rule to a single active site configuration and generate all possible products

        Called by the method self.apply rules

        :param configuration_id: the unique identifier for this configuration
        :param rule: a Rule db object
        """
        #: A list of lists. Each main list corresponds to a rule part and is populated by all the molecules in
        # the active site that match that rule part. The same active site molecule may match different rule parts
        reactants_lists = []
        for part in self.rule_parts[rule.id]:
            reactants_lists.append([])
            for mol in self.G.nodes[node_id]["mols"]:
                if self.mol_to_part_matches[mol][part]:
                    reactants_lists[-1].append(mol)

        for i, reactants in enumerate(itertools.product(*reactants_lists)):
            products_possibilities_data = self.get_products_for_reactants_and_rule(reactants, rule)
            # print(products_possibilities_data)

            for products, reaction_data in products_possibilities_data.items():
                configuration_mols = (self.G.nodes[node_id]["mols"] - set(reactants)) | products
                new_node_id = self.set_and_get_configuration_id(configuration_mols)
                ignore = False
                if new_node_id in self.G[node_id]:
                    if new_node_id == self.G[node_id][new_node_id][0]["start"]:
                        # ignore reversed direction
                        continue
                    else:
                        for _, d in self.G[node_id][new_node_id].items():
                            if rule.id == d["rule_id"] and reaction_data["max_bond_length"] > d["max_bond_length"]:
                                ignore = True
                                # print("edge exists", d["max_bond_length"], reaction_data["max_bond_length"])

                if ignore:
                    continue
                edge_no = self.G.add_edge(node_id, new_node_id)
                edge = self.G.edges[node_id, new_node_id, edge_no]


                # rcs_idx_r = {self.G.nodes[node_id]["mols"].index(mol): list(idxs)
                             # for mol, idxs in reaction_data["rcs_idx_r"].items()}
                # rcs_idx_p = {self.G.nodes[new_node_id]["mols"].index(mol): list(idxs)
                             # for mol, idxs in reaction_data["rcs_idx_p"].items()}
                edge["start"] = node_id
                edge["target"] = new_node_id
                edge["edge_no"] = edge_no
                edge["max_bond_length"] = reaction_data["max_bond_length"]
                edge["score"] = reaction_data["score"]
                edge["iteration"] = iteration
                edge["rule_id"] = rule.id
                edge["rcs_ams"] = reaction_data["rcs_ams"]
                # edge["rcs_idx_r"] = rcs_idx_r
                # edge["rcs_idx_p"] = rcs_idx_p

    def _get_next_node_to_explore(self) -> Optional[int]:
        """return next node to be explored or None if there are None or max number of nodes was reached"""
        if not self.to_explore:
            logger.info(f"{self.G.number_of_nodes()} configurations ({len(self.explored)} explored).")
            self.prediction.progress_status = "Searching {}/{}".format(
                    len(self.explored), self.prediction.max_nodes)
            self.prediction.save()
            n_in_a_row = 1 if len(self.explored) < 100 else 10

            if len(self.explored) < self.prediction.max_nodes:
                start_nodes = (1, 2) if self.explore_from_r else (2, 1)
                self.explore_from_r = not self.explore_from_r

                for start_node in start_nodes:
                    for k in nx.single_source_dijkstra_path_length(self.G, start_node, weight="score"):
                        if k not in self.explored:
                            self.to_explore.add(k)
                            if len(self.to_explore) >= n_in_a_row:
                                break
                    else:
                        continue
                    break
        if self.to_explore:
            next_node = self.to_explore.pop()
            self.explored.add(next_node)
            self.G.nodes[next_node]["explored"] = True
            return next_node
        return


    def _set_nodes_depth_and_relabel(self) -> Dict[int, str]:
        """
        set depth of nodes as shortest path to the Reactant and relabel nodes accordingly

        return node label mapping
        """
        r_node_id = 1
        p_node_id = 2
        depths_r = dict(nx.shortest_path_length(self.G, r_node_id))
        depths_p = dict(nx.shortest_path_length(self.G, p_node_id))
        nodes_at_length_r: Dict[int, int] = collections.defaultdict(int)
        labels_mapping = {1: "R", 2: "P"}

        for node in nx.node_connected_component(self.G, r_node_id):
            depth = depths_r[node]
            self.G.nodes[node]["depth"] = depth
            nodes_at_length_r[depth] += 1
            if node not in labels_mapping:
                labels_mapping[node] = "{}{}".format(depth, common.utils.label_int_to_letters(nodes_at_length_r[depth]))

        # relabel nodes that are only connect to the products, if they exist
        if len(labels_mapping) < len(self.G.nodes):
            print("R and P subgraphs are not connected")

            max_r_depth = max(depths_r.values())
            max_p_depth = max(depths_p.values())
            nodes_at_inverted_depth = collections.defaultdict(int)

            for node in nx.node_connected_component(self.G, p_node_id):
                depth = depths_p[node]
                inverted_depth = 1 + max_r_depth + max_p_depth - depth
                nodes_at_inverted_depth[inverted_depth] += 1
                self.G.nodes[node]["depth"] = inverted_depth
                if node not in labels_mapping:
                    labels_mapping[node] = "p{}{}".format(
                        inverted_depth, common.utils.label_int_to_letters(nodes_at_inverted_depth[inverted_depth]))

        nx.relabel_nodes(self.G, labels_mapping, copy=False)
        return labels_mapping

    def find_bonds_diff(self, mol1, mol2):
        """
        Return bonds that exist in mol1 but not mol2 and vice versa

        Assumes mols have the same atoms and idxs are sorted equally.
        """
        output = []
        for mol_a, mol_b in [(mol1, mol2), (mol2, mol1)]:
            output.append([bond for bond in mol_a.GetBonds()
                           if not mol_b.GetBondBetweenAtoms(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())])
        return output

    def _set_edge_new_and_broken_bonds(self, edge):
        """
        add a property to each edge with broken and new bonds info

        assumes combined_mols in edge are in the same order
        """
        cleaved_bonds, formed_bonds = self.find_bonds_diff(edge["start_combined_mol"], edge["end_combined_mol"])
        edge["new_bonds"] = [(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()) for bond in formed_bonds]
        edge["cleaved_bonds"] = [(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()) for bond in cleaved_bonds]

    def _write_network_json_file(self, prune=True):
        """
        Write the network to a file.

        Json seems to work best when including more data like the curly arrows
        """
        node_labels_mapping = self._set_nodes_depth_and_relabel()
        json_dict = {"elements": [], "info": {}}
        json_dict["info"]["pdb_id"] = self.prediction.structure.pdb_id

        if False:# prune:
            nodes_to_keep = {node for node, data in self.G.nodes(data=True) if data.get("explored")}
            print(nodes_to_keep)
            for node, data in self.G.nodes(data=True):
                print(data)
                depths = {self.G.nodes[neigh].get("depth") for neigh in self.G.neighbors(node)}
                node_depth = data.get("depth")
                if node_depth + 1 in depths and node_depth-1 in depths:
                    nodes_to_keep.add(node)
            subgraph = self.G.subgraph(nodes_to_keep)
        else:
            subgraph = self.G

        logger.info("saving svgs")
        for node, data in subgraph.nodes(data=True):
            data_to_save = {k: v for k, v in data.items()
                            if k in {"is_initial_reactant", "is_initial_product", "explored"}}
            json_dict["elements"].append({"data": {"id": node, **data_to_save}})
            self.save_configuration_svg(node)
            # self.save_configuration_mol(node)
        logger.info("finished")
        logger.info("Saving {} nodes".format(len(subgraph.nodes())))

        for _, _, data in subgraph.edges(data=True):
            start, target = [node_labels_mapping[data[pos]] for pos in ("start", "target")]
            edge_id = "{}_{}_{}".format(start, target, data["edge_no"])

            # remove self edges
            if start == target:
                continue

            json_dict["elements"].append(
                    {"data": {
                        "id": edge_id,
                        "source": start, "target": target,
                        "rule_id": data["rule_id"],
                        "max_bond_length": data["max_bond_length"],
                        "score": data["score"],
                        "iteration": data["iteration"],
                        "rcs_ams": data["rcs_ams"],
                        # "rcs_idx_r": data["rcs_idx_r"],
                        # "rcs_idx_p": data["rcs_idx_p"],
                        # "new_bonds": data["new_bonds"],
                        # "cleaved_bonds": data["cleaved_bonds"],  # wip todo
                        }})

        self.prediction.json_dict_to_output = json_dict
        prediction_output = self.prediction.output

        logger.info("saving paths summary to info")
        prediction_output.save_paths_summary_to_info()
        logger.info("finished")
        logger.info("graph stats in json: {}".format(json_dict["info"]))

        info_json_filename = self.prediction.data_folder.joinpath("info.json")
        with open(info_json_filename, "w") as info_json_file:
            info_json_file.write(json.dumps(json_dict["info"]))

        network_json_filename = self.prediction.data_folder.joinpath("network.json")
        with open(network_json_filename, "w") as network_json_file:
            network_json_file.write(json.dumps(json_dict))

    def save_configuration_mol(self, configuration_id: int, combined_mol: Optional[Chem.Mol] = None, save_3d=True):
        """save a mol file with this active site 2D or 3D configuration"""
        if combined_mol is None:
            combined_mol = Chem.RWMol(common.utils.combine_mols_list(list(self.G.nodes[configuration_id]["mols"])))
            # combined_mol = self.G.nodes[configuration_id]["combined_mol"]
        mol_pathfilename = self.prediction.mol_folder / "{}.mol".format(configuration_id)
        Chem.MolToMolFile(combined_mol, mol_pathfilename.as_posix(), confId=0)#int(save_3d))
        # logger.info("Saved {} mol file for configuration: {}.".format("3D" if save_3d else "2D", configuration_id))

    def save_configuration_svg(self, configuration_id: int):
        """save a 2D curly arrow scheme svg file with this configuration"""
        mols = []
        ams = []
        for mol in self.G.nodes[configuration_id]["mols"]:
            mol = Chem.rdmolops.RemoveAllHs(mol)
            for atom in mol.GetAtoms():
                ams.append(atom.GetAtomMapNum())
                atom.SetAtomMapNum(0)
                atom.SetIsotope(0)
                atom.SetProp("ref", "4")
            mol.UpdatePropertyCache()
            mols.append(mol)
        combined_mol = Chem.RWMol(common.utils.combine_mols_list(mols))
        
        drawer = Chem.Draw.rdMolDraw2D.MolDraw2DSVG(450, 450)
        opts = drawer.drawOptions()
        opts.includeAtomTags = True
        # opts.atomHighlightsAreCircles = True
        opts.bondLineWidth = 2
        drawer.DrawMolecule(combined_mol)
        drawer.TagAtoms(combined_mol)
        drawer.FinishDrawing()
        svg_text = drawer.GetDrawingText()

        # add atom map info to atom-selector
        svg_tree = etree.fromstring(bytes(svg_text, encoding='iso-8859-1'))
        namespaces = {"ns": "http://www.w3.org/2000/svg"}
        circles = svg_tree.xpath("//ns:circle[contains(@class, 'atom-selector')]", namespaces=namespaces)
        if not circles:
            ## rdkit does not save this info consistently
            circles = svg_tree.xpath("//ns:ellipse[contains(@class, 'atom-')]", namespaces=namespaces)

        # for no, circle in enumerate(circles):
            # circle.attrib["class"] += f" am-{ams[no]}"
        for no, am in enumerate(ams):
            circles[no].attrib["class"] += f" am-{am}"
        svg_text = etree.tostring(svg_tree, pretty_print=True).decode(encoding='iso-8859-1')

        svg_pathfilename = self.prediction.svg_folder / "{}.svg".format(configuration_id)
        with open(svg_pathfilename.as_posix(), 'w') as svg_file:
            svg_file.write(svg_text)
        # logger.info("Saved svg file for configuration: {}.".format(configuration_id))


    def copy_chirality_information(self, mol_from, mol_to):
        """
        Copy atom and bond direction information from one RDKIT mol to the other

        Atom and bond indexes must be the same in the two molecules
        """
        for atom in mol_from.GetAtoms():
            mol_to.GetAtomWithIdx(atom.GetIdx()).SetChiralTag(atom.GetChiralTag())
        for bond in mol_from.GetBonds():
            mol_to.GetBondWithIdx(bond.GetIdx()).SetBondDir(bond.GetBondDir())
            mol_to.GetBondWithIdx(bond.GetIdx()).SetStereo(bond.GetStereo())


    def add_3d_conformer_to_inital_mols_from_pdb(self):
        """
        add a second conformer to initial_mols with 3d information coming from the PDB

        if molecule is not present do not add 3D conformer
        """
        pathlib.Path("/tmp/mol/").mkdir(exist_ok=True)
        for n, (db_molecule, mol) in enumerate(self.prediction.db_molecule_to_scheme_mol.items()):

            # mol with the pdb mol and the 3d coordinates
            pdb_residue_mol = self.prediction.db_molecule_to_pdb_residue_mol[db_molecule]

            # find common substructure between the scheme_mol and the pdb_mol
            tmp_mol = Chem.Mol(mol)
            for atom in tmp_mol.GetAtoms():
                if atom.GetSymbol() in {"*", }:
                    atom.SetAtomicNum(6)
                atom.SetIsotope(atom.GetAtomicNum())
            for atom in pdb_residue_mol.GetAtoms():
                atom.SetIsotope(atom.GetAtomicNum())
            Chem.rdmolops.SanitizeMol(tmp_mol)
            # need to embed first otherwise it sometimes causes problems during the minimisation
            Chem.AllChem.EmbedMolecule(tmp_mol)
            pdb_conformer = pdb_residue_mol.GetConformer()
            if tmp_mol.GetNumAtoms() == 1:
                tmp_mol.GetConformer().SetAtomPosition(0, pdb_conformer.GetAtomPosition(0))
            else:
                mcs = Chem.rdFMCS.FindMCS((tmp_mol, pdb_residue_mol), timeout=2,
                                          atomCompare=Chem.rdFMCS.AtomCompare.CompareIsotopes)

                mcs2pdb = pdb_residue_mol.GetSubstructMatch(mcs.queryMol)
                mcs2tmp = tmp_mol.GetSubstructMatch(mcs.queryMol)
                mol_to_pdb_idx = dict(zip(mcs2tmp, mcs2pdb))

                mp = Chem.AllChem.MMFFGetMoleculeProperties(tmp_mol)
                ff = Chem.AllChem.MMFFGetMoleculeForceField(tmp_mol, mp, confId=0)

                # get coords of overlapping atoms from PDB and freeze these atoms in the optimisation
                tmp_conformer = tmp_mol.GetConformer()
                for atom in tmp_mol.GetAtoms():
                    if (atom_idx := atom.GetIdx()) in mol_to_pdb_idx:
                        pdb_coord = pdb_conformer.GetAtomPosition(mol_to_pdb_idx[atom_idx])
                        tmp_conformer.SetAtomPosition(atom_idx, pdb_coord)
                        ff.MMFFAddPositionConstraint(atom_idx, .1, 100)

                Chem.MolToMolFile(tmp_mol, "/tmp/mol/pre_{}.mol".format(n))
                ff.Minimize(maxIts=2000)
                Chem.MolToMolFile(tmp_mol, "/tmp/mol/end_{}.mol".format(n))

            mol.AddConformer(tmp_mol.GetConformer(), assignId=True)

            Chem.MolToMolFile(pdb_residue_mol, "/tmp/mol/pdb_residue_{}.mol".format(n))
            Chem.MolToMolFile(tmp_mol, "/tmp/mol/tmp_mol_{}.mol".format(n))


class PredictionOutput():

    def __init__(self, prediction, initial_json_dict=None):
        self.prediction = prediction
        # use initial_json if dict is already in memory (does not come from file)
        self.initial_json_dict = initial_json_dict
        self.info_json_dict = {}
        # print("Reading prediction for ", self.prediction.entry)


    @cached_property
    def graph(self):
        """a nx.Graph object with these prediction results"""
        if self._graph_json_content is not None:
            graph = nx.MultiGraph()
            for element in self._graph_json_content["elements"]:
                data = element["data"]
                element_id = data["id"]
                if isinstance(element_id, int) or "_" not in element_id:  # element is a node
                    graph.add_node(element_id, **data)
                    if "is_initial_reactant" in data:
                        self.initial_reactant_node_id = element_id
                    if "is_initial_product" in data:
                        self.initial_product_node_id = element_id
                else:                      # element is an edge
                    start, end, i = element_id.split("_")
                    graph.add_edge(start, end, key=i, **data)
            return graph
        else:
            return None

    @property
    def exists(self):
        print("reading", self.prediction.entry)
        return bool(self._graph_json_info)

    @property
    def info(self):
        """return a dict with information on the prediction mechanism"""
        return self._graph_json_info

    @property
    def pdb_id(self):
        return self.info["pdb_id"]

    @cached_property
    def other_rules(self):
        """ids of the rules that are seen in other mechanisms (independently of seen here or not)"""
        other_rules = {rule_id for rule_id in self.rules_found if rule_id not in self.prediction.only_own_rules}
        return other_rules

    @cached_property
    def own_edges(self):
        return [(s, e, k) for s, e, k, data in self.graph.edges(data=True, keys=True)
                if data.get("rule_id", "") in self.prediction.own_rules]

    @cached_property
    def own_graph(self):
        """a nx.Graph that contains only edges and nodes accessible with these mechanism own rules"""
        return self.graph.edge_subgraph(self.own_edges)

    @cached_property
    def blind_edges(self):
        """include rules that are seen in other mechanism (could be seen in this mechanism or not)"""
        return [(s, e, k) for s, e, k, data in self.graph.edges(data=True, keys=True)
                if data.get("rule_id", "") in self.other_rules]

    @cached_property
    def blind_graph(self):
        """a nx.Graph that contains only edges that must be seen in other edges"""
        return self.graph.edge_subgraph(self.blind_edges)

    def shortest_reaction_path(self, graph):
        """return the subgraph that contains the smallest path between R and P"""
        try:
            return nx.shortest_path(graph, self.initial_reactant_node_id, self.initial_product_node_id)
        except (nx.NetworkXNoPath, nx.NodeNotFound):
            return []

    def graph_nedges(self, graph):
        initial_edges = graph.edges(data=True, keys=True)
        nedges_graph = graph.copy()
        for s, e, k, data in initial_edges:
            new_node = "{}_{}_{}".format(s, e, k)
            nedges_graph.add_node(new_node, **data)
            nedges_graph.add_edge(s, new_node)
            nedges_graph.add_edge(new_node, e)
        nedges_graph.remove_edges_from(initial_edges)
        return nedges_graph

    def shortest_reaction_paths_nedges(self, graph):
        try:
            out = list(nx.all_shortest_paths(self.graph_nedges(graph), self.initial_reactant_node_id, self.initial_product_node_id))
            return out
        except (nx.NetworkXNoPath, nx.NodeNotFound):
            return []

    def shortest_reaction_paths_depths(self, graph):
        return [max([data.get("iteration", 0) for _, _,  data in graph.subgraph(node_ids).edges(data=True)])
                for node_ids in self.shortest_reaction_paths_nedges(graph)]

    def shortest_reaction_paths_scores(self, graph):
        return [max([data.get("score", 0) for _, _,  data in graph.subgraph(node_ids).edges(data=True)])
                for node_ids in self.shortest_reaction_paths_nedges(graph)]

    def shortest_reaction_paths_max_bond_lengths(self, graph):
        return [max([data.get("max_bond_length", 0) for _, _,  data in graph.subgraph(node_ids).edges(data=True)])
                for node_ids in self.shortest_reaction_paths_nedges(graph)]

    def shortest_reaction_paths_depths_str(self, graph):
        shortest_reaction_paths_depths = self.shortest_reaction_paths_depths(graph)
        if not shortest_reaction_paths_depths:
            return "-"
        else:
            min_depth = min(shortest_reaction_paths_depths)
            max_depth = max(shortest_reaction_paths_depths)
            if min_depth != max_depth:
                return "{} - {}".format(min_depth, max_depth)
            else:
                return str(min_depth)

    ## properties for template; could be done dinamically
    # all
    def shortest_reaction_paths_depths_str_all(self):
        try:
            return self.info["shortest_reaction_paths_depths_str_all"]
        except KeyError:
            return self.shortest_reaction_paths_depths_str(self.graph)

    def shortest_reaction_paths_n_all(self):
        try:
            return self.info["shortest_reaction_paths_n_all"]
        except KeyError:
            return len(self.shortest_reaction_paths_nedges(self.graph))

    def shortest_reaction_path_length_all(self):
        try:
            return self.info["shortest_reaction_path_length_all"]
        except KeyError:
            return len(self.shortest_reaction_path(self.graph))

    # own
    def shortest_reaction_paths_depths_str_own(self):
        try:
            return self.info["shortest_reaction_paths_depths_str_own"]
        except KeyError:
            return self.shortest_reaction_paths_depths_str(self.own_graph)

    def shortest_reaction_paths_n_own(self):
        try:
            return self.info["shortest_reaction_paths_n_own"]
        except KeyError:
            return len(self.shortest_reaction_paths_nedges(self.own_graph))

    def shortest_reaction_path_length_own(self):
        try:
            return self.info["shortest_reaction_path_length_own"]
        except KeyError:
            return len(self.shortest_reaction_path(self.own_graph))

    # blind
    def shortest_reaction_paths_depths_str_blind(self):
        try:
            return self.info["shortest_reaction_paths_depths_str_blind"]
        except KeyError:
            return self.shortest_reaction_paths_depths_str(self.blind_graph)

    def shortest_reaction_paths_n_blind(self):
        try:
            return self.info["shortest_reaction_paths_n_blind"]
        except KeyError:
            return len(self.shortest_reaction_paths_nedges(self.blind_graph))

    def shortest_reaction_path_length_blind(self):
        try:
            return self.info["shortest_reaction_path_length_blind"]
        except KeyError:
            return len(self.shortest_reaction_path(self.blind_graph))

    def number_of_explored_nodes(self):
        try:
            return self.info["number_of_explored_nodes"]
        except KeyError:
            return len([node for node, data in self.graph.nodes(data=True) if data.get("explored", False)])

    def number_of_nodes(self):
        try:
            return self.info["number_of_nodes"]
        except KeyError:
            return len(self.graph.nodes)

    def save_paths_summary_to_info(self):
        properties = [
            "number_of_explored_nodes",
            "number_of_nodes",
            # "number_of_edges",
            "shortest_reaction_paths_depths_str_all",
            "shortest_reaction_paths_n_all",
            "shortest_reaction_path_length_all",
            "shortest_reaction_paths_depths_str_own",
            "shortest_reaction_paths_n_own",
            "shortest_reaction_path_length_own",
            "shortest_reaction_paths_depths_str_blind",
            "shortest_reaction_paths_n_blind",
            "shortest_reaction_path_length_blind",
        ]
        for property in properties:
            print(property)
            self._graph_json_content["info"][property] = self.__getattribute__(property)()
            print("f")

    @cached_property
    def rules_found(self):
        """ids of the rules found in this prediction"""
        return {data["rule_id"] for s, e, data in self.graph.edges(data=True)}

    @property
    def number_of_rules(self):
        """the total number of unique rules in the graph."""
        return len(self.rules_found)

    @cached_property
    def _graph_json_content(self):
        """return contents of graph json file into a python dict"""
        if self.initial_json_dict is not None:
            return self.initial_json_dict
        self._graph_filepath = self.prediction.data_folder / "network.json"
        if self._graph_filepath.exists():
            with open(self._graph_filepath, "r") as graph_file:
                return json.loads(graph_file.read())
        else:
            return None

    @cached_property
    def _graph_json_info(self):
        """return contents of info json file into a python dict"""
        if self.info_json_dict:
            return self.info_json_dict
        self._graph_filepath = self.prediction.data_folder / "info.json"
        try:
            with open(self._graph_filepath, "r") as graph_file:
                return json.loads(graph_file.read())
        except FileNotFoundError:
            return {}


def merge_rule_parts(rule_parts, repeated_ams_groups):
    """merge rule parts that match the same reactant"""
    new_rule_part = Chem.RWMol(common.utils.combine_mols_list(rule_parts))
    new_rule_part_am_to_atom = {atom.GetAtomMapNum(): atom for atom in new_rule_part.GetAtoms()}
    atoms_to_delete = set()
    for rule_ams in repeated_ams_groups:
        rule_ams = rule_ams[:]
        atom_to_keep = new_rule_part_am_to_atom[rule_ams.pop(0)]
        for rule_am in rule_ams:
            atom_to_delete = new_rule_part_am_to_atom[rule_am]
            atoms_to_delete.add(atom_to_delete.GetIdx())
            bonds = atom_to_delete.GetBonds()
            if bonds and atom_to_keep.GetAtomicNum() == 0:
                common.utils.reset_atom_query(atom_to_keep, 6)
            for bond in bonds:
                other_atom = bond.GetOtherAtom(atom_to_delete)
                if not new_rule_part.GetBondBetweenAtoms(atom_to_keep.GetIdx(), other_atom.GetIdx()):
                    new_rule_part.AddBond(atom_to_keep.GetIdx(), other_atom.GetIdx(), bond.GetBondType())
    for atom in sorted(atoms_to_delete, reverse=True):
        new_rule_part.RemoveAtom(atom)
    return new_rule_part
