"""
Functions used by the predict mechanim algorithm.
"""

# python imports
import collections
import logging
from typing import Dict, Iterable, Tuple

# rdkit imports
from rdkit import Chem
from rdkit.Geometry import Point2D, Point3D


logger = logging.getLogger("script")


def mol_is_proton(mol):
    """
    Check if mol is a proton

    :param mol: a rdkit Mol object representing a molecule
    :return: a boolean with the result of the test
    """
    return mol.GetNumAtoms() == 1 and mol.GetAtoms()[0].GetSymbol() == "H"


def first_proton(mol):
    """
    return the first proton in the mol

    :param mol: a rdkit Mol object representing a molecule
    :return atom: a rdkit Atom object representing the proton
    """
    for atom in mol.GetAtoms():
        if atom.GetSymbol() == "H":
            return atom


def mol_is_water_species(mol):
    """
    Check if mol is a water molecule, hydroxide or hydronium

    :param mol: a rdkit Mol object representing a molecule
    :return: a boolean with the result of the test
    """
    return mol.GetNumHeavyAtoms() == 1 and any([atom.GetSymbol() == "O" for atom in mol.GetAtoms()])


def mol_is_hydronium(mol):
    """
    Check if mol is a hydronium

    :param mol: a rdkit Mol object representing a molecule
    :return: a boolean with the result of the test
    """
    return mol.GetNumHeavyAtoms() == 1 and any(
        [atom.GetSymbol() == "O" and atom.GetFormalCharge() == 1 for atom in mol.GetAtoms()])


def mol_is_water_or_proton(mol):
    """
    Check if mol is a water molecule or a proton

    :param mol: a rdkit Mol object representing a molecule
    :return: a boolean with the result of the test
    """
    return mol_is_proton(mol) or mol_is_water_species(mol)


def find_n_matching_mols(mol_list_a: Iterable[Chem.Mol], mol_list_b: Iterable[Chem.Mol], mol_maps=None, comp=None):
    """Find matching mols. Will match more than one mol to another mol after masking previous matches"""
    # comparison is based on an aromatized version of the molecules
    if comp is None:
        comp = {mol: Chem.RWMol(mol) for group in [mol_list_a, mol_list_b] for mol in group}
        [Chem.SanitizeMol(mol) for mol in comp.values()]
        for mol in comp.values():
            for atom in mol.GetAtoms():
                atom.SetIsotope(atom.GetAtomicNum())
                atom.SetAtomMapNum(0)

    # first calculate all vs all similarity
    similarities = []
    mol_maps = {} if mol_maps is None else mol_maps
    for mol_a in mol_list_a:
        for mol_b in mol_list_b:
            mcs = Chem.rdFMCS.FindMCS([comp[mol_a], comp[mol_b]], timeout=2,
                atomCompare=Chem.rdFMCS.AtomCompare.CompareIsotopes)
            if mcs.queryMol is not None:
                similarities.append((mol_a, mol_b, mcs.queryMol))

    # then pair molecules starting with the most similar ones
    if similarities:
        similarities.sort(key=lambda x: -x[2].GetNumAtoms())
        mol_a, mol_b, query_mol = similarities[0]
        # print(Chem.MolToSmarts(query_mol))
        # logger.debug("Paired {} with {}. {}, {}".format(Chem.MolToSmiles(mol_a), Chem.MolToSmiles(mol_b),
                                                        # substructure.GetNumAtoms(), Chem.MolToSmiles(substructure)))
        order_a = comp[mol_a].GetSubstructMatch(query_mol)
        order_b = comp[mol_b].GetSubstructMatch(query_mol)
        idx_mapping = {order_a[no]: order_b[no] for no in range(len(order_a))}
        atom_num_map_mapping = {mol_a.GetAtomWithIdx(idx_a).GetAtomMapNum():
                                mol_b.GetAtomWithIdx(idx_b).GetAtomMapNum()
                                for idx_a, idx_b in idx_mapping.items()}
        if (mol_a, mol_b) not in mol_maps:
            mol_maps[(mol_a, mol_b)] = (idx_mapping, atom_num_map_mapping)
        for idx_a, idx_b in idx_mapping.items():
            comp[mol_a].GetAtomWithIdx(idx_a).SetIsotope(1000)
            comp[mol_b].GetAtomWithIdx(idx_b).SetIsotope(2000)

        return find_n_matching_mols(mol_list_a, mol_list_b, mol_maps, comp)
    else:
        return mol_maps


def find_matching_mols(mol_list_a: Iterable[Chem.Mol], mol_list_b: Iterable[Chem.Mol])\
        -> Dict[Chem.Mol, Tuple[Chem.Mol, Dict[int, int], Dict[int, int]]]:
    """
    Find the matching mols in two list of mols

    Returns a Dict where mols of the first list are keys and the values are a tuple with the mol
    in the second list and two mappings. The first mapping is based on the the atom idx,
    the second is based on the the atom map number

    :param mol_list_a: a list of rdkit Mol objects
    :param mol_list_b: a list of rdkit Mol objects
    :return: a Dict with the matches and atom mappings
    """
    # comparison is based on an aromatized version of the molecules
    comp = {mol: Chem.RWMol(mol) for group in [mol_list_a, mol_list_b] for mol in group}
    [Chem.SanitizeMol(mol) for mol in comp.values()]

    # first calculate all vs all similarity
    similarities = []
    for mol_a in mol_list_a:
        for mol_b in mol_list_b:
            # explicit comparison for single atoms since mcs only calculated for molecules
            if mol_a.GetNumAtoms() == 1 and mol_b.GetNumAtoms() == 1:
                substructure = mol_a
            # map protons to hydronium
            elif mol_is_proton(mol_a) and mol_is_hydronium(mol_b) or\
                    mol_is_proton(mol_b) and mol_is_hydronium(mol_a):
                substructure = mol_a if mol_is_proton(mol_a) else mol_b
            else:
                # timeout is necessary because the algorithm gets stuck sometimes
                mcs = Chem.rdFMCS.FindMCS([comp[mol_a], comp[mol_b]], timeout=2)
                substructure = Chem.MolFromSmarts(mcs.smartsString)
            if substructure.GetNumAtoms():
                similarities.append((mol_a, mol_b, substructure))

    # then pair molecules starting with the most similar ones
    similarities.sort(key=lambda x: -x[2].GetNumAtoms())
    pairs = {}
    no = 0
    while similarities:
        mol_a, mol_b, substructure = similarities[0]
        logger.debug("Paired {} with {}. {}, {}".format(Chem.MolToSmiles(mol_a), Chem.MolToSmiles(mol_b),
                                                        substructure.GetNumAtoms(), Chem.MolToSmiles(substructure)))
        no += 1
        if mol_is_proton(mol_a) and mol_is_proton(mol_b):
            idx_mapping = {mol_a.GetAtoms()[0].GetIdx(): mol_b.GetAtoms()[0].GetIdx()}
            atom_num_map_mapping = {mol_a.GetAtoms()[0].GetIdx(): mol_b.GetAtoms()[0].GetIdx()}
        elif mol_is_proton(mol_a) and mol_is_hydronium(mol_b) or\
                mol_is_proton(mol_b) and mol_is_hydronium(mol_a):
            first_proton_a, first_proton_b = first_proton(mol_a), first_proton(mol_b)
            idx_mapping = {first_proton_a.GetIdx(): first_proton_b.GetIdx()}
            atom_num_map_mapping = {first_proton_a.GetAtomMapNum(): first_proton_b.GetAtomMapNum()}
        else:
            order_a = comp[mol_a].GetSubstructMatch(substructure)
            order_b = comp[mol_b].GetSubstructMatch(substructure)
            idx_mapping = {order_a[no]: order_b[no] for no in range(len(order_a))}
            atom_num_map_mapping = {mol_a.GetAtomWithIdx(idx_a).GetAtomMapNum():
                                    mol_b.GetAtomWithIdx(idx_b).GetAtomMapNum()
                                    for idx_a, idx_b in idx_mapping.items()}
        pairs[mol_a] = (mol_b, idx_mapping, atom_num_map_mapping)
        similarities = [sim for sim in similarities if sim[0] is not mol_a and sim[1] is not mol_b]

    return pairs


def compare_mols(mol_a: Chem.Mol, mol_b: Chem.Mol, atom_compare=Chem.rdFMCS.AtomCompare.CompareElements,
                 bond_compare=Chem.rdFMCS.BondCompare.CompareOrder) -> float:
    """
    return the percentage of similar atoms between two Chem.Mol objects
    """
    mcs = Chem.rdFMCS.FindMCS([mol_a, mol_b], atomCompare=atom_compare, bondCompare=bond_compare, timeout=2)
    return mcs.numAtoms/max(mol_a.GetNumAtoms(), mol_b.GetNumAtoms())


def get_atom_map_to_point_coord(mol: Chem.Mol, conformer_id: int = 0) -> Dict[Chem.Atom, Point3D]:
    """
    return a dict that maps the atom maps of this mol to 3d coordinates

    :param mol: a rkdit Mol object with one or more conformers
    """
    try:
        conformer = mol.GetConformer(conformer_id)
    except ValueError:
        logger.warning("{} has no conformer {}".format(Chem.MolToSmiles(mol), conformer_id))
        return {}
    Point = Point3D if conformer.Is3D() else Point2D
    no_coords = 3 if conformer.Is3D() else 2
    return {atom.GetAtomMapNum(): Point(*list(conformer.GetAtomPosition(atom.GetIdx()))[:no_coords])
            for atom in mol.GetAtoms()}
