"""Contains the ReactionResult class"""

# python imports
import logging
from typing import Collection, Optional

# rdkit imports
from rdkit import Chem

logger = logging.getLogger("script")


class ReactionResult():
    """class to hold and manipulate the result of an rdkit reaction"""

    # discard messages
    LOG_DISCARD_BOND_CARBON_RESIDUES = "Bond formed with carbon atom of catalytic residue"
    LOG_DISCARD_UNSTABLE_PRODUCTS = "Product contains unstable substructure"
    LOG_DISCARD_SANITIZE_ERROR = "This product could not be sanitized: "
    LOG_DISCARD_REACTION_WITH_R_GROUP = "This product involves an R-group"
    LOG_DISCARD_ONLY_INERT_ATOMS_OF_RXN_SUBSTRATE = "This product involves only inert atoms the substrate"

    def __init__(self, reactants: Collection[Chem.Mol], products: Collection[Chem.Mol], rule, prediction_run=None):
        """
        initialises a ReactionResult object

        Methods in the class assume that all the mols in reactants have unique atom maps
        And that isotope information is used as backup for atom mapping

        :param reactants: A collection with the reactants used in the reaction

        :param products: A list of mols generated by the RunReactants or EnumerateLibrary rdkit functions
        """
        self.reactants = reactants
        self.rule = rule
        self.predict_mechanism = prediction_run
        self.reactant = reactants

        self.products, self.max_bond_length, self.reaction_centres_am = self._quick_parse(products)

    # @profile
    def _quick_parse(self, products):
        """
        just extract the info necessary for the calculation of prioritization
        """
        new_bonds_from_rule = self.predict_mechanism.rule_to_bond_changes_am_heavy[self.rule.id][0]
        self.rule_am_to_product_am = {}

        rcs_ams = set()
        for product in products:
            for atom in product.GetAtoms():
                if atom.GetAtomicNum() > 1 and atom.HasProp("old_mapno"):
                    rule_am = atom.GetIntProp("old_mapno")
                    atom.ClearProp("old_mapno")
                    product_am = atom.GetIsotope()
                    self.rule_am_to_product_am[rule_am] = product_am
                    atom.SetAtomMapNum(product_am)
                    rcs_ams.add(product_am)

        # for quick max distance use the pdb coordinates
        max_bond_length = 0
        for bond in new_bonds_from_rule:

            atom1_am, atom2_am = bond
            bond_length_1 = (self.predict_mechanism.am_to_3d_coord[self.rule_am_to_product_am[atom1_am]] -
                           self.predict_mechanism.am_to_3d_coord[self.rule_am_to_product_am[atom2_am]]).Length()

            ams = (self.rule_am_to_product_am[atom1_am], self.rule_am_to_product_am[atom2_am])
            bond_length = self.predict_mechanism.ams_to_distance[tuple(sorted(ams))]

            max_bond_length = max(max_bond_length, bond_length)

        return products, max_bond_length, rcs_ams

    def get_bond_changes(self):
        new_and_broken_bonds = [set(), set()]
        for no, data_dict in enumerate((self.p_to_idx_to_r_idx, self.r_to_idx_to_p_idx)):
            for mol, idx_to_other_idx in data_dict.items():
                for bond in mol.GetBonds():
                    idx1 = bond.GetBeginAtomIdx()
                    idx2 = bond.GetEndAtomIdx()
                    if idx1 in idx_to_other_idx and idx2 in idx_to_other_idx:
                        other1, other_idx1 = idx_to_other_idx[idx1]
                        other2, other_idx2 = idx_to_other_idx[idx2]
                        if not (other1 is other2 and other1.GetBondBetweenAtoms(other_idx1, other_idx2)):
                            new_and_broken_bonds[no].add((mol, idx1, idx2, other1, other_idx1, other2, other_idx2))
        return new_and_broken_bonds

    def get_strict_reaction_centres(self):
        """
        return am of atoms that belong to bonds that were formed or cleaved in this reaction
        """
        reactant_bonds_ams = set()
        product_bonds_ams = set()
        for product in self.products:
            for bond in product.GetBonds():
                product_bonds_ams.add(
                        (frozenset((bond.GetBeginAtom().GetAtomMapNum(), bond.GetEndAtom().GetAtomMapNum())), bond.GetBondTypeAsDouble())
                )
        for reactant in self.reactants:
            for bond in reactant.GetBonds():
                reactant_bonds_ams.add(
                        (frozenset((bond.GetBeginAtom().GetAtomMapNum(), bond.GetEndAtom().GetAtomMapNum())), bond.GetBondTypeAsDouble())
                )
        strict_reaction_centres = set()
        for bond in reactant_bonds_ams ^ product_bonds_ams:
            strict_reaction_centres.update(bond[0])
        return strict_reaction_centres

    def minimise_products_2d_conformation(self):
        """
        Clean the 2D coordinates of the products
        """
        for product in self.products:
            product.UpdatePropertyCache()
            conformer_2d = product.GetConformer(0)
            coord_map = {atom.GetIdx(): Point2D(conformer_2d.GetAtomPosition(atom.GetIdx()))
                         for atom in product.GetAtoms() if atom.GetSymbol() != "H"}
            new_conformer_id = Chem.rdDepictor.Compute2DCoords(product, clearConfs=False, coordMap=coord_map)
            new_conformer = product.GetConformer(new_conformer_id)
            for atom in product.GetAtoms():
                conformer_2d.SetAtomPosition(atom.GetIdx(), new_conformer.GetAtomPosition(atom.GetIdx()))
            product.RemoveConformer(new_conformer_id)

    def minimise_products_3d_structure(self):
        """
        do an energy minimisation of the 3d conformation of the products and update the conformer
        """
        for product in self.products:
            # copy mol, cannot minise product directly because it adds aromaticity
            product_copy = Chem.Mol(product)
            Chem.SanitizeMol(product_copy)
            mp = Chem.AllChem.MMFFGetMoleculeProperties(product_copy, mmffVariant='MMFF94s')
            ff = Chem.AllChem.MMFFGetMoleculeForceField(product_copy, mp, confId=1)
            if ff is None:
                logger.warning("Force Field does not recognize molecule: {} ".format(Chem.MolToSmiles(product_copy)))
                continue
            skip = False
            for atom in product_copy.GetAtoms():
                if atom.GetSymbol() in ["Cl"]:
                    logger.warning("Skipping minimisation because atom causes minimisation to fail: {} ".format(
                        Chem.MolToSmiles(product_copy)))
                    skip = True
                if (atom.GetAtomMapNum() not in self.reaction_centres_atom_maps) and (atom.GetSymbol() != "H"):
                    # atom is not a reaction centre, place a heavy constrain
                    ff.MMFFAddPositionConstraint(atom.GetIdx(), 0.1, 5)
                elif atom.GetSymbol() != "H":
                    # atom is a heavy atom reaction centre, place soft constraint
                    ff.MMFFAddPositionConstraint(atom.GetIdx(), 0.1, 1)
            if not skip:
                ff.Minimize(maxIts=2000)  # , forceTol=1, energyTol=1e-1)
                for atom in product.GetAtoms():
                    product.GetConformer(1).SetAtomPosition(
                        atom.GetIdx(), product_copy.GetConformer(1).GetAtomPosition(atom.GetIdx()))

    def detect_sterochemistry_from_3d_structure(self):
        """
        assign new stereochemistry flags based on the 3d structure
        """
        for product in self.products:
            Chem.rdmolops.AssignAtomChiralTagsFromStructure(product, confId=1)
            for bond in product.GetBonds():
                # this is required otherwise DetectBondStereoChemistry will not change these bonds
                if bond.GetStereo() == Chem.rdchem.BondStereo.STEREOANY:
                    bond.SetStereo(Chem.rdchem.BondStereo.STEREONONE)
            Chem.rdmolops.DetectBondStereoChemistry(product, product.GetConformer(1))

    def get_bonds_away_from_rxn_rc(self) -> Optional[int]:
        """
        return the number of bonds away this reaction is from overall rxn rcs,

        return none if reaction does not involve any of the substrates
        """
        # if self.get_strict_reaction_centres() & self.predict_mechanism.prediction.step_rc_ams:
            # return 0
        if self.reaction_centres_am & self.predict_mechanism.prediction.step_rc_ams:
            return 1
        else:
            return None
        # WIP
        # maybe add specific scores for different distances [-2, -1, 0, 2, ...] for example
        # or [ -2, -1, 0, 0, ..]
        # if mol_is_water_or_proton(step_mol):
        # continue

    def get_max_bond_length(self, allow_rotations: bool = True) -> float:
        """
        return the max bond length for this reaction_result

        :param allow_rotations: if true some kind of rudimentary rotation algorithm is applied
        """
        max_bond_length = 0.0

        for product, idx1, idx2, r1, r1_idx, r2, r2_idx in self.new_bonds:
            correction = 0.0
            r_mols = (r1, r2)
            r_idxs = (r1_idx, r2_idx)
            pos_to_compare = []
            for i, rc_idx in enumerate((idx1, idx2)):
                r_atom = r_mols[i].GetAtomWithIdx(r_idxs[i])
                r_conformer = r_mols[i].GetConformer(1)
                # position of terminal atoms not important
                # fixes problems with wrong placement of protons and oxygen in carboxylic groups, for example
                if len(r_atom.GetBonds()) == 1 and allow_rotations:
                    other_atom_idx = r_atom.GetBonds()[0].GetOtherAtom(r_atom).GetIdx()

                    bond_length = Chem.rdMolTransforms.GetBondLength(r_conformer, r_idxs[i], other_atom_idx)
                    # bond_length = dist_matrix[rc_idx][other_atom_idx]
                    correction -= bond_length * 0.5
                    pos_to_compare.append(r_conformer.GetAtomPosition(other_atom_idx))
                else:
                    pos_to_compare.append(r_conformer.GetAtomPosition(r_idxs[i]))
            bond_length = (pos_to_compare[0] - pos_to_compare[1]).Length() + correction
            max_bond_length = max(max_bond_length, bond_length)
        return max_bond_length

