# coding=utf-8
"""Includes the Protein and PDB models which store the reference sequence and structure for the entry"""

# python imports
import collections

# django imports
from django.db import models
from django.utils.html import format_html
from simple_history.models import HistoricalRecords
from django.db.models import Prefetch, OuterRef, F

# atlas imports
from common.models import Entry
import common.models


class PdbQuerySet(models.QuerySet):
    """Some predefined querysets for the PDB model"""

    def annotate_domains(self):
        """annotates with information of catalytic cath domains"""

        return self.select_related("structure").prefetch_related(
            Prefetch(
                "structure__domain_set",
                queryset=common.models.Domain.objects.filter(residue_chains__is_reference=True).select_related("cath"),
                to_attr="domains"))


class Pdb(models.Model):
    """A link from a PDB structure to the M-CSA protein model"""
    protein = models.ForeignKey("Protein", on_delete=models.CASCADE)
    structure = models.ForeignKey("Structure", on_delete=models.PROTECT)
    is_reference = models.BooleanField(null=True)
    assembly = models.IntegerField(null=True)
    history = HistoricalRecords()

    objects = PdbQuerySet.as_manager()

    def __str__(self):
        return "{}-ref:{}".format(self.structure.pdb_id, self.is_reference)

    def save(self, *args,  **kwargs):
        if self.assembly is None:
            self.assembly = self.structure.biological_assembly
        super().save(*args, **kwargs)


class ProteinQuerySet(models.QuerySet):
    """Some predefined querysets for the Entry model"""

    def sequences(self, sequence_ids, is_reference=True):
        """Filters Proteins that have these sequences"""
        return self.filter(sequences__in=sequence_ids).order_by("-is_reference")  # , is_reference=is_reference)

    def annotate_pdbs(self, is_reference=True):
        """Annotates proteins with the pdbs and structure objects"""
        return self.prefetch_related(
            Prefetch(
                "pdb_set",
                queryset=common.models.Pdb.objects.filter(is_reference=is_reference)
                               .select_related("structure").annotate_domains(),
                to_attr="pdbs"))


class Protein(models.Model):
    """A protein is defined as a list of sequences (uniprot ids)"""
    entry = models.ForeignKey("Entry", on_delete=models.CASCADE)
    name = models.CharField(max_length=255, blank=False, verbose_name="Name")
    is_reference = models.BooleanField(verbose_name="Is Reference?", db_index=True)
    notes = models.TextField(verbose_name='Notes', blank=True)
    YES_NO_UNKNOWN_CHOICES = (
        ("unknown", "Unknown"),
        ("yes", "yes"),
        ("no", "no"),
    )
    is_alosteric = models.CharField(max_length=10, choices=YES_NO_UNKNOWN_CHOICES, default="unknown")
    known_to_moonlight = models.BooleanField(verbose_name="Known to Moonlight", default=False)
    sequences = models.ManyToManyField('Sequence')
    # related to automatic annotation
    # default values are adequate to curated proteins
    is_automated = models.BooleanField(verbose_name="Annotated Automatically?", default=False)
    conserved_active_site = models.BooleanField(default=True)
    same_ec = models.BooleanField(default=True)

    history = HistoricalRecords()

    objects = ProteinQuerySet.as_manager()

    # class Meta:
    #     ordering = ['entry', '-is_reference']

    def __str__(self):
        return "{0.name} - {0.uniprot_str}".format(self)

    @property
    def uniprot_str(self):
        return ", ".join([str(s) for s in self.sequences.all()[:5]])

    @property
    def uniprot_text_links(self):
        sequences = [s.uniprot_text_link for s in self.sequences.all()[:6]]
        text = format_html(", ".join(sequences[:5]) + " (...)" if len(sequences) > 5 else "")
        return text

    @property
    def uniprot_species_text_links(self):
        return format_html(", ".join([s.species.uniprot_text_link for s in self.sequences.all()]))

    @classmethod
    def create_automated_objects(cls):
        """Creates proteins of homologous sequences with the same EC and catalytic residues of the curated mechanism"""
        uniprot2id = dict(common.models.Sequence.objects.reviewed().get_idx2id())
        uniprot2id.update(dict(common.models.Sequence.objects.reference().get_idx2id()))
        entries = Entry.objects.mcsa()
        for protein in cls.objects.filter(is_reference=True, entry__in=entries).prefetch_related("sequences").order_by("entry__macie_id"):
            # print(protein)
            ecs = set(common.models.EC.objects.filter(reaction__entry__protein=protein).values_list("id", flat=True))
            ecs.update(set(common.models.EC.objects.filter(sequence__protein=protein).values_list("id", flat=True)))
            uniprot_resid = collections.defaultdict(dict)
            residues = common.models.Residue.objects.filter(active_site__entry=protein.entry)
            for resseq in common.models.ResidueSequence.objects.residues(residues).prefetch_related("sequence__ecs", "reference_sequence"):
                hom_ecs = set([ec.id for ec in resseq.sequence.ecs.all()])
                if (hom_ecs & ecs) or resseq.reference_sequence_id == resseq.sequence_id:
                    uniprot_resid[resseq.reference_sequence.uniprot_id][(resseq.uniprot_id, resseq.residue_id)] = resseq

            for ref_uniprot_id, hom_dict in uniprot_resid.items():
                conservation_list = collections.defaultdict(list)
                for (hom_uniprot_id, hom_residue_id), resseq in hom_dict.items():
                    ref_resseq = uniprot_resid[ref_uniprot_id][(ref_uniprot_id, hom_residue_id)]
                    conservation_list[hom_uniprot_id].append(ref_resseq.code == resseq.code)
            # create only one automated protein linked to all the sequences
            sequences_to_add = set()
            automated_protein, _ = Protein.objects.get_or_create(
                entry=protein.entry,
                name="automated",
                notes="Same EC as the reference protein. All catalytic residues are conserved",
                is_reference=False,
                is_automated=True,
                conserved_active_site=True,
                same_ec=True)
            for uniprot_id, conservation in conservation_list.items():
                if all(conservation):
                    sequences_to_add.add(uniprot2id[uniprot_id])
            automated_protein.sequences.set(sequences_to_add)
            # print(automated_protein, len(sequences_to_add))

        return

    def find_repeated_proteins(self):
        macie_entries = Entry.objects.mcsa().exclude(id=self.entry.id)
        existing_proteins = Protein.objects.filter(entry__in=macie_entries, is_automated=False)
        repeated_proteins = existing_proteins.filter(sequences__in=self.sequences.all())
        return repeated_proteins
